/*
 *  Commandant.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSException.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSArray.h>
#include "Commandant.h"
#include "Sound.h"
#include "Missile.h"
#include "GGConcreteShip.h"
#include "World.h"
#include "Tableau.h"
#include "utile.h"
#include "GGOrbiteur.h"
#include "Sched.h"
#include "Laser.h"
#include "Console.h"
#include "GGSky.h"
#include "GGRepere.h"
#include "Preference.h"


#define VAISSEAU 1
#define ORBITEUR 2

Commandant *theCommandant = nil;
NSString *GGNotificationShipDoHyperJump = @"GGNotificationShipDoHyperJump";

static inline Wing createWings(real a, real b,real c, real x, real y, real z,
			       real alpha, real beta)
{
  Wing w;
  initVect(&w.pos, a, b, c);
  initVect(&w.n, x, y, z);
  w.alpha = -alpha;
  w.beta = -beta;
  return w;
}
			       

@implementation Commandant
- init
{
  [super init];
  _flags.locality = 1;
  numShip = 0;
  destination = nil;
  modePilote = Manuel;
  theCommandant = self;
  carburant = 10000;
  fric = 100000;
  indexSystem = 0;
  speedWished = 0.0;

  tracked = YES;

#ifdef GGSIM
  /*grandes ailes*/
  wings[0] = createWings(5, 0, 0,
			 0, 1, 0,
			 2, 5);
  wings[1] = createWings(-5, 0, 0,
			 0, 1, 0,
			 2, 5);

  /*ailerons de roulis */
  wings[2] = createWings(5, 0, 0,
			 0, 1, 0,
			 0, 1);
  wings[3] = createWings(-5, 0, 0,
			 0, 1, 0,
			 0, 1);

  /*ailerons de profondeur */
  wings[4] = createWings(0, 0, -5,
			 0, 1, 0,
			 0, 1);

  /*drive */
  wings[5] = createWings(0, 0, -5,
			 1, 0, 0,
			 0, 3);

  /*gouverne*/
  wings[6] = createWings(0, 1, -5,
			 1, 0, 0,
			 0, 1);

  initVect(&inertie, 10, 10, 10);
#endif
  return self;
}

- (void)encodeWithCoder:(NSCoder *)encoder
{
#define GGEncode(obj) [encoder encodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]

  [super encodeWithCoder: encoder];	
  GGEncode(numShip);
  GGEncode(modePilote);
  GGEncode(destination);
  GGEncode(hinterFlag);
  GGEncode(speedWished);
  GGEncode(sleepTimeHinter);
  GGEncode(timerAction);
  GGEncode(altitude);
  GGEncode(hauteurTerrain);
  GGEncode(nearPlanete);
  GGEncode(wishSystemIndex);
}

- (id)initWithCoder:(NSCoder *)decoder
{
  [super initWithCoder: decoder];
  GGDecode(numShip);
  GGDecode(modePilote);
  destination = [decoder decodeObject];
  GGDecode(hinterFlag);
  GGDecode(speedWished);
  GGDecode(sleepTimeHinter);
  GGDecode(timerAction);
  GGDecode(altitude);
  GGDecode(hauteurTerrain);
  GGDecode(nearPlanete);
  GGDecode(wishSystemIndex);

  theCommandant = self;

  return self;
}


- (ModePilotage) modePilote
{
  return modePilote;
}


- (int) wishSystemIndex
{
  return wishSystemIndex;
}

- (void) setWishSystemIndex: (int) index
{
  wishSystemIndex = index;
}

- (void) hyperJump
{
  if( indexSystem != wishSystemIndex )
    {
      [[NSNotificationCenter defaultCenter]
	postNotificationName: GGNotificationShipDoHyperJump
	object: self];
      //FIXME: add the cloud here
      [theWorld jumpToSystemIndex: wishSystemIndex];
      indexSystem = wishSystemIndex;
    }
  else
    [self notify: @"can't jump to the same system"];
}

- (GGSpaceObject *) destination
{
  return destination;
}

- (void) notify: (NSString *) format,...
{
  va_list ap;

  va_start (ap, format);
  consoleLogv (format, ap);
  va_end (ap);
  [theLoop setTimeSpeed: 1];
}

- (void) atDestination: (NSNotification *)aNotification
{
  [self setModePilotage: Manuel];
  [self notify: @"Ship at Destination"];
}


- (void) setDestination: (GGSpaceObject *) newDest
{
  if( newDest == destination ) return;
  destination =  newDest;
  [destination enregistreForRemoving: self
	       with: @selector(removeDestinationWithNotification:)];
  if( modePilote == AutoPilote )
    [self autoPilotTo: newDest];
}

- (void) removeDestinationWithNotification:  (NSNotification *) aNotification
{
  if( [aNotification object] == destination )
    [self setDestination: nil];
}

- (void) changeModePilot
{
  switch(modePilote)
    {
    case Manuel:
      [self setModePilotage: Hinter];
      break;
    case AutoPilote:
      [self setModePilotage: Manuel];
      break;
    case Hinter:
      [self setModePilotage: AutoPilote];
      break;
    default:
      consoleLog(@"mode pilotage inconnu");
      break;
    }
}

- (void) autoPilotTo: (GGSpaceObject *)dest
{
  [self setDestination: dest];
  [self setModePilotage: Manuel];
  [self setModePilotage: AutoPilote];
}


- (void) setTrack: (BOOL) val
{
}


- (void) nextTarget
{
  NSArray *ships = [theWorld ships];
  unsigned size;
  if( numShip >= (size = [ships count]) )
    numShip = 0;
  if( size == 0 )
    {
      consoleLog(@"No ship to target!");
    }
  else
    {
      [self setTarget: [ships objectAtIndex: numShip]];
      numShip++;
    }
}

- (GGShip *) fireMissile
{
  GGShip *mis = [super fireMissile];
  if( mis != nil )
    [theWorld setAuxiliaryCamera: mis];
  return mis;
}

- (double) speedWished
{
  return speedWished;
}

- (void) startHinterMode
{
  GGOrbiteur *closestOrbiteur = [theTableau closestOrbiteur];
  PetitGalileen pg;

  if( closestOrbiteur == nil )
    return;

  [closestOrbiteur positionSpeedRelativesTo: self
		   decal: NULL
		   in: &pg];
  speedWished = sqrt(prodScal(&pg.speed, &pg.speed));

#ifdef DEBUG
  if( thePref.easyPilot )
    speedWished = 0.0;
#endif

  [theLoop iaSchedule: self
	   withSelector: @selector(hintShip)
	   arg: nil];
  modePilote = Hinter;
  sleepTimeHinter = 0;
  hinterFlag = 0;
}

- (void) stopHinterMode
{
  [self stopPousse];
  [[theLoop iaScheduler] deleteTarget: self
			withSelector: @selector(hintShip)];
  modePilote = Manuel;
}

- (void) hintShip
{
  GGOrbiteur *closestOrbiteur = [theTableau closestOrbiteur];
  double constanteFacteur = 0;
  double normeDiff;
  PetitGalileen pg;
  Vect3D wish;

  if( closestOrbiteur == nil )
    {
      [self stopHinterMode];
      return;
    }

  [closestOrbiteur positionSpeedRelativesTo: self
		   decal: NULL
		   in: &pg];
  mulScalVect(Direction(self), speedWished, &wish);
  addVect(&wish, &pg.speed, &wish);

  normeDiff = prodScal(&wish, &wish);
  if( normeDiff > EPS*EPS )
    {
      normeDiff = sqrt(normeDiff);
      divScalVect(&wish, normeDiff, &wish);
      [self setPousse: &wish];
      [self setPower: normeDiff/(GACC(self)*deltaTime)];
    }
  else
    [self stopPousse];

  NSDebugMLLogWindow(@"hinter", @"accelere %d", hinterFlag);

  if( hinterFlag & (GGAccelere | GGDecelere) )
    {
      timerAction += realDeltaTime;
      constanteFacteur = 8+2*timerAction;
    }

  if( sleepTimeHinter > 0 )
    {
      sleepTimeHinter -= realDeltaTime;
    }
  else if( hinterFlag & GGAccelere )
    {
      if( speedWished < 0 && speedWished > -1 )
	{
	  speedWished = 0;
	  sleepTimeHinter = 2.0;
	}
      else
	{
	  double K = pow(constanteFacteur, realDeltaTime);
	  if( speedWished < 0 )
	    speedWished /= K;
	  else if (speedWished > 0 )
	    speedWished *= K;
	  else
	    speedWished = 1;
	}
    }
  else if( hinterFlag & GGDecelere )
    {
      if( speedWished > 0 && speedWished < 1 )
	{
	  speedWished = 0;
	  sleepTimeHinter = 2.0;
	}
      else
	{
	  double K = pow(constanteFacteur, realDeltaTime);
	  if( speedWished < 0 )
	    speedWished *= K;
	  else if (speedWished > 0 )
	    speedWished /= K;
	  else
	    speedWished = -1;
	}
    }
}

- (void) stopEngine
{
  [self setModePilotage: Manuel];
  [self stop];
}

- (void) setModePilotage: (ModePilotage) theNewMode
{
  if( modePilote == theNewMode )
    return;

  switch(modePilote)
    {
    case AutoPilote:
      [self stopAutoPilot];
      break;
    case Hinter:
      [self stopHinterMode];
      break;
    default:
      break;
    }

  NSDebugMLLog(@"hinter", @"newMode = %d", theNewMode);

  switch(theNewMode)
    {
    case AutoPilote:
      if( destination != nil )
	{
	  AutoPilotCommon *mis;
	  //	NSLog(@"cible = %@", cible);
	  if( [destination isKindOfClass: [Station class]] )
	    mis = [FlightToStation destination: (Station *)destination
				   ship: self];
	  else
	    mis = [AutoPilotCommon autoPilotWithTarget: destination
				   andShip: self];
	  [self setIA: mis];
	  modePilote = AutoPilote;
	  break;
	}
      //else, we go to next case :
    case Manuel:
      modePilote = Manuel;
      break;
    case Hinter:
      [self startHinterMode];
      break;
    default:
      break;
    }
}


- (void) accelere
{
  NSDebugMLLog(@"hinter", @"modePilote %d", modePilote);
  if( modePilote != Hinter)
    [super accelere];
  else
    {
      hinterFlag |= GGAccelere;
      timerAction = 0;
    }
}

- (void) stopAccelere
{
  [super stopAccelere];
  hinterFlag &= ~GGAccelere;
}

- (void) decelere
{
  if( modePilote != Hinter )
    [super decelere];
  else
    {
      hinterFlag |= GGDecelere;
      timerAction = 0;
    }
}

- (void) stopDecelere
{
  [super stopDecelere];
  hinterFlag &= ~GGDecelere;
}
    
- (void) setSpeedWished: (real) val
{
  speedWished = val;
}

- (void) setNearPlanete: (BOOL) flag
{
  nearPlanete = flag;
}

- (void) setAltitude: (double) val
{
  nearPlanete = YES;
  altitude = val;
}

- (void) setHauteurTerrain: (double) val
{
  nearPlanete = YES;
  hauteurTerrain = val;
}

- (BOOL) nearPlanete
{
  return nearPlanete;
}

- (double) altitude
{
  NSAssert(nearPlanete, NSInternalInconsistencyException);
  return altitude;
}

- (double) hauteurTerrain
{
  NSAssert(nearPlanete, NSInternalInconsistencyException);
  return hauteurTerrain;
}

- (void) tryGlobalCoordinates: obj
{
  GGRepere * _repere = [manager repere];
  double d;
  d = (prodScal(Point(self), Point(self)));
  if( d > distanceCameraRepere2 )
    {
      NSDebugMLLog(@"updaterep", @"obj : %@ rep : %@", 
		   [self getStringPosition],
		   [_repere getStringPosition]);
      [_repere updateOrigine: self];
    }
}



#ifdef GGSIM

- (void) setXMouvement: (float) val
{
  real sina = sin(val*M_PI/4);
  real cosa = cos(val*M_PI/4);
  wings[2].n.z = -sina;
  wings[2].n.x = 0;
  wings[2].n.y = cosa;
  wings[3].n.z = sina;
  wings[3].n.x = 0;
  wings[3].n.y = cosa;
}

- (void) setYMouvement: (float) val
{
  real sina = sin(val*M_PI/4);
  real cosa = cos(val*M_PI/4);
  wings[4].n.z = sina;
  wings[4].n.x = 0;
  wings[4].n.y = cosa;
  //  NSLog(@"fuck %@", stringOfVect(&wings[4].n));
}

- (void) setRolMouvement: (float) val
{
  real sina = sin(val*M_PI/4);
  real cosa = cos(val*M_PI/4);
  wings[6].n.z = -sina;
  wings[6].n.x = cosa;
  wings[6].n.y = 0;
}

- (void) updatePosition
{
  int i;
  Vect3D grav = {0, 0, -10};
  Vect3D v;

  NSDebugMLLogWindow(@"ailes", @"speed %@", stringOfVect(Speed(self)));
  transposeProduitVect3D(Matrice(self), Speed(self), &v);
  //  [self addResultante: &grav];
  gAcc = 10;
  gFrein = 4;

  NSDebugMLLogWindow(@"ailes", @"speed %@", stringOfVect(&v));

  for( i = 0; i <= 6; ++i )
    {
      Vect3D temp, moment, MG, result;
      mulScalVect(&v, wings[i].alpha, &temp);
      addLambdaVect(&temp, 
		    (wings[i].beta-wings[i].alpha)*prodScal(&v, &wings[i].n),
		    &wings[i].n, &temp);
      mkVectFromPoint(&wings[i].pos, (Vect3D *)&GGZeroVect, &MG);
      prodVect(&temp, &MG, &moment);
      produitMatriceVecteur3D(Matrice(self), &temp, &result);
      [self addMomentum: &moment
  	    resultante: &result];
      NSDebugMLLogWindow(@"ailes", @"%@\n%@", stringOfVect(&temp),
			 stringOfVect(&moment));
    }
  [super updatePosition];
}
#else

#ifdef DEBUG
#define coef 1
- (void) updatePosition
{
  if( thePref.easyPilot )
    {
      if( modeInertie & GGLeft )
	addLambdaVect(Direction(self), realDeltaTime*coef, Gauche(self),
		      Direction(self));
      if( modeInertie & GGRight )
	addLambdaVect(Direction(self), -realDeltaTime*coef, Gauche(self),
		      Direction(self));
      if( modeInertie & GGUp )
	addLambdaVect(Direction(self), -realDeltaTime*coef, Haut(self),
		      Direction(self));
      if( modeInertie & GGDown )
	addLambdaVect(Direction(self), +realDeltaTime*coef, Haut(self),
		      Direction(self));
      if( modeInertie & GGStrafLeft )
	addLambdaVect(Haut(self), +realDeltaTime*coef, Gauche(self),
		      Haut(self));
      if( modeInertie & GGStrafRight )
	addLambdaVect(Haut(self), -realDeltaTime*coef, Gauche(self),
		      Haut(self));
      addLambdaVect(Point(self), deltaTime*speedWished, Direction(self),
		    Point(self));

      [self normalise];
      
    }
  else
    [super updatePosition];
}
#endif

#endif
@end
