/*
 *  Console.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: April 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <GL/gl.h>
#include <Foundation/NSObjCRuntime.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSException.h>
#include <Foundation/NSData.h>
#include <Foundation/NSDebug.h>
#include "config.h"
#include "Metaobj.h"
#include "Console.h"
#include "GGLoop.h"
#include "World.h"
#include "GGRepere.h"
#include "GGFont.h"
#include "GGInput.h"


//static Console *theConsole;
NSLog_printf_handler *oldHandler;

void myLog(NSString *str)
{
  Console *theConsole = [theGGInput console];
  (*oldHandler)(str);
  [theConsole logString: str];
}

void consoleLogv (NSString* format, va_list args)
{
  Console *theConsole = [theGGInput console];
  [theConsole logString: 
		[NSString stringWithFormat: format arguments: args]];
}

void 
consoleLog (NSString* format, ...)
{
  va_list ap;

  va_start (ap, format);
  consoleLogv (format, ap);
  va_end (ap);
}


@implementation Console
+ (void) initialize
{
  if( self == [Console class] )
    {
//        oldHandler = _NSLog_printf_handler;

//        _NSLog_printf_handler = myLog;
    }
}

- (id) initWithFrame: (NSRect)frameRect
{
  [super initWithFrame: frameRect];

  array = [NSMutableArray new];
  ASSIGN(layoutFont, [GGFont fontWithFile: DefaultFontName
			   atSize: 12]);
  [self clear];
  return self;
}

- (void) dealloc
{
  RELEASE(array);
  RELEASE(layoutFont);
  //  NSWarnMLog(@"fuck!!!");
  [super dealloc];
}

- (void) _addData: (NSMutableData *) data
	 withSize: (int) size
{
  if( size > 0 )
    {
      [data setLength: size*sizeof(InterLeavedT2V3)];
      [array addObject: data];
      ligne -= [layoutFont lineSpacing];
    }
}

- (void) setTransluentBackground: (BOOL) flag
{
  transluentBackground = flag;
}

- (void) logString: (NSString *)str
{
  NSMutableData *data = [NSMutableData dataWithCapacity: 
					 4*40*sizeof(InterLeavedT2V3)];
  
  InterLeavedT2V3 buffer[4];
  int i, size;
  int len = [str length];
  const unsigned char *cStr = [str cString];
  NSPoint current = NSMakePoint(8, ligne);
  int widthConsole = [self frame].size.width;

  size = 0;
  for( i = 0; i < len; ++i)
    {
      int width;
      switch(cStr[i])
	{
	case  '\n':
	  [self _addData: data
		withSize: size];

	  data = [NSMutableData dataWithCapacity: 
				   4*40*sizeof(InterLeavedT2V3)];
	  size = 0;
	  current.y = ligne;
	  current.x = 8;
	  break;
	case ' ':
	  current.x += [layoutFont spaceWidth];
	  break;
	default:
	  width = [layoutFont getGlyphInfo: buffer
			forChar: cStr[i]
			atOrigin: current];
  	  NSAssert3(width > 0, @"char with width 0  %d :'%@' %d", i, str, 
  		    (int)cStr[i]);
	  
	  [data appendBytes: buffer
		length: sizeof(buffer)];

	  current.x += width;

	  if( current.x > widthConsole-8 )
	    {
	      [self _addData: data
		    withSize: size];
	      data = [NSMutableData dataWithCapacity: 
				      4*40*sizeof(InterLeavedT2V3)];
	      size = 0;
	      current.y = ligne;
	      current.x = 8;
	    }
	  size += 4;
	  break;
	}
    }
  [self _addData: data
	withSize: size];
}

- (void) log: (NSString *)format, ...
{
  va_list ap;
  NSString *str;

  va_start (ap, format);
  str = [NSString stringWithFormat: format
		  arguments: ap];
  [self logString: str];
  va_end (ap);
}

- (void) clear
{
  NSRect frameRect = [self frame];
  [array  removeAllObjects];
  ligne = frameRect.size.height - 14;
}

- (void) drawRect: (NSRect) theRect
{
  int i;
  int n = [array count];

  if( transluentBackground )
    {
      NSRect rect = [self frame];
      glColor4f(0.5, 0.5, 0.5, 0.7);
      glRecti(0, 0, rect.size.width, rect.size.height);
    }

  if( n > 0)
    {
      NSData *data[n];
      
      [layoutFont selectFont];
      glColor3f(1, 1, 1);
      [array getObjects: data];

      if( ligne < 8 )
	glTranslatef(0, -ligne+8, 0);


      for( i = 0; i < n; ++i)
	{
	  int size = [data[i] length]/sizeof(InterLeavedT2V3);
	  glInterleavedArrays(GL_T2F_V3F, 0, [data[i] bytes]);
	  glDrawArrays(GL_QUADS, 0, size);
	}
      [layoutFont endSelectFont];
    }

}

@end
