/*
 *  Explosion.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDebug.h>
#include "GGLoop.h"
#include "Explosion.h"
#include "GGRepere.h"
#include "utile.h"
#include "Sched.h"
#include "Tableau.h"

#include <math.h>
#include <time.h>

static Vect3D randomVect(float max)
{
  Vect3D res;
  res.x = (vrnd()-0.5)*max;
  res.y = (vrnd()-0.5)*max;
  res.z = (vrnd()-0.5)*max;
  return res;
}


@implementation Explosion
- (void) __initTriangles: (double) temp
{
  int i;

  _flags.graphical = YES;
  nTriangles = 15;
  triangles = malloc(nTriangles*sizeof(Triangle));


  for( i = 0; i < nTriangles; ++ i)
    {
      triangles[i].A = randomVect(temp);
      triangles[i].B = randomVect(temp);
      triangles[i].C = randomVect(temp);
      SetZeroVect(triangles[i].pos);
      triangles[i].vit = randomVect(temp);
      triangles[i].couleur.r = vrnd();
      triangles[i].couleur.g = vrnd();
      triangles[i].couleur.b = vrnd();
      triangles[i].lifeTime = 3*vrnd();
    }
  collisionCounter = -1;
  masse = 1e6;
}

- 	initAt: (GGSpaceObject *)obj
    withEnergy: (float) val
{
  double temp = sqrt(val);

  [self initAt: obj];
  [self __initTriangles: temp];

  return self;
}

/*The inherited implementation of encodeWithCoder is the good one,
  for the moment
*/

- initWithCoder: (NSCoder *) decoder
{
  [super initWithCoder: decoder];
  [self __initTriangles: 1000];
  return self;
}

+ explosionAt: (GGSpaceObject *) obj
   withEnergy: (float) val
{
  Explosion *blast;
  NSDebugMLLog(@"Explosion", @"blast = %g", val);
  blast = [self alloc];
  return AUTORELEASE([blast initAt: obj
			    withEnergy: val]);
}

- (void) dealloc
{
  NSDebugMLLog(@"Explosion", @"Explosion libere");
  free(triangles);
  [super dealloc];
}

- (void) updatePosition
{
  int i;
  BOOL ok=NO;
  for( i = 0; i < nTriangles; ++i )
    {
      Triangle *T = &triangles[i];
      if( T->lifeTime > 0 )
	{
	  ok = YES;
	  T->lifeTime -= deltaTime;
	  addLambdaVect(&T->pos, deltaTime, &T->vit, &T->pos);
	}
    }

  masse /= exp(deltaTime);

  if( ok )
    [super updatePosition];
  else
    [self destroy];

}

- (BOOL) shouldCache
{
  return NO;
}

- (BOOL) shouldBeClipped
{
  return NO;
}

- (void) rasterise
{
  int i;

  NSDebugMLLogWindow(@"Explosion", @"on rasterise");
  glDisable(GL_LIGHTING);
  for( i = 0; i < nTriangles; ++i )
    {
      Triangle *T = &triangles[i];
      if( T->lifeTime > 0 )
	{
	  glPushMatrix();
	  glTranslated(T->pos.x, T->pos.y, T->pos.z); 
	  glBegin(GL_TRIANGLES);
	  glColor4d(T->couleur.r,T->couleur.g,T->couleur.b,1);
	  glVertex3d(T->A.x, T->A.y, T->A.z);
	  glVertex3d(T->B.x, T->B.y, T->B.z);
	  glVertex3d(T->C.x, T->C.y, T->C.z);
	  glEnd();
	  glPopMatrix();
	}

    }
  glEnable(GL_LIGHTING);
}

- (void) beginSchedule
  /*We don't need to call [super beginSchedule] because we don't need other
   *Scheduling facilities
   */
{
  [self setMode: GGInertialMode];
}
@end

