/*
 *  GGArray.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "GGArray.h"
#include <Foundation/NSDebug.h>



@implementation GGArray
- initWithCapacity: (unsigned) val
{
  [super init];
  lock = 0;
  theArray = RETAIN([NSMutableArray arrayWithCapacity: val]);
  newObject = RETAIN([NSMutableArray arrayWithCapacity: 10]);
  trashObject = RETAIN([NSMutableArray arrayWithCapacity: 10]);
  return self;
}

+ arrayWithCapacity: (unsigned) val
{
  GGArray *obj;

  obj = [self alloc];

  return AUTORELEASE([obj initWithCapacity: val]);
}

- (void) dealloc
{
  RELEASE(theArray);
  RELEASE(newObject);
  RELEASE(trashObject);
  [super dealloc];
}

- (void) addObject: (id) obj
{
  if( lock )
    {
      [newObject addObject: obj];
      NSDebugMLLog(@"array", @"Lock is useful");
    }
  else
    [theArray addObject: obj];
}

- (void) removeObject: (id) obj
{
  if( lock )
    {
      NSDebugMLLog(@"array", @"Lock is useful");
      [trashObject addObject: obj];
    }
  else
    [theArray removeObject: obj];
}

- (void) __clean
{
  if( lock == 0 )
    {
      [theArray removeObjectsInArray: trashObject];
      [theArray addObjectsFromArray: newObject];

      [trashObject removeAllObjects];
      [newObject removeAllObjects];
    }
}

- (void) lock
{
  lock++;
}

- (void) unlock
{
  lock--;
  [self __clean];
}

- (void) makeObjectsPerformSelector:(SEL) sel
		 withObject: obj
{
  lock++;
  [theArray makeObjectsPerformSelector: sel
	    withObject: obj];
  lock--;

  /* can be locked many times */
  [self __clean];
}

- (NSMutableArray *) getArray
{
  return theArray;
}

- (int) count
{
  //FIXME : ce rsultat n'a pas beaucoup de sens, mais bon.
  return [theArray count];
}
@end
