/* 	-*-ObjC-*- */
/*
 *  GGBlock.h
 *
 *  Copyright (c) 2002 by Frdric De Jaeger
 *  
 *  Date: January 2002
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GGBlock_h
#define __GGBlock_h

#include "type.h"

@class GGBlock;
@class GGScene;
@class Texture;

typedef struct __subblock
{
  GGBlock			*block;
  //  short				err;
  Vect3D			pos;
}SubBlock;

#define BRANCHEMENT 16
#define LNBRANCHEMENT 4
#define SIZEMIP (BRANCHEMENT+1)
#define SIZEEPS (LNBRANCHEMENT+2)
#define SIZEBOX (BRANCHEMENT*2)

typedef struct __boxf
{
  short int lomin, lomax, lamin, lamax, hmin, hmax;
}Boxf;

typedef struct __boxExt
{
  Boxf theBox;
  unsigned short roughness;
  unsigned char cache;
}BoxExt;


typedef struct __frame
{
  Vect3D		pos;
  real			size;
  int			i, j;
  Vect3D		decal;   
}Frame;

#include "GGBump.h"
@class NSMutableArray;

typedef struct __blockContext
{
  Vect3D		posObs;
  Vect3D		posLocalFrame;
  int			nFrame;
  Frustum		frustum;
  NSMutableArray	*forLocal, *localStudy;
  NSMutableArray	*forGlobal;
  Texture		*texture;
}BlockContext;

@interface GGBlock : NSObject
{
  BlockContext		*manager;
  SubBlock		fils[BRANCHEMENT][BRANCHEMENT];

  /* data associates with this block */
  VertexData		data[SIZEMIP][SIZEMIP];
  BoxExt		boundingBox[SIZEBOX][SIZEBOX];


  /* pointer to adjacent blocks */
@public
  GGBlock		*haut, *bas, *droite, *gauche;
@protected
  
  /* pointer to the father, NULL means we are the root */
  GGBlock		*father;
  NSObject<GGBump>	*bump;

  /*position of the block in space*/
  Frame			frame;

  /* position of the center of the box :
   */
  Vect3D		origin;  
  /*position of the center of the box in the local frame
   */


  int			nFrame, frameCreation;

  /* current Level Of Detail of this block */
  int			lod;
  double		floatMipLevel;

  /* depth in the tree */
  int			depth;

  /* rugosity of the terrain, help to compute lod depending of the position
     of the observer */
  unsigned short	epsilon[SIZEEPS];

  /* cached value of the number of lateral sons that are split
     The problem is.  It's not possible to split a son if the neighboors are not
     enough splitted.  So we cache this information so that we can test 
     efficiently if we can split or not */
  unsigned short int	nhaut, nbas, ndroite, ngauche;
  unsigned short	nSonSplit; //total number of sons splitted.
  BOOL			alreadyVisited, local;

  /* position in the reality */
  double		xmin, ymin, width, height;
}

+ topBlockWithManager:(BlockContext *)man
		 bump: (NSObject<GGBump> *)aBump
		   at: (double) _xmin
		  and: (double) _ymin
		width: (double) _width
	       height: (double) _height;
- (void) rasterize;
+ (void) initGLContextBeforeDrawing: (Texture *)theTexture;
+ (void) finish;

#ifdef DEBUG
+ freeze;
- (void) checkSons;
+ (void)debugBefore;
+ (void)debugAfter;
+ (void) debugAfterGlobal;
+ (void) debugAfterLocal;
+goLevelU;
+ goUp;
+goDown;
+goLeft;
+goRight;
+goLevelD;
#endif
@end

void secondPassBis(GGBlock *block);

#endif
