/*
 *  GGBump.m
 *
 *  Copyright (c) 2002 by Frdric De Jaeger
 *  
 *  Date: January 2002
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <Foundation/NSObject.h>
#include <Foundation/NSException.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDebug.h>
#include "GGBlock.h"
#include "utile.h"
#include "GG3D.h"
#include "perlin.h"
#include "Metaobj.h"

#define NHARM 8
#define SCALEF 3000
#define SCALEAlt 1000
#define OCTAVE 2
#define ALPHACOEF 2.5

static inline void dataAt(double radius, double theta, double phi, 
			  VertexData *pdata, double *rel)
{
  const double gradFact = SCALEF*SCALEAlt;
  double a, b, c;
  double h;
  double cost;
  double r;
  double grad[3];
  Vect3D grad2;
  Vect3D ur;
  Vect3D ut, uf;
  Vect3D n;
  Vect3D *pos, *norm;
  pos = &pdata->v;
  norm = &pdata->norm;

  NSCParameterAssert(-M_PI/2 <= theta && theta <= M_PI/2 && 0 <= phi && 
		     phi <= 2*M_PI);


  cost = cos(theta);

  a = cost*cos(phi);
  b = cost*sin(phi);
  c = sin(theta);

//    pdata->texCoord.x = a;
//    pdata->texCoord.y = b;
//    pdata->texCoord.z = c;

  r = radius;
//   h=SCALEAlt*PerlinNoise2D(SCALEF*phi, SCALEF*theta, ALPHACOEF, OCTAVE, NHARM);
//   grad[0] = 0;
//   grad[1] = 0;
//   grad[2] = 0;
  h=SCALEAlt*PerlinNoise3DWithGradscale(SCALEF*a, SCALEF*b, SCALEF*c, ALPHACOEF,
					OCTAVE, NHARM, grad, 1);
  if( h < 0 )
    {
      h = 0;
      initVect(norm, a, b, c);
    }
  else
    {
      r += h;
      initVect(&grad2, gradFact*grad[0], gradFact*grad[1], gradFact*grad[2]);

      //we are sure there is no error here...
      if( theta == M_PI/2 ) 
	{
	  initVect(&ur, 0, 0, 1);
	  initVect(&uf, 1, 0, 0);
	  initVect(&ut, 0, 1, 0);
	}
      else if( theta == -M_PI/2 )
	{
	  initVect(&ur, 0, 0, -1);
	  initVect(&uf, 1, 0, 0);
	  initVect(&ut, 0, -1, 0);
	}
      else
	{
	  initVect(&ur, a, b, c);
	  initVect(&uf, -sin(phi), cos(phi), 0);
	  initVect(&ut, -sin(theta)*cos(phi), -sin(theta)*sin(phi), cos(theta));
	}


      

      /* vecteur normal  la surface en coordonnes sphrique 
	 (u_r, u_\phi, u_\theta)*/
      initVect(&n, r, -prodScal(&grad2, &uf), -prodScal(&grad2, &ut));

      mulScalVect(&ur, n.x, norm);
      addLambdaVect(norm, n.y, &uf, norm);
      addLambdaVect(norm, n.z, &ut, norm);
  
      normaliseVect(norm);
    }

  if(rel)
    {
      //we keep double precision as far as we can :
      pos->x = (r*a) - rel[0];
      pos->y = (r*b) - rel[1];
      pos->z = (r*c) - rel[2];
    }      
  else
    {
      pos->x = (r*a);
      pos->y = (r*b);
      pos->z = (r*c);
    }


}

#if 1
static inline void vertexAtInDouble(double radius, double theta, double phi, 
				    double res[3])
{
  double a, b, c;
  double h;
  double cost;
  double r;

  NSCParameterAssert(-M_PI/2 <= theta && theta <= M_PI/2 && 0 <= phi && 
		     phi <= 2*M_PI);


  cost = cos(theta);

  a = cost*cos(phi);
  b = cost*sin(phi);
  c = sin(theta);

  r = radius;
//   h=SCALEAlt*PerlinNoise2D(SCALEF*phi, SCALEF*theta, ALPHACOEF, 
//  OCTAVE, NHARM);
  h=SCALEAlt*PerlinNoise3D(SCALEF*a, SCALEF*b, SCALEF*c, ALPHACOEF, 
			   OCTAVE, NHARM);
  if( h < 0 )
    {
      h = 0;
    }
  r += h;
  res[0] = (float)(r*a);
  res[1] = (float)(r*b);
  res[2] = (float)(r*c);
}
#endif 

#if 0
static inline void vertexAtInDouble(double radius, double theta, double phi, 
				    double res[3])

{
  VertexData data;
  dataAt(radius, theta, phi, &data, NULL);
  res[0] = data.v.x;
  res[1] = data.v.y;
  res[2] = data.v.z;
}
#endif


@implementation PerlinBump
- init
{
  [super init];
  addToCheck(self);
  return self;
}

+ bump
{
  return AUTORELEASE([[self alloc] init]);
}

- (void) dataAt: (double) x
	    and: (double) y
	     in: (VertexData *)pdata
       relative: (double *)orig
{
  double grad[3];
  Vect3D *pos, *norm;
  //  VectCol *col;
  pos = &pdata->v;
  norm = &pdata->norm;
  //  col = &pdata->col;

  pos->x = 1000*x;
  pos->y = 1000*y;
  pos->z = 1000*PerlinNoise3DWithGradscale(x, y, 1.24, ALPHACOEF, 
					   OCTAVE, NHARM, grad, 1e-6);
  //  pos->z = 1000*PerlinNoise2DWithGradscale(x, y, 2, 1.8, 10, grad, 1e-6);
//    pos->z = 1000*PerlinNoise2DPeriodiqueXWithGradscale(x, y, 2.2, 15, 
//  							  M_PI,
//  							  grad, 
//  							  scale);
  if( pos->z < 0)
    {
      pos->z = 0;
      initVect(norm, 0, 0, 1);
    }
  else
    {
      norm->x = -grad[0];
      norm->y = -grad[1];
      norm->z = 1;
      normaliseVect(norm);
    }

}

- (void) vertexAt: (double) x
	      and: (double) y
	 inDouble: (double [3])dest
{}

@end

@implementation PerlinBumpSphere
- (void) dataAt: (double) x
	    and: (double) y
	     in: (VertexData *)pdata
       relative: (double *)orig
{
  return dataAt(radius, y, x, pdata, orig);
}

- (void) vertexAt: (double) x
	      and: (double) y
	 inDouble: (double [3])dest
{
  return vertexAtInDouble(radius, y, x, dest);
}


- initWithRadius: (double) r
{
  [super init];
  radius = r;
  return self;
}

- init
{
  return [self initWithRadius: 10000.0];
}

+ bumpWithRadius: (double) r
{
  return AUTORELEASE([[self alloc] initWithRadius: r]);
}
@end

#if 0
@implementation PerlinBumpCube
- initWithNorm: (Vect3D )v
	   dir: (Vect3D )d
{
  [super init];
  dirnorm = v;
  dirx = d;
  prodVect(&v,&d, &diry);
  return self;
}

- init
{
  Vect3D v = {0, 0, 1};
  Vect3D w = {1, 0, 0};
  return [self initWithNorm: v
	       dir: w];
}

+ bumpWithNorm: (Vect3D)v
	   dir: (Vect3D)w
{
  return AUTORELEASE([[self alloc] initWithNorm: v
				   dir: w]);
}

- (void) dataAt: (double) x
	    and: (double) y
	     in: (VertexData *)pdata
       relative: (double *)orig
{
  double a, b, c;
  double res;
  double grad[3];
  Vect3D *pos, *norm;
  pos = &pdata->v;
  norm = &pdata->norm;

  a = dirnorm.x+x*dirx.x+y*diry.x;
  b = dirnorm.y+x*dirx.y+y*diry.y;
  c = dirnorm.z+x*dirx.z+y*diry.z;
  res = 1000*PerlinNoise3DWithGradscale(3*a, 3*b, 3*c, ALPHACOEF, 
					OCTAVE, NHARM, grad, 1e-6);

  initVect(pos, 5000*a, 5000*b, 5000*c);
  
  if( res < 0)
    {
      *norm = dirnorm;
      res = 0;
    }
  else
    {
      Vect3D grad2;
      Vect3D norm2;
      initVect(&grad2, grad[0], grad[1], grad[2]);
      mulScalVect(&grad2, 3*1000, &grad2);
      initVect(&norm2, -prodScal(&grad2, &dirx), 
	       -prodScal(&grad2, &diry), 5000);
	       
      addLambdaVect(pos, res, &dirnorm, pos);

      mulScalVect(&dirx, norm2.x, norm);
      addLambdaVect(norm, norm2.y, &diry, norm);
      addLambdaVect(norm, norm2.z, &dirnorm, norm);
      normaliseVect(norm);
    }

}
@end
#endif

@implementation PerlinBumpSphereTop
void convertTopToSpherical(double x, double y, double res[2])
{
  double theta, phi;
  if( x+y <= 0)
    {
      if( y <= x) /*first sector*/
	{
	  //we know y <= 0
	  double fy;  // =fabs(y) = -y

	  fy = -y;

	  if( fy <= 1e-8 )
	    {
	      theta=M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = (M_PI/4.0)*(y+2.0);
	      phi = (M_PI/4.0)*((x/fy)+1);
	    }
	}
      else /*4th sector */
	{
	  //we know x <= 0
	  double fx;
	  fx = -x;
	  if( fx <= 1e-8 )
	    {
	      theta=M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = (M_PI/4.0)*(x+2.0);
	      phi = (M_PI/4.0)*(7-(y/fx));
	    }
	}
    }
  else 
    {
      if( y <= x ) /*second sector */
	{
	  //we know x >= 0
	  if( x <= 1e-8)
	    {
	      theta=M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = (M_PI/4.0)*(2.0-x);
	      phi = (M_PI/4.0)*((y/x)+3);
	    }
	}
      else  /*3rd*/
	{
	  //y >= 0
	  if( y <= 1e-8 )
	    {
	      theta=M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = (M_PI/4.0)*(2.0-y);
	      phi = (M_PI/4.0)*(5-(x/y));
	    }
	}
    }
  res[0] = theta;
  res[1] = phi;
}

- (void) dataAt: (double) x
	    and: (double) y
	     in: (VertexData *)pdata
       relative: (double *)orig
{
  double res[2];
  convertTopToSpherical(x, y, res);
  dataAt(radius, res[0], res[1], pdata, orig);
}

- (void) vertexAt: (double) x
	      and: (double) y
	 inDouble: (double [3])dest
{
  double spher[2];
  convertTopToSpherical(x, y, spher);
  vertexAtInDouble(radius, spher[0], spher[1], dest);
}

@end

@implementation PerlinBumpSphereBottom
static void convertBottomToSpherical(double x, double y, double res[2])
{
  double theta, phi;
  if( x+y <= 0)
    {
      if( y <= x) /*first sector*/
	{
	  //we know y <= 0
	  double fy;  // =fabs(y) = -y

	  fy = -y;

	  if( fy <= 1e-8 )
	    {
	      theta=-M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = -(M_PI/4.0)*(y+2.0);
	      phi = (M_PI/4.0)*(7 - (x/fy));
	    }
	}
      else /*4th sector */
	{
	  //we know x <= 0
	  double fx;
	  fx = -x;
	  if( fx <= 1e-8 )
	    {
	      theta=-M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = -(M_PI/4.0)*(x+2.0);
	      phi = (M_PI/4.0)*(1+(y/fx));
	    }
	}
    }
  else 
    {
      if( y <= x ) /*second sector */
	{
	  //we know x >= 0
	  if( x <= 1e-8)
	    {
	      theta=-M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = -(M_PI/4.0)*(2.0-x);
	      phi = (M_PI/4.0)*(5-(y/x));
	    }
	}
      else  /*3rd*/
	{
	  //y >= 0
	  if( y <= 1e-8 )
	    {
	      theta=-M_PI/2;
	      phi = 0;
	    }
	  else
	    {
	      theta = -(M_PI/4.0)*(2.0-y);
	      phi = (M_PI/4.0)*(3+(x/y));
	    }
	}
    }
  res[0] = theta;
  res[1] = phi;
}

- (void) dataAt: (double) x
	    and: (double) y
	     in: (VertexData *)pdata
       relative: (double *)orig
{
  double spher[2];
  convertBottomToSpherical(x, y, spher);
  dataAt(radius, spher[0], spher[1], pdata, orig);
}


- (void) vertexAt: (double) x
	      and: (double) y
	 inDouble: (double [3])dest
{
  double spher[2];
  convertBottomToSpherical(x, y, spher);
  vertexAtInDouble(radius, spher[0], spher[1], dest);
}

@end
