/*
 *  GGConcreteShip.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSNotification.h>
#include "GG3D.h"
#include "GGInput.h"
#include "GGConcreteShip.h"
#include "utile.h"
#include "parseASC.h"
#include "Texture.h"
#include "Tableau.h"
#include "Station.h"
#include "World.h"
#include "Task.h"
#include "Sched.h"

void printData(GGdata data[], int size)
{
  int i;

  for (i = 0; i< size; ++i)
    {
      //      fprintf(stderr, "col[%d] : (%f, %f, %f, %f)\n", i, data[i].col.r,
      //             data[i].col.g, data[i].col.b, data[i].col.a);
      //      fprintf(stderr, "norm : (%f, %f, %f)\n", data[i].normal.x,
      //       data[i].normal.y,data[i].normal.z);
      fprintf(stderr, "pos : (%.3f, %.3f, %.3f)\n", data[i].vertex.x,
	      data[i].vertex.y,data[i].vertex.z);
    }
  fprintf(stderr, "sizeof(GGdata) = %d : %d\n", sizeof(GGdata), 
	  10*sizeof(GLfloat));
}
      
@implementation GGTetra /*it is a missile, now*/
- init
{
  Vect3D data[] = {
    {0.2, 0., 0.0},
    {-.2, 0, 0.0},
    {0.0, 0.2, 0.},
    {0.0, 0.0, 1}
  };

  int face[] = {0, 1, 2, 3, 0, 1};
  
  VectCol cl[] = {
    {1.0, 0, 0, 1},
    {1.0, 0, 0, 1},
    {1.0, 0, 0, 1},
    {0.0, 1, 0, 1},
    {0.0, 1, 0, 1},
    {0.0, 1, 0, 1}
  };

  int i, inc;

  Vect3D tp1, tp2, tp3;

  self = [super init];
  if( self == nil )
    return nil;

  dimension = 4;
  masse = 1000;
  gAcc = 1000;  /* 100 g should be enough */
  gFrein = 30;

  mkVectFromPoint(&(data[0]), &(data[1]), &tp1);
  mkVectFromPoint(&(data[0]), &(data[2]), &tp2);

  prodVect(&tp1, &tp2, &tp3);
  ggplot[0].normal = tp3;
  ggplot[1].normal = tp3;
  ggplot[2].normal = tp3;

  
  for( i = 0; i < sizeof(cl)/sizeof(VectCol); ++i )
    {
      ggplot[i].col = cl[i];
      ggplot[i].vertex = data[face[i]];
    }

  inc = -1;
  for( i = 1; i < sizeof(cl)/sizeof(VectCol)-1; ++i )
    {
      mkVectFromPoint(&(data[face[i]]), &(data[face[i-inc]]), &tp1);
      mkVectFromPoint(&(data[face[i]]), &(data[face[i+inc]]), &tp2);
      prodVect(&tp1, &tp2, &tp3);

      normaliseVect(&tp3);
      ggplot[i+1].normal = tp3;
      inc = -inc;
    }
  //  printData(ggplot, 6);

  [self setMode: GGInertialMode];
  timer = [GGTimer timerWithObject: self
		   message: @selector(endOfFuel:)];
  [timer scheduleForDate: gameTime + 60];
  RETAIN(timer);
  

  return self;
}

- (void) dealloc
{
  [timer invalidate];
  RELEASE(timer);
#ifdef DEBUG
  tetra = nil;
#endif
  [super dealloc];
}

- (void) endOfFuel: (GGTimer *) timer
{
  [self doExplosion];
}

- (void) atDestination: (NSNotification *)aNotification
{
  [self doExplosion];
}

- (void) rasterise
{
  int i;
#if 0
  {
    GGMatrix4 mat;
    glGetFloatv(GL_MODELVIEW_MATRIX, (real *)&mat);
    NSDebugMLLogWindow(@"Matrice", @"ModelView \n%@", stringOfMatrice(&mat));
    NSDebugMLLogWindow(@"Matrice", @"%@ \n%@",
		     self, stringOfMatrice(Matrice(self)));
    NSDebugMLLogWindow(@"Matrice", @"render \n%@",
		     stringOfMatrice(&renderMatrix));
  }
#endif


  {
    GLfloat amb[] = {0.6, 0.4, 0.4, 0.8};
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, amb);
  }

  glBegin(GL_TRIANGLE_STRIP);
  for ( i = 0; i < 6; ++i )
    {
      glColor3f(ggplot[i].col.r,
		ggplot[i].col.g,
		ggplot[i].col.b);
      glNormal3f(ggplot[i].normal.x,
      		 ggplot[i].normal.y,
      		 ggplot[i].normal.z);
      glVertex3f(ggplot[i].vertex.x,
		 ggplot[i].vertex.y,
		 ggplot[i].vertex.z);
    }
  glEnd();
}

- (float) energyForExplosion
{
  return 10000;
}

@end

@implementation GGEnnemi
- (void) setDefaultIA
{
  [self setIA:[StupidIA stupidIA: self]];
}
@end
