#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include "GGFont.h"
#include "GGGlyphLayout.h"
#include "Metaobj.h"
#include "Texture.h"
#include "Metaobj.h"
#include "Resource.h"

#include <ft2build.h>
#include FT_FREETYPE_H

static FT_Library library;

#include "config.h"

NSString *DefaultFontName = GGDefaultFontName;

@implementation GGFont
+ (void) initialize
{
  static int first = 1;
  FT_Error error;

  if (first )
    {
      first = 0;
      error = FT_Init_FreeType(&library);

      NSAssert(! error, @"unable to init freetype library");
    }
}

typedef unsigned char ligneFont[256];
static inline unsigned char getBit(unsigned char *buffer, int bit)
{
  /* there must be an assembler instruction to do this job...*/
  int index = bit >> 3;
  int bitIndex = bit & 7;
  return (buffer[index]&(1<<(7-bitIndex)) ? 1 : 0);
}

static inline int log2i(unsigned int l)
{
  /*there must exist a faster way!!*/
  int i;
  for( i = 1; l > i; i <<= 1);
  return i;
}

- initWithFile: (NSString *) fileName
	atSize: (int) sizeWished
{
  struct FT_FaceRec_ 	*face;
  ligneFont *dataGlyph;
  FT_Error error;
  int i;
  int widthFace, heightFace;
  int currentX;
  int currentY;
  int currentHeight;
  FT_GlyphSlot slot;
  int indexMax, indexMin;

  [super init];

  face = NULL;


  widthFace = sizeWished;
  heightFace = sizeWished;

  NSDebugMLLog(@"GGFont", @"opening %@ atsize : %d", fileName,
	       sizeWished);
  error = FT_New_Face( library, [fileName cString], 0, &face);
  NSAssert1(!error, @"can't open font file %@", fileName);

#ifdef DEBUG
  if( face->num_fixed_sizes > 0)
    {
      NSDebugMLLog(@"GGFont", @"there is fixed size glyphs : %d !", 
		  face->num_fixed_sizes);
      for(i = 0; i < face->num_fixed_sizes; ++i)
	NSDebugMLLog(@"GGFont", @"(%d %d)", face->available_sizes[i].width,
		    face->available_sizes[i].height);
//        widthFace = face->available_sizes[0].width;
//        heightFace = face->available_sizes[0].height;
      //      NSAssert(widthFace <= 16, @"width face > 16!!");
    }
  else
    NSDebugMLLog(@"GGFont", 
		@"no fixed size found, using %d %d", widthFace, heightFace);
#endif

  error = FT_Set_Pixel_Sizes(face, widthFace, heightFace);

  if( error)
    {
      NSAssert(face->num_fixed_sizes > 0, @"can't set font size");
      widthFace = face->available_sizes[0].width;
      heightFace = face->available_sizes[0].height;
      error = FT_Set_Pixel_Sizes(face, widthFace, heightFace);
      NSAssert(!error, @"can't set font size");
    }

  dataGlyph = NSZoneMalloc([self zone], 256*sizeof(ligneFont));
  NSAssert(dataGlyph != NULL, @"can't allocate Memory");

  slot = face->glyph;

  currentY = 0;
  currentX = 0;
  currentHeight = 0;

  for( i = 1; i < 256; ++i )
    {
      FT_UInt glyph_index;
      glyph_index = FT_Get_Char_Index(face, i);
      if( glyph_index == 0)
	continue;
      error = FT_Load_Char(face, i, FT_LOAD_RENDER);
      if( ! error && slot->metrics.height != 0)
	{
	  int k, l;
	  int pitch = slot->bitmap.pitch;
	  int index = 0;
	  if( currentX + slot->bitmap.width > 256 )
	    {
	      currentX = 0;
	      currentY += currentHeight;
	      currentHeight = 0;
	      NSAssert(currentY+slot->bitmap.rows < 256, 
		       @"too small texture!!");
	    }

	  if( slot->bitmap.rows > currentHeight )
	    currentHeight = slot->bitmap.rows;

	  if( slot->bitmap.pixel_mode == ft_pixel_mode_grays )
	    for( l = 0; l < slot->bitmap.rows; ++l)
	      {
		memcpy(dataGlyph[currentY+l]+currentX, 
		       slot->bitmap.buffer+index,
		       slot->bitmap.width);
		index += pitch;
	      }
	  else if( slot->bitmap.pixel_mode == ft_pixel_mode_mono )
	    for( l = 0; l < slot->bitmap.rows; ++l)
	      {
		for( k = 0; k < slot->bitmap.width; ++k)
		  {
		    dataGlyph[currentY+l][currentX+k] = 
		      255*getBit(slot->bitmap.buffer+index, k);
		  }
		index += pitch;
	      }
	  else
	    NSAssert(0, @"unknown pixel mode!!");

	  glyphInfo[i].s1 = (float) currentX/256;
	  glyphInfo[i].s2 = (float) (currentX+slot->bitmap.width)/256;
	  glyphInfo[i].width = slot->bitmap.width;
	  glyphInfo[i].height = slot->bitmap.rows;
	  glyphInfo[i].left = slot->bitmap_left;
	  glyphInfo[i].top = slot->bitmap_top;
	  glyphInfo[i].t1 = (float) currentY/256;
	  glyphInfo[i].t2 = (float) (currentY+slot->bitmap.rows)/256;
	  glyphInfo[i].advanceX = slot->advance.x>>6;
	  glyphInfo[i].defined = YES;

	  if( slot->bitmap_top > ascend )
	    {
	      ascend = slot->bitmap_top;
	      indexMax = i;
	    }

	  if( slot->bitmap_top-slot->bitmap.rows < descend )
	    {
	      descend = slot->bitmap_top-slot->bitmap.rows;
	      indexMin = i;
	    }

	  currentX += slot->bitmap.width;
	  
	}
      }

  lineSpacing = face->size->metrics.height>>6;
  spaceWidth = glyphInfo['M'].width;

  NSDebugMLLog(@"GGFont", @"max ascent = %d, min descend = %d", 
	       indexMax, indexMin);
  NSDebugMLLog(@"GGFont", @"ascend = %d descend = %d", ascend, descend);
  NSDebugMLLog(@"GGFont", @"lineSpacing = %d spaceWidth = %d", 
	      lineSpacing, spaceWidth);

  texture = [Texture alloc];
  [texture initForFont: dataGlyph 
	   width: 256
	   height: 256];
  NSZoneFree([self zone], dataGlyph);
  if( face != NULL )
    FT_Done_Face(face);

  drawMethod = (int (*)(id, SEL, int, NSPoint))
    [self methodForSelector:   @selector(_drawGlyph:atOrigin:)];

  addToCheck(self);
  return self;
}

- (void) selectFont
{
  glPushAttrib(GL_ENABLE_BIT);
  glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  //  glBlendFunc(GL_ONE, GL_ONE);

  glBindTexture(GL_TEXTURE_2D, [texture idTexture]);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
}

- (void) endSelectFont
{
  glDisable(GL_TEXTURE_2D);
  //  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glPopClientAttrib();
  glPopAttrib();
}

+ fontWithFile: (NSString *)aString
	atSize: (int) sizeWished
{
  GGFont 	*aFont;
  id 		obj;
  NSString	*nameOfFont;

  if( sizeWished == -1 )
    sizeWished = 18;

  nameOfFont = [aString stringByAppendingFormat: @"%d", sizeWished];
  obj = [theResourceManager resourceWithName: nameOfFont];
  if( obj != nil )
    return obj;

  aFont = [self alloc];
  [aFont initWithFile: [[theResourceManager fontPath] 
			 stringByAppendingPathComponent: aString]
	 atSize: sizeWished];
  [theResourceManager addResource: aFont
		      withName: nameOfFont];
  AUTORELEASE(aFont);
  return aFont;
}

+ fontWithFile: (NSString *) aString
{
  return [self fontWithFile: aString
	       atSize: -1];
}

+ (void) setDefaultFont: (GGFont *) f
{
  NSAssert(f != nil, NSInternalInconsistencyException);
  [theResourceManager setDefaultFont: f];
}

+ defaultFont
{
  GGFont *f;
  f = [theResourceManager defaultFont];

  if( f == nil )
    [theResourceManager setDefaultFont: [self fontWithFile: DefaultFontName]];

  return [theResourceManager defaultFont];
}

+ defaultButtonFont
{
  return [self fontWithFile: DefaultFontName
	       atSize: 18];
}

+ defaultTitleFont
{
  return [self fontWithFile: DefaultFontName
	       atSize: 18];
}

+ defaultOptionFont
{
  return [self fontWithFile: DefaultFontName
	       atSize: 14];
}

+ defaultSmallFont
{
  return [self fontWithFile: DefaultFontName
	       atSize: 10];
}

- (void) dealloc
{
  RELEASE(texture);
  [super dealloc];
}

- (int) getGlyphInfo: (InterLeavedT2V3 [4]) ilArray
	     forChar: (int) ascii
	    atOrigin: (NSPoint) origin
{
  float s1, t1, s2, t2;
  float x1, x2, y1, y2;
  NSAssert1(0<= ascii && ascii < 256, @"ascii code too big %d", ascii);

  s1 = glyphInfo[ascii].s1;
  s2 = glyphInfo[ascii].s2;
  t1 = glyphInfo[ascii].t1;
  t2 = glyphInfo[ascii].t2;

  x1 = origin.x+glyphInfo[ascii].left;
  x2 = origin.x+glyphInfo[ascii].left + glyphInfo[ascii].width;
  y1 = origin.y+glyphInfo[ascii].top - glyphInfo[ascii].height;
  y2 = origin.y+glyphInfo[ascii].top;

  ilArray[0].s = s1;
  ilArray[0].t = t1;
  ilArray[0].x = x1;
  ilArray[0].y = y2;
  ilArray[0].z = 0;

  ilArray[1].s = s1;
  ilArray[1].t = t2;
  ilArray[1].x = x1;
  ilArray[1].y = y1;
  ilArray[1].z = 0;

  ilArray[2].s = s2;
  ilArray[2].t = t2;
  ilArray[2].x = x2;
  ilArray[2].y = y1;
  ilArray[2].z = 0;

  ilArray[3].s = s2;
  ilArray[3].t = t1;
  ilArray[3].x = x2;
  ilArray[3].y = y2;
  ilArray[3].z = 0;

  return glyphInfo[ascii].advanceX;

}

- (Texture *) texture
{
  return texture;
}

- (int) lineSpacing
{
  return lineSpacing;
}

- (int) spaceWidth
{
  return spaceWidth;
}

- (int) ascend
{
  return ascend;
}

- (int) descend
{
  return descend;
}

- (int) bearingX
{
  return glyphInfo['M'].left;
}

- (int) _drawGlyph: (int) ascii
	   atOrigin: (NSPoint) origin
{
  float s1, t1, s2, t2;
  float x1, x2, y1, y2;
  NSAssert1(0<= ascii && ascii < 256, @"ascii code too big %d", ascii);

  s1 = glyphInfo[ascii].s1;
  s2 = glyphInfo[ascii].s2;
  t1 = glyphInfo[ascii].t1;
  t2 = glyphInfo[ascii].t2;

  x1 = origin.x+glyphInfo[ascii].left;
  x2 = origin.x+glyphInfo[ascii].left + glyphInfo[ascii].width;
  y1 = origin.y+glyphInfo[ascii].top - glyphInfo[ascii].height;
  y2 = origin.y+glyphInfo[ascii].top;
  
  glTexCoord2f(s1, t1);
  glVertex2f(x1, y2);

  glTexCoord2f(s1, t2);
  glVertex2f(x1, y1);

  glTexCoord2f(s2, t2);
  glVertex2f(x2, y1);


  glTexCoord2f(s2, t1);
  glVertex2f(x2, y2);

  return glyphInfo[ascii].advanceX;
}

static inline BOOL changeAttributeWithChar(unsigned char c)
{
  switch(c)
    {
    case 'b':
      glColor3f(0, 0, 1);
      break;
    case 'r':
      glColor3f(1, 0, 0);
      break;
    case 'w':
      glColor3f(1, 1, 1);
      break;
    case 'B':
      glColor3f(0, 0, 0);
      break;
    case 'y':
      glColor3f(1, 1, 0);
      break;
    case 'g':
      glColor3f(0, 1, 0);
      break;
    case '@':
      return 0;
    default:
      return -1;
    }
  return 1;
}

- (void) drawString: (NSString *) str
	   atOrigin: (NSPoint) origin
{
  int 			strlength = [str length];
  const unsigned char 	*cStr = [str cString];
  int 			i;
  int			tmp;
  NSPoint 		current;


  current = origin;
  
  [self selectFont];
  glBegin(GL_QUADS);
  for(i = 0; i < strlength; ++i)
    {
      int width;
      switch(cStr[i])
	{
	case  '\n':
	  current.y -= [self lineSpacing];
	  current.x = origin.x;
	  break;
	case ' ':
	  current.x += [self spaceWidth];
	  break;
	case '@':
	  tmp = changeAttributeWithChar(cStr[i+1]);
	  if( tmp == 1 ) /* standard escapement*/
	    {
	      i++;
	      break;
	    }
	  else if ( tmp == 0 ) /* '@' escaped */
	    {
	      i++;  // but continue to next case
	    }
	  /* unknow, print '@', continue to next case
	   */

	default:
	  width = (*drawMethod)(self, @selector(_drawGlyph:atOrigin:),
				cStr[i], current);
	  /*	  width = [self _drawGlyph: cStr[i]
			atOrigin: current];*/
	  //	  NSAssert(width > 0, @"char with width 0");
      
	  current.x += width;
	  break;
	}
    }

  
  glEnd();
  [self endSelectFont];

}
@end
