/*
 *  GGGlyphLayout.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: April 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSException.h>
#include "Metaobj.h"
#include "GGGlyphLayout.h"
#include "GGFont.h"
#include "Texture.h"

#include "Tableau.h"

typedef struct __glyphChunk
{
  int 		next;
  VectCol	col;
}GlyphChunk;

static inline BOOL colWithChar(unsigned char c, VectCol *ret)
{
#define glColor3f(x, y, z) ret->r = x;ret->g=y;ret->b=z;ret->a=1
  switch(c)
    {
    case 'b':
      glColor3f(0, 0, 1);
      break;
    case 'r':
      glColor3f(1, 0, 0);
      break;
    case 'w':
      glColor3f(1, 1, 1);
      break;
    case 'B':
      glColor3f(0, 0, 0);
      break;
    case 'y':
      glColor3f(1, 1, 0);
      break;
    case 'g':
      glColor3f(0, 1, 0);
      break;
    case '@':
      return 0;
    default:
      return -1;
    }
  return 1;
#undef glColor3f
}


@implementation GGGlyphLayout
- _setString: (NSString *)str
{
  int i;
  NSPoint current;
  const unsigned char *cStr = [str cString];
  int strlength = [str length];  
  InterLeavedT2V3 *ila;
  GlyphChunk *block;

  if( ilArray == NULL )
    ilArray = NSZoneMalloc([self zone], 4*(strlength+2)*sizeof(InterLeavedT2V3));
  else
    {
      ilArray = NSZoneRealloc([self zone], ilArray, 
			      4*(strlength+2)*sizeof(InterLeavedT2V3));
    }

  NSAssert(ilArray != NULL, @"No Memory!!!");


  multiLigne = NO;
  current = NSZeroPoint;
  nChunk = 1;
  //  size = 0;
  boundRect = NSZeroRect;
  ila = ilArray;
  block = (GlyphChunk *)ila;
  initRGBA(&block->col, 1, 1, 1, 1);
  block->next = 0;
  ila+=4;
  
  for(i = 0; i < strlength; ++i)
    {
      int width;
      switch(cStr[i])
	{
	case  '\n':
	  multiLigne = YES;
	  current.y -= [layoutFont lineSpacing];
	  current.x = 0;
	  break;
	case ' ':
	  current.x += [layoutFont spaceWidth];
	  break;
	case '@':
	  {
	    VectCol temp;
	    int res;
	    res = colWithChar(cStr[i+1], &temp);
	    if( res == 1 ) //standard escape
	      {
		if( i == 0 )
		  block->col = temp;
		else  //create a newBlock
		  {
		    block = (GlyphChunk *)ila;
		    block->next = 0;
		    block->col = temp;
		    ila += 4;
		    nChunk++;
		  }
		i++;
		break;
	      }
	    else if( res == 0 ) // '@'
	      {
		i++;
		// don't break;
	      }
	    /*else print '@'*/
	  }
	default:
	  width = [layoutFont getGlyphInfo: ila
			forChar: cStr[i]
			atOrigin: current];
//  	  width = [layoutFont getGlyphInfo: ilArray+4*size
//  			forChar: cStr[i]
//  			atOrigin: current];
//  	  NSAssert3(width > 0, @"char with width 0  %d :'%@' %d", i, str, 
//  		    (int)cStr[i]);
      
	  current.x += width;
	  ila += 4;
	  block->next++;
	  //	  size++;
	  break;
	}

      if( boundRect.size.width < current.x  )
	boundRect.size.width = current.x;

    }

  boundRect.size.height = -current.y+[layoutFont ascend] - [layoutFont descend];
  boundRect.origin.x = [layoutFont bearingX];
  boundRect.origin.y = current.y+[layoutFont descend];

  [self setAlignement: BottomAlignement];

  ASSIGN(string, str);


  addToCheck(self);
  return self;
}

- initWithFont: (GGFont *) font
	string: (NSString *) str
{
  [super init];
  ASSIGN(layoutFont, font);
  [self _setString: str];
  addToCheck(self);


  return self;
}

- (void) updateString: (NSString *)str
{
  [self _setString: str];
}

- (void) setAlignement: (Alignement) newAli
{
  alignement = newAli;
}

+ glyphLayoutWithString: (NSString *)s
	       withFont: (GGFont *)f 
{
  GGGlyphLayout *obj;
  obj = [self alloc];
  obj = [obj initWithFont: f
	     string: s];
  if( obj->nChunk == 1 && ((GlyphChunk *)obj->ilArray)->next == 0 )
    {
      obj->empty = YES;
    }
  return AUTORELEASE(obj);
}

+ glyphLayoutWithString: (NSString *)s
{
  return [self glyphLayoutWithString: s
	       withFont: [GGFont defaultFont]];
}


- (void) dealloc
{
  RELEASE(layoutFont);
  RELEASE(string);
  if( ilArray != NULL )
    NSZoneFree([self zone], ilArray);
  [super dealloc];
}

- (void) drawAtPoint: (NSPoint) aPoint
{
  // FIXME:  Factorise a ailleurs que cela ne soit fait qu'une seule fois
  if( empty ) return;
  [layoutFont selectFont];

  glPushMatrix();
  glTranslatef(aPoint.x, aPoint.y, 0);
  if( multiLigne )
    switch(alignement)
      {
      case TopAlignement:
	glTranslatef(0, -[layoutFont ascend], 0);
	break;
      case CenterAlignement:
	glTranslatef(0, boundRect.size.height / 2, 0);
	break;
      case BottomAlignement:
	glTranslatef(0, boundRect.size.height - [layoutFont ascend], 0);
	break;
      }
  else
    switch(alignement)
      {
      case TopAlignement:
	glTranslatef(0, -[layoutFont ascend], 0);
	break;
      case CenterAlignement:
	break;
      case BottomAlignement:
	glTranslatef(0, -[layoutFont descend], 0);
	break;
      }
    

  [self fastDraw];

  glPopMatrix();
  
  [layoutFont endSelectFont];

  NSDebugMLLogWindow(@"Glyph", @"%@ = %d", string, nChunk);
}

- (void) fastDraw
{
  int i;
  GlyphChunk *chunk;
  InterLeavedT2V3 *ila;

  if( empty ) return;

  chunk = (GlyphChunk *)ilArray;
  ila = ilArray+4;
  for( i = 0; i < nChunk; ++i )
    {
      if( i > 0)
	glColor4fv((real *)&chunk->col);
      glInterleavedArrays(GL_T2F_V3F, 0, ila);
      glDrawArrays(GL_QUADS, 0, 4*chunk->next);
      ila += (chunk->next) * 4;
      chunk = (GlyphChunk *)ila;
      ila += 4;
    }
}
//  {
//    glInterleavedArrays(GL_T2F_V3F, 0, ilArray);
//    glDrawArrays(GL_QUADS, 0, 4*size);
//  }

- (void) bleuargh
{
  glPushAttrib(GL_ENABLE_BIT);
  glEnable(GL_TEXTURE_2D);

  glBindTexture(GL_TEXTURE_2D, [[layoutFont texture] idTexture]);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  glColor3f(1, 1, 1);

  glBegin(GL_QUADS);

  glTexCoord2f(0, 1);
  glVertex2f(0, 0);

  glTexCoord2f(1, 1);
  glVertex2f(256, 0);
  glTexCoord2f(1, 0);
  glVertex2f(256, 256);
  glTexCoord2f(0, 0);
  glVertex2f(0, 256);
  glEnd();

  glDisable(GL_TEXTURE_2D);
  glPopAttrib();
}

- (NSString *) description
{
  return string;
}

- (NSRect) frameRect
{
  return boundRect;
}
@end
