/*
 *  GGIcon.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2002
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSDictionary.h>


#include "GGWidget.h"
#include "Metaobj.h"
#include "NSWindow.h"
#include "utile.h"
#include "Texture.h"
#include "GGGlyphLayout.h"
#include "GGIcon.h"
#include "Resource.h"

@interface Texture (IconSupport)
+ loadIconSet;
@end

#define NAMERES @"TextureIcon"

@interface IconManagerForTexture : NSObject
{
  Texture 		*tex;
  NSMutableDictionary	*dic;
  int x, y;
  int curHeight;
}
+ iconManagerForTexture;
- (NSRect) rectForIconName: (NSString *) name;
- (Texture *) texture;
@end

#define SIZETEXT 256.0

@implementation IconManagerForTexture

static IconManagerForTexture *iconManagerForTexture;

+ iconManagerForTexture;
{
  if( iconManagerForTexture == nil )
    {
      iconManagerForTexture = [[self alloc] init];
      [theResourceManager addForCleaning: self];
    }

  return iconManagerForTexture;
}

+ (void) clean
{
  DESTROY(iconManagerForTexture);
}

- init
{
  [super init];
  tex = [[Texture alloc] initWithData: NULL
		   width: (int)SIZETEXT
		   height: (int)SIZETEXT
		   channels: 3
		   mipmap: NO];
  dic = [NSMutableDictionary new];

  addToCheck(self);
		   
  return self;
}

- (void) dealloc
{
  RELEASE(tex);
  RELEASE(dic);
  [super dealloc];
}

- (NSRect) rectForIconName: (NSString *) name
{
  int width, height, ncomp;
  char *subImage;
  NSValue *rect;
  NSRect theRect;
  NSString *fileName;
  rect = [dic objectForKey: name];
  if( rect != nil )
    {
      return [rect rectValue];
    }

  fileName = [[theResourceManager texturePath] 
	       stringByAppendingPathComponent: name];
  subImage = rawImageFromJPGFileMalloc([fileName cString], &width, 
				       &height, &ncomp);

  NSAssert(ncomp == 3 && subImage != NULL, @"texture data file corrupted");


  if( x+width > SIZETEXT )
    {
      x=0;
      y += curHeight;
      curHeight = 0;
    }
      
  NSAssert(y+height <= SIZETEXT && x+width <= SIZETEXT,
	   @"Icon too big or not enough memory");

  [tex setSubImage: subImage
       format: GL_RGB
       xoffset: x
       yoffset: y
       width: width
       height: height];

  theRect = NSMakeRect((float)x/SIZETEXT,
		       (float)y/SIZETEXT,
		       (float)width/SIZETEXT,
		       (float)height/SIZETEXT);
  rect = [NSValue valueWithRect: theRect];
  [dic setObject: rect
       forKey: name];

  x += width;
  if( height > curHeight )
    curHeight = height;

  return theRect;
}

- (Texture *) texture
{
  return tex;
}
@end

@implementation GGIcon
- initWithTexture: (Texture *) text
	     rect: (NSRect) aRect
	    scale: (float) scale
{
  [super init];

  ASSIGN(texture, text);
  width = SIZETEXT*aRect.size.width*scale;
  height = SIZETEXT*aRect.size.height*scale;
  xmin = NSMinX(aRect);
  ymin = NSMinY(aRect);
  xmax = NSMaxX(aRect);
  ymax = NSMaxY(aRect);

  
  addToCheck(self);
  return self;
}

+ iconWithName: (NSString *) name
	 scale: (float) scale
{
  NSRect aRect;
  GGIcon *icon;
  IconManagerForTexture *im;

  im = [IconManagerForTexture iconManagerForTexture];

  aRect = [im rectForIconName: name];
  
  icon = [self alloc];
  [icon initWithTexture: [im texture]
	rect: aRect
	scale: scale];
  return AUTORELEASE(icon);
}

+ iconWithName: (NSString *) name
{
  return [self iconWithName: name
	       scale: 1];
}

- (void) dealloc
{
  RELEASE(texture);
  [super dealloc];
}

- (void) drawAtPoint: (NSPoint) aPoint
{
  // FIXME:  Factorise a ailleurs que cela ne soit fait qu'une seule fois
  //FIXME : cache everything.
  glPushAttrib(GL_ENABLE_BIT | GL_TEXTURE_BIT);
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, [texture idTexture]);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
  glBegin(GL_QUADS);

  glTexCoord2f(xmin, ymin);
  glVertex2f(aPoint.x, aPoint.y);

  glTexCoord2f(xmax, ymin);
  glVertex2f(aPoint.x+width, aPoint.y);

  glTexCoord2f(xmax, ymax);
  glVertex2f(aPoint.x+width, aPoint.y+height);

  glTexCoord2f(xmin, ymax);
  glVertex2f(aPoint.x, aPoint.y+height);
  
  glEnd();

  glPopAttrib();
}

- (NSRect) frameRect
{
  return NSMakeRect(0, 0, width, height);
}

@end
