/*
 *  GGInput.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <SDL/SDL.h>
#include <stdio.h>
#include <stdlib.h>

#include <Foundation/NSException.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDebug.h>
#include "GGInput.h"
#include "GGConcreteShip.h"
#include "GGLoop.h"
#include "GGRepere.h"
#include "World.h"
#include "Tableau.h"
#include "Metaobj.h"
#include "utile.h"
#include "Missile.h"
#include "Sound.h"
#include "Commandant.h"
#include "Preference.h"
#include "GGGlyphLayout.h"
#include "GGFont.h"
#include "GGSky.h"
#include "ShipWin.h"
#include "StarView.h"
#include "TopView.h"
#include "Console.h"
#include "SystemView.h"
#include "GGScene.h"
#include "GGMenu.h"
#include "GGWidget.h"
#include "Sched.h"
#include "Station.h"

#include <math.h>
#include <time.h>

GGInput *theGGInput;
@class GGBlock;

@interface NextDestView : NSView
{
  GGTimer *timer;
}
+ nextDestViewWith: (NSArray *) array;
@end

@implementation NextDestView
static BOOL NextDestViewOpened;

- initWith: (NSArray *) array
{
  int n;

  if( !NextDestViewOpened && (n = [array count]) > 0 )
    {
      GGSpaceObject *objs[n];
      GGMenu *menu;
      int i;
      [self initWithFrame: NSMakeRect(200, 100, 200, 400)];
      [array getObjects: objs];
      menu = [GGMenu menu];
      [menu setFrameOrigin: NSMakePoint(10, 10)];

      for(i = 0; i< n; ++i)
	{
	  [menu addItemWithTitle: [objs[i] description]
		target: self
		action: @selector(setDestination:)
		tag: objs[i]];
	}
      [menu endItem];
      [self addSubview: menu];

      NextDestViewOpened = YES;
      timer = [GGTimer scheduledTimerWithTimeInterval: 5
		       target: self
		       selector: @selector(removeFromSuperview)
		       repeats: NO];
      RETAIN(timer);
    }
  else
    {
      RELEASE(self);
      return nil;
    }

  return self;
}

- (void) dealloc
{
  NextDestViewOpened = NO;
  [timer invalidate];
  RELEASE(timer);
  [super dealloc];
}

- (void) drawRect: (NSRect) theRect
{
  NSRect rect = [self frame];
  glColor4f(0.5, 0.5, 0.5, 0.8);
  glRecti(0, 0, rect.size.width, rect.size.height);
}

- (void) setDestination: (GGBouton *)bo
{
  GGSpaceObject *obj = (GGSpaceObject *)[bo tag];
  [[theWorld commandant] autoPilotTo: obj];
  [self removeFromSuperview];
}

- (void) keyDown: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_t: 
      {
	Commandant *c;
	c = [theWorld commandant];
	[c autoPilotTo: [c target]];
	[self removeFromSuperview];
      }
      break;
    default:
      [super keyDown: pev];
      break;
    }
}

+ nextDestViewWith: (NSArray *) array
{
  return AUTORELEASE([[self alloc] initWith: array]);
}
@end


@implementation GGInput

- init
{
  //  [super initWithFrame: NSMakeRect(0, 0.1*HEIGHT, WIDTH, HEIGHT*0.9)];
  ViewTableau *view;
  [super initWithFrame: NSMakeRect(0, 0, WIDTH, HEIGHT)];
  ship = nil;
  trackMode = 0;
  tableauDeBord = [GGTableau new];
  [tableauDeBord setFrameRect: frameWindow];
  
  view = [ViewTableau new];
  [[self contentView] addSubview: view];

  theConsole = RETAIN([view console]);

  RELEASE(view);

  stackView = [NSMutableArray new];
  guiState = GUI_SpaceView;

  addToCheck(self);

  theGGInput = self;

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(loadingAGame:)
    name: GGNotificationBeforeLoadingGame
    object: nil];

  [self setDrawBackground: NO];
  return self;
}

//  - (void) opening
//  {
//    WinTableau *tab;
//    tab = [WinTableau window];
//    [tab open];
//  }

- (void) dealloc
{
  RELEASE(tableauDeBord);
  RELEASE(ship);
  RELEASE(theConsole);
  RELEASE(currentView);
  RELEASE(stackView);
  RELEASE(externCamera);
  theGGInput = nil;
  [super dealloc];
}

- (void) loadingAGame: obj
{
  station = nil;
}

- (Console *)console
{
  return theConsole;
}

- (void) setInStation: (Station *) s
{
  station = s;
}

- (void) printChose
{
  glViewport(frameWindow.origin.x, frameWindow.origin.y, 
	     frameWindow.size.width, frameWindow.size.height);
  glDepthMask(GL_FALSE);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_FOG);
  glDisable(GL_LIGHTING);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,frameWindow.size.width,0,frameWindow.size.height,-1.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  [tableauDeBord drawTableau];
}

- (void) setFrameWindow: (NSRect) aRect
{
  NSDebugMLLog(@"Resize", @"new : %g %g", aRect.size.width, 
	       aRect.size.height);
  [super setFrameWindow: aRect];
  [tableauDeBord setFrameRect: aRect];
  [theWorld updateProjection];
}

- (void) manageCameraMovement
{
  if( cameraMove )
    {
      if( cameraMove & GGRight )
	addLambdaVect(Direction(externCamera), 0.8*realDeltaTime, 
		      Gauche(externCamera),
		      Direction(externCamera));
      if( cameraMove & GGLeft )
	addLambdaVect(Direction(externCamera), -0.8*realDeltaTime, 
		      Gauche(externCamera),
		      Direction(externCamera));
      if( cameraMove & GGDown )
	addLambdaVect(Direction(externCamera), 0.8*realDeltaTime, 
		      Haut(externCamera),
		      Direction(externCamera));
      if( cameraMove & GGUp )
	addLambdaVect(Direction(externCamera), -0.8*realDeltaTime, 
		      Haut(externCamera),
		      Direction(externCamera));

      if( cameraMove & GGAccelere )
	distanceCamera *= exp(-realDeltaTime);
      if( cameraMove & GGDecelere )
	distanceCamera *= exp(realDeltaTime);

      if( distanceCamera < 2*Dimension(ship) )
	distanceCamera = 2*Dimension(ship);

      normaliseMatrice(Matrice(externCamera));
      mulScalVect(Direction(externCamera), -distanceCamera, 
		  Point(externCamera));
    }
}

- (void) drawWindow
{
  NSDebugMLLogWindow(@"GGInput", @"val %d", guiState);
  if( externCamera )
    [self manageCameraMovement];
  if( (windowAbove != nil && [windowAbove opacity] >= 1)
       || currentView != nil)
    {
//        glViewport(0, 0, WIDTH, HEIGHT);
//        glClearColor(0, 0, 0, 0);
//        glDisable(GL_SCISSOR_TEST);
//        glClear(GL_COLOR_BUFFER_BIT);
    }
  else
    {
      if( station )
	{
	  [self initWindowGL];
	  glClearColor(0, 0, 0, 0);
	  glClear(GL_COLOR_BUFFER_BIT);
	  glPushMatrix();
	  glTranslatef(WIDTH/2, HEIGHT/2, 0);
	  glRotatef(105, 0, 0, 1);
	  glScalef(3, 3, 3);
	  glColor3f(1, 0, 0);
	  [[GGFont defaultFont]
	    drawString: @"Hello"
	    atOrigin: NSZeroPoint];
	  glPopMatrix();
	}
      else
	[theWorld drawScene: frameWindow];
      [self printChose];
    }
  [super drawWindow];
}

- (void) display
{
  if( !open )
    return;
  [self drawWindow];
  [windowAbove display];
}


- (void) setShip: (Commandant *) unShip
{
  ASSIGN(ship, unShip);
}

- (GGShip *) ship
{
  return ship;
}

- (void) newShip
{
  Vect3D dec = { 100, 0, 0};
  GGEnnemi *unShip;
  unShip = [GGEnnemi shipWithFileName: @"viper.shp"];
  [unShip setPosition: &dec];
  [theWorld addShip: unShip];
  [[theWorld repere] addSubNode: unShip];
  [unShip scheduleAsLocal];
  [unShip setDefaultIA];
}

- (void) changeView
{
  if( externCamera == nil )
    {
      distanceCamera = 3*Dimension(ship);
      externCamera = [GGSpaceObject new];
      [ship addSubNode: externCamera];
      //      [externCamera setNodePere: (GGNode*)ship];
      Point(externCamera)->z = -distanceCamera;
      [theWorld setMainCamera: externCamera];
    }
  else
    {
      [theWorld setMainCamera: [theWorld commandant]];
      [externCamera destroy];
      DESTROY(externCamera);
    }
}

- (void) externCameraTurnLeft
{
  cameraMove |= GGLeft;
}

- (void) externCameraTurnRight
{
  cameraMove |= GGRight;
}

- (void) externCameraTurnUp
{
  cameraMove |= GGUp;
}

- (void) externCameraTurnDown
{
  cameraMove |= GGDown;
}

- (void) externCameraGetCloser
{
  cameraMove |= GGAccelere;
}

- (void) externCameraGetFarther
{
  cameraMove |= GGDecelere;
}

- (void) stopExternCameraTurnLeft
{
  cameraMove &= ~GGLeft;
}

- (void) stopExternCameraTurnRight
{
  cameraMove &= ~GGRight;
}

- (void) stopExternCameraTurnUp
{
  cameraMove &= ~GGUp;
}

- (void) stopExternCameraTurnDown
{
  cameraMove &= ~GGDown;
}

- (void) stopExternGetCloser
{
  cameraMove &= ~GGAccelere;
}

- (void) stopExternGetFarther
{
  cameraMove &= ~GGDecelere;
}

- (void) popAllView
{
  while( currentView != nil )
    [currentView close];
}

- (void) keyDown: (Event *)pev
{
  if( pev->key.keysym.mod & KMOD_SHIFT )
    {
      switch(pev->key.keysym.sym)
	{
	case SDLK_F1:
	  [theLoop setTimeSpeed: 1];
	  break;
	case SDLK_F2:
	  [theLoop setTimeSpeed: 10];
	  break;
	case SDLK_F3:
	  [theLoop setTimeSpeed: 100];
	  break;
	case SDLK_F4:
	  [theLoop setTimeSpeed: 1000];
	  break;
	case SDLK_F5:
	  [theLoop setTimeSpeed: 10000];
	  break;
	case SDLK_F10:
	  [theLoop setTimeSpeed: 0];
	  break;
	default:
	  break;
	}
    }
  else
    {
      switch(pev->key.keysym.sym)
	{
	case SDLK_F6:
	  {
	    ShipWindow *sw =  [ShipWindow new];
	    [sw open];
	    RELEASE(sw);
	  }
	  break;
	case SDLK_F1:
	  if( guiState == GUI_SpaceView )
	    [self changeView];
	  else if( guiState != GUI_StationView )
	    [self popAllView];
	  break;
	case SDLK_F2:
	  if( guiState != GUI_StationView )
	    {
	      SystemView *sv;
	      [self popAllView];
	      sv = [SystemView systemView];
	      [sv open];
	    }
	  break;
	case SDLK_F3:
	  if( guiState != GUI_StationView )
	    {	
	      StarView *sw;
	      [self popAllView];
	      sw = [StarView starView];
	      [sw open];
	    }	
	  break;
	case SDLK_F4:
	  if( station )
	    {
	      [self popAllView];
	      [station openView];
	    }
	  break;
	case SDLK_F9:
	  [theWorld trySaveBordel];
	  break;
	case SDLK_F10:
	  [theWorld tryLoadBordel];
	  break;
	default:
	  [super keyDown: pev];
	  break;
	}
#ifdef DEBUG
      if( wizard )
	switch(pev->key.keysym.sym)
	  {

	  case SDLK_o:
	    [GGBlock freeze];
	    break;
	  case SDLK_q:
	    [theLoop stopMainLoop];
	    break;
	  case SDLK_KP8:
	    [GGBlock goUp];
	    break;
	  case SDLK_KP2:
	    [GGBlock goDown];
	    break;
	  case SDLK_KP4:
	    [GGBlock goLeft];
	    break;
	  case SDLK_KP6:
	    [GGBlock goRight];
	    break;
	  case SDLK_KP5:
	    [GGBlock goLevelD];
	    break;
	  case SDLK_KP9:
	    [GGBlock goLevelU];
	    break;
	  case SDLK_u:
	    {
	      extern BOOL lookForClosest;
	      lookForClosest = YES;
	    }
	    break;
	  default:
	    break;
	  }
#endif
    }

  if( station == nil )
    {
      if( ! (pev->key.keysym.mod & KMOD_SHIFT ) )
	switch(pev->key.keysym.sym)
	  {
	  case SDLK_s:
	    [ship setSpeedWished: 0.0];
	    break;
	  case SDLK_LEFT:
	    if( externCamera )
	      [self externCameraTurnLeft];
	    else
	      [ship turnLeft];
	    break;	
	  case SDLK_RIGHT:    
	    if( externCamera )
	      [self externCameraTurnRight];
	    else
	      [ship turnRight];
	    break;
	  case SDLK_DOWN:
	    if( externCamera )
	      [self externCameraTurnDown];
	    else
	      [ship turnDown];
	    break;
	  case SDLK_UP:
	    if( externCamera )
	      [self externCameraTurnUp];
	    else
	      [ship turnUp];
	    break;
	  case SDLK_PAGEUP:
	    if( externCamera )
	      [self externCameraGetFarther];
	    break;
	  case SDLK_PAGEDOWN:
	    if( externCamera )
	      [self externCameraGetCloser];
	    break;
	  case SDLK_n:
	    //	      [ship nextDestination];
	    {
	      NextDestView *view = [NextDestView nextDestViewWith: 
						   [theWorld allOrbiteur]];
	      if( view )
		{
		  [[self contentView] addSubview: view];
		  [self makeFirstResponder: view];
		}
	    }
	    break;
	  case SDLK_t:
	    [ship nextTarget];
	    break;
	  case SDLK_r:
	    [ship setCarburant: 100000.0];
	    break;
	  case SDLK_h:
	    [ship hyperJump];
	    break;
	  case SDLK_TAB:
	    [ship changeModePilot];
	    break;
	  case SDLK_m:
	    [ship fireMissile];
	    break;
	  case SDLK_l:
	    [ship lockMissile];
	    break;
	  case SDLK_SPACE:
	    [ship startFireLaser];
	    break;
	  case SDLK_x:
	    [self newShip];
	    break;
	  case SDLK_a:
	    [ship setPower: 1.0];
	    [ship accelere];
	    break;
	  case SDLK_z:
	    [ship setPower: 1.0];
	    [ship decelere];
	    break;
	  case SDLK_COLON:
	    [ship rolLeft];
	    break;
	  case SDLK_EXCLAIM:
	    [ship rolRight];
	    break;
	  default:
	    break;
	  }
    }

}

- (void) keyUp: (Event *)pev
{
  if( guiState != GUI_SpaceView )
    return;
  switch(pev->key.keysym.sym)
    {
    case SDLK_LEFT:
      if( externCamera )
	[self stopExternCameraTurnLeft];
      else
	[ship stopTurnLeft];
      break;	
    case SDLK_RIGHT:    
      if( externCamera )
	[self stopExternCameraTurnRight];
      else
	[ship stopTurnRight];
      break;
    case SDLK_DOWN:
      if( externCamera )
	[self stopExternCameraTurnDown];
      else
	[ship stopTurnDown];
      break;
    case SDLK_UP:
      if( externCamera )
	[self stopExternCameraTurnUp];
      else
	[ship stopTurnUp];
      break;
    case SDLK_PAGEUP:
      if( externCamera )
	[self stopExternGetFarther];
      break;
    case SDLK_PAGEDOWN:
      if( externCamera )
	[self stopExternGetCloser];
      break;
    case SDLK_a:
      [ship stopAccelere];
      break;
    case SDLK_z:
      [ship stopDecelere];
      break;
    case SDLK_l:
      [ship stopManche];
      break;
    case SDLK_SPACE:
      [ship stopFireLaser];
      break;
    case SDLK_COLON:
      [ship stopRolLeft];
      break;
    case SDLK_EXCLAIM:
      [ship stopRolRight];
      break;
    default:
      [super keyUp: pev];
      break;
    }
}

- (void) mouseDown: (Event *) pev
{
  if( guiState != GUI_SpaceView )
    return;
  if( pev->button.button == SDL_BUTTON_RIGHT )
    {
      trackMode = !trackMode;
      if( !trackMode )
	{
	  [theTableau setJoyXPos: 0];
	  [ship setXMouvement: 0];
	  [theTableau setJoyYPos: 0];
	  [ship setYMouvement: 0];
	}
    }
  else if (pev->button.button == SDL_BUTTON_LEFT )
    {
      /*try to select object in screen*/
      NSPoint 		res 	= NSMakePoint(pev->button.x, pev->button.y);
      ProjectionParam	pp;
      GGSpaceObject	*enumerator;
      GGSpaceObject	*closest;
      GGSpaceObject 	*camera;
      int 		distCliq = 10000;
      
      camera = [theWorld camera];
      enumerator = [theWorld firstObject];
      [theWorld getProjectionParam: &pp];

      pp.x = 0;
      pp.y = 0;
      pp.width = frameWindow.size.width;
      pp.height = frameWindow.size.height;

      NSDebugMLLog(@"InputSelect", @"mouse %g %g", res.x, res.y);

      closest = nil;


      while( (enumerator = [enumerator nextObject]) != nil )
	{
	  Vect3D screenPos;
	  NSDebugMLLog(@"InputSelect", @"obj : %@", enumerator);
	  if( [enumerator windowsCoordinatesForCamera: camera
			  parameter: &pp
			  result: &screenPos] )
	    {
	      int temp = fabs(screenPos.x-res.x) + fabs(screenPos.y-res.y);
	      NSDebugMLLog(@"InputSelect", @"pos %@ %g %g", enumerator,
			   screenPos.x, screenPos.y);
	      if( temp < 20 && temp < distCliq )
		{
		  distCliq = temp;
		  closest = enumerator;
		}
	    }
	}

      [ship setDestination: closest];
      //      NSLog(@"closest = %@", closest);
    }
  else
    [super mouseDown: pev];
}

  

- (void) mouseMoved: (Event *) pev
{
  if( guiState != GUI_SpaceView )
    return;
  NSDebugMLLog(@"Move", @"souris bouge");
  if( trackMode )
    {
      NSSize size = frameWindow.size;
      float x, y;
      x = (float)(pev->motion.x) / size.width - 0.5;
      if( fabs(x) <= 0.2 )
	x = 0;
      else if( x >= 0 )
	{
	  x = (x-0.2)*3;
	}
      else
	x = (x+0.2)*3;
      x = GGClamp(x);
      y = (float)(pev->motion.y) / size.height - 0.5;
      if( fabs(y) <= 0.2 )
	y = 0;
      else if( y >= 0 )
	{
	  y = (y-0.2)*3;
	}
      else
	y = (y+0.2)*3;
      y = GGClamp(y);

      [theTableau setJoyXPos: x];
      [ship setXMouvement: x];
      [theTableau setJoyYPos: y];
      [ship setYMouvement: y];

//        [ship genMove: pev->motion.xrel
//  	    and: pev->motion.yrel];
    }
  else
    [super mouseMoved: pev];
}

- (void) joystickChange: (Event *)pev
{
  float temp;
  if( guiState != GUI_SpaceView )
    return;
  switch(pev->type)
    {
    case    SDL_JOYAXISMOTION:
      NSParameterAssert(pev->jaxis.which == 0);
      
      switch(pev->jaxis.axis)
	{
	case 0: // X axis
	  temp = (float)(pev->jaxis.value)/32768;
	  [theTableau setJoyXPos: temp];
	  [ship setXMouvement: temp];
	  break;
	case 1:  // Y axis
	  temp = (float)(pev->jaxis.value)/32768;
	  [theTableau setJoyYPos: temp];
	  [ship setYMouvement: temp];
	  break;
	case 2: // Z axis (rolling)
	  temp = (float)(pev->jaxis.value)/32768;
	  [ship setRolMouvement: temp];
	  break;
	case 3: // gaz
	  temp = (float)(pev->jaxis.value)/32768;
	  [ship setPower: (1-temp)/2.0];
	  break;
	default:
	  NSDebugMLLog(@"Joystick", @"SDL_JOYAXISMOTION %d val %d",
		       pev->jaxis.axis, pev->jaxis.value);
	  break;
	}
      break;
    case    SDL_JOYBALLMOTION:
      //NSDebugMLLog(@"Joystick", @"SDL_JOYBALLMOTION");
      break;
    case    SDL_JOYHATMOTION:
      //      NSDebugMLLog(@"Joystick", @"SDL_JOYHATMOTION");
      break;
    case    SDL_JOYBUTTONDOWN:
      switch(pev->jbutton.button)
	{
	case 0:
	  [ship startFireLaser];
	  break;
	case 1:
	  [ship fireMissile];
	  break;
	default:
	  NSDebugMLLog(@"Joystick", @"SDL_JOYBUTTONDOWN %d %d", 
		       pev->jbutton.button, pev->jbutton.state);
	  break;
	}
      break;
    case    SDL_JOYBUTTONUP:
      if( pev->jbutton.button == 0 )
	[ship stopFireLaser];
      break;
    default:
      [super joystickChange: pev];
      break;
    }

}

- (BOOL) respondToEvent
{
  return 1;
}

- (void) addTopView: (TopView *) aView
{
  if( currentView != nil )
    [[self contentView]
      removeSubview: currentView];
  else
    {
      [theWorld setDrawWorld: NO];
      [self looseFocus];
    }
  [stackView addObject: aView];
  ASSIGN(currentView, aView);
  [[self contentView]
    addSubview: aView];
  guiState = [aView state];
  //  [aView open];
}

- (void) aTopViewIsClosing: (TopView *) aView
{
  NSDebugMLLog(@"GGInput", @"called with %@", aView);

  if( aView == currentView )
    {
      [stackView removeLastObject];
      if( [stackView count] >= 1 )
	{
	  TopView *temp;
	  temp = [stackView lastObject];
	  ASSIGN(currentView, temp);
	  [[self contentView]
	    addSubview: temp];
	  NSDebugMLLog(@"GGInput", @"someone in the stack");
	  guiState = [currentView state];
	}
      else
	{
	  DESTROY(currentView);
	  [theWorld setDrawWorld: YES];
	  [self takeFocus];
	  NSDebugMLLog(@"GGInput", @"noone in the stack");
	  guiState = GUI_SpaceView;
	}
    }
  else
    NSWarnMLog(@"bizarre");
}

@end
