/* 	-*-ObjC-*- */
/*
 *  GGLoop.h
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GGLoop_h
#define __GGLoop_h

#include "GG3D.h"
#include <Foundation/NSObject.h>
#include <SDL/SDL.h>

@class GGRepere;
@class GGTableau;
@class GGSpaceObject;
@class GGSky;
@class GGArray;
@class NSWindow;
@class NSMutableArray;
@class Sched;
@class TaskSched;
@class World;
@class GGInput;
@class Sound;

@protocol InputHandler
- (BOOL) handleEvent: (void *)event;
@end

@class NSAutoreleasePool;

struct __loopState
{
  Sched			*fastScheduler, *slowScheduler, *iaScheduler;
  TaskSched		*taskSched;
};

@interface GGLoop : NSObject
{
  SDL_Joystick		*joy;
  struct __loopState	state;
  TaskSched		*noSaveSched;
  float		   	timeSpeed;
  BOOL             	shouldStop;
  World			*world;
  GGInput		*input;
  NSWindow		*window;
  NSWindow		*preference;
  Sound			*sound;
  short			xMouse, yMouse;
  BOOL			displayMouseCursor;
  BOOL              	fog;
  int 			cont;
  int		   	winId;
  int			pollMode;
  BOOL			grab;
  BOOL			enterMode;
  BOOL			timeSpeedChange;
  NSAutoreleasePool	*myPool;
  double		lastTime;
}

+ new: (int) argc arg: (char *[]) argv;
- (float) timeSpeed;
- setTimeSpeed: (float) val;
- (double) fps;
- (struct __loopState *)pstate;
- (void) setState: (struct __loopState *)pstate;

- (void) setInput: (GGInput *)val;
- (GGInput *) input;
- (void) setSky: (GGSky *) s;
- (void) setWin: (int) aWin;
- (void) setMainWindow: (NSWindow *)aWin;
- (void) pushWindow: (NSWindow *) aWin;
- (void) popWindow: (NSWindow *)aWin;
- (void) setSoundFlag: (BOOL) val;

- (void) setPollMode: (int) theMode;

- (void) stopMainLoop;
- (void) runLoop;
- (void) manageEvent;
- (BOOL) isStop;

- (void) beginGrab;
- (void) endGrab;
- (int) queryPointer: (int *)px and: (int *) py;

- (Sched *) slowScheduler;
- (Sched *) fastScheduler;
- (Sched *) iaScheduler;
- (TaskSched *) taskSched;
- (TaskSched *) taskSchedNoSave;

- (void) slowSchedule: (NSObject *)obj
	withSelector: (SEL) s
		 arg: arg;
- (void) fastSchedule: (NSObject *)obj
	withSelector: (SEL) s
		 arg: arg;
- (void) iaSchedule: (NSObject *)obj
      withSelector: (SEL) s
	       arg: arg;
- (void) unSchedule: (NSObject *) obj;

- (GGInput *) input;

@end

extern double deltaTime;	//time between two frame measure in game time
extern double gameTime;		// time in the game
extern double realDeltaTime; 	// real time between two frames
extern double currentTime;	// real system current time.
extern GGLoop *theLoop;

extern NSString *GGScreenSizeChange;

extern int  WIDTH;
extern int  HEIGHT;

extern BOOL wizard;


#define WaitForEvent   			0
#define DontWaitForEvent		1

#define Event SDL_Event

#endif
