/*
 *  GGLoop.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <Foundation/NSException.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSData.h>
#include <Foundation/NSException.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSValue.h>
#include "GGLoop.h"
#include "GGSpaceObject.h"
#include "GGInput.h"
#include "utile.h"
#include "Tableau.h"
#include "Metaobj.h"
#include "NSWindow.h"
#include "NSView.h"
#include "Sched.h"
#include "Preference.h"
#include "World.h"
#include "Sound.h"

int  WIDTH=800;
int  HEIGHT=600;

static float fogcolor[4]={1.0,1.0,1.0,1.0};
double deltaTime, realDeltaTime, gameTime, currentTime;
BOOL wizard;


GGLoop *theLoop = nil;
static Class spaceObjCacheClass;

NSString *GGScreenSizeChange = @"GGScreenSizeChange";

static void resize( unsigned int width, unsigned int height )
{
  NSSize oldSize = NSMakeSize(WIDTH, HEIGHT);
  NSValue *oldData = [NSValue valueWithSize: oldSize];
  NSDebugFLLog(@"Resize", @"old : (%d, %d) new : (%d %d)", WIDTH, HEIGHT,
	       width, height);
  WIDTH = width;
  HEIGHT = height;
  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGScreenSizeChange object: oldData];
}


@interface GGLoop (Private)
- (void )initGLContext;
- (void) initInput;
@end

#define PlayMode  	0
#define UIMode		1


@implementation GGLoop
+ (void)initialize
{
  if( self == [GGLoop class])
    {
      spaceObjCacheClass = [GGSpaceObject class];
    }
}

- initWithArgc: (int) argc
	  argv: (char *[]) argv
{
  int i;
  [super init];
  if( theLoop != nil )
    NSWarnLog(@"GGLoop should be instanciated only once");
  theLoop = self;
  addToCheck(self);

  shouldStop = 0;
  state.fastScheduler = [Sched new];
  state.slowScheduler = [Sched new];
  state.iaScheduler = [Sched new];
  state.taskSched = [TaskSched new];
  noSaveSched = [TaskSched new];
  grab = NO;
  fog = 0;
  enterMode = 0;
  timeSpeed = 1.0;
  timeSpeedChange = 0;
  xMouse = 0;
  yMouse = 0;
  displayMouseCursor = YES;
  pollMode = DontWaitForEvent;
  preference = RETAIN([UserPreference window]);
  window = nil;
  cont = 1;


  wizard = NO;
  for( i = 0; i < argc; ++i )
    {
      if( strcmp(argv[i], "-D") == 0 )
	wizard = YES;
    }

  if( wizard )
    NSLog(@"Starting in Wizard mode");
    
  [self initGLContext];
  [self initInput];
  if( thePref.sound )
    sound = [Sound new];
  else
    sound = nil;

  input = [GGInput new];
  world = [World new];
  [input open];
  //  [self addInputHandler: input];

  return self;
}

+ new: (int) argc 
  arg: (char *[]) argv
{
  return [[self alloc] initWithArgc: argc
		       argv: argv];
}

- (void) dealloc
{
  [sound endSound];
  if( joy != NULL )
    SDL_JoystickClose(joy);
  RELEASE(sound);
  RELEASE(window);
  RELEASE(preference);
  RELEASE(input);
  RELEASE(world);
  RELEASE(state.fastScheduler);
  RELEASE(state.slowScheduler);
  RELEASE(state.iaScheduler);
  RELEASE(state.taskSched);
  RELEASE(noSaveSched);
  RELEASE(myPool);
  [[NSNotificationCenter defaultCenter] removeObserver: self];
  [super dealloc];

  theLoop =nil;
}

- (void) initInput
{
  int njoy, i;
  njoy = SDL_NumJoysticks();
  for( i = 0; i < njoy; ++ i )
    {
      NSDebugMLLog(@"Joystick", @"joystick %d %s", i+1, SDL_JoystickName(i));
    }
  if( njoy > 0 )
    {
      SDL_JoystickEventState(SDL_ENABLE);
      joy=SDL_JoystickOpen(0);
      NSParameterAssert(joy != NULL);
      NSDebugMLLog(@"Joystick", 
		   @"Number of Axes: %d\n", SDL_JoystickNumAxes(joy));
      NSDebugMLLog(@"Joystick", 
		   @"Number of Buttons: %d\n", SDL_JoystickNumButtons(joy));
      NSDebugMLLog(@"Joystick", 
		   @"Number of Balls: %d\n", SDL_JoystickNumBalls(joy));
    }
  else
    {
      joy = NULL;
      NSDebugMLLog(@"Joystick", @"No Joystick found");
    }
}

- (void )initGLContext
{
  GLfloat mat_specular[] = { 0.2, 0.4, 0.3, 1.0 };
  GLfloat ambient[] = { 0.4, 0.4, 0.4, 1.0};
  GLfloat mat_shininess[] = { 50.0 };

  glShadeModel(GL_FLAT);

  /*  glEnable(GL_BLEND);*/
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_CULL_FACE);
   
  //  glEnable(GL_FOG);
  glFogi(GL_FOG_MODE,GL_EXP);
  glFogfv(GL_FOG_COLOR,fogcolor);
  glFogf(GL_FOG_DENSITY,0.1);
  glHint(GL_FOG_HINT,GL_NICEST);


  //   glShadeModel (GL_SMOOTH);

  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
  glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);

  glEnable(GL_DEPTH_TEST);
  {
    NSString *glExt;
    NSArray *ext;
    glExt = [NSString stringWithCString: glGetString(GL_EXTENSIONS)];
    ext = [glExt componentsSeparatedByString: @" "];
    NSDebugMLLog(@"gl", @"Extension %@", [ext componentsJoinedByString: @"\n"]);
    glExt = [NSString stringWithCString: glGetString(GL_VERSION)];
    NSDebugMLLog(@"gl", @"Version : %@", glExt);
  }
}

- (GGInput *) input
{
  return input;
}

- (void) setMainWindow: (NSWindow *)aWin
{
  ASSIGN(window, aWin);
}

- (void) pushWindow: (NSWindow *)aWin
{
  if( window != nil )
    {
      [window looseFocus];
      [window pushOnTheTop: aWin];
    }
  else
    {
      [self setMainWindow: aWin];
    }
  [aWin takeFocus];
}

- (void) popWindow: (NSWindow *)aWin
{
  [aWin looseFocus];
  if( aWin == window )
    {
      NSAssert([aWin windowBelow] == nil, NSInternalInconsistencyException);
      [self setMainWindow: nil];
    }
  else
    {
      NSWindow *below = [aWin windowBelow];
      NSDebugMLLog(@"NSWindow", @"pop %@", aWin);
      NSAssert(below != nil, NSInternalInconsistencyException);
      [below popWindow];
      [below takeFocus];
    }
}

- (void) setInput: (GGInput *)val
{
  ASSIGN(input, val);
}

- setTimeSpeed: (float) val
{
  if( val != timeSpeed )
    {
      timeSpeedChange = 1;
      timeSpeed = val;
    }
  return self;
}

- (struct __loopState *)pstate
{
  return &state;
}

- (void) setState: (struct __loopState *)pstate
{
  DESTROY(state.fastScheduler);
  DESTROY(state.slowScheduler);
  DESTROY(state.iaScheduler);
  DESTROY(state.taskSched);

  state = *pstate;
}

- (float) timeSpeed
{
  return timeSpeed;
}

- (double) fps
{
  static double base;
  static double fps = 0;
  static int count = 0;

  if( count == 0 )
    {
      base = currentTime;
    }
  if( currentTime - base >= 1.0 )
    {
      fps = (double) count/(currentTime-base);
      count = 0;
      base = currentTime;
    }
  count ++;
  return fps;
}


//  - (void) setPollMode: (int) theMode
//  {
//    if( theMode == pollMode )
//      return;
//    if( theMode == WaitForEvent )
//      {
//        pollMode = UIMode;
//        [world setPaused: YES];  
//        if( grab )
//  	{
//  	  NSDebugMLLog(@"grab", @"on end grab");
//  	  hasGrabbed = YES;
//  	  [self endGrab];
//  	}
//      }
//    else
//      {
//        modeUI = PlayMode;
//        [world setPaused: NO];
//        if( hasGrabbed )
//  	{
//  	  hasGrabbed = NO;
//  	  [self beginGrab];
//  	}
//      }
//  }

- (void) setPollMode: (int) theMode
{
  pollMode = theMode;
}

- (void) setSky: (GGSky *) s
{
  [world setSky: s];
}

- (void) setWin: (int) aWin
{
  winId = aWin;
}

- (void) setSoundFlag: (BOOL) val
{
  if( val && sound == nil )
    sound = [Sound new];
  else if(!val && sound != nil )
    {
      [sound endSound];
      DESTROY(sound);
    }
}

static void takeAScreenShot(void)
{
  unsigned char *data;
  data = NSZoneMalloc(NSDefaultMallocZone(), WIDTH*HEIGHT*4);

  glPixelStorei(GL_PACK_ALIGNMENT,1);
  glReadPixels(0, 0, WIDTH, HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, data);
  writeRGBTofile(WIDTH, HEIGHT, data);
  NSZoneFree(NSDefaultMallocZone(), data);
}

-(int) handleEvent: (Event *)pevent
{
  switch (pevent->type) {
  case SDL_KEYDOWN:
    //    val = XLookupString(&(pevent->xkey), buffer, 10, &retour, &status);
    switch (pevent->key.keysym.sym)
      {
      case SDLK_ESCAPE:
	if( ![preference isOpen] )
	  {
	    [preference open];
	    return 1;
	  }
	break;
      case SDLK_f:
	fog=!fog;
	return 1;
      case SDLK_RETURN:
	enterMode = !enterMode;
	return 1;
      case SDLK_F12:
	takeAScreenShot();
	return 1;
      default:
	break;
      }
    break;
  case SDL_MOUSEMOTION:
    {
      xMouse = pevent->motion.x;
      yMouse = pevent->motion.y;
      if( grab )
	{
	  NSDebugMLLog(@"Motion", @"%d %d %d %d", pevent->motion.x,
		       pevent->motion.y, pevent->motion.xrel, 
		       pevent->motion.yrel);
	  if( abs(xMouse-WIDTH/2) + abs(yMouse-HEIGHT/2) > 100 )
	    {
	      SDL_WarpMouse(WIDTH/2, HEIGHT/2);
	    }
	  if( abs(pevent->motion.xrel) + abs(pevent->motion.yrel) > 50 )
	    {
	      return 1;
	    }
	}
    }
    return 0;
  case SDL_VIDEORESIZE:
    resize( pevent->resize.w, pevent->resize.h);
    return 1;
  case SDL_QUIT:
    [self stopMainLoop];
    return 1;
  }
  return 0;
}


- (int) dispatchEvent: (Event *) pev
{

  if( [self handleEvent: pev] )
    return 1;
  if( window != nil )
    return [[window windowOnTop] handleEvent: pev];

  return 0;
}


- (void) drawMouseCursor
{
  int tmp = HEIGHT-yMouse;
  if( displayMouseCursor )
    {
      glViewport(0, 0, WIDTH, HEIGHT);
      glDisable(GL_SCISSOR_TEST);
      glLoadIdentity();
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0, WIDTH, 0, HEIGHT, -1, 1);
      glMatrixMode(GL_MODELVIEW);
      glColor3f(0.8, 1.0, 1.0);
      glBegin(GL_LINES);
      glVertex2i(xMouse-10, tmp-10);
      glVertex2i(xMouse+10, tmp+10);
      glVertex2i(xMouse-10, tmp+10);
      glVertex2i(xMouse+10, tmp-10);
      glEnd();
    }
}

- (void) stopMainLoop
{
  cont = 0;
}

- (BOOL) isStop
{
  return cont == 0;
}


- (void) runLoop
{
  char buffer[20];
  double tp;
  SDL_Event event;
  lastTime = gettimesec();
  gameTime = 0.0;
  
  cont = 1;

  while (cont) {
    NSAutoreleasePool *pool;
    pool = [NSAutoreleasePool new];
    currentTime = tp = gettimesec();    

    if( pollMode == WaitForEvent )
      {
	lastTime = tp;
      }

    realDeltaTime = tp - lastTime;
    deltaTime = timeSpeed*realDeltaTime;
    gameTime += deltaTime;
    lastTime = tp;

    if( enterMode )
      {
	//FIXME : buffer overflow here 
	fgets(buffer, 20, stdin);
	lastTime = gettimesec(); //suspend time
      }

    [state.fastScheduler launchAll];

    if( timeSpeedChange )
      {
	timeSpeedChange = 0;
	[state.slowScheduler launchAll];
      }

    //[state.slowScheduler launchAll];
    [state.slowScheduler launch];
    [state.iaScheduler launchAll];
    [state.taskSched run];
    [noSaveSched run];
    NSDebugLLogWindow(@"TaskSched", @"%@", state.taskSched);
    NSDebugLLogWindow(@"TaskSched", @"%@", noSaveSched);

#ifdef DEBUG
    NSDebugLLogWindow(@"Sched", @"slow : %@",state.slowScheduler);
    NSDebugLLogWindow(@"Sched", @"fast : %@",state.fastScheduler);
    NSDebugLLogWindow(@"Sched", @"ia : %@", state.iaScheduler);
    NSDebugLLogWindow(@"Channel", @"bla \"%@\"\nblou %d", [MessageManager class],
		      obsTable);
#endif
    [window display];

    [self drawMouseCursor];

    SDL_GL_SwapBuffers();

    while ( SDL_PollEvent(&event) > 0 && cont)
	[self dispatchEvent: &event];

    if( pollMode == WaitForEvent )
      {
	SDL_WaitEvent(&event );
	[self dispatchEvent: &event];
      }
    printGLError(nil);

    RELEASE(pool);
  }
}

- (BOOL) gotEvent
{
  NSDebugMLLog(@"Stub", @"here");
  return SDL_PollEvent(NULL) > 0;
}

- (SDLKey) getKeyCode: (int *) val
{
  SDL_Event event;
  NSDebugMLLog(@"Stub", @"here");
  if( SDL_WaitEvent(&event) == 0 )
    {
      NSWarnMLog(@"error");
      return SDLK_UNKNOWN;
    }
  if( event.type == SDL_KEYDOWN )
    {
      *val = 1;
      NSDebugMLLog(@"SDLKey", @"press keysym %d", (int)(event.key.keysym.sym));
      return event.key.keysym.sym;
    }
  else if (event.type == SDL_KEYUP )
    {
      *val = 0;
      NSDebugMLLog(@"SDLKey", @"release keysym %d", (int)(event.key.keysym.sym));
      return event.key.keysym.sym;
    }
  else
    {
      NSDebugMLLog(@"SDLKey", @"other event %d", (int)(event.type));
      return SDLK_UNKNOWN;
    }
}

- (void) manageEvent
{
  SDL_Event event;
  while(SDL_PollEvent(&event) && cont)
    {
      [self dispatchEvent: &event];
    }
}


- (void) show
{
  double tp;
  NSDebugMLLog(@"Stub", @"window = %@", window);
  currentTime = tp = gettimesec();

  if( pollMode == WaitForEvent )
    {
      lastTime = tp;
    }

  realDeltaTime = tp - lastTime;
  deltaTime = timeSpeed*realDeltaTime;
  gameTime += deltaTime;
  lastTime = tp;
  if( myPool != nil )
    RELEASE(myPool);
  myPool = [NSAutoreleasePool new];
  [window display];

  [self drawMouseCursor];

  SDL_GL_SwapBuffers();
  
}

- (void) beginGrab
{
  NSDebugMLLog(@"grab", @"grab");
//    XGrabPointer(disp, win, False, ButtonPressMask | PointerMotionMask ,
//  	       GrabModeAsync, GrabModeAsync, win, None , CurrentTime);
  SDL_WM_GrabInput(SDL_GRAB_ON);
  displayMouseCursor = NO;
  grab = YES;
}
- (void) endGrab
{
  NSDebugMLLog(@"grab", @"ungrab");
//    XUngrabPointer(disp, CurrentTime);
  SDL_WM_GrabInput(SDL_GRAB_OFF);
  grab = NO;
  displayMouseCursor = YES;
}

- (int) queryPointer: (int *)px and: (int *) py
{
//    int bidon;
//    unsigned fuck;
//    Window foo;
//    XQueryPointer(disp, win, &foo, &foo, &bidon, &bidon, px, py, &fuck);
  *px = xMouse;
  *py = yMouse;
//    if( abs(*px-WIDTH/2) > WIDTH/4 || abs(*py-HEIGHT/2) > HEIGHT/4)
//      {
//        XWarpPointer(disp, None, win, 0, 0, 0, 0, WIDTH/2, HEIGHT/2);
//        return 0;
//      }
//    else
    return 1;
}

- (Sched *) slowScheduler
{
  return state.slowScheduler;
}

- (Sched *) fastScheduler
{
  return state.fastScheduler;
}

- (Sched *) iaScheduler
{
  return state.iaScheduler;
}

- (TaskSched *) taskSched
{
  return state.taskSched;
}

- (TaskSched *) taskSchedNoSave
{
  return noSaveSched;
}

- (void) slowSchedule: (NSObject *)obj
	 withSelector: (SEL) s
		  arg: arg
{
  [state.slowScheduler addTarget: obj
	withSelector: s
	arg: arg];
}

- (void) fastSchedule: (NSObject *)obj
	withSelector: (SEL) s
		 arg: arg
{
  [state.fastScheduler addTarget: obj
		withSelector: s
		arg: arg];
}

- (void) iaSchedule: (NSObject *)obj
      withSelector: (SEL) s
	       arg: arg
{
  [state.iaScheduler addTarget: obj
		withSelector: s
		arg: arg];
}

- (void) unSchedule: (NSObject *) obj
{
  [state.fastScheduler deleteTarget: obj];
  [state.slowScheduler deleteTarget: obj];
  [state.iaScheduler deleteTarget: obj];
}

@end
