/* 	-*-ObjC-*- */
/*
 *  GGOrbiteur.h
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GGOrbiteur_h
#define __GGOrbiteur_h

#include <GL/glu.h>
#include "GGSpaceObject.h"


@class Texture;
@class NSArray;
@class System;

typedef struct __parametresOrbitaux
{
  real			orbiteRadius;	// rayon orbitale
  real			l0;		// phase  t=0
  //  real			l;		// phase courante
  real			omegaOrbite;	// speed angulaire orbitale

  real			inclinaison1;
  real			inclinaison2;
  real			omegaPropre;	// speed angulaire de rvolution propre
  //  real			phi;		// phase courante de rvolution propre
}ParametresOrbitaux;

typedef enum __GGTypeOrbiteur
{
  GGStandard,
  GGLocal,
  GGLocalStrong,
}GGtypeOrbiteur;

#include "GGNode.h"

@interface GGOrbiteur : GGSpaceObject
{
  //  Matrice4D	cacheRep;
  GGOrbiteur    	*pere;
  NSMutableArray	*fils;
  real			tailleSysteme;
  /* paramtres orbitaux autour du pre :
   * Les orbites sont toutes dans le plan de l'cliptique.
   * et circulaire avec le pre au centre du cercle.
   */

  ParametresOrbitaux	orbite;

  Vect3D		_nord, _haut, _droit;
  GGtypeOrbiteur	typeOrbiteur;
  BOOL			proche;
  unsigned char		exactCalcul;
  @public
  System		*surroundingSystem;
  real			distanceCamera2;
  GGNode		*repereTranslation; //not retained
  GGSpaceObject		*repereRotatif;     //not retained
}
- initSystemWithContentOfFile: (NSString *)fileName;
- initRandomSystemForStar: (int) index;
- initWithDictionary: (NSDictionary *)dic
		name: (NSString *)_name;
- setRandomSystemForStar: (int) index;

- (real) tailleSysteme;
- (NSArray *) getAllOrbiteur;
- (void) addSon: (GGOrbiteur *)aSon;
- (NSArray *)sons;
- (GGOrbiteur *) orbiteurPere;

- (BOOL) isOriginOfNode;


- (BOOL) lightSource;

- (void) setDistanceLocal: (real) val;
- (void) setDistanceGlobal: (real) val;
- (void) setOmegaOrbite: (real) val;
- (void) setParametresOrbitaux: (ParametresOrbitaux*) _orbite;
- (ParametresOrbitaux *) parametresOrbitaux;
- (void) calculeTailleSysteme;
- (void) miseAJourSysteme;
- (void) getParametresOrbitaux: (ParametresOrbitaux *)pparam;
- closestFromRelative: (Vect3D *)pos;
- getClosestFrom: (Vect3D *) pV distance: (real *)ret;
- (NSArray *) visibleOrbiteursGlobal;

- (void) startBeingLocal;
- (void) stopBeingLocal;

- (void) startBeingOrigineOfLocale: (GGNode *)aNode;
- (void) stopBeingOrigineOfLocale;
- (void) stopBeingMainLocale;
- (void) startBeingMainLocale;

- (void) beginBeingSuperLocal;
- (void) endBeingSuperLocal;

- (double) anomalie;

- (System *) surroundingSystem;
- (GGNode *) subSystemFrame;
@end

extern NSString *GGOrbiteurBeingSuperLocal;
extern NSString *GGOrbiteurEndBeingSuperLocal;
void debugOrbiteurWindow(NSString *name);
#endif
