/*
 *  GGRepere.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include "Metaobj.h"
#include "GGLoop.h"
#include "GGSpaceObject.h"
#include "GGShip.h"
#include "GGRepere.h"
#include "utile.h"
#include "Tableau.h"
#include "GGOrbiteur.h"
#include "Preference.h"
#include "World.h"
#include "Missile.h"
#include "utile.h"
#include "Commandant.h"


@implementation GGRepere
- init
{
  [super init];

  modeRepere = RepereNormal;
  return self;
}

- (BOOL) shouldSave
{
  return YES;
}

- (id)replacementObjectForArchiver:(NSArchiver *)anArchiver
{
  return [GGArchiver  ggArchiver: [NSString  stringWithFormat: @"reploc:"]];
}

- (void) updateOrigine: (GGSpaceObject *) obj
  /*
   *modifie l'origine du repre en supposant que obj vit dans les coordonnes
   *locales dudit repre. Cela donne des subtilits dans la formule de 
   *changement de repre.
   * Si le repre est centr sur une plante, il faut repasser en repre
   * flottant. De tout faon, obj est ncessairement la camra, donc
   * ce ne peut jamais tre une plante.
   */
{
#ifdef DEBUG
  {
    Vect3D pos, vit;
    pos = [obj getPositionSpeedAbsolues: &vit];
    NSDebugMLLog(@"updaterep", @"pos : %@ speed : %@", 
		 stringOfVect2(pos), stringOfVect2(vit));
  }
#endif

  NSAssert1([self isParentOf: obj], @"%@ is not a son of self", obj);

  if( modeRepere == ReperePlanete )
    {
      Vect3D i = {1, 0, 0};
      Vect3D j = {0, 1, 0};
      Vect3D k = {0, 0, 1};
      GGGalileen gal0, gal1;
//        Vect3D posObj;
  
      //  SPEEDUP:
  
      [obj galileanTransformationTo: self
	   in: &gal0];
      transposeGalileen(Galileen(self), &gal1);
      //      transformePoint4D(&gal1.matrice, &gal0.matrice.position, &posObj);
      gal1.matrice.position = gal0.matrice.position;
//        transposeProduitVect3D(Matrice(self), Speed(self), &vcom);
//        mulScalVect(&vcom, -1, &vcom);

//        gal0.matrice.gauche = i;
//        gal0.matrice.haut = j;
//        gal0.matrice.direction = k;
//        gal0.speed = vcom;

      [self galileanTransformationNodeWith: &gal1];
      Matrice(self)->gauche = i;
      Matrice(self)->haut = j;
      Matrice(self)->direction = k;
      
      
      NSDebugMLLog(@"updaterep", @"speed should be nul %@", 
		   stringOfVect(Speed(self)));
      NSDebugMLLog(@"updaterep", @"position should be nul %@", 
		   stringOfVect(Point(self)));
    }
  else
    {
      [self translateNodeTo: obj];
      [self rectifyAxes];
      [self setMode: GGInertialMode];
    }

#ifdef DEBUG
  {
    Vect3D pos, vit;
    pos = [obj getPositionSpeedAbsolues: &vit];
    NSDebugMLLog(@"updaterep", @"pos : %@ speed : %@", 
		 stringOfVect(Point(self)), stringOfVect(Speed(self)));
  }
#endif
  //  NSDebugMLLog(@"updaterep", @"position : %@", [self getStringPosition]);
}


- (void) willDrawSons: (Camera *)cam
{
  //      glShadeModel(GL_FLAT);
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  glDepthRange(0, 0.94);
  glDepthFunc(GL_LESS);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(Focal,WIDTH/(float)HEIGHT,1,300e3);
  glMatrixMode(GL_MODELVIEW);
}


- (void) beginSchedule
  /*I don't wan't to call super because here it has no meaning (according
   *to the job that does super in the current implementation
   */
{
  [self setMode: GGInertialMode];
}

- (void) changeLocalCoordinates: (GGGalileen *) pgal
{
  [super changeLocalCoordinates: pgal];
  [self rectifyAxes];
}

@end
