/*
 *  GGScrollView.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: August 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include "GGScrollView.h"
#include "GGLoop.h"
#include "GGWidget.h"

@implementation GGScrollView
- initWithFrame: (NSRect) aRect
{
  [super initWithFrame: aRect];
  posDocView = 0;
  return self;
}

- (void) dealloc
{
  [theLoop unSchedule: self];
  RELEASE(documentView);
  [super dealloc];
}

- (void) setDocumentView: (NSView *)aView
{
  NSRect frameView;
  NSRect frame;

  if( aView == documentView )
    return;

  if( documentView != nil )
    [self removeSubview: documentView];
  ASSIGN(documentView, aView);

  if( aView == nil )
    {
      scroll = NO;
      return;
    }

  [documentView setFrameOrigin: NSMakePoint(0, 0)];

  [self addSubview: documentView];

  frameView = [documentView frame];
  frame = [self frame];

  NSDebugMLLog(@"Scroll", @"scroll = %d", (int) scroll);
  if( frameView.size.height > frame.size.height )
      scroll = YES;
  else
    scroll = NO;
  heightDocument = frameView.size.height;
  [self scrollToBeginningOfDocument];
}

- (void) _setPosDocView: (float) val
{
  NSRect frame;
  NSParameterAssert(scroll);
  frame = [self frame];

  if( val + frame.size.height > heightDocument )
    val = heightDocument - frame.size.height;
  if( val < 0 )
    val = 0;

  posDocView = val;
  [documentView setFrameOrigin: NSMakePoint(0, -posDocView)];
}

- (void) _moveDown
{
  [self _setPosDocView: posDocView-100*realDeltaTime];
}

- (void) _moveUp
{
  [self _setPosDocView: posDocView+100*realDeltaTime];
}

- (void) scrollToBeginningOfDocument
{
  if( scroll )
    [self _setPosDocView: 1e5];
}

- (void) scrollUp: (BOOL) val
{
  if( scroll && documentView != nil )
    {
      NSRect frame = [self frame];
      if( heightDocument - posDocView > frame.size.height )
	{
	  if( val )
	    {
	      [theLoop fastSchedule: self
		       withSelector: @selector(_moveUp)
		       arg: nil];
	    }
	  else
	    {
	      [self _setPosDocView: posDocView+100];
	    }
	}
    }
}


- (void) scrollDown: (BOOL) val
{
  if( scroll && documentView != nil )
    {
      if( posDocView >= 20 )
	{
	  if( val )
	    {
	      [theLoop fastSchedule: self
		       withSelector: @selector(_moveDown)
		       arg: nil];
	    }
	  else
	    {
	      [self _setPosDocView: posDocView-100];
	    }
	}
    }
}


- (void) keyDown: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_UP:
      [self scrollUp: YES];
      break;
    case SDLK_DOWN:
      [self scrollDown: YES];
      break;
    case SDLK_PAGEUP:
      [self scrollUp: NO];
      break;
    case SDLK_PAGEDOWN:
      [self scrollDown: NO];
      break;
    default:
      [super keyDown: pev];
      break;
    }
}

- (void) keyUp: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_UP:
    case SDLK_DOWN:
      [theLoop unSchedule: self];
      break;
    default:
      [super keyUp: pev];
      break;
    }
}


- (void) drawRect: (NSRect) _theRect
{
  NSRect theRect = [self frame];
  NSDebugMLLog(@"GGScrollView", @"size %@", theRect);
  glColor3fv(GGBorderColor);
  glBegin(GL_LINE_LOOP);
  glVertex2f(0,0);
  glVertex2f(theRect.size.width-1, 0);
  glVertex2f(theRect.size.width-1,theRect.size.height-1);
  glVertex2f(0,theRect.size.height-1);
  glEnd();
  [super drawRect: theRect];
}
@end
