/* 	-*-ObjC-*- */
/*
 *  GGShip.h
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GGShip_h
#define __GGShip_h

#include "GGSpaceObject.h"

#define GGLeft 0x1
#define GGRight 0x2
#define GGUp 0x4
#define GGDown 0x8
#define GGIncTime 0x10
#define GGDecTime 0x20
#define GGAccelere 0x40
#define GGDecelere 0x80
#define GGManche 0x100
#define GGWishDirection 0x200
#define GGStrafLeft 0x400
#define GGStrafRight 0x800
#define GGPousse 0x1000
#define GGLaser 0x2000


@class GGRepere;
@class NSMutableDictionary;
@class Laser;
@class ModelObject;
@class Item;
@class NSNotification;
@class GGTask;

#define GACC(x) ((x)->gAcc)
#define GFREIN(x) ((x)->gFrein)
#define Pousse(x) (&((x)->resultante))
#define Momentum(x) (&((x)->momentum))

#define MAXLASER 5
#define MAXMISSILE 10

typedef enum __shipErrorCode
{
  NoError = 0,
  NoMoney,
  GunBusy,
  MissilePylonBusy,
  NoPlaceLeft,
  EquipmentNotFound
}ShipErrorCode;

@interface GGShip : GGSpaceObject
{
  @public
  NSMutableArray 	*equipement;
  NSDictionary		*property;
  ModelObject 		*model;
  GGShip		*target;  //NOT retained anymore, can cause loop.
  float			gAcc, gFrein;
  GGTask	  	*ia;
  Vect3D		resultante, momentum;
  Vect3D		inertie;
  int           	modeInertie;
  float			angleManche, _cosam, _sinam;
  float			speedRotation;
  float			powerRotation;
  float			power;
  float			xMouvement, yMouvement, rolMouvement;
  float			masseCoque;
  float			capaciteFret;
  float			masseFret;

  int			hyperDriveType;
  int			pylonRocket;
  int			pylonLaser;

  double		bouclier, endurance;
  double		carburant;	// Need some precision
  double		fric;
  float			consommation;

  Item			*missiles[MAXMISSILE];

  int			currentLaser;
  int			currentMissile;
  int			nbrMissile;
  float  		powerLaser;
  Item			*laserGun[MAXLASER];
  double		timeStartFire;
  double		timeLaserRelaxation;
  Laser			*laser;

  double		lastTime;
  Vect3D		wishDir;
  Vect3D		pousse;

  Vect3D		posReacteur1, posReacteur2;
  BOOL			invincible;
  BOOL			tracked; //if tracked by commandant

  /*index of current system:*/
  int				indexSystem;

}
/*designated initializer:
 */
- init;
+ shipWithFileName: (NSString *)name;

- (ShipErrorCode) buyItem: (Item *)it;
- (ShipErrorCode) sellItem: (Item *)it;
- (ShipErrorCode) addItem: (Item *)it;
- (ShipErrorCode) removeItem: (Item *)it;
- (int) nbrItem: (Item *)it;

- (float) rangeForHyperJump;
- (int) indexSystem;
- (void) setIndexSystem: (int) index;

- (NSObject<IA> *) ia;
- (void) setIA: (NSObject <IA> *)aIa;

- (GGShip *) target;
- (void) setTarget: (GGSpaceObject *) newTarget;

- (void) notify: (NSString *) format,...;
- (void) updatePosition;

- (void) setPower: (float) newPower;
- (float) power;

- (void) atDestination: (NSNotification *)aNotification;
- (void) looseTarget: (GGSpaceObject *)obj;

- (void) setTrack: (BOOL) val;
- (BOOL) tracked;
- (void) stopEngine;
- (void) stopAutoPilot;
- (void) stop;
- (void) setXMouvement: (float) val;
- (void) setYMouvement: (float) val;
- (void) setRolMouvement: (float) val;
- (void) turnLeft;
- (void) stopTurnLeft;
- (void) turnRight;
- (void) stopTurnRight;
- (void) turnUp;
- (void) stopTurnUp;
- (void) turnDown;
- (void) stopTurnDown;
- (void) rolLeft;
- (void) stopRolLeft;
- (void) rolRight;
- (void) stopRolRight;
//- (void) genMove: (int) dx and: (int) dy;
- (void) accelere;
- (void) stopAccelere;
- (void) decelere;
- (void) stopDecelere;
- (void) startManche;
- (void) stopManche;

- (void) setPousse: (Vect3D *)dir;
- (void) stopPousse;
- (void) setAngleManche: (float) alpha;
- (void) startWishDirection;
- (void) stopWishDirection;
- (void) setWishDirection: (Vect3D *)vect
	      speedManche:(float) val;
- (void) setSpeedManche: (float) val;

- (void) lockMissile;
- (GGShip *) fireMissile;
- (void) startFireLaser;
- (void) stopFireLaser;

- (void) doExplosion;
- (float) energyForExplosion;
- (void) setMasseCoque: (float) theMasse;
- (void) setInvincibility: (BOOL) flag;

- (double) carburant;
- (void) setCarburant: (double) val;

- (void) scheduleAsLocal;
- (void) scheduleAsGlobal;
- (void) switchToLocal:(GGRepere *)_repere;
- (void) switchToGlobal: (GGRepere *) _repere;
- (void) tryLocalCoordinates: obj;
- (void) tryGlobalCoordinates: obj;

- (void) addMomentum: (Vect3D *)v
	  resultante: (Vect3D *)w;
- (void) addResultante: (Vect3D *)v;
@end

//extern NSString *GGLaserItem;
//extern NSString *

extern NSString *GGSwitchToLocal;
extern NSString *GGSwitchToGlobal;
extern NSString *GGNotificationShipLeaveCurrentSystem;
extern NSString *GGNotificationShipEnterCurrentSystem;
extern NSString *GGNotificationShipBecomeActivated;
extern NSString *GGNotificationShipBecomeInactivated;

NSString *shipMsgFromErrorCode(ShipErrorCode code);

#endif
