/*
 *  GGSky.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSData.h>
#include <Foundation/NSNotification.h>

#include "Metaobj.h"
#include "GGSky.h"
#include "GGInput.h"
#include "GG3D.h"
#include "GGSpaceObject.h"
#include "utile.h"
#include "Preference.h"
#include "GGRepere.h"
#include "Texture.h"
#include "World.h"
#include "GGGlyphLayout.h"
#include "Tableau.h"
#include "GGFont.h"

static double setColor(double mag)
{
  double val;
  if( mag < 0.1 )
    val = 1.1-mag;
  else
    {
      val = (2/mag);
      val = (val < 0 ? 0 : val);
      if( val >= 0.6 )
	val =  sqrt(val);
    }
  //  return val;
  return (val > 1 ? 1 : val);
}


void parseAStar(const char *chaine, Star *pStar, NSMutableData *bigString)
{
  static int line = 0;
  const char *tmp;
  char nom[50];
  int val;
  double x, y, z;
  double mag;
  chaine = strchr(chaine, ':');
  chaine++;
  tmp = strchr(chaine, ':');

  NSCParameterAssert(tmp-chaine < 50);

  strncpy(nom, chaine, tmp-chaine); /*FIXME:too bad if tmp-chaine > 50 */
  nom[tmp-chaine] = '\0';
  pStar->index = [bigString length];
  [bigString appendBytes: nom
	     length: tmp-chaine+1];
  pStar->stringLayout = nil;
  chaine = tmp+1;
  if( (val=sscanf(chaine, "%le:%le:%le:%le", &x, &y, &z, &mag)) != 4 )
    {
      NSLog(@"erreur de parsing ligne %d val = %d name =%s", line+1, val, nom);
      abort();
    }
  pStar->xd = x;
  pStar->yd = y;
  pStar->zd = z;
  pStar->mag0 = mag;
  
  line++;
}


@implementation GGSky
- initAtOrigin: (Vect3D) origin
{
  int i;
  char chaine[200];
  FILE *input;
  [super init];
  isCompile = NO;
  input = fopen("data/model/stars.gg", "r");
  NSAssert(input != NULL, @"can't find stars");

  fontStar = RETAIN([GGFont fontWithFile: DefaultFontName
			   atSize: 12]);
  bigString = [NSMutableData new];
  for ( i = 0; i < SIZEMAX && !feof(input) ; ++i )
    {
      fgets(chaine, 200, input);
      parseAStar(chaine, &(theStars[i]), bigString);
      theStars[i].indexStar = i;
    }
  NSAssert1(i < SIZEMAX, @"trop de parsing %d", i);
  nStars = i;
  addToCheck(self);
  [self genereSkyAt: origin];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(smallTextureFlagChanged)
    name: @"SmallTextureFlagChanged"
    object: nil];

  return self;
}

- (void) dealloc
{
  int i;
  [[NSNotificationCenter defaultCenter]
    removeObserver: nil 
    name: nil 
    object: self];
  RELEASE(bigString);
  RELEASE(fontStar);
  RELEASE(dev);
  RELEASE(der);
  RELEASE(ba);
  RELEASE(ha);
  RELEASE(ga);
  RELEASE(dr);
  for( i = 0; i < nStars; ++i)
    {
      RELEASE(theStars[i].stringLayout);
    }
  [super dealloc];
}

- (void) smallTextureFlagChanged
{
  DESTROY(dev);
  DESTROY(der);
  DESTROY(ba);
  DESTROY(ha);
  DESTROY(ga);
  DESTROY(dr);
  //FIXME
  [self genereSkyAt: GGZeroVect];
}

- (void) genereSkyForIndex: (int) index
{
  Vect3D pos = {theStars[index].xd, theStars[index].yd, theStars[index].zd};
  [self genereSkyAt: pos];
}

typedef unsigned char starFace[SIZETEXTURESKY];

static void setZero(NSZone *zone, starFace **pface)
{
  *pface = NSZoneCalloc(zone, SIZETEXTURESKY, sizeof(starFace));
  NSCParameterAssert(*pface != NULL);
}

- (void) genereSkyAt: (Vect3D) origin
{
  starFace	* devant;
  starFace	*derriere;
  starFace	*droite;
  starFace	*gauche;
  starFace	*haut;
  starFace	*bas;
  int sizeTextureSky;
  int i;
  NSZone	*zone = [self zone];

  setZero(zone, &devant);
  setZero(zone, &derriere);
  setZero(zone, &droite);
  setZero(zone, &gauche);
  setZero(zone, &haut);
  setZero(zone, &bas);

  if( thePref.SmallTexture )
    {
      sizeTextureSky = 256;
    }
  else
    sizeTextureSky = 512;

  for( i = 0; i < nStars; ++i )
    {
      Star 	s 	= theStars[i];
      double 	norm 	= sqrt(square(s.xd-origin.x)+square(s.yd-origin.y) +
			       square(s.zd-origin.z));
      double	distSol = sqrt(square(s.xd)+square(s.yd) + square(s.zd));
      double 	x, y, z;
      int 	xi, yi;

      if( norm <= 1 )
	{
	  theStars[i].intensity = 0.0;
	  DESTROY(theStars[i].stringLayout);
	  continue;
	}

      x 	= (s.xd-origin.x) / norm;
      y 	= (s.yd-origin.y) / norm;
      z 	= (s.zd-origin.z) / norm;

      initVect(&theStars[i].pos, x, y, z);
      theStars[i].intensity = setColor(theStars[i].mag0 + 
				       5*log10(distSol/norm));

      if( theStars[i].intensity > 0.7 )
	{
	  ASSIGN(theStars[i].stringLayout,
		 [GGGlyphLayout glyphLayoutWithString:
				  [NSString stringWithCString: [self cNameOf: i]]
				withFont: fontStar]);
	}
      else
	DESTROY(theStars[i].stringLayout);

      NSParameterAssert(theStars[i].intensity <= 1.0);
//        {
//  	unsigned char c = theStars[i].intensity*255;
//  	NSLog(@"%@ %d %g", theStars[i].name, (int)c, theStars[i].intensity);
//        }

      if( fabs(x) > fabs(y) )
	{
	  if( fabs(x) > fabs(z) )
	    {
	      if( x > 0 )
		{
		  xi = (y / x + 1)/2*(sizeTextureSky-1);
		  yi = (z / x + 1)/2*(sizeTextureSky-1);
		  devant[yi][xi] = theStars[i].intensity*255;
		}
	      else
		{
		  xi = (y / x + 1)/2*(sizeTextureSky-1);
		  yi = (z / x + 1)/2*(sizeTextureSky-1);
		  derriere[yi][xi] = theStars[i].intensity*255;
		}
	    }
	  else
	    {
	      if( z > 0 )
		{
		  xi = (x / z + 1)/2*(sizeTextureSky-1);
		  yi = (y / z + 1)/2*(sizeTextureSky-1);
		  haut[yi][xi] = theStars[i].intensity*255;
		}
	      else
		{
		  xi = (x / z + 1)/2*(sizeTextureSky-1);
		  yi = (y / z + 1)/2*(sizeTextureSky-1);
		  bas[yi][xi] = theStars[i].intensity*255;
		}

	    }
	}
      else
	{
	  if( fabs(y) > fabs(z) )
	    {
	      if( y > 0 )
		{
		  xi = (x / y + 1)/2*(sizeTextureSky-1);
		  yi = (z / y + 1)/2*(sizeTextureSky-1);
		  gauche[yi][xi] = theStars[i].intensity*255;
		}
	      else
		{
		  xi = (x / y + 1)/2*(sizeTextureSky-1);
		  yi = (z / y + 1)/2*(sizeTextureSky-1);
		  droite[yi][xi] = theStars[i].intensity*255;
		}
	    }
	  else
	    {
	      if( z > 0 )
		{
		  xi = (x / z + 1)/2*(sizeTextureSky-1);
		  yi = (y / z + 1)/2*(sizeTextureSky-1);
		  haut[yi][xi] = theStars[i].intensity*255;
		}
	      else
		{
		  xi = (x / z + 1)/2*(sizeTextureSky-1);
		  yi = (y / z + 1)/2*(sizeTextureSky-1);
		  bas[yi][xi] = theStars[i].intensity*255;
		}

	    }

	}
    }

  if( dev != nil )
    {
      [dev changeStarField: &(devant[0][0])
	   width: sizeTextureSky
	   height: sizeTextureSky];
      [der changeStarField: &(derriere[0][0])
	   width: sizeTextureSky
	   height: sizeTextureSky];
      [ha changeStarField: &(haut[0][0])
	   width: sizeTextureSky
	   height: sizeTextureSky];
      [ba changeStarField: &(bas[0][0])
	   width: sizeTextureSky
	   height: sizeTextureSky];
      [dr changeStarField: &(droite[0][0])
	   width: sizeTextureSky
	   height: sizeTextureSky];
      [ga changeStarField: &(gauche[0][0])
	   width: sizeTextureSky
	   height: sizeTextureSky];
    }
  else
    {

      dev = [[Texture alloc] initStartField: &(devant[0][0])
			     width: sizeTextureSky
			     height: sizeTextureSky];
      der = [[Texture alloc] initStartField: &(derriere[0][0])
			     width: sizeTextureSky
			     height: sizeTextureSky];
      ha = [[Texture alloc] initStartField: &(haut[0][0])
			    width: sizeTextureSky
			    height: sizeTextureSky];
      ba = [[Texture alloc] initStartField: &(bas[0][0])
			    width: sizeTextureSky
			    height: sizeTextureSky];
      dr = [[Texture alloc] initStartField: &(droite[0][0])
			    width: sizeTextureSky
			    height: sizeTextureSky];
      ga = [[Texture alloc] initStartField: &(gauche[0][0])
			    width: sizeTextureSky
			    height: sizeTextureSky];
    }
  NSZoneFree(zone, devant);
  NSZoneFree(zone, derriere);
  NSZoneFree(zone, haut);
  NSZoneFree(zone, bas);
  NSZoneFree(zone, droite);
  NSZoneFree(zone, gauche);
}

- (void) drawSky: (Camera *) cam
{
  int i;
  if( thePref.stars )
    {
      if( isCompile )
	{
	  glCallList(idList);
	}
      else
	{
	  idList = glGenLists(1);
	  NSDebugMLLog(@"gl", @"idList = %u", idList);
	  glNewList(idList, GL_COMPILE_AND_EXECUTE);

	  glDepthMask(GL_TRUE);
	  glDepthRange(1, 1);
	  glDepthFunc(GL_ALWAYS);
	  glEnable(GL_DEPTH_TEST);
      

	  glDisable(GL_CULL_FACE);
	  glEnable(GL_TEXTURE_2D);
	  glBindTexture(GL_TEXTURE_2D, [dev idTexture]);
	  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
      
	  //	  glColor3d(1, 1, 1);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0, 0);
	  glVertex3f(1, -1, -1);
	  glTexCoord2f(1, 0);
	  glVertex3f(1, 1, -1);
	  glTexCoord2f(1, 1);
	  glVertex3f(1, 1, 1);
	  glTexCoord2f(0, 1);
	  glVertex3f(1, -1, 1);
	  glEnd();

	  glBindTexture(GL_TEXTURE_2D, [der idTexture]);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0, 0);
	  glVertex3f(-1, 1, 1);
	  glTexCoord2f(1, 0);
	  glVertex3f(-1, -1, 1);
	  glTexCoord2f(1, 1);
	  glVertex3f(-1, -1, -1);
	  glTexCoord2f(0, 1);
	  glVertex3f(-1, 1, -1);
	  glEnd();

	  glBindTexture(GL_TEXTURE_2D, [ha idTexture]);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0, 0);
	  glVertex3f(-1, -1, 1);
	  glTexCoord2f(1, 0);
	  glVertex3f(1, -1, 1);
	  glTexCoord2f(1, 1);
	  glVertex3f(1, 1, 1);
	  glTexCoord2f(0, 1);
	  glVertex3f(-1, 1, 1);
	  glEnd();

	  glBindTexture(GL_TEXTURE_2D, [ba idTexture]);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0, 0);
	  glVertex3f(1, 1, -1);
	  glTexCoord2f(1, 0);
	  glVertex3f(-1, 1, -1);
	  glTexCoord2f(1, 1);
	  glVertex3f(-1, -1, -1);
	  glTexCoord2f(0, 1);
	  glVertex3f(1, -1, -1);
	  glEnd();

	  glBindTexture(GL_TEXTURE_2D, [ga idTexture]);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0, 0);
	  glVertex3f(-1, 1, -1);
	  glTexCoord2f(1, 0);
	  glVertex3f(1, 1, -1);
	  glTexCoord2f(1, 1);
	  glVertex3f(1, 1, 1);
	  glTexCoord2f(0, 1);
	  glVertex3f(-1, 1, 1);
	  glEnd();

	  glBindTexture(GL_TEXTURE_2D, [dr idTexture]);
	  glBegin(GL_QUADS);
	  glTexCoord2f(0, 0);
	  glVertex3f(1, -1, 1);
	  glTexCoord2f(1, 0);
	  glVertex3f(-1, -1, 1);
	  glTexCoord2f(1, 1);
	  glVertex3f(-1, -1, -1);
	  glTexCoord2f(0, 1);
	  glVertex3f(1, -1, -1);
	  glEnd();


	  glEnable(GL_CULL_FACE);
	  glDisable(GL_TEXTURE_2D);

	  glDepthRange(0.0, 1.0);
	  glDepthFunc(GL_LESS);
	  glDisable(GL_DEPTH_TEST);
	  glEndList();
	  isCompile = YES;
	}
    }
  else
    {
      glDepthMask(GL_TRUE);
      glClearColor(0, 0, 0, 1.0);
      glClearDepth(1.0);
      glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
      glDepthMask(GL_FALSE);
    }
  if( thePref.nameStar && cam->param->label)
    {
      NSSize size = NSMakeSize(cam->param->width, cam->param->height);
      glColor3f(0.6, 0.6, 0.6);
      
      pushStandardRep(size);
      [fontStar selectFont];
      NSDebugMLLogWindow(@"GGSky", @"mat\n%@", stringOfMatrice(&cam->modelView));
      for( i = 0; i < nStars; ++i)
	{
	  if( theStars[i].stringLayout!= nil )
	    {
	      Vect3D pos = theStars[i].pos;
	      Vect3D res;
	      if( getWinCoordinates(cam, &pos, &res) )
		{
		  res.x += 5;
		  res.y -= 5;
		  glLoadIdentity();
		  glTranslatef(res.x, res.y, 0);
		  [theStars[i].stringLayout fastDraw];
		  NSDebugMLLogWindow(@"GGSky", @"%@\n  %@", 
				     stringOfVect(&pos),
				     stringOfVect(&res));
		}
	    }
	}
      [fontStar endSelectFont];
      popStandardRep();
    }
}

- (Star *) stars: (int*)size
{
  if( size != NULL )
    *size = nStars;
  return theStars;
}

- (char *) names
{
  return (char *)[bigString bytes];
}

- (char *) cNameOf:(int) index
{
  NSParameterAssert(0<= index && index < nStars);
  return (char *)[bigString bytes] + theStars[index].index;
}

- (char *) cNameOfStar: (Star *)s
{
  return [self cNameOf: s->indexStar];
}

@end
