/* 	-*-ObjC-*- */
/*
 *  GGSpaceObject.h
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GGSpaceObject_h
#define __GGSpaceObject_h

#include "GG3D.h"
#include "GGLoop.h"

enum __SpaceObjectMode {
  GGNormalMode = 0x1,
  GGInertialMode = 0x2,
  GGInertialRotationalMode = 0x4
};


#define Galileen(x)(&((x)->repere))
#define Matrice(x) (&((x)->repere.matrice))
#define Speed(x) (&((x)->repere.speed))
#define Omega(x) (&((x)->omega))
#define Point(x) (&((x)->repere.matrice.position))
#define Haut(x) (&((x)->repere.matrice.haut))
#define Gauche(x) (&((x)->repere.matrice.gauche))
#define Direction(x) (&((x)->repere.matrice.direction))
#define Dimension(x) ((x)->dimension)
#define Masse(x) ((x)->masse)
#define CollisionCounter(x) ((x)->collisionCounter)


#define IdentityMatrix 2
#define TranslationMatrix 1
#define GeneralMatrix 0

@class GGRepere;
@class GGSpaceObject;
@class GGShip;
@class GGNode;
@class GGGlyphLayout;

@protocol IA
- (PetitGalileen *) positionRelativeCible;
- (void) setShip: (GGShip *) aShip;
- (void) setTarget: (GGSpaceObject *)aTarget;
- (double) estimatedDateAtDestination;
@end

@protocol WorldManager
- (GGSpaceObject *)camera;
- (GGRepere *)repere;
@end

@protocol GGNode
- (GGNode *) nodePere;
- (void) setNodePere: (GGSpaceObject *) obj;
- (BOOL) isParentOf: (GGSpaceObject *) obj;
- (void) enregistreForRemoving: dest
			  with: (SEL) s;
- (void) enregistre: obj
	       with: (SEL) s;
- (void) drawWithCamera: (Camera *)cam;
- (void) changeLocalCoordinates: (GGGalileen *)gal;
- (NSString *) displayObject: (int) i;
@end


@interface GGSpaceObject : NSObject <GGNode, NSCopying, NSCoding>
{
  @public
  GGSpaceObject		*nodePere;
  NSMutableArray	*filsNode;
  id<WorldManager>	manager;
  GGGalileen		repere;
  Vect3D		omega;		//rotational speed in local coordinates.
  real			distanceLocal2, distanceGlobal2;
  real			dimension;
  float			masse;
  int			collisionCounter;

  int           	mode;
  GLuint		compileListId;
  NSString		*defaultName;
  GGGlyphLayout		*cacheNameRep;
  struct {
    unsigned 			isCamera:1;
    unsigned 			matrixType:2;
    unsigned			visible:1;
    unsigned			alive:1;
    unsigned			shouldCache:1;
    unsigned			cacheCorrect:1;
    unsigned			clipable:1;
    unsigned 			majRepere:1;
    unsigned			moreThingsToDraw:1;
    unsigned			stateChange:1;
    unsigned			locality:2;  /*0 outside system
					       1 in system but outside local 
					         frame
					       2 in local frame
					     */  
    unsigned			activated:1; /* 1 when the object shoould update
						its position frequently
						because being observed*/
    unsigned			isFilsEmpty;
    unsigned			graphical;  /* mean that rasterize need to
					       be called*/
  }_flags;
}
- initAt: (GGSpaceObject *)obj;
+ spaceObjectAt: (GGSpaceObject *) obj;
- (GGNode *) nodePere;
- (void) setNodePere: (GGSpaceObject *)obj;
- (void) setManager: (id<WorldManager>) newManager;
- (id<WorldManager>) manager;
- (void) removeFromSuperNode;
- (GGSpaceObject *)ancestor;
- (BOOL) isParentOf: (GGSpaceObject *)obj;
- (BOOL) shouldSave;
- (BOOL) saveChannel;

- (BOOL) shouldBeClipped;
- (BOOL) shouldCache;
- (BOOL) shouldDrawMoreThings;
- (BOOL) isLocal;
- (unsigned) locality;
- (void) setLocality: (unsigned)val;
- (void) invalidCache;

- (NSString *) signature;
- objectFromSignature;

- (void) setDefaultName: (NSString *)aName;
- (void) setName: (NSString *)aName;

/* Les mthodes suivantes n'ont pas besoin d'tre redfinie dans des 
 * sous-classes.
 */
- (void) setVisible: (BOOL) val;
- (void) setMode: (int) val;
- (void) setCameraFlag: (BOOL) val;

- (void) enregistre: dest
	       with: (SEL) s;
- (void) enregistreForRemoving: dest
			  with: (SEL) s;
- (void) desenregistre: dest;
- (void) reallyDestroy;
- (void) destroy;

/* call before the object is about to be removed from the current system
   It will post a Notification
*/
- (void) looseMainSystem;


/*mthodes qui font joujou avec la gomtrie locale de l'objet
 */
- (real) proximite;
- (Vect3D *) getDirection;
- (Vect3D *) getHaut;
- (Vect3D *) getSpeed;
- (void) getGalileen:(GGGalileen *) ret;
//- (void) setGlobalPosition: (Vect3D *) pv;
- (void) setPosition:(Vect3D *) pv;
- (GGMatrix4 *)getMatrix;
- (void) normalise;
- (void) centerObject;

- (NSString *)getStringPosition;
- (NSString *)name;

- (BOOL) manageCollision;
- (void) chocWithEnergy: (real) val;
- (void) collisionOccursWith: (GGSpaceObject *) other;



/* L, on arrive  des mthodes assez foireuses qui grent le sytme
 * de repre, de matrice, de changement de vue. 
 */


- (void) rotationMatrixTo: (GGSpaceObject *) obj
		       in: (GGMatrix4 *)res;
- (void) galileanTransformationTo: (GGSpaceObject *)obj
			       in: (GGGalileen *)pgal;
- (void) galileanTransformationFromAncestor: (GGSpaceObject *)ancestor
					 in: (GGGalileen *)pgal;
- (void) rotationMatrixFromAncestor: (GGSpaceObject *)ancestor
				 in: (GGMatrix4 *)res;

- (Vect3D) getPositionAbsolu;
- (Vect3D) getPositionSpeedAbsolues: (Vect3D *)pspeed;

/*
 *La smantique de cette mthode est:
 * On calcule la position et la speed  self relativement  obj.
 * Le rsultat est converti dans le systme de coordonne dans lequel obj vie
 * !!!IL NE S'AGIT PAS DES COORDONNEES LOCALES DE OBJ !!!
 * plutt celle du pere de obj.
 */
- (void) positionSpeedRelativesTo: (GGSpaceObject *)obj
			    decal: (Vect3D *)v
			       in: (PetitGalileen *)pgl;

/* Cette mthode calcule la position de self dans repere. A ne pas confondre 
 * avec la mthode prcdente
 */
- (void) positionInRepere: (GGSpaceObject *)repere
		       in: (Vect3D *)pv;

/*Fais la meme chose que positionSpeedRelativesTo:in: mais le resultat
  est mesur dans les cooordonnes de repere.  Elle est plus complete que la
  mthode prcdente :
*/
- (void) positionSpeedInRepere: (GGSpaceObject *)repere
			    in: (PetitGalileen *)pgl;

//- (void) multLocalCoordinate;

/* this method is called mostly by nodes to change the coordinates of the 
   object when the frame changes.  If one want to attach an object to the 
   origin of a node (bad idea but it is used in Planete.m), one can override
   this method to do nothing
*/
- (void) changeLocalCoordinates: (GGGalileen *) gal;

/*mthode trs utile qui convertit les coordonnes de self en 
  coordonnes fentre.  Le type de camera est dans param.
  renvoie YES si on screen, NO sinon
*/
- (BOOL) windowsCoordinatesForCamera: (GGSpaceObject *)camera
			   parameter: (ProjectionParam *)param
			      result: (Vect3D *)res;


/* dplace la position courante le long du vecteur speed.
 * La position peut-tre locale o globale.
 */
- (void) updatePosition;
- (void) drawWithCameraNoTransform: (Camera *)cam
			      from: (GGSpaceObject *)from;
- (void) drawWithCamera: (Camera *)cam;
- (void) drawGraphProjection: (ProjectionParam *)ppp;
- (void) rasterise;
- (void) rasterise: (Camera *)cam;


/*Fonctions diverses et varies.
 */




/*Should be call after all the initialisation of the objet
 */
- (void) enterUniverse;

/*Should be call after all the initialisation of the objet, in
 *order it can organise it's own scheduling.  
 * For a ship, it when it will be part of the current system.
 */
- (void) beginSchedule;

/* Should be call when an object should be not schedule (because it is
 *  not in the current system any more, ...) all the schedules begun
 *  in [ - beginSchedule] should be ended here 
 */
- (void) endSchedule;



- (NSString *) displayObject:(int) ind;
@end

#define monAddLambdaVect(a, b, c, d) do {	\
  addLambdaVect((a), (b), (c), (d));		\
  majRepere = 1;				\
}while(0)

#define DOLABEL [self drawLabel]

extern NSString *GGNotificationRemoveFromCurrent;
#define GGEncode(obj) [encoder encodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]
#define GGDecode(obj) [decoder decodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]

#endif
