/*
 *  GGSpaceObject.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSCoder.h>
#include "GGLoop.h"
#include "GGSpaceObject.h"
#include "GGInput.h"
#include "GGRepere.h"
#include "Metaobj.h"
#include "utile.h"
#include "Tableau.h"
#include "Preference.h"
#include "Sched.h"
#include "World.h"
#include "GGNode.h"
#include "GGFont.h"
#include "Commandant.h"

#include <math.h>
#include <time.h>
//test

NSString *GGNotificationRemoveFromCurrent = 
@"GGNotificationRemoveFromCurrent";



@implementation GGSpaceObject
- init
{
  [super init];
  nodePere = nil;
  _flags.isCamera = NO;
  _flags.matrixType = GeneralMatrix;
  _flags.majRepere = NO;
  _flags.visible = 0;
  _flags.graphical = NO;
  _flags.locality = 0;
  [self centerObject];
  dimension = 1;
  _flags.alive = 1;
  masse = 1;
  collisionCounter = 0;
  _flags.shouldCache = [self shouldCache];
  _flags.clipable = [self shouldBeClipped];
  _flags.moreThingsToDraw = [self shouldDrawMoreThings];
  _flags.cacheCorrect = NO;

  _flags.isFilsEmpty = YES;

  mode = 0;
  addToCheck(self);
  return self;
}

- initAt: (GGSpaceObject *)obj
{
  [self init];
  *Galileen(self) = *Galileen(obj);
  return self;
}

+ spaceObjectAt: (GGSpaceObject *) obj
{
  GGSpaceObject *newObj;
  newObj = [self alloc];
  return AUTORELEASE([newObj initAt: obj]);
}

- (id) copyWithZone: (NSZone *)zone
{
  GGSpaceObject *copy;
  int n;
  copy = (GGSpaceObject *)NSCopyObject(self, 0, zone);
  copy->manager = nil;
  copy->nodePere = nil;
  copy->_flags.cacheCorrect = NO;
  copy->defaultName = [defaultName copy];
  copy->cacheNameRep = nil;
  copy->collisionCounter = 0;
  copy->mode = 0;
  copy->_flags.isCamera = NO;
  copy->_flags.matrixType = GeneralMatrix;
  copy->_flags.visible = 0;
  copy->_flags.alive = 1;
  copy->_flags.shouldCache = [copy shouldCache];
  copy->_flags.clipable = [copy shouldBeClipped];
  copy->filsNode = nil;
  copy->_flags.isFilsEmpty = YES;

  n = [filsNode count];
  if( n > 0 )
    {
      GGSpaceObject *objs[n];
      int i;

      [filsNode getObjects: objs];

      for(i=0; i<n; ++i)
	{
	  objs[i] = [objs[i] copyWithZone: zone];
	  [copy addSubNode: objs[i]];
	  RELEASE(objs[i]);
	}
    }
  
  return copy;
}

- (void)encodeWithCoder:(NSCoder *)encoder
{

  unsigned tmp;
#define GGEncode(obj) [encoder encodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]
#define GGEncodeFlag(f) tmp = (unsigned int)(_flags.f); GGEncode(tmp)

  //  [super encodeWithCoder: encoder];
  NSDebugMLLog(@"Coding", @"%@ son of %@", self, nodePere);
  GGEncode(nodePere);
  GGEncode(manager);

  //  [encoder encodeConditionalObject: nodePere];
  //  [encoder encodeConditionalObject: manager];
  GGEncode(repere);
  GGEncode(omega);
  GGEncode(distanceLocal2);
  GGEncode(distanceGlobal2);
  GGEncode(dimension);
  GGEncode(masse);
  GGEncode(collisionCounter);
  GGEncode(mode);
  GGEncode(defaultName);
  GGEncodeFlag(isCamera);
  GGEncodeFlag(matrixType);
  GGEncodeFlag(visible);
  GGEncodeFlag(alive);
  GGEncodeFlag(shouldCache);
  GGEncodeFlag(clipable);
  GGEncodeFlag(majRepere);
  GGEncodeFlag(moreThingsToDraw);
  GGEncodeFlag(stateChange);
  GGEncodeFlag(locality);
  GGEncodeFlag(activated);
  GGEncodeFlag(graphical);
}

- (id)initWithCoder:(NSCoder *)decoder
{
  unsigned tmp;
  GGNode *pere;
#define GGDecode(obj) [decoder decodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]
#define GGDecodeFlag(f) GGDecode(tmp); _flags.f = tmp
 
  pere = [decoder decodeObject];
  manager = [decoder decodeObject];

  filsNode = nil;
  _flags.isFilsEmpty = YES;

  GGDecode(repere);
  GGDecode(omega);
  GGDecode(distanceLocal2);
  GGDecode(distanceGlobal2);
  GGDecode(dimension);
  GGDecode(masse);
  GGDecode(collisionCounter);
  GGDecode(mode);
  GGDecode(defaultName);
  GGDecodeFlag(isCamera);
  GGDecodeFlag(matrixType);
  GGDecodeFlag(visible);
  GGDecodeFlag(alive);
  GGDecodeFlag(shouldCache);
  GGDecodeFlag(clipable);
  GGDecodeFlag(majRepere);
  GGDecodeFlag(moreThingsToDraw);
  GGDecodeFlag(stateChange);
  GGDecodeFlag(locality);
  GGDecodeFlag(activated);
  GGDecodeFlag(graphical);
  _flags.cacheCorrect = 0;
  if( pere != nil )
    {
      NSDebugMLLog(@"Coding", @"%@ wants %@", self, pere);
      [pere addSubNode: self];
    }
  return self;
}


- signature
{
  return nil;
}

- objectFromSignature
{
  return self;
}

- (BOOL) shouldCache
{
  return YES;
}

- (BOOL) shouldBeClipped
{
  return YES;
}

- (BOOL) shouldDrawMoreThings
{
  return NO;
}

- (BOOL) shouldSave
{
  return YES;
}

- (BOOL) saveChannel
{
  return YES;
}

- (BOOL) isLocal
{
  return _flags.locality == 2;
}

- (unsigned) locality
{	
  return _flags.locality;
}

- (void) setLocality: (unsigned)val
{
  _flags.locality = val;
}

- (void) invalidCache
{
  if( _flags.cacheCorrect )
    {
      _flags.cacheCorrect = NO;
      glDeleteLists(compileListId, 1);
    }
  _flags.shouldCache = [self shouldCache];
  _flags.clipable = [self shouldBeClipped];
  _flags.moreThingsToDraw = [self shouldDrawMoreThings];
}

- (GGSpaceObject *) nodePere
{
  return nodePere;
}

- (void) setNodePere: (GGSpaceObject *)obj
{
  if( obj != nil )
    {
      NSAssert( nodePere == nil, NSInternalInconsistencyException);
      //      [self beginSchedule];
      [self setManager: [obj manager]];
    }
  else
    {
      NSAssert( nodePere != nil, NSInternalInconsistencyException);
      //      [self endSchedule];
      [self setManager: nil];
    }
  nodePere = obj;
}

- (void) setManager: (id<WorldManager>) newManager
{
  manager = newManager;
  [filsNode makeObjectsPerformSelector: @selector(setManager:)
	    withObject: newManager];
}


- (id<WorldManager>) manager
{
  return manager;
}

- (void) removeFromSuperNode
{
  //  NSAssert(nodePere != nil, NSInternalInconsistencyException);
  if( nodePere != nil )
    [nodePere removeSubNode: self];
  else
    NSWarnMLog(@"%@ has no superNode", self);
}

- (void) setDefaultName: (NSString *)aName
{
  ASSIGN(defaultName, aName);
  DESTROY(cacheNameRep);
}

- (void) setName: (NSString *)val
{
  [self setDefaultName: val];
}

- (NSString *)description
{
  NSString *tmp;
  if( defaultName != nil )
    {
#ifdef DEBUG
      if( self->_flags.activated )
	return [NSString stringWithFormat: @"@r%@@w", defaultName];
      else
#endif
	return defaultName;
    }
  else if( (tmp = [self name]) != nil )
    return tmp;
  else
    return [super description];
}

- (void) centerObject
{
  initWithIdentityTransfo(Galileen(self));
  SetZeroVect(omega);
}

- (unsigned char) matrixType
{
  return _flags.matrixType;
}

- (BOOL) isParentOf: (GGSpaceObject *)obj
{
  while(obj != nil)
    {
      if( obj == self )
	return YES;
      obj = obj->nodePere;
    }
  return NO;
}

- (void) setMode: (int) val
{
  int oldmode;
  oldmode = mode;
  mode = val;
  if( nodePere == nil )
    return;

  if( !oldmode && mode )
    {
      NSDebugMLLog(@"ModeInertie", @"start sched  for %@",
		   self);
      [theLoop fastSchedule: self
	       withSelector: @selector(updatePosition)
	       arg: nil];
    }
  else if( oldmode && !mode )
    {
      NSDebugMLLog(@"ModeInertie", @"STOP sched for %@",
		   self);
      [[theLoop fastScheduler]   deleteTarget: self
				 withSelector: @selector(updatePosition)];
    }
}

- (void) setVisible: (BOOL) val
{
  _flags.visible = val;
}

- (void) setCameraFlag: (BOOL) val
{
  _flags.isCamera = val;
}

- (Vect3D *) getDirection
{
  return Direction(self);
}

- (Vect3D *) getHaut
{
  return Haut(self);
}

- (Vect3D *) getSpeed
{
  return Speed(self);
}

- (void) getGalileen:(GGGalileen *)ret
{
  *ret = *Galileen(self);
}

- (real) proximite
{
  return 10*dimension;
}

- (GGSpaceObject *)ancestor
{
  if( nodePere == nil )
    return self;
  else
    return [nodePere ancestor];
}

- (Vect3D) getPositionAbsolu
{
  //SPEEDUP : too slow
  GGGalileen gl;
  [self galileanTransformationTo: [self ancestor] in: &gl];
  return gl.matrice.position;
}

- (Vect3D) getPositionSpeedAbsolues: (Vect3D *)pspeed
{
  //SPEEDUP : too slow
  GGGalileen gl;
  [self galileanTransformationTo: [self ancestor] in: &gl];

  *pspeed = gl.speed;
  return gl.matrice.position;
}

- (void) positionInRepere: (GGSpaceObject *)_repere
		       in: (Vect3D *)pv
{
  //SPEEDUP : speed up this
  GGGalileen gal;
  [self galileanTransformationTo: _repere
	in: &gal];
  *pv = gal.matrice.position;
}

- (void) positionSpeedRelativesTo: (GGSpaceObject *)obj
			    decal: (Vect3D *)decal
			       in: (PetitGalileen *)pgl
{
  //SPEEDUP : speed up this
  GGGalileen gal;
  PetitGalileen pg, pgres;
  NSAssert(nodePere != nil, NSInternalInconsistencyException);
  [nodePere galileanTransformationTo: [obj nodePere]
	       in: &gal];
  NSDebugMLLogWindow(@"Galileen", @"Matrice\n%@Speed : %@", 
		     stringOfMatrice(&gal.matrice), 
		     stringOfVect(&gal.speed));
  pg.speed = *Speed(self);
  if( decal )
    {
      addVect(decal, Point(self), &pg.point);
    }
  else
    pg.point = *Point(self);

  transformeGalileen(&gal, &pg, &pgres);
  
  diffVect(&pgres.point, Point(obj), &pgl->point);
  diffVect(&pgres.speed, Speed(obj), &pgl->speed);
}

- (void) positionSpeedInRepere: (GGSpaceObject *)_repere
			    in: (PetitGalileen *)pgl
{
  GGGalileen gal;
  [self galileanTransformationTo: _repere
	in: &gal];
  pgl->point = gal.matrice.position;
  pgl->speed = gal.speed;
}


- (void) galileanTransformationTo: (GGSpaceObject *)obj
			       in: (GGGalileen *)pgal
{
  if( [self isParentOf: obj] )
    [obj galileanTransformationFromAncestor: self
	 in: pgal];

  else
    {
      GGGalileen gal;
      NSAssert2(nodePere != nil, @"nodePere == nil, self = %@ obj = %@",
		self, obj);
      [nodePere galileanTransformationTo: obj
		in: &gal];
      produitGalileen(&gal, Galileen(self), pgal);
    }
}

- (void) galileanTransformationFromAncestor: (GGSpaceObject *)ancestor
					 in: (GGGalileen *)pgal
{
  if( ancestor == self )
    initWithIdentityTransfo(pgal);
  else
    {
      GGGalileen gal;
      NSAssert(nodePere != nil, @"nodePere == nil");
      [nodePere galileanTransformationFromAncestor: ancestor
		in: &gal];
      transposeProduitGalileen(Galileen(self), &gal, pgal);
      NSDebugMLLogWindow(@"Galileen", @"%@ %@", self, nodePere);
      NSDebugMLLogWindow(@"Galileen", @"sub\n%@\nself\n%@\nresult\n%@",
			 stringOfMatrice(&gal.matrice),
			 stringOfMatrice(Matrice(self)),
			 stringOfMatrice(&pgal->matrice));
    }
}


- (void) rotationMatrixFromAncestor: (GGSpaceObject *)ancestor
				 in: (GGMatrix4 *)res
{
  if( ancestor == self )
    initWithIdentityMatrix(res);
  else
    {
      if( _flags.matrixType == GeneralMatrix )
	{
	  GGMatrix4 mat;
	  [nodePere rotationMatrixFromAncestor: ancestor
		    in: &mat];
	  transposeProduit3D(Matrice(self), &mat, res);
	}
      else
	[nodePere rotationMatrixFromAncestor: ancestor
		  in: res];
    }
}

- (void) rotationMatrixTo: (GGSpaceObject *) obj
		       in: (GGMatrix4 *)res
{
  if( [self isParentOf: obj] )
    {
      [obj rotationMatrixFromAncestor: self
	   in: res];
    }
  else
    {
      if( _flags.matrixType == GeneralMatrix )
	{
	  GGMatrix4 mat;
	  NSAssert(nodePere != nil, @"nodePere == nil");
	  [nodePere rotationMatrixTo: obj
		    in: &mat];
	  produitMatrice3D(&mat, Matrice(self), res);
	}
      else
	[nodePere rotationMatrixTo: obj
		  in: res];
    }
}

- (void) setPosition:(Vect3D *) pv
{
  *Point(self) = *pv;
}

- (GGMatrix4 *)getMatrix
{
  return Matrice(self);
}


- (void) dealloc
{
  NSDebugMLLog(@"Alloc", @"space object %@ released!!!", self);
  [self removeAllSons];
  RELEASE(filsNode);

  RELEASE(defaultName);
  RELEASE(cacheNameRep);
  [MessageManager
    removeObserver: nil 
    name: nil 
    object: self];
  [MessageManager removeObserver: self];
  [theLoop unSchedule: self];
  [self invalidCache];
  
  [super dealloc];
}

- (void) enregistre: dest
	       with: (SEL) s
{
  [MessageManager addObserver: dest 
		  selector: s 
		  name: @"Destruction"
		  object: self];
  NSDebugMLLog(@"Notification", @"%@ surveille %@ avec %@", dest, self,
	       NSStringFromSelector(s));
}

- (void) enregistreForRemoving: dest
			  with: (SEL) s
{
  [MessageManager addObserver: dest 
		  selector: s 
		  name: GGNotificationRemoveFromCurrent
		  object: self];
  NSDebugMLLog(@"Notification", @"%@ surveille %@ avec %@", dest, self,
	       NSStringFromSelector(s));
}


- (void) desenregistre: dest
{
  //    [[NSNotificationCenter defaultCenter] removeObserver: dest	
  //  					name:@"Destruction"	
  //  					object: self];
}

- (void) looseMainSystem
{
  _flags.locality = 0;
  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGNotificationRemoveFromCurrent
    object: self];
}


- (void) reallyDestroy
{
  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGNotificationRemoveFromCurrent 
    object: self];
  [[NSNotificationCenter defaultCenter]
    postNotificationName: @"Destruction" 
    object: self];
  [theLoop unSchedule: self];
}

- (void) destroy
{
  _flags.alive = 0;
  [theWorld addToKillPool: self];
  [theLoop unSchedule: self];
}

- (void) normalise
{
  normaliseMatrice(Matrice(self));
}

- (void) updatePosition
{
  if( mode & GGInertialMode )
    {
      addLambdaVect(Point(self), deltaTime, Speed(self), Point(self));
    }
  if( (mode & GGInertialRotationalMode) /*&& 
					  prodScal(Omega(self), Omega(self)) > 1e-6*/ )
    {
      real fact;
      Vect3D temp;

      fact = (self == theCommandant ? realDeltaTime : deltaTime);
      _flags.majRepere = YES;

      // Direction is moving
      mulScalVect(Gauche(self), Omega(self)->y, &temp);
      addLambdaVect(&temp, -Omega(self)->x, Haut(self), &temp);
      addLambdaVect(Direction(self), fact, &temp, Direction(self));

      // Haut

      mulScalVect(Gauche(self), -Omega(self)->z, &temp);
      addLambdaVect(&temp, Omega(self)->x, Direction(self), &temp);
      addLambdaVect(Haut(self), fact, &temp, Haut(self));

    }
  if( _flags.majRepere )
    {
      [self normalise];
      _flags.majRepere = 0;
    }
  if( collisionCounter > 0 )
    collisionCounter--;
}

- (NSString *)name
{
  return defaultName;
}

- (void) rasterise
{
}

- (void) rasterise: (Camera *)cam
{
  [self rasterise];
}

- (void) rasteriseMoreThings
{
  [self subclassResponsibility: _cmd];
}

static inline BOOL isClipped(real fov, real x, real z, real dimension)
{
  real om2 = x*x + z*z;
  real v2 = 1+fov*fov;
  real fx = fabs(x);
  return fx >= fov*z && v2*om2 - square(fx*fov+z) >= v2*square(dimension);
}

- (void) drawWithCameraNoTransform: (Camera *)cam
			      from: (GGSpaceObject *)from
{
  Vect3D posInCam;
  if( !_flags.visible ) return;

//   NSDebugMLLogWindow(@"Laser", @"%@ empty %s", self,
// 		     (_flags.isFilsEmpty ? "YES" : "NO"));

  /*test if object is visible*/

  if( cam != NULL && _flags.clipable && _flags.isFilsEmpty )
    {

      //      transformePoint4D(&cam->modelView, &GGZeroVect, &posInCam);
      posInCam = cam->modelView.position;  
      if( posInCam.z < EPS || 
	  isClipped(cam->param->xFoc, posInCam.x, posInCam.z, dimension) ||
	  isClipped(cam->param->yFoc, posInCam.y, posInCam.z, dimension) )
	{
	  NSDebugMLLogWindow(@"Clip", @"%@ is clipped", self);
	  return;
	}
      else
	NSDebugMLLogWindow(@"Clip", @"%@ is not clipped", self);
    }

  if( _flags.graphical )
    {

      if( _flags.shouldCache )
	{
	  if( _flags.cacheCorrect )
	    glCallList(compileListId);
	  else
	    {
	      compileListId = glGenLists(1);
	      NSDebugMLLog(@"gl", @"idList for %@ = %u", self, compileListId);
	      glNewList(compileListId, GL_COMPILE_AND_EXECUTE);
	      [self rasterise: cam];
	      glEndList();
	      _flags.cacheCorrect = 1;
	    }
	}
      else  
	[self rasterise: cam];

      if( _flags.moreThingsToDraw )
	[self rasteriseMoreThings];

      {
	Vect3D res;
	if( cam != NULL && cam->param->label && defaultName != nil 
	    && getWinCoordinates(cam, &GGZeroVect, &res ) )
	  {
	    if( cacheNameRep == nil )
	      {
		ASSIGN(cacheNameRep, [GGGlyphLayout 
				       glyphLayoutWithString: [self name]
				       withFont: [GGFont fontWithFile: 
							   DefaultFontName
							 atSize: 14]]);
	      }
	    glPushAttrib(GL_ENABLE_BIT);
	    glDisable(GL_LIGHTING);
	    glDisable(GL_DEPTH_TEST);
	    pushStandardRep(NSMakeSize(cam->param->width, cam->param->height));
	    glColor3f(0.5, 0.9, 0.5);
	    glTranslatef(res.x+5, res.y-5, 0);
	    [cacheNameRep drawAtPoint: NSZeroPoint];
	    popStandardRep();
	    glPopAttrib();
	    NSDebugMLLogWindow(@"label", @"%@ isdrawing", self);
	  }
      }
    }

  if( ! _flags.isFilsEmpty )
    {
      int i;
      int n = [filsNode count];
      GGSpaceObject *objs[n];
      NSAssert(filsNode != nil, NSInternalInconsistencyException);

      [self willDrawSons: cam];
      [filsNode getObjects: objs];

      for( i = 0; i < n; ++i)
	{
	  //optimize a little bit for invisible object.
	  if( objs[i] != from && objs[i]->_flags.visible )
	    [objs[i] drawWithCamera: cam];
	  //	  	  [objs[i] drawWithCamera: NULL];
	}
      [self hasDrawnSons: cam];
    }
}

- (void) drawWithCamera: (Camera *)cam
{
  Camera camFils;
  NSDebugMLLogWindow(@"Matrice", @"%@\n%@", self, 
  		     stringOfMatrice(Matrice(self)));
  NSDebugMLLogWindow(@"MatrixType", @"%@ %d", self, (int)_flags.matrixType);

  if( !_flags.visible || (!_flags.graphical && _flags.isFilsEmpty) ) return;


  glPushMatrix();
  glMultMatrix((real *)Matrice(self));

  produitMatrice(&cam->modelView, Matrice(self), &camFils.modelView);
  camFils.param = cam->param;
  camFils.prev = cam;
#ifdef DEBUG
  camFils.owner = self;
#endif

  [self drawWithCameraNoTransform: &camFils
	from: nil];


  //  if( _flags.matrixType != IdentityMatrix )
  glPopMatrix();
}

- (void) drawGraphProjection: (ProjectionParam *)ppp
{
  unsigned save;
  Camera cam;
  GGMatrix4 qid;
  pmat p = (pmat)&qid;

  NSDebugMLLogWindow(@"MatriceAncestor", @"%@\n%@", self, 
		     stringOfMatrice(Matrice(self)));

  initWithIdentityMatrix(&qid);
  initWithIdentityMatrix(&cam.modelView);

  p[0][0] = -1;
  p[2][2] = -1;

  glLoadMatrix((real *) &qid);
  cam.param = ppp;
#ifdef DEBUG
  cam.owner = self;
#endif
  
  // HACK
  save = _flags.graphical;
  _flags.graphical = NO;
  [self ancestorDraw: &cam
	from: nil];
  _flags.graphical = save;

  glDisable(GL_LIGHTING);
}

- (BOOL) windowsCoordinatesForCamera: (GGSpaceObject *)camera
			   parameter: (ProjectionParam *)param
			      result: (Vect3D *)res
{
  GGGalileen gal;
  Camera p;
  [self galileanTransformationTo: camera
	in: &gal];

  p.param = param;
  p.modelView = gal.matrice;

  return getWinCoordinates(&p, &GGZeroVect, res);
}


- (void) changeLocalCoordinates: (GGGalileen *) pgal
  /* obj a boug dans son repre galilen et on le recentre.
   * gal code le nouveau repre galilen dans l'ancien.
   */
{
  GGGalileen res;

  produitGalileen(pgal, Galileen(self), &res);
  *Galileen(self) = res;
}

- (NSString *)getStringPosition
{
  return stringOfVect(Point(self));
}

- (void) beginSchedule
{
}

- (void) endSchedule
{
}

- (BOOL) manageCollision
{
  return NO;
}

- (void) chocWithEnergy: (real) val
{
}

- (void) collisionOccursWith: (GGSpaceObject *) other
{
  Vect3D diffPos, diffVit;
  Vect3D uniteX;
  real x, a, c, norme;
  float masseOther = Masse(other);

  if( CollisionCounter(self) || CollisionCounter(other) )
    return;

  diffVect(Point(self), Point(other), &diffPos);
  diffVect(Speed(self), Speed(other), &diffVit);

  uniteX = diffPos;
  norme = prodScal(&uniteX, &uniteX);
  if( norme < EPS )
    {
      Vect3D temp = {1, 0, 0};
      uniteX = temp;
    }
  else
    {
      norme = sqrt(norme);
      divScalVect(&uniteX, norme, &uniteX);
    }

  NSDebugMLLog(@"Collision2", @"uniteX : %@", stringOfVect2(uniteX));

  x = prodScal(&diffVit, &uniteX);

  a = 2*x*masse/(masse + masseOther);
  c = x*(masse-masseOther)/(masseOther+masse);

  [self chocWithEnergy: masse*SQ(c-x)];
  [other chocWithEnergy: masseOther*SQ(a)];

  addLambdaVect(Speed(self), (c-x), &uniteX, Speed(self));
  addLambdaVect(Speed(other), a, &uniteX, Speed(other));
  CollisionCounter(self) = 10;
  CollisionCounter(other) = 10;
}

- (void) enterUniverse
{
}

- (NSString *) displayObject:(int) ind
{
  NSMutableString *str = [NSMutableString string];
  int i, n;
  
  for( i = 0; i  < ind; ++i )
    [str appendString: @"---"];

  [str appendFormat: @"%@ %@ %@\n", [self description], 
       stringOfVect(Point(self)), stringOfVect(Speed(self))];

  if( ! _flags.isFilsEmpty )
    {
      NSAssert(filsNode != nil, NSInternalInconsistencyException);
      n = [filsNode count];

      if( n > 0 )
	{
	  GGSpaceObject *objs[n];
	  [filsNode getObjects: objs];
      
	  for( i = 0; i < n; ++i)
	    [str appendString: [objs[i] displayObject: ind + 1]];
	}
    }

  return str;
}


@end

