/*
 *  GGWidget.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>

#include "GGWidget.h"
#include "Metaobj.h"
#include "NSWindow.h"
#include "utile.h"
#include "Texture.h"
#include "GGGlyphLayout.h"
#include "GGFont.h"
#include "GGIcon.h"

float GGBorderColor[3] = BorderColor;


typedef enum __buttonSate
{
  Pushed = 1,
  Released = 0
}ButtonState;

@implementation GGBouton
- init
{
  [super init];
  target = nil;
  action = NULL;
  state = Released;
  str = nil;
  tag = -1;
  type = NSMomentaryPushButton;
  border = YES;
  return self;
}

+ button
{
  return AUTORELEASE([self new]);
}

- (void) setButtonType: (NSButtonType)aType
{
  type = aType;
}

- (void) setTag: (int) obj
{
  tag = obj;
}

- (int) tag
{
  return tag;
}


- (void) setType: (NSButtonType) theType
{
  type = theType;
}

- (void) _computeFrame
{
  BOOL ok = NO;
  NSSize aSize;
  NSRect temp, tempStr;
  float ystart;

  if( border )
    {
      aSize = NSMakeSize(20, 20);
      ystart = 10;
    }
  else
    {
      aSize = NSZeroSize;
      ystart = 0;
    }

  if( strLayout != nil )
    {
      tempStr = [strLayout frameRect];
      aSize.width += tempStr.size.width;
      aSize.height += tempStr.size.height;
      ok = YES;
    }

  if( icon != nil )
    {
      float val;
      temp = [icon frameRect];
      val = temp.size.width + (border ? 10:0);
      if( val > aSize.width )
	aSize.width = val;
      aSize.height += temp.size.height+(strLayout?2:0);
      ok = YES;
    }

  if( strLayout != nil )
    {
      textOrigin = NSMakePoint(aSize.width/2.0 - tempStr.size.width/2.0,   
			       ystart);
      ystart += tempStr.size.height + 2;
    }
  if( icon != nil )
    iconOrigin = NSMakePoint(aSize.width/2.0 - temp.size.width/2.0, 
			     ystart);
      
  if( ok )
    {
      NSDebugMLLog(@"GGBouton", NSStringFromSize(aSize));
      [self setFrameSize: aSize];
    }
}

- (void) setBorder: (BOOL) val
{
  border = val;
  [self _computeFrame];
}

- (void) setIcon: (GGIcon *) _icon
{
  ASSIGN(icon, _icon);
  [self _computeFrame];
}

- (void) setTitle: (NSString *) theStr
	 withFont: (GGFont *) f
{
  GGGlyphLayout *temp;
  ASSIGN(str, theStr);
  temp = [GGGlyphLayout glyphLayoutWithString: str
			withFont: f];
  [temp setAlignement: BottomAlignement];
  ASSIGN(strLayout, temp);
  [self _computeFrame];
}

- (void) setTitle: (NSString *) theStr
{
  [self setTitle: theStr
	withFont: [GGFont defaultButtonFont]];
}

- (void) dealloc
{
  RELEASE(icon);
  RELEASE(str);
  RELEASE(strLayout);
  [super dealloc];
}

- (BOOL) isOpaque
{
  return YES;
}

- (void) drawRect: (NSRect) theRect
{
  NSRect aRect = [self frame];
  if( border || state != Released || highlighted )
    {
      if( highlighted && state == Released )
	glColor3f(0.7, 0.3, 0.3);
      else
	glColor3f(0.8, 0.5, 0.4);
      glPushAttrib(GL_POLYGON_BIT);
      glPolygonMode(GL_FRONT, state == Pushed || highlighted 
		    ? GL_FILL : GL_LINE);

      glRecti(1, 1, aRect.size.width-1, aRect.size.height-1);
      glPopAttrib();
    }

  if( icon )
    {
      glColor4f(1, 1, 1, 0.4);
      [icon drawAtPoint: iconOrigin];
    }

  if( strLayout != nil )
    {
      glColor3f(1, 1, 1);
      [strLayout drawAtPoint: textOrigin];
    }

  [super drawRect: theRect];
}

- (void) invertState
{
  state = (state == Pushed ? Released : Pushed);
}

- (void) mouseDown: (Event *) pev
{
  NSDebugMLLog(@"GGBouton", @"mouse down in GGBouton %@!!", self);
  switch(type)
    {
    case NSMomentaryPushButton:
      state = Pushed;
      break;
    case NSToggleButton:
      [self invertState];
      [target performSelector: action withObject: self];
      break;
    }
}

- (void) mouseUp: (Event *) pev
{
  NSDebugMLLog(@"GGBouton", @"mouse up in GGBouton %@!!", self);
  switch(type)
    {
    case NSMomentaryPushButton:
      state = Released;
      [target performSelector: action withObject: self];
      break;
    case NSToggleButton:
      break;
    }

}

- (void) setTarget: obj
{
  target = obj;
}

- (void) setAction: (SEL) aSel
{
  action = aSel;
}

- (void) setState: (int) theState
{
  state = theState;
}

- (int) state
{
  return state;
}

- (void) viewWillMoveToWindow: (NSWindow *)win
{
  [super viewWillMoveToWindow: win];

  if( _window == nil )
    highlighted = NO;
}

- (void)  wasViewUnderMouse
{
  highlighted = NO;
}

- (void) isViewUnderMouse
{
  highlighted = YES;
}
  

@end

@implementation GGString
- initWithFrame: (NSRect) aFrame
{
  [super initWithFrame: aFrame];
  [self setString: @"aString"];
  return self;
}

- (BOOL) isOpaque
{
  return YES;
}

- (void) dealloc
{
  RELEASE(str);
  RELEASE(strLayout);
  [super dealloc];
}

- (void)setString: (NSString *)aStr
{
  [self setString: aStr
	withFont: [GGFont defaultFont]];
}

- (void) updateString: (NSString *)Astr
{
  NSRect aRect;
  [strLayout updateString: Astr];
  ASSIGN(str, Astr);
  aRect = [strLayout frameRect];
  [self setFrameSize: aRect.size];
}

- (void) setString: (NSString *) aStr
	  withFont: (GGFont *) f
{
  NSRect aRect;
  GGGlyphLayout *temp;
  ASSIGN(str, aStr);
  temp = [GGGlyphLayout glyphLayoutWithString: str
			withFont: f];
  aRect = [temp frameRect];
  ASSIGN(strLayout, temp);
  [self setFrameSize: aRect.size];
}

- (void) drawRect: (NSRect) theRect
{
  if( strLayout != nil )
    {
      NSDebugMLLog(@"GGString", @"rect %@ %@", NSStringFromRect(theRect),
		   str);
      glColor3f(1, 1, 1);
      [strLayout drawAtPoint: NSZeroPoint];
    }
  [super drawRect: theRect];
  
}

- (NSString *)description
{
  if( str == nil )
    return [super description];
  else
    return str;
}
@end


@implementation GGOption
- init
{
  [super init];
  bo = AUTORELEASE([GGBouton new]);
  [bo setFrame: NSMakeRect(0, 2, 15, 15)];
  [bo setType: NSToggleButton];
  [self addSubview: bo];

  str = AUTORELEASE([GGString new]);
  [str setString: @"optionString"
       withFont: [GGFont defaultOptionFont]];
  [str setFrame: NSMakeRect(30, 0, 200, 20)];
  [self addSubview: str];

  [self setFrameSize: NSMakeSize(200, 35)];
  
  return self;
}

- (void) setType: (NSButtonType) aType
{
  [bo setType: aType];
}

- (void) setTitle: (NSString *)theStr
{
  NSRect rect;
  NSSize size;
  [str updateString: theStr];
  rect = [str frame];
  size.width = rect.size.width+30;
  size.height = rect.size.height;
  [self setFrameSize: size];
}

+ optionWithStr: (NSString *) theStr
{
  GGOption *opt;
  opt = AUTORELEASE([self new]);
  [opt setTitle: theStr];
  return opt;
}

- (void) setTarget: obj
{
  [bo setTarget: obj];
}

- (void) setAction: (SEL) aSel
{
  [bo setAction: aSel];
}

- (void) setState: (int) state
{
  [bo setState: state];
}

- (void) setTag: (int) obj
{
  [bo setTag: obj];
}

- (void) setButtonType: (NSButtonType)aType
{
  [bo setButtonType: aType];
}

- (void) mouseDown: (Event *)pev
{
  NSPoint point = {pev->button.x, pev->button.y};
  NSPoint local;
  local = [self convertPoint: point fromView: nil];
  NSDebugMLLog(@"MousePress", @"%@ global %@ local %@", str,
	       NSStringFromPoint(point),
	       NSStringFromPoint(local));
  [super mouseDown: pev];
}

@end

