/*
 *  Item.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <Foundation/NSObject.h>
#include <Foundation/NSException.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSCoder.h>
#include "Metaobj.h"
#include "Item.h"

typedef struct __ItemData
{
  ItemType	type;
  int 		index;
  NSString 	*name;
  int		masse;
  int		price;
}ItemData;

static ItemData itemLst[] = {
  {Mine, 1, @"XB74 Proximity Mine", 1, 27},
  {BombeEnergie, 2, @"Energy Bomb", 4, 17000},
  {Missile, 3, @"KLT60 Homing Missile", 1, 45},
  {Missile, 4, @"LV111 Smart Missile", 1, 63},
  {Missile, 5, @"NN500 Naval Missile", 1, 90},
  {Missile, 6, @"MV1 Assault Missile", 1, 117},
  {Missile, 7, @"MV2 Assault Missile", 1, 144},
  {Other, 8, @"Missile Viewer", 0, 45},
  {Other, 9, @"Laser Cooling Booster", 1, 360},
  {LaserPulsion, 10, @"1mw Pulse Laser-red", 1, 810},
  {LaserPulsion, 11, @"5mw Pulse Laser-yel", 3, 2700},
  {LaserRayon, 12, @"1mw Beam Laser-red", 5, 6300},
  {LaserRayon, 13, @"4mw Beam Laser-amb", 20, 11000},
  {LaserRayon, 14, @"20mw Beam Laser-wht", 75, 22000},
  {LaserRayon, 15, @"100mw Beam Lsr-m blu", 200, 54000},
  {LaserRayon, 16, @"Sm Plasma Accel-L blu", 500, 270000},
  {Other, 17, @"Auto Targetter", 0, 72},
  {LaserRayon, 18, @"Lg Plasma Accel-L blu", 900, 540000},
  {LaserRayon, 19, @"30mw Mining Lsr-m blu", 10, 10000},
  {Other, 20, @"MB4 Mining Machine", 30, 5850},
  {Other, 21, @"Auto Refueler", 1, 90},
  {Other, 22, @"Atmospheric Shielding", 1, 180},
  {Other, 23, @"Cargo Bay Life Support", 1, 540},
  {Other, 24, @"Energy Booster Unit", 5, 9900},
  {Other, 25, @"Escape Capsule", 5, 18000},
  {Other, 26, @"Extra Passenger Cabin", 5, 1350},
  {Other, 27, @"Fuel Scoop", 6, 3150},
  {Other, 28, @"Cargo Scoop Conver.", 2, 3600},
  {Other, 29, @"Tractor Beam C Scoop", 1, 8550},
  {Other, 30, @"Hull Auto Repair Sys.", 40, 15000},
  {Other, 31, @"Shield Generator", 4, 2250},
  {Other, 32, @"Chaff Dispenser", 1, 72},
  {ECM, 33, @"E.C.M. System", 2, 1080},
  {ECM, 34, @"Naval E.C.M. System", 2, 13000},
  {Other, 35, @"Automatic Pilot", 1, 1260},
  {Other, 36, @"Navigation Computer", 0, 162},
  {Other, 37, @"Combat Computer", 0, 270},
  {Other, 38, @"Radar Mapper", 1, 900},
  {Other, 39, @"Scanner", 1, 630},
  {Other, 40, @"Hyper Cloud Anal", 1, 1575},
  {Hyperdrive, 41, @"Interplanetary Drive", 4, 4500},
  {Hyperdrive, 42, @"Class 1 Hyperdrive", 10, 7200},
  {Hyperdrive, 43, @"Class 2 Hyperdrive", 25, 13000},
  {Hyperdrive, 44, @"Class 3 Hyperdrive", 45, 27000},
  {Hyperdrive, 45, @"Class 4 Hyperdrive", 80, 54000},
  {Hyperdrive, 46, @"Class 5 Hyperdrive", 150, 108000},
  {Hyperdrive, 47, @"Class 6 Hyperdrive", 250, 216000},
  {Hyperdrive, 48, @"Class 7 Hyperdrive", 400, 432000},
  {Hyperdrive, 49, @"Class 8 Hyperdrive", 600, 864000},
  {Hyperdrive, 50, @"Class 1 Military Drive", 6, 22000},
  {Hyperdrive, 51, @"Class 2 Military Drive", 12, 45000},
  {Hyperdrive, 52, @"Class 3 Military Drive", 24, 90000}
};


@implementation Item
+ (NSArray *) itemList
{
  int i;
  NSMutableArray *tmp = [NSMutableArray array];
  for( i = 0; i < sizeof(itemLst)/sizeof(ItemData); ++i)
    [tmp addObject: [self itemWithName: itemLst[i].name
			  type: itemLst[i].type
			  arg: itemLst[i].index
			  prixAchat: itemLst[i].price
			  prixVente: itemLst[i].price
			  masse: 1000*itemLst[i].masse]];
  return tmp;
}

- init
{
  [super init];
  prixAchat = -1;
  prixVente = -1;
  quantite = -2;
  addToCheck(self);
  return self;
}

- initWithName: (NSString *) aName
	  type: (int) aType
	   arg: (int) aArg
     prixAchat: (int) aPrixAchat
     prixVente: (int) aPrixVente
	 masse: (int) m
{
  [self init];
  ASSIGN(name, aName);
  type = aType;
  arg1 = aArg;
  prixAchat = aPrixAchat;
  prixVente = aPrixVente;
  masse = m;
  return self;
}



- (void) dealloc
{
  RELEASE(name);
  [super dealloc];
}

- (void)encodeWithCoder:(NSCoder *)encoder
{
#define GGEncode(obj) [encoder encodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]
  GGEncode(type);
  GGEncode(arg1);
  GGEncode(arg2);
  GGEncode(arg3);
  GGEncode(masse);
  GGEncode(prixAchat);
  GGEncode(prixVente);
  GGEncode(quantite);
  GGEncode(name);
}

- (id)initWithCoder:(NSCoder *)decoder
{
#define GGDecode(obj) [decoder decodeValueOfObjCType: @encode(typeof(obj))\
			       at: &obj]
  

  GGDecode(type);
  GGDecode(arg1);
  GGDecode(arg2);
  GGDecode(arg3);
  GGDecode(masse);
  GGDecode(prixAchat);
  GGDecode(prixVente);
  GGDecode(quantite);
  GGDecode(name);
  return self;
}


- (void) setItemName: (NSString *)_name
{
  ASSIGN(name, _name);
}

- (BOOL) isEqual: (id) object
{
  Item *obj;
  NSAssert([object isKindOfClass: [Item class]], 
	   NSInternalInconsistencyException);
  obj = (Item *)object;

  return (obj->type == type && obj->arg1 == arg1);
  // &&
  //	  obj->arg2 == arg2 && obj->arg3 == arg3);
}

+ itemWithName: (NSString *) aName
	  type: (int) aType
	   arg: (int) aArg
     prixAchat: (int) aPrixAchat
     prixVente: (int) aPrixVente
	 masse: (int) m
{
  Item *i = [self alloc];
  [i initWithName: aName
     type: aType
     arg: aArg
     prixAchat: aPrixAchat
     prixVente: aPrixVente
     masse: m];
  return AUTORELEASE(i);
}

+ itemWithName: (NSString *) aName
	  type: (int) aType
	   arg: (int) aArg
     prixAchat: (int) aPrixAchat
     prixVente: (int) aPrixVente
{
  return [self itemWithName:aName
	       type: aType
	       arg: aArg
	       prixAchat: aPrixAchat
	       prixVente: aPrixVente
	       masse: 1000];
}

+ itemWithType: (ItemType) _type
{
  Item *i;
  i = [Item new];
  i->type = _type;
  return AUTORELEASE(i);
}

- (NSString *) description
{
  if( name != nil )
    return name;
  else
    return [super description];
}
@end
