/*
 *  Laser.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include "GGLoop.h"
#include "Laser.h"
#include "GGRepere.h"
#include "utile.h"
#include "Sched.h"
#include "Preference.h"
#include "World.h"
#include "Tableau.h"

#include <math.h>
#include <time.h>


@implementation Laser
- initLaserAt: (GGSpaceObject *) obj
	power: (double) pow
     lifeTime: (float) lt
{
  [super init];
  _flags.graphical = YES;
  canon = obj;
  [obj enregistreForRemoving: self
       with: @selector(removeCanonWithNotification:)];
  collisionCounter = -1;
  if( lt > 0 )
    {
      [GGTimer scheduledTimerWithTimeInterval: gameTime + lt
	       target: self
	       selector: @selector(timeToDie:)
	       repeats: NO];
    }
  power = pow;
  if( pow < 5e6 )
    initRGBA(&col, 1, 0, 0, 0.5);
  else if (pow < 10e6)
    initRGBA(&col, 0, 1, 1, 0.5);
  else
    initRGBA(&col, 1, 1, 1, 0.5);
  
  return self;
}

- (void) encodeWithCoder: (NSCoder *)encoder
{
  [super encodeWithCoder: encoder];
  GGEncode(col);
  GGEncode(power);
  GGEncode(canon);
}

-  initWithCoder: (NSCoder *) decoder
{
  [super initWithCoder:  decoder];
  GGDecode(col);
  GGDecode(power);
  canon = [decoder decodeObject];
  
  return self;
}

- (void) removeCanonWithNotification: (NSNotification *)aNotification
{
#ifndef NS_BLOCK_ASSERTIONS
  id obj = [aNotification object];
#endif
  NSAssert(obj == canon, NSInternalInconsistencyException);
  [self destroy];
}

+ fireLaserAt: (GGSpaceObject *) obj
	power: (float) pow
     lifeTime: (double) lt
{
  Laser *laser;
  laser = [self alloc];
  laser = [laser initLaserAt: obj
		 power: pow
		 lifeTime: lt];
  [obj  addSubNode: laser];
  [laser enterUniverse];
  [laser setVisible: YES];
  [laser beginSchedule];
  [theWorld manageLaser: laser];
  return AUTORELEASE(laser);
}

- (void) dealloc
{
  NSDebugMLLog(@"Laser", @"Destruction");
  [super dealloc];
}

- (void) timeToDie: (GGTimer *) aTimer
{
  [self destroy];
}

- (BOOL) shouldCache
{
  //  return YES;
  return NO;
}

- (BOOL) shouldBeClipped
{
  return NO;
}

- (void) rasterise
{
  NSDebugMLLogWindow(@"Laser", @"on dessine");
  glColor4fv((GLfloat*)&col);

  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_LIGHTING);
  if( thePref.GLBlend )
    glEnable(GL_BLEND);
  glBegin(GL_TRIANGLES);
  glVertex3f(1, -10, 0);
  glVertex3f(-1, -10, 0);
  glVertex3f(0, 0, 50);
  glEnd();
  
  glPopAttrib();
}

- (float) power
{
  return power;
}
@end

