/*
 *  Metaobj.m
 *
 *  Copyright (c) 1999 Free Software Foundation, Inc.
 *  
 *  Author: Frdric De Jaeger
 *  Date: May 2000
 * 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSThread.h>
#include <Foundation/NSNotification.h>
#include "Metaobj.h"
#include <stdio.h>
#define MAXOBJ 10000

NSObject *listobj[MAXOBJ];
static int mode;
static int state;
static NSRecursiveLock *monLock = nil;

@interface ObjectControl : NSObject
+ alloc;
- (void) dealloc;
@end

void setMode(int newmode)
{
  mode = newmode;
}
/*
void erreur(const char *format, ...)
{
     va_list theva;
     va_start(theva, format);
     fprintf(stderr, "Erreur : ");
     vfprintf(stderr, format, theva);
     fprintf(stderr, "\n");
     va_end(theva);
     abort();
//     exit(1);
}
*/

void initMetaObj()
{
     int i;
     [ObjectControl poseAsClass: [NSObject class]];
     setMode(CheckMarked);
     fprintf(stderr, "Debut de l'initialisation du controleur d'allocation\n");
     for( i = 0; i < MAXOBJ; ++i)
	  listobj[i] = nil;
}

BOOL exist(NSObject *p)
{
  int i;
  [monLock lock];
  for( i = 0; i < MAXOBJ; ++i )
    if( p == listobj[i] )
      {
	[monLock unlock];
	return 1;
      }
  [monLock unlock];
  return 0;
}

void add(NSObject *p)
{
     int i;
     
     if( state || exist(p) )
       {
	 //	 puts("Cration");
	 return;
       }
     [monLock lock];
     i = 0;
     while(i < MAXOBJ && listobj[i] != nil)
	  i++;
     if( i == MAXOBJ )
	  NSLog(@"Trop d'allocation");
     else
     {
#ifdef HardAllocCheck
	  printf("Allocation  l'indice %d de %s.\n", i, [p name]);
#endif
	  listobj[i] = p;
     }
     [monLock unlock];
}

void clearObject(NSObject *p)
{
     int i;
     if( state )
       {
	 //	 puts("collision de dstruction");
	 return;
       }
     [monLock lock];
     state = 1;
     i = 0;
     while(i < MAXOBJ && listobj[i] != p)
	  i++;
     if( i == MAXOBJ )
       {
	 if( mode & CheckAll )
	   {
	     /*	     NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	     NSLog(@"Pointeur non trouve %@\n", p);
	     //	     fprintf(stderr, "Pointeur non trouve : %x %s\n",  p,
	     //		     [[p description] cString]);
		     //	     erreur("Pointeur non trouv");
		     [pool release];*/
	   }
       }
     else
       {
	 if( mode & LogFree )
	   fprintf(stderr, "bidule libr : %s\n", [[[p class] 
						      description]
						     cString]);
	 listobj[i] = nil;

       }
     state = 0;
     [monLock unlock];
}

void checkMetaObj(void)
{
     int i;
     NSAutoreleasePool *pool;
     puts("Dbut du controle mmoire");
     [monLock lock];
     state = 1;
     pool = [[NSAutoreleasePool alloc] init];
     for( i = 0; i < MAXOBJ; ++i)
	  if( listobj[i] != nil )
	    {
	      fprintf(stderr, "Erreur, objet non dtruit  %d\n", i);
	      fprintf(stderr, " : %s.\n", [[listobj[i] description]cString]);
	      //	      NSLog(@"Objet non deruit : %@\n", [listobj[i] class]);
	    }
     printf("Fin du controle.\n");
     state = 0;
     [pool release];
     [monLock unlock];
}

int getIndice(NSObject *p)
{
  int i;
  for( i = 0; i < MAXOBJ; ++i)
    if( listobj[i] == p )
      return i;
  return -1;
}


@implementation ObjectControl
+ (void) initialize
{
  static BOOL first = YES;
  if( self == [ObjectControl class] && first)
    {
      first = NO;
      [[NSNotificationCenter defaultCenter] 
	addObserver: self 
	selector: @selector(becomeMultiThread)
	name: NSWillBecomeMultiThreadedNotification
	object: nil];
    }
}

+ (void) becomeMultiThread
{
  NSLog(@"on est multithreade!!!");
  monLock = [NSRecursiveLock new];
}

+ alloc
{
  id obj;
  obj = [super alloc];
  if( mode & CheckAll )
    add(obj);
#ifdef HardAllocDebug
  fprintf(stderr, "%s a t allou\n", [obj name]);
#endif
  return obj;
}

- markObject
{
  if( mode & CheckMarked )
    add(self);
  return self;
}

- copy
{
  ObjectControl *obj;
  obj = [super copy];
  if( mode & CheckAll )
    add(obj);
  return obj;
}

- (void)dealloc
{
#ifdef HardAllocDebug
  fprintf(stderr, "%s a t libr\n", [self name]);
#endif
  clearObject(self);
  [super dealloc];
}
  
@end

