/*
 *  Missile.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>

#include <Foundation/NSException.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSValue.h>

#include "GGShip.h"
#include "Missile.h"
#include "Tableau.h"
#include "Metaobj.h"
#include "utile.h"
#include "GGNode.h"
#include "Sched.h"
#include "World.h"
#include "Commandant.h"

static const real Konst = 1;
static const real DistCont = 20;
//static const real AlmostCollinear = 0.0001;
static const real MaxTangente = 3;
static const float CoefSecu=1.0;
static const real PowerOptimale = 0.9;
static const real CorrectionPower = 1.5;

enum {
  NothingYet = 0,
  Accelerating = 1,
  Freining = 2,
  FreiningHard = 3,
  OutSideSystem = 4
}modePilotage;

NSString *GGMissileArrive = @"GGMissileArrive";

@interface AutoPilotCommon (PrivateMethod)
- (void) notifyArrival;
//designated initializer :
- initWithTarget: (id) aTarget
	 andShip: (GGShip *) aShip
	   decal: (Vect3D *) pdecal
	   speed: (real) v
	distance: (real) _d
	    type: (TypeAutoPilote) _type;

- (void) setTarget: (id)aTarget;
- (void) setType: (TypeAutoPilote) newType;

@end

@implementation AutoPilotCommon
+ autoPilotWithTarget: (id) aTarget
	      andShip: (GGShip *) aShip;
{
  if( aTarget == nil || aShip == nil )
    return nil;
  return AUTORELEASE([[self alloc] 
		       initWithTarget: aTarget
		       andShip: aShip
		       decal: NULL
		       speed: -1
		       distance: -1
		       type: ModeVaisseau]);
}

+ autoPilotWithTarget: (id) aTarget
	      andShip: (GGShip *) aShip
		decal: (Vect3D *) pdecal
		speed: (real) v
	     distance: (real) _d
{
  if( aTarget == nil || aShip == nil )
    return nil;
  return AUTORELEASE([[self alloc]
		       initWithTarget: aTarget
		       andShip: aShip
		       decal: pdecal
		       speed: v
		       distance: _d
		       type: ModeVaisseau]);
}


+ missilePilotWithTarget:(id) aTarget
		 andShip: (GGShip *) aShip
{
  if( aTarget == nil || aShip == nil )
    return nil;
  return AUTORELEASE([[self alloc] 
		       initWithTarget: aTarget
		       andShip: aShip
		       decal: NULL
		       speed: -1
		       distance: -1
		       type: ModeMissile]);
}


- (void)  _initSchedule
{
  
  if([ship locality])
    {
      if( ship->_flags.activated )
	{
	  [theLoop slowSchedule: self 
		   withSelector: @selector(miseAJourTimer)
		   arg: nil];
	  [theLoop iaSchedule: self 
		   withSelector: @selector(pilote)
		   arg: nil];
	}
      else
	{
	  lastTime = -1;
	  [theLoop slowSchedule: self 
		   withSelector: @selector(slowPilote)
		   arg: nil];
	}
    }
  else
    {
      if( endTaskTimer->datef == GGDistantFuture )
	{
	  NSWarnMLog(@"don't know how to compute the time to destination for %@...",
		     ship);
	  [endTaskTimer rescheduleForDate: gameTime+1e4];
	}
      modeFlight = OutSideSystem;
    }
}  

- initWithTarget: (id) aTarget
	 andShip: (GGShip *) aShip
	   decal: (Vect3D *) pdecal
	   speed: (real) v
	distance: (real) _d
	    type: (TypeAutoPilote) _type
{
  [super initWithShip: aShip];
  
  NSParameterAssert(aShip != nil);
  modeFlight = NothingYet;
  [self setTarget: aTarget];
  [self setType: _type];

  [endTaskTimer setObject: self
		message: @selector(timerExpired:)];
  [endTaskTimer scheduleForDate: GGDistantFuture];

  CollisionCounter(ship) = 20;
  addToCheck(self);

  SetZeroVect(decal);

  if( type == ModeVaisseau )
    {
      admissibleSpeed = (v >= 0 ? v : 0);
      if( _d >= 0 )
	distanceTarget = _d;
      if( pdecal )
	decal = *pdecal;
    }

  [self _initSchedule];

  [MessageManager
    addObserver: self
    selector: @selector(targetNotification:)
    name: nil
    object: aTarget];

  [MessageManager 
    addObserver: self
    selector: @selector(shipJump:)
    name: GGNotificationShipDoHyperJump
    object: aShip];
  [MessageManager 
    addObserver: self
    selector: @selector(shipBecomeActivated:)
    name: GGNotificationShipBecomeActivated
    object: aShip];
  [MessageManager 
    addObserver: self
    selector: @selector(shipBecomeInactivated:)
    name: GGNotificationShipBecomeInactivated
    object: aShip];
  [MessageManager
    addObserver: self
    selector: @selector(leaveCurrentSystem:)
    name: GGNotificationLeavingSystem
    object: nil];
  [MessageManager
    addObserver: self
    selector: @selector(enteringNewSystem:)
    name: GGNotificationEnteringSystem
    object: nil];
  return self;
}

- (void)encodeWithCoder:(NSCoder *)encoder
{
  [super encodeWithCoder: encoder];
  GGEncode(theTarget);
  GGEncode(modeFlight);
  GGEncode(type);
  GGEncode(gal);
  GGEncode(timeOfLeaving);
  GGEncode(signature);

  GGEncode(distanceTarget);
  GGEncode(decal);
  GGEncode(admissibleSpeed);
}

- (id)initWithCoder:(NSCoder *)decoder
{
  [super initWithCoder: decoder];
  theTarget = [decoder decodeObject];
  GGDecode(modeFlight);
  GGDecode(type);
  GGDecode(gal);
  GGDecode(timeOfLeaving);
  GGDecode(signature);

  GGDecode(distanceTarget);
  GGDecode(decal);
  GGDecode(admissibleSpeed);
  return self;
}

- (void) doinvalidate
{
  theTarget = nil;
  [super doinvalidate];
}

- (void) shipBecomeInactivated: (NSNotification *) aNotification
{
  [theLoop unSchedule: self];

  [self _initSchedule];
  //  [ggtimer suspend];
}

- (void) shipBecomeActivated: (NSNotification *) aNotification
{
  [theLoop unSchedule: self];
  [self _initSchedule];
  //  [ggtimer rescheduleForDate: [self estimatedDateAtDestination]];
}

- (void) leaveCurrentSystem: (NSNotification *)aNotification
{
  if( modeFlight != OutSideSystem ) /* We are remove from the current system
				     */
    {
      modeFlight = OutSideSystem;
      [endTaskTimer rescheduleForDate: [self estimatedDateAtDestination]];
      [theLoop unSchedule: self];
      timeOfLeaving = gameTime;
      signature = [theTarget signature];
      if( signature != nil )
	{
	  RETAIN(signature);
	  theTarget = nil;
	}
    }
}

- (void) enteringNewSystem:(NSNotification *)aNotification
{
  NSNumber *number = [[aNotification userInfo]objectForKey: @"Name"];
  NSParameterAssert(modeFlight == OutSideSystem);
  NSParameterAssert(number!=nil);
  NSDebugMLLog(@"AutoPilot", @"%@ are here", ship);
  if( [ number intValue] == [ship indexSystem] )
    /* So we are back in our system
     */
    {
      PetitGalileen pg;
      // we recover the target
      modeFlight = NothingYet;
      NSDebugMLLog(@"AutoPilot", @"self = %@ target=%@", self, theTarget);
      if( theTarget == nil && signature != nil )
	theTarget = [signature objectFromSignature];
      if( theTarget == nil )
	{
	  NSWarnMLog(@"theTarget is now nil!!, sig was %@", signature);
	  [self invalidate];
	  return;
	}
      // we position the ship

      if( [endTaskTimer date] > timeOfLeaving )
	{
	  [theTarget positionSpeedRelativesTo: ship
		     decal: NULL
		     in: &pg];
	  mulScalVect(&pg.point, (gameTime - timeOfLeaving)/
		      ([endTaskTimer date]-timeOfLeaving),&pg.point);
	  addVect(Point(ship),&pg.point,Point(ship));
	}
      [self _initSchedule];
    }
}

- (void) shipJump: (NSNotification *)aNotification
{
  if( [aNotification object] == ship )
    {
      NSDebugMLLog(@"AutoPilot", @"%@ removing %@ as a target", ship,
		   theTarget);
      [ship looseTarget: theTarget];
    }
}

- (void) targetNotification: (NSNotification *) aNotification
{
  if( [[aNotification name] isEqualToString: GGNotificationShipDoHyperJump] )
    {
      GGShip *_ship = ship;
      GGSpaceObject *_target = theTarget;
      [self invalidate];
      [_ship looseTarget: _target];
    }
}

- (void) timerExpired: (GGTimer *)timer
{
  NSDebugMLLog(@"Missile", @"timer expired");
  [self notifyArrival];
}

- (void) dealloc
{
  RELEASE(signature);
  [super dealloc];
}

- (void) setType: (TypeAutoPilote) newType
{
  type = newType;
  if( theTarget != nil )
    {
      if( newType == ModeMissile )
	distanceTarget = Dimension(theTarget);
      else
	distanceTarget = [theTarget proximite];
    }
}

- (void) setTarget: (id) obj
  /* not retained */
{
  NSAssert(theTarget == nil, NSInternalInconsistencyException);

  if( [obj isKindOfClass: [NSString class]] )
    {
      theTarget = [(NSString *)obj objectFromSignature];

      //we keep track of our destination.  We are probably not in the current
      //system.  We may get back in the current system and be prepared
      //to recover our concrete target.
      if( theTarget == nil )
	{
	  signature = [obj copy];
	  NSLog(@"sig = %@, obj = %@", signature, obj);
	}
    }
  else
    theTarget = obj;

  [theTarget enregistre: self
	     with:@selector(_removeWithNotification:)];
  [self setType: type];
}

- (void) _removeWithNotification: (NSNotification *)aNotification
{
  id obj = [aNotification object];
  if( theTarget == obj )
    {
      GGShip *_ship = ship;
      GGSpaceObject *_target = theTarget;
      [self invalidate];
      [_ship looseTarget: _target];
    }
  if( ship == obj )
    [self invalidate];
}

- (void) notifyArrival
{
  BOOL bla;
  GGShip *myship;
  /* we save the ship because this will destroy it :
   */
  myship = ship;

  bla = (fatherTask == nil);
  [ship stop];
  [self invalidate];

  //FIXME ugly hack :
  if( bla )
    {
      [[NSNotificationCenter defaultCenter]
	postNotificationName: GGNotificationArriveToStation
	object: myship];
    }

  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGMissileArrive
    object: myship];
  /*after this, we are destroyed generically.
    don't do anything...
  */
}

- (void) miseAJourTimer
{
  if( theTarget == nil || ship == nil)
    {
      NSDebugLLog(@"Missile", @"no Target");
      [self invalidate];
      return;
    }

  [endTaskTimer rescheduleForDate: [self estimatedDateAtDestination]];
}



- (void) pilote
{
  GGShip		*obj = ship;
  Vect3D 		i, j, tmp, res;
  real 			tant, norm;
  real 			vScalI;		// \vec V\scal \vec \i
  real 			vit;		//Norme du vecteur speed
  real			veff;		//speed vers la cible
  real			distCible;
  real			theTant;	//Valeur de la tangente choisie
  real                  gAccObj, gFreinObj;
  real			dmax;

  int                   sens;
  BOOL			ok = YES;

  if( theTarget == nil || ship == nil)
    {
      NSDebugLLog(@"Missile", @"no Target");
      [self invalidate];
      return;
    }

  [theTarget positionSpeedRelativesTo: ship
	     decal: &decal
	     in: &gal];
  NSDebugLLogWindow(@"Missile", @"%@ %@", theTarget, obj);
  NSDebugLLogWindow(@"Missile", @"fatherTask = %@", fatherTask);
  NSDebugLLogWindow(@"Missile", @"relative Pos : %@, speed %@", 
		    stringOfVect2(gal.point),
		    stringOfVect2(gal.speed));

  //    SetZeroVect(i);
  //    diffVect(&i, &(gal.point), &i);
  mulScalVect(&gal.speed, -1, &gal.speed);
  i = gal.point;
  norm = prodScal(&i, &i);
  norm = sqrt(norm);
  distCible = norm;

  if( distCible < distanceTarget ) /*we are at destination*/
    {
      [self notifyArrival];
      return;
    }

  divScalVect(&i, norm, &i);
  NSDebugLLogWindow(@"Missile", 
		    [NSString stringWithFormat: @"distance %f", norm]);


  NSDebugLLogWindow(@"Missile", @"i = %@", stringOfVect(&i));
  vScalI = prodScal(&i, &(gal.speed));
  vit = sqrt(prodScal(&(gal.speed), &(gal.speed)));


  if( vit > EPS ) /*je panne rien  ce code...*/
    {
      if( vScalI > 0 ) /*je vais plus ou moins vers ma cible */
	{
	  double	cosa, sina;	//angle between speed and direction
	  double	HM, OH;
	  cosa = vScalI/vit;
	  if( cosa <= 1 )
	    sina = sqrt(1-cosa*cosa);
	  else
	    {
	      cosa = 1;
	      sina = 0;
	    }
	  HM = sina*distCible;
	  if( HM < distanceTarget )
	    /* the trajectory will encounter the target sometime in the future*/
	    {
	      double delta = sqrt(distanceTarget*distanceTarget - HM*HM);
	      double OC;
	      OH = cosa*distCible;
	      OC = OH - delta;
	      if( OC  < deltaTime * vit )
		{
		  /* we will touch the target before next frame */
		  double timeEff = OC/vit;
		  NSDebugMLLog(@"Missile", @"risque de collision %g %g",
			       HM, distanceTarget);
		  NSDebugMLLog(@"Missile", @"OH = %g\nOC = %g delta = %g", 
			       OH, OC, delta);
		  NSDebugMLLog(@"Missile", @"target hit");
		  addLambdaVect(Point(ship), OC/vit, &(gal.speed),
				Point(ship));
		  if( type == ModeMissile )
		    {
		      NSDebugMLLog(@"Missile", @"distance %g", 
				   (double)sqrt(distance2Vect(Point(theTarget),
							      Point(ship))));
		    }
		  else
		    {
		      if( vit < GFREIN(ship)*timeEff )
			diffVect(Speed(ship), &gal.speed, Speed(ship));
		      else
			addLambdaVect(Speed(ship), -timeEff*GFREIN(ship)/vit,
				      &gal.speed, Speed(ship));
		    }
		  [self notifyArrival];
		  return;
		}
	    }
	}
      veff = vScalI;
    }
  else 
    {
      vScalI = 0;
      vit = 0;
      veff = 0;
    }


  switch(type){
  case ModeMissile:
    NSDebugLLogWindow(@"Missile", @"Mode Missile");
    if( modeFlight != Accelerating )
      {
	NSDebugMLLog(@"Missile", @"ModeMissile -> Acceleration");
	[obj stop];
	[obj accelere];
	modeFlight = Accelerating;
	[obj setPower: 1];
      }
    sens = 1;
    break;
  case ModeVaisseau:
    //FIXME: prevent a Warning
    dmax = 0;
    gAccObj = GACC(obj);
    gFreinObj = GFREIN(obj);
    distCible -= distanceTarget;
      
    NSDebugLLogWindow(@"Missile", @"obj proximite = %g %g", 
		      [theTarget proximite], distanceTarget);
    NSDebugLLogWindow(@"Missile", @"Mode Vaisseau");
    NSDebugLLogWindow(@"Missile", @"veff= %g distCible= %g\n admissibleSpeed=%g", 
		      veff, distCible, admissibleSpeed);
    if ( veff < admissibleSpeed || 
	 (dmax = (veff*veff - admissibleSpeed*admissibleSpeed )/
	  (2*CoefSecu*gFreinObj)) < 0.7*distCible )
      {
	NSDebugLLogWindow(@"Missile", @"Acceleration dmax = %g", dmax);
	if( modeFlight != Accelerating )
	  {
	    NSDebugMLLog(@"Missile", @"ModeVaisseau -> Acceleration");
	    modeFlight = Accelerating;
	    [obj stop];
	    [obj accelere];
	    [obj setPower: 1];
	  }
	sens = 1;
      }
    else /* dmax has just been calculated!!*/
      {
	//should be greater than 0.7*CoefSecu
	real power = CoefSecu* dmax/distCible;

	NSDebugLLogWindow(@"Missile", @"Freinage power : %e", power);


	power = (power - PowerOptimale) * CorrectionPower+PowerOptimale;

	NSDebugLLogWindow(@"Missile", @"real power : %e ", power);
	NSDebugLLogWindow(@"Missile", @"dmax = %g", dmax);
	if( modeFlight != Freining )
	  {
	    NSDebugMLLog(@"Missile", @"-> Freinage");
	    [obj stop];
	    [obj decelere];
	    modeFlight = Freining;
	  }
	[obj setPower: power];
	sens = -1;
      }
    break;
  default:
    sens = 1;
    break;
  }


  if( fabs(vScalI) < EPS * vit )
    /* $ <V, i> est presque droit
     */
    {
      NSDebugLLogWindow(@"Missile", @"V.i \\simeq 0");
      theTant = -sens*MaxTangente;
    }
  else
    {
      if( vScalI <= 0 )
	{
	  NSDebugLLogWindow(@"Missile", @"missile part dans le mauvais sens");
	  //      [obj setWishDirection: &i];
	  ok = NO;
	}
      prodVect(&i, &(gal.speed), &tmp);
      NSDebugLLogWindow(@"Missile", @"tmp = %@", stringOfVect(&tmp));
      if( prodScal(&tmp, &tmp) <= vit*EPS )
	{
	  NSDebugLLogWindow(@"Missile", @"direction quasi optimale");
	  [obj setWishDirection: &i
	       speedManche:0.01];
	  return;
	}
      prodVect(&tmp, &i, &j);
      normaliseVect(&j);
      tant = prodScal(&j, &(gal.speed))/prodScal(&i, &(gal.speed));
      tant = fabs(tant);
      //        tant = -Konst*sqrt(tant);
      {
	double tmpd;
	tmpd = sqrt(tant);
	tant = -Konst*tmpd*sqrt(tmpd);
      }
      if( tant < -MaxTangente )
	tant = -MaxTangente;
      theTant = sens*tant;
    }

  NSDebugLLogWindow(@"Missile", @"tant = %f", theTant);
  res = i;
  addLambdaVect(&res, theTant, &j, &res);
  if( ok )
    [obj setWishDirection: &res
	 speedManche: fabs(theTant)];
  else
    [obj setWishDirection: &res
	 speedManche: 1];
  NSDebugLLogWindow(@"Missile", @"res %@", stringOfVect(&res));
  NSDebugLLogWindow(@"Missile", @" power : %g flag : %x", (double)obj->power, 
		    obj->modeInertie);
  
}


- (PetitGalileen *) positionRelativeCible
{
  return &gal;
}

- (double) estimatedDateAtDestination
{
  if( ship == nil || theTarget == nil )
    return -1;
  else
    {
      real d, delta, v;
      [theTarget positionSpeedRelativesTo: ship
		 decal: NULL
		 in: &gal];
      d = normeVect(&gal.point);
      v = normeVect(&gal.speed);
      
      if( type == ModeVaisseau && modeFlight == Freining)
	{	
	  if( v > EPS )
	    delta = 2*d/v;
	  else
	    delta = GGDistantFuture;
	}
      else
	delta = sqrt(2*d/GACC(ship));
      return gameTime + delta;
    }
}

- (NSString *) description
{
  return [NSString stringWithFormat: @"%@ tracking %@", ship, theTarget];
}

- (BOOL) shouldSave
{
  return YES;
}

- (BOOL) saveChannel
{
  return YES;
}

- (void) slowPilote
{
  PetitGalileen pg;
  if( theTarget == nil || ship == nil)
    {
      NSDebugLLog(@"Missile", @"no Target");
      [self invalidate];
      return;
    }


  if( lastTime > 0 )
    {
      double dateAtDest;
      dateAtDest = [endTaskTimer date];
      NSAssert(dateAtDest > 0, NSInternalInconsistencyException);
      NSAssert(lastTime < dateAtDest, NSInternalInconsistencyException);
      [theTarget positionSpeedRelativesTo: ship
		 decal: NULL
		 in: &pg];
      mulScalVect(&pg.point, (gameTime - lastTime)/
		  ([endTaskTimer date]-lastTime),&pg.point);
      addVect(Point(ship),&pg.point,Point(ship));
    }
  else
    {
      double dateAtDest;
      dateAtDest = [self estimatedDateAtDestination];
      NSAssert(dateAtDest > 0, NSInternalInconsistencyException);

      [endTaskTimer rescheduleForDate: dateAtDest];
    }

  lastTime = gameTime;
}
@end

