/*
 *  Model3DS.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSString.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSDebug.h>
#include "Model3DS.h"
#include "parse3ds.h"
#include "Metaobj.h"

@implementation Model3DS
- initWith3DStudioFile: (NSString *)str
	withDictionary: (NSDictionary *)dico
{
  NSString *val;
  file = parse3ds([str cString]);
  if( file == NULL )
    {
      NSWarnMLog(@"Unable to load 3DS file %@", str);
      RELEASE(self);
      return nil;
    }
  initVect(&min, -100, -100, -100);
  initVect(&max, 100, 100, 100);
  scale = 1;
  if( dico != nil  && (val = [dico objectForKey:@"Scale"]) != nil )
    {
      scale = [val floatValue];
    }


  if( dico != nil  && (val = [dico objectForKey:@"Swap"]) != nil )
    {
      swap = [val boolValue];
      if( swap )
	{
	  GGMatrix4 val = { {0, 0, 0}, 0, 
			    {0, 0, 0}, 0, 
			    {0, 0, 0}, 0, 
			    {0, 0, 0}, 1};
	  pmat pp = (pmat) &val;

	  pp[0][1] = pp[1][2] = pp[2][0] = 1;

	  mat = val;
	  
	}
    }

  addToCheck(self);
  return self;
}

- (void) dealloc
{
  if( file )
    destroy(file);
  [super dealloc];
}

- (void) rasterise
{
  if( swap )
    glMultMatrix((real *)&mat);
  if( scale != 1 )
    glScalef(scale, scale, scale);
  rasterise(file);
}

- (Vect3D *) boundingBox: (Vect3D *)pax
{
  *pax = max;
  return &min;
}


+ modelWith3DStudioFile: (NSString *)str
	 withDictionary: (NSDictionary *)dico
{
  Model3DS *model;
  model = [self alloc];
  model = [model initWith3DStudioFile: str
		 withDictionary: dico];
  return AUTORELEASE(model);
}

@end
