#include <Foundation/NSArray.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSData.h>
#include <Foundation/NSValue.h>

#include "GGLoop.h"
#include "Metaobj.h"
#include "NSView.h"
#include "NSWindow.h"
#include "Preference.h"
#include "GGWidget.h"

@interface NSView (GGExtension)
- (void) wasViewUnderMouse;
- (void) isViewUnderMouse;
@end

@implementation NSView (GGExtension)
- (void) wasViewUnderMouse{}
- (void) isViewUnderMouse{}
@end

@interface GGView : NSView
{}
@end

@implementation GGView
- (BOOL) isOpaque
{
  return YES;
}

- initWithFrame: (NSRect) aRect
{
  [super initWithFrame: aRect];
  //  [self setAutoresizesSubviews: NO];
  return self;
}
@end

@interface NSWindow (Private)
- (void) setFirstOpaque: (NSWindow *) aWin;
@end


@implementation NSWindow
+ window
{
  return AUTORELEASE([self new]);
}

- (NSView *)firstResponder
{
  return firstResponder;
}

- (void) makeFirstResponder: (NSView *)aView
{
//   if( (id) aView == self )
//     aView = nil;
  firstResponder = aView;
}


- (void) flushWindow
{
}

- (void) setViewsNeedDisplay: (BOOL) flag
{
}

- initWithFrame: (NSRect) aRect
{    
  NSView *aView;
  [super init];
  frameWindow = aRect;
  firstResponder = nil;
  contentView = nil;
  open = 0;
  windowOnTop = self;
  windowAbove = nil;
  windowBelow = nil;
  firstOpaque = nil;
  doBackGround = YES;
  couleur = GGMakeCol(0, 0, 0, 0);
  grabeur = nil;
  metaView = [GGView new];
  [metaView setFrame: NSMakeRect(0, 0, frameWindow.size.width, 
				 frameWindow.size.height)];
  [metaView viewWillMoveToWindow: self];
  [self setContentView: AUTORELEASE([[NSView alloc] initWithFrame: aRect])];
  aView = [self contentView];
  [aView setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];

  [self setOpacity: 0.5];
  addToCheck(self);

  [[NSNotificationCenter defaultCenter] addObserver: self
					selector: @selector(screenSizeChanged:)
					name: GGScreenSizeChange
					object: nil];

  return self;
}

//  - init
//  {
//    NSRect aRect = NSMakeRect(0, 0, WIDTH, HEIGHT);
//    [self initWithFrame: aRect];
//    return self;
//  }

- init
{
  NSRect aRect = NSMakeRect(0, HauteurTableauBord, WIDTH, 
			    HEIGHT-HauteurTableauBord);
  [self initWithFrame: aRect];
  return self;
}



- (void) addCloseButton
{
  GGBouton *bo;

  bo = AUTORELEASE([GGBouton new]);
  [bo setFrame: NSMakeRect(frameWindow.size.width-100, 10, 100, 35)];
  [bo setTitle: @"close"];
  [bo setTarget: self];
  [bo setAction: @selector(close)];
  [bo setAutoresizingMask: NSViewMinXMargin | NSViewMaxYMargin];

  [[self contentView] addSubview: bo];

}

- (void) setFrameWindow: (NSRect) newFrame
{
  frameWindow = newFrame;
  [metaView setFrameSize: newFrame.size];
  //  [contentView setFrameSize: frameWindow.size];
}

- (void) screenSizeChanged: (NSNotification *)obj
{
  NSRect newFrame = frameWindow;
  float cx, cy;
  NSValue * old = [obj object];
  NSSize oldSize = [old sizeValue];
  NSDebugMLLog(@"Resize", @"old : %g %g new : %d %d",
	       oldSize.width, oldSize.height, WIDTH, HEIGHT);
  newFrame.size.width *= cx = WIDTH/oldSize.width;
  newFrame.size.height *= cy = HEIGHT/oldSize.height;
  newFrame.origin.x *= cx;
  newFrame.origin.y *= cy;
  NSDebugMLLog(@"Resize", @"new : %g %g", newFrame.size.width, 
	       newFrame.size.height);
  [self setFrameWindow: newFrame];
  

  NSDebugMLLog(@"Resize", @"new : %g %g new : %d %d",
	       frameWindow.size.width, frameWindow.size.height, WIDTH, HEIGHT);
}

- (void) setOpacity:(float) val
{
  real opacity;
  NSAssert(!open, @"try to set opacity when already open");
  opacity = val;
  opacity = (opacity < 0 ? 0 : opacity);
  opacity = (opacity > 1 ? 1 : opacity);
  couleur.a = opacity;
  NSDebugMLLog(@"Opacity", @"opacity %g arg %g", opacity, val);
  if (opacity == 1 && firstOpaque == nil )
    {
      [self setFirstOpaque: self];
    }
}

- (real) opacity
{
  if( thePref.GLBlend )
    return couleur.a;
  else
    return 1.0;
}

- (void) setBackgroundColor: (VectCol) val
{
  couleur.r = val.r;
  couleur.g = val.g;
  couleur.b = val.b;
  [self setOpacity: val.a];
}

- (BOOL) isOpen
{
  return open;
}

- (void) opening
{
}

- (void) open
{
  if( !open )
    {
      open = YES;
      [theLoop pushWindow: self];
      [self opening];
    }
}

- (void) closing
{
}

- (void) close
{
  if( open )
    {
      open = NO;
      RETAIN(self);
      [theLoop popWindow: self];
      [self closing];
      RELEASE(self);
    }
}

- (void)dealloc 
{
  /*Inutile, ce cause des segfault  la fin du programme
   */
  //  if( open )
  //    [self close];
  RELEASE(viewUnderMouse);
  RELEASE(grabeur);
  RELEASE(windowAbove);
  RELEASE(metaView);
  [[NSNotificationCenter defaultCenter] removeObserver: self];
  [super dealloc];
}

- (void) setContentView: (NSView*)aView
{
  if (aView == nil)
    {
      aView = AUTORELEASE([NSView new]);
    }
  if (contentView != nil)
    {
      [contentView removeFromSuperview];
    }
  contentView = aView;
  [contentView setAutoresizingMask: (NSViewWidthSizable 
				      | NSViewHeightSizable)];
  [metaView addSubview: contentView];
  [contentView resizeWithOldSuperviewSize: [contentView frame].size]; 
  [contentView setFrameOrigin: [metaView bounds].origin];

  NSAssert1 ([[metaView subviews] count] == 1,
    @"window's view has %d	 subviews!", [[metaView subviews] count]);

  [contentView setNextResponder: self];
}

- (NSView *)contentView
{
  return contentView;
}

- (void) setFrameInRect: (NSRect) aRect
	       atOrigin: (NSPoint) aPoint;
{
  glLoadIdentity();
  glTranslatef(aPoint.x, aPoint.y, 0);
  glScissor((GLint)(frameWindow.origin.x+NSMinX(aRect)), 
	    (GLint)(frameWindow.origin.y+NSMinY(aRect)),
	    (GLint)(NSWidth(aRect)), (GLint)(NSHeight(aRect)));
}


- (void) setGLFrameInRect: (NSRect) _aRect
{
  float fx, fy;
  NSRect aRect;
  //  NSDebugMLLog(@"NSWindow", @"rect = %@", NSStringFromRect(_aRect));
  fx = 1;//(float)WIDTH/VirtualWidth;
  fy = 1;//(float)HEIGHT/VirtualHeight;
  aRect.origin = NSMakePoint(fx*NSMinX(_aRect), fy*NSMinY(_aRect));
  aRect.size = NSMakeSize(fx*NSWidth(_aRect), fy*NSHeight(_aRect));
  glScissor((GLint)(frameWindow.origin.x+NSMinX(aRect)), 
	    (GLint)(frameWindow.origin.y+NSMinY(aRect)), 
	    (GLint)(NSWidth(aRect)), (GLint)(NSHeight(aRect)));
  glViewport((GLint)(frameWindow.origin.x+NSMinX(aRect)), 
	    (GLint)(frameWindow.origin.y+NSMinY(aRect)), 
	    (GLint)(NSWidth(aRect)), (GLint)(NSHeight(aRect)));
  /*
  [theTableau addLogString: NSStringFromRect(_aRect)];
  [theTableau addLogString: NSStringFromRect(aRect)];
  */
}

- (void) initWindowGL
{
  glViewport(frameWindow.origin.x, 
	     frameWindow.origin.y, 
	     frameWindow.size.width,
	     frameWindow.size.height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.,frameWindow.size.width,0,frameWindow.size.height,-1.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glDisable(GL_DEPTH_TEST);
  glDepthMask(GL_FALSE);
}

- (void) setDrawBackground:(BOOL) f
{
  doBackGround = f;
}


- (void) drawWindow
{
  NSDebugMLLog(@"display", @"debut display");
  [self initWindowGL];
  glEnable(GL_SCISSOR_TEST);
  if( thePref.GLBlend && couleur.a < 1 )
    {
      glEnable(GL_BLEND);
    }

  if( doBackGround )
    {
      [self setFrameInRect: NSMakeRect(0, 0, frameWindow.size.width, 
				       frameWindow.size.height)
	    atOrigin: NSZeroPoint];

      glColor4fv((real *)&couleur);
      glRecti(0,0,frameWindow.size.width,frameWindow.size.height);
    }

  [contentView display];
  glDisable(GL_SCISSOR_TEST);
  glDisable(GL_BLEND);
}

- (void) display
{
  if( !open )
    return;
  if( thePref.GLBlend && firstOpaque != nil && firstOpaque != self )
    {
      [firstOpaque display];
    }
  else if( !thePref.GLBlend && windowOnTop != self )
    [windowOnTop display];
  else
    {
      [self drawWindow];
      [windowAbove display];
    }
}

- (BOOL) handleEvent: (void *)event
{
  SDL_Event 		ev = *(SDL_Event *)event;
  int 			x, y;
  NSPoint 		thePoint;
  NSView		*hitView;
  NSDebugMLLog(@"NSWindow", 
	       @"self : %@ grabeur : %@ type : %d\nfirst responder %@", 
	       self,  grabeur, ev.type, firstResponder);
  if( !open )
    return 0;
//    if( windowOnTop != self )
//      return [windowOnTop handleEvent: event];

  [theLoop queryPointer: &x 
	   and: &y];
	   //  x = ev.xbutton.x;
	   //  y = ev.xbutton.y;
  thePoint = NSMakePoint((float)(x-frameWindow.origin.x),
			 (HEIGHT - (float)y-frameWindow.origin.y));
  
  if( grabeur && [grabeur window]  == nil )
    DESTROY(grabeur);

  if( viewUnderMouse && [viewUnderMouse window] == nil )
    DESTROY(viewUnderMouse);

  if( grabeur != nil )
    hitView = grabeur;
  else
    {
      hitView = [contentView hitTest: thePoint];
      if( hitView != viewUnderMouse )
	{
	  if( viewUnderMouse )
	    [viewUnderMouse wasViewUnderMouse];
	  ASSIGN(viewUnderMouse, hitView);
	  if( viewUnderMouse )
	    [viewUnderMouse isViewUnderMouse];
	}
    }	  

  if( hitView == nil )
    {
      if( windowBelow != nil )
	return [windowBelow handleEvent: event];
      else
	return 0;
    }
  
  NSDebugMLLog(@"NSWindow", @"hitView %@", hitView);
  switch(ev.type)
    {
    case SDL_KEYDOWN:
      NSDebugMLLog(@"Event", 
		   @"self : %@ grabeur : %@ type : %d\nfirst responder %@", 
		   self,  grabeur, ev.type, firstResponder);
      if( firstResponder != nil )
	{
	        
	  NSDebugMLLog(@"Event", @"firstResponder->next = %@", 
		       [firstResponder nextResponder]);


	  [firstResponder keyDown: &ev];
	}
      else
	[hitView keyDown: &ev];
      break;
    case SDL_KEYUP:
      if( firstResponder != nil )
	[firstResponder keyUp: &ev];
      else
	[hitView keyUp: &ev];
      break;
    case SDL_MOUSEBUTTONDOWN:
      NSDebugMLLog(@"MousePress", @"%@ point %@", hitView,
		   NSStringFromPoint(thePoint));
      ev.button.x = thePoint.x;
      ev.button.y = thePoint.y;
      [hitView mouseDown: &ev];
      ASSIGN(grabeur, hitView);
      break;
    case SDL_MOUSEBUTTONUP:
      ev.button.x = thePoint.x;
      ev.button.y = thePoint.y;
      [hitView mouseUp: &ev];
      DESTROY(grabeur);
      break;
    case SDL_MOUSEMOTION:
      ev.motion.x = thePoint.x;
      ev.motion.y = thePoint.y;
      [hitView mouseMoved: &ev];
      break;
    case    SDL_JOYAXISMOTION:
    case    SDL_JOYBALLMOTION:
    case    SDL_JOYHATMOTION:
    case    SDL_JOYBUTTONDOWN:
    case    SDL_JOYBUTTONUP:
      [hitView joystickChange: &ev];
      break;
    default:
      return 0;
    }
  return 1;

}

- (void) setFirstOpaque: (NSWindow *) aWin
{
  if( aWin == nil && [self isOpaque] )
    aWin = self;
  firstOpaque = aWin;
  [windowBelow setFirstOpaque: aWin];
}

- (void) setWinOnTheTop: (NSWindow *) aWin
{
  windowOnTop = aWin;
  [windowBelow setWinOnTheTop: aWin];
}

- (NSWindow *) windowOnTop
{
  return windowOnTop;
}

- (BOOL) isOpaque
{
  return couleur.a >= 1;
}

- (NSWindow *) getFirstOpaque
{
  if( firstOpaque )
    return firstOpaque;
  else if( [self isOpaque] )
    return self;
  else
    return nil;
}

- (void) pushOnTheTop: (NSWindow *)aWin
{
  [windowOnTop pushWindow: aWin];
}

- (void) pushWindow: (NSWindow *) aWin
{
  if( aWin == windowAbove )
    return;
  if( windowAbove != nil )
    {
      [self setFirstOpaque: nil];
      [self setWinOnTheTop: self];
    }

  ASSIGN(windowAbove, aWin);
  [aWin setWindowBelow: self];

  if( aWin == nil )
    return;

  [self setFirstOpaque: [aWin getFirstOpaque]];
  [self setWinOnTheTop: [aWin windowOnTop]];
}

- (NSWindow *) popWindow
{
  NSWindow *retVal = nil;
  ASSIGN(retVal, windowAbove);
  DESTROY(windowAbove);

  [self setFirstOpaque: nil];
  [self setWinOnTheTop: self];

  [retVal setWindowBelow: nil];
  NSDebugMLLog(@"poping", @"poping from %@ value %@", self, retVal);

  return AUTORELEASE(retVal);
}

- (void) grabEvent: (NSView *) aView
{
//    [self makeFirstResponder: aView];
//    grabeur = aView;
}

- (void) unGrabEvent
{
//    grabeur = nil;
}

- (void) setWindowBelow: (NSWindow *)win
{
  windowBelow = win;
}

- (NSWindow *) windowBelow
{
  return windowBelow;
}

- (void) takeFocus
{
}

- (void) looseFocus
{
}
@end
