/*
 *  Planete.m
 *
 *  Copyright (c) 2002 by Frdric De Jaeger
 *  
 *  Date: February 2002
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glx.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSException.h>
#include <Foundation/NSArray.h>
#include "GGInput.h"
#include "Planete.h"
#include "GGRepere.h"
#include "utile.h"
#include "GG3D.h"
#include "Tableau.h"
#include "Commandant.h"
#include "perlin.h"
#include "GGBlock.h"
#include "Texture.h"
#include "Preference.h"

#define PRECISION 20

@interface GenTexture : NSObject
{}
@end
@implementation GenTexture
+ (VectCol) colorBoxPointdv: (double [3]) val
{
  double n, h, r, s, t;
  r = val[0];
  s = val[1];
  t = val[2];
  n = sqrt(r*r+s*s+t*t);
  r /= n;
  s /= n;
  t /= n;
  h=1000*PerlinNoise3D(10*r, 10*s, 10*t, 3, 2, 10);
  return genColFromHeight(h);
}
@end

@interface LocalPlanetBlock : GGSpaceObject
{
@public
  BlockContext *context;  // not retained!!
  Vect3D	col;
}
+ localBlock: (BlockContext *) _c;
@end

static void rasterizeArray(NSMutableArray *array)
{
  int n;
  n = [array count];
  NSDebugLLogWindow(@"Scene", @"nblock = %d", n);
  if( n > 0)
    {
      GGBlock *blocks[n];
      int i;

      [array getObjects: blocks];
      for(i = 0; i < n; ++i)
	secondPassBis(blocks[i]);
    }
  [array removeAllObjects];
}


@implementation LocalPlanetBlock
+ localBlock: (BlockContext *) _c
{
  LocalPlanetBlock *obj;
  obj = [self new];
  obj->context = _c;
  obj->_flags.graphical = YES;
  [obj setVisible: YES];
  
  return AUTORELEASE(obj);
}

- copyWithZone: (NSZone *)zone
{
  [NSException raise: NSInternalInconsistencyException
	       format: @"%@ should not be copied", self];
  return nil;
}

- (BOOL) shouldCache
{
  return NO;
}

- (BOOL) shouldBeClipped
{
  return NO;
}

- (void) rasterise: (Camera *)cam
{

  glColor3fv((real*) &col);
  [GGBlock initGLContextBeforeDrawing: context->texture];

  
  rasterizeArray(context->forLocal);

  [GGBlock finish];
#ifdef DEBUG
  [GGBlock debugAfterLocal];
  [GGBlock debugAfter];
#endif
}

- (void) changeLocalCoordinates: (GGGalileen *) gal
{
  NSDebugMLLog(@"Scene", @"change of repere local, we do nothing");
}
@end

static inline int indexForVect(Vect3D *v)
{
  int res = 0;
  if( v->x < 0 )
    res |= 0x4;
  if( v->y < 0 )
    res |= 0x2;
  if( v->z < 0 )
    res |= 0x1;
  return res;
}

static void makeFrustum(Frustum *f, Camera *pcam)
{
  Plan devant, gauche, droite, haut, bas;
  GGMatrix4 *mat = &pcam->modelView;


#define Cam (*pcam)


  /* Compute frustum :
   */

  droite = mkPlan(1, 0, Cam.param->xFoc, 0);
  gauche = mkPlan(-1, 0, Cam.param->xFoc, 0);
  haut = mkPlan(0, -1, Cam.param->yFoc, 0);
  bas = mkPlan(0, 1, Cam.param->yFoc, 0);
  devant = mkPlan(0, 0, 1, Cam.param->zmin);

  transformePlan(mat, &droite, &f->droite);
  transformePlan(mat, &gauche, &f->gauche);
  transformePlan(mat, &haut, &f->haut);
  transformePlan(mat, &bas, &f->bas);
  transformePlan(mat, &devant, &f->devant);

  f->idro = indexForVect(&f->droite.v);
  f->igau = indexForVect(&f->gauche.v);
  f->ihau = indexForVect(&f->haut.v);
  f->ibas = indexForVect(&f->bas.v);
  f->idev = indexForVect(&f->devant.v);

}

@implementation Planete
#define theTexture context.texture

- (void) buildSphere
{
  NSObject<GGBump> *  bump;
  bump = [PerlinBumpSphere bumpWithRadius: radius];
  
  b1 = RETAIN([GGBlock topBlockWithManager: &context
		       bump: bump
		       at: 0
		       and: -M_PI/4
		       width: M_PI/2
		       height: M_PI/2]);
  b2 = RETAIN([GGBlock topBlockWithManager: &context
		       bump: bump
		       at: M_PI/2
		       and: -M_PI/4
		       width: M_PI/2
		       height: M_PI/2]);
  b3 = RETAIN([GGBlock topBlockWithManager: &context
		       bump: bump
		       at: M_PI
		       and: -M_PI/4
		       width: M_PI/2
		       height: M_PI/2]);
  b4 = RETAIN([GGBlock topBlockWithManager: &context
		       bump: bump
		       at: 3*M_PI/2
		       and: -M_PI/4
		       width: M_PI/2
		       height: M_PI/2]);

  b5 = RETAIN([GGBlock topBlockWithManager: &context
 		       bump: [PerlinBumpSphereTop bumpWithRadius: radius]
 		       at: -1
 		       and: -1
 		       width: 2
 		       height: 2]);

  b6 = RETAIN([GGBlock topBlockWithManager: &context
 		       bump: [PerlinBumpSphereBottom bumpWithRadius: radius]
 		       at: -1
 		       and: -1
 		       width: 2
 		       height: 2]);

  b1->gauche = b4;
  b2->gauche = b1;
  b3->gauche = b2;
  b4->gauche = b3;

  b1->droite = b2;
  b2->droite = b3;
  b3->droite = b4;
  b4->droite = b1;

  b1->haut = b5;
  b2->haut = b5;
  b3->haut = b5;
  b4->haut = b5;

  b5->bas = b1;
  b5->droite = b2;
  b5->haut = b3;
  b5->gauche = b4;

  b1->bas = b6;
  b2->bas = b6;
  b3->bas = b6;
  b4->bas = b6;

  b6->bas = b4;
  b6->droite = b3;
  b6->haut = b2;
  b6->gauche = b1;

  NSDebugMLLog(@"Sphere", @"%@ builded", self);

  [self invalidCache];
}

- (void) destroySphere
{
  DESTROY(b1);
  DESTROY(b2);
  DESTROY(b3);
  DESTROY(b4);
  DESTROY(b5);
  DESTROY(b6);
  NSDebugMLLog(@"Sphere", @"%@ destroyed", self);
  [self invalidCache];
}

- (void) _initIVar
{
  context.forLocal = [NSMutableArray new];
  context.localStudy = [NSMutableArray new];
  context.forGlobal = [NSMutableArray new];
}

- (void) __privatePlanetePostInit
{
  distanceLocal2 = SQ(2*radius);
  distanceGlobal2 = SQ(3*radius);
  dimension = radius;
  masse = 4.0/3.0*M_PI*SQ(radius)*1000;

  CollisionCounter(self) = -1;
}


- initWithDictionary: (NSDictionary *)dic
		name: (NSString *)_name
  /* We know that [super initWithDictionary] will call [self init]
   */
{
  NSString *val;
  [super initWithDictionary: dic
	 name: _name];

  if( (val = [dic objectForKey:@"radius"]) != nil )
    {
      NSDebugMLLog(@"dico", @"radius = %@", val);
      radius = [val doubleValue];
    }

  if( (val = [dic objectForKey:@"LightSource"]) != nil )
    {
      NSDebugMLLog(@"dico", @"LightSource = %@", val);
      [self setLightSource: [val isEqualToString: @"YES"]];
    }

  if( (val = [dic objectForKey:@"texture"]) != nil )
    {
      NSDebugMLLog(@"dico", @"texture = %@", val);
      if( thePref.GLTexture)
	{
	  ASSIGN(textureName, val);
	}
    }

  [self __privatePlanetePostInit];

  return self;
}

- init
{
  Vect3D _col = { 0.8, 0.8, 0.8 };
  [super init];


  col = _col;
  radius = 1e3;

  [self __privatePlanetePostInit];
  [self _initIVar];
  
  return self;
}


- (id) copyWithZone: (NSZone *)zone
{
  Planete *copy;
  copy = [super copyWithZone: zone];
  copy->lb = nil;
  RETAIN(copy->theTexture);
  RETAIN(copy->textureName);
  [copy _initIVar];
  return copy;
}


- (void) dealloc
{
  RELEASE(context.forLocal);
  RELEASE(context.forGlobal);
  RELEASE(context.localStudy);
  RELEASE(b1);
  RELEASE(b2);
  RELEASE(b3);
  RELEASE(b4);
  RELEASE(b5);
  RELEASE(b6);
  RELEASE(textureName);
  RELEASE(theTexture);
  [super dealloc];
}

- setRandomSatAt: (real) orbiteRadius
{
  NSString *texName[] = {
    @"callisto.jpg",
    @"charon.jpg",
    @"europa.jpg",
    @"moon.jpg",
    @"neptune.jpg"};
  ParametresOrbitaux prm =
  {
    0,
    0,
    0,

    0.,
    0,
    0
  };
  prm.orbiteRadius = orbiteRadius;
  radius = (vrnd()*5+1)*1e6;
  prm.omegaPropre = (vrnd() * 2 + 1)*2e-5;
  prm.l0 = vrnd()*2*M_PI;
  [self setParametresOrbitaux: &prm];
  ASSIGN(textureName, texName[irnd(sizeof(texName)/sizeof(NSString *))]);

  [self __privatePlanetePostInit];

  return self;
}

- setRandomPlaneteAt: (real) orbiteRadius
{
  int nSat, i;
  NSString *texName[] = {
    @"callisto.jpg",
    @"charon.jpg",
    @"europa.jpg",
    @"moon.jpg",
    @"neptune.jpg"};
  ParametresOrbitaux prm =
  {
    0,
    0,
    0,

    0.,
    0,
    0
  };
  prm.orbiteRadius = orbiteRadius;
  radius = (vrnd()*10+1)*5e6;
  prm.omegaPropre = (vrnd() * 2 + 1)*2e-5;
  prm.l0 = vrnd()*2*M_PI;
  [self setParametresOrbitaux: &prm];
  ASSIGN(textureName, texName[irnd(sizeof(texName)/sizeof(NSString *))]);

  nSat = irnd(3*log(radius)/log(1e6));

  orbiteRadius = radius*10;

  for( i = 0; i < nSat; ++i )
    {
      Planete *p;
      p = [Planete new];
      [p setName: [NSString stringWithFormat: @"%@ %c", self, 'a'+i]];
      [p setRandomSatAt: orbiteRadius];
      [p setOmegaOrbite: omegaOrbite(radius, orbiteRadius)];
      [self addSon: p];
      RELEASE(p);
      orbiteRadius *= 2+5*vrnd();
    }

  [self __privatePlanetePostInit];

  return self;
}

- setRandomSystemForStar: (int) index
  /*use Titus-Bode law to generate random system
   */
{
  real a = 0.4;
  int n;
  int i;
  real	orbiteRadius;
  [super setRandomSystemForStar: index];
  radius = 5e9;
  [self setLightSource: YES];

  n = irnd(10);
  orbiteRadius = (100+ vrnd()*10)*radius;
  
  for( i = 0; i < n; ++i )
    {
      Planete *p;
      p = [Planete new];
      [p setName: [NSString stringWithFormat: @"%@-%d", self, i+1]];
      [p setRandomPlaneteAt: a*orbiteRadius];
      [p setOmegaOrbite: omegaOrbite(radius, orbiteRadius)];
      [self addSon: p];
      RELEASE(p);
      orbiteRadius *= 1.8+5*vrnd();

      if( i > 0 )
	a = 0.4+(0.3)*(2<<(i-1));
      else
	a = 0.7;
    }

  [self __privatePlanetePostInit];
  return self;
}

- (BOOL) lightSource
{
  return lightSource;
}

- (void) setLightSource: (BOOL) val
{
  lightSource = val;
}

- (void) cameraBeginBeingClose
{
  if( textureName )
    [self setTexture: [Texture cubeMapTextureWithFileName:  textureName]];
}

- (void) cameraStopBeingClose
{
  [self setTexture: nil];
}

- (void) setTexture: (Texture *)t
{
  ASSIGN(theTexture, t);
  [self invalidCache];
}


- (BOOL) shouldCache
{
  return b1 == nil;
}

- (BOOL) shouldBeClipped
{
  return NO;
}

- (void) beginBeingSuperLocal
{
  GGNode *aNode;
  repereRotatif = aNode = AUTORELEASE([GGNode new]);
  [repereRotatif setDefaultName: 
		  [NSString stringWithFormat: @"repereRotatif attache a %@",
			    self]];
  *Galileen(repereRotatif) = *Galileen(self);
  [repereTranslation addSubNode: repereRotatif];
  NSAssert([repereRotatif nodePere] == [self nodePere], 
	   NSInternalInconsistencyException);

  [aNode reparent: self];
  initWithIdentityTransfo(Galileen(self));
  SetZeroVect(*Omega(self));

  [super beginBeingSuperLocal];

  [aNode setMode: GGInertialRotationalMode];

  //we don't need to retain it
  NSAssert(lb == nil, NSInternalInconsistencyException);
  lb = [LocalPlanetBlock localBlock: &context];
  lb->col = col;
  
  [[manager repere] addSubNode: lb];
  [self buildSphere];
}

- (void) endBeingSuperLocal
{
  GGNode *aNode = (GGNode *)repereRotatif;
  NSAssert(repereRotatif != nil, NSInternalInconsistencyException);
  NSAssert(lb != nil, NSInternalInconsistencyException);
  [lb removeFromSuperNode];
  lb = nil;
  [aNode removeFromSuperNodeAndReparentSons];
  repereRotatif = nil;
  [super endBeingSuperLocal];
  [theCommandant setNearPlanete: NO];
  [self destroySphere];
}

- (real) proximite
{
  return 2*radius;
}

- (void) rasterise: (Camera *)cam
{
  glColor3f(col.x, col.y, col.z);
  if( b1 == nil )
    {
      GLUquadricObj *data;

      data = gluNewQuadric();
      gluQuadricNormals(data, GLU_SMOOTH);
      gluQuadricOrientation(data, GLU_OUTSIDE);


      [GGBlock initGLContextBeforeDrawing: theTexture];
      if( lightSource )
	{
	  glDisable(GL_LIGHTING);
	  glDisable(GL_DEPTH_TEST);
	}

      gluSphere(data, radius, PRECISION, PRECISION);
      gluDeleteQuadric(data);
      [GGBlock finish];
    }
  else
    {
      GGSpaceObject *obj;
      //compute the position of the camera in object coordinates;
  
      Camera camBis;
      camBis.param = cam->param;

      produitMatrice(&cam->modelView, Matrice(self), &camBis.modelView);

      NSDebugMLLogWindow(@"blaa", @"%@", stringOfMatrice(Matrice(self)));

      makeFrustum(&context.frustum, &camBis);

      transposeTransformePoint4D(&camBis.modelView, &GGZeroVect, &context.posObs);
      obj = [manager repere];
      context.nFrame ++;
      context.posLocalFrame = *Point(obj);
      NSDebugLLogWindow(@"Lod", @"manager = %@\nrepere = %@\n%@\ncampos : %@",
			manager, obj, stringOfVect(Point(obj)),
			stringOfVect2(context.posObs));
      NSDebugLLogWindow(@"Lod", @"owner camera %@", cam->owner);
      NSDebugLLogWindow(@"Lod", @"mat repere non local ;\n%@",
			stringOfMatrice(&cam->modelView));



      NSDebugMLLogWindow(@"GGBlock", @"nFrame = %d pos : %@\n%@\n%@",
			 context.nFrame, stringOfVect(&cam->modelView.position),
			 stringOfVect(&context.posObs), stringOfVect(Point(self)));

#ifdef DEBUG
      [GGBlock debugBefore];
#endif
      [GGBlock initGLContextBeforeDrawing: theTexture];

      [b1 rasterize];
      [b2 rasterize];
      [b3 rasterize];
      [b4 rasterize];
      [b5 rasterize];
      [b6 rasterize];
  
      {
	int n = [context.localStudy count];
	if( n > 0 )
	  {
	    GGBlock *blocks[n];
	    int i;
	    [context.localStudy getObjects: blocks];

#if 1
	    //FIXME
	    // We need the position of the camera in the local frame
	    // So we look for the transformation matrix that has been already
	    // computed.  But we need to go far into the camera stack...
	    transposeTransformePoint4D(&cam->prev->prev->modelView, 
				       &GGZeroVect, 
				       &context.posObs);
	    makeFrustum(&context.frustum, cam->prev->prev);
#endif

	    NSDebugLLogWindow(@"Lod", @"owner camera %@", cam->prev->owner);
	    NSDebugLLogWindow(@"Lod", @"manager = %@\nrepere = %@\n%@\ncampos : %@",
			      manager, obj, stringOfVect(Point(obj)),
			      stringOfVect2(context.posObs));

	    NSDebugLLogWindow(@"Lod", @"mat repere local ;\n%@\nposCom: %@",
			      stringOfMatrice(&cam->prev->modelView),
			      stringOfVect(Point([manager camera])));


	    for(i = 0; i < n; ++i)
	      [blocks[i] rasterize];
	  }
	[context.localStudy removeAllObjects];
      }

#ifdef DEBUG
      [b1 checkSons];
      [b2 checkSons];
      [b3 checkSons];
      [b4 checkSons];
      [b5 checkSons];
      [b6 checkSons];
#endif

      rasterizeArray(context.forGlobal);
      [GGBlock finish];
#ifdef DEBUG
      [GGBlock debugAfterGlobal];
      if( ! lb)
	[GGBlock debugAfter];
#endif
      //  rasterizeArray(context.forLocal);
    }
}


@end
