/*
 *  Preference.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSUserDefaults.h>

#include "Metaobj.h"
#include "GGWidget.h"
#include "Preference.h"
#include "NSWindow.h"
#include "World.h"
#include "GGMenu.h"


typedef struct __nameOption
{
  NSString 	*str;
  BOOL		*pval;
  NSObject	*obj;
  SEL		s;
  NSString	*notification;
}NameOption;


UserPref thePref;


@interface Pref (Private)
+ (Pref *)pref;
- (void) loadOption;
- (void) saveOption;
@end


static Pref *__thePref;

@implementation Pref
+ pref
{
  if(__thePref  == nil )
    {
      __thePref = [self new];
    }
  return __thePref;
}

+ (void) loadOption
{
  if( !__thePref )
    [Pref pref];
  [__thePref loadOption];
}


+ (void) saveOption
{
  if( !__thePref )
    [Pref pref];
  [__thePref saveOption];
}

+ (void) clean
{
  DESTROY(__thePref);
}

- init
{
  NameOption options[] = 
  {
    {@"GL_BLEND", &(thePref.GLBlend), nil, NULL, nil},
    {@"GL_TEXTURE_2D", &(thePref.GLTexture), nil, NULL, nil},
    {@"GL_SMOOTH", &(thePref.GLSmooth), nil, NULL, nil},
    {@"Enable Label", &(thePref.label), nil, NULL, nil},
    {@"Name Of Star", &(thePref.nameStar), nil, NULL, nil},
    {@"Stars", &(thePref.stars), nil, NULL, nil},
    {@"Frame Rate", &(thePref.frameRate), nil, NULL, nil},
    {@"Sound", &(thePref.sound), theLoop, @selector(setSoundFlag:), nil},
#ifdef DEBUG
    {@"Debug Memory", &(thePref.DebugMemoire), nil, NULL, nil},
    {@"Easy Pilot", &(thePref.easyPilot), nil, NULL, nil},
#endif
    {@"Mipmaping", &(thePref.GLMipmap), nil, NULL, nil},
    {@"Light @rLocale", &(thePref.LightLocal), nil, NULL, nil},
    {@"Small Texture", &(thePref.SmallTexture), nil, NULL, 
     @"SmallTextureFlagChanged"},
    {@"Direction Hinter", &(thePref.DirectionHinter), nil, NULL, nil},
    {@"Full Screen", &(thePref.FullScreen), nil, NULL, nil},
    {@"High Quality", &(thePref.HighQuality), nil, NULL, nil},
  };

  [super init];


  opts = NSZoneMalloc([self zone], sizeof(options));

  memcpy(opts, options, sizeof(options));

  nOption = sizeof(options)/sizeof(NameOption);

  return self;
}

- (void) dealloc
{
  NSZoneFree([self zone], opts);
  [super dealloc];
}

- (void) loadOption
{
  int i;
  NSUserDefaults		*def = [NSUserDefaults standardUserDefaults];
  for( i = 0; i< nOption; ++ i)
    {
      *opts[i].pval = [def boolForKey: opts[i].str];
    }
    
}

- (void) saveOption
{
  NSUserDefaults		*def = [NSUserDefaults standardUserDefaults];
  int i;

  for(i = 0; i<nOption; ++i)
    [def setBool: *opts[i].pval
	 forKey: opts[i].str];
  [def synchronize];
}

@end


@implementation UserPreference
- initUI
{
  GGOption 		*opt;
  GGMenu 		*menu;
  int i;

  [self addCloseButton];
  menu = [GGMenu menu];
  [menu setFrameOrigin: NSMakePoint(100,100)];

    
  for(i = __thePref->nOption-1; i>=0 ; --i)
    {
      opt = [GGOption optionWithStr: __thePref->opts[i].str];
      [opt setState: !!*(__thePref->opts[i].pval)];
      [opt setAction: @selector(setOption:)];
      [opt setTag: i];
      [opt setTarget: self];
      [menu addMenuItem: opt];
      NSDebugMLLog(@"Pref", @"%@ %d", __thePref->opts[i].str, 
		   *(__thePref->opts[i].pval));
    }
  [menu endItem];
  [[self contentView] addSubview: menu];

  //  [self setOpacity: 1.0];

  return self;
}

- init
{
  NSRect aRect = NSMakeRect(0, 0, WIDTH, HEIGHT);
  [self initWithFrame: aRect];
  [self initUI];
  return self;
}

- (void) opening
{
  NSDebugMLLog(@"Pref", @"window below %@", windowBelow);
  [theWorld setPaused: YES];
}


- (void) save
{
  [[Pref pref] saveOption];
}

- (void) closing
{
  NSDebugMLLog(@"Pref", @"closing");
  [theWorld setPaused: NO];
  [self save];
}


- (void) dealloc
{
  [self save];
  [super dealloc];
}

- (void) setOption: (GGBouton *)bou
{
  NameOption *opts;
  int state = !![bou state];
  int i = [bou tag];
  opts = [Pref pref]->opts;
  *opts[i].pval = state;
  if( opts[i].obj != nil )
    [opts[i].obj performSelector: opts[i].s withObject: (id) state ];
  if( opts[i].notification != nil )
    [[NSNotificationCenter defaultCenter]
      postNotificationName: opts[i].notification 
      object: self];

  NSDebugMLLog(@"Pref", @"i = %d %@ %d %@", i, opts[i].str, state,
	       opts[i].obj);
}


- (void) keyDown: (Event*)theEvent
{
  NSDebugMLLog(@"Pref", @"keyDown");

  switch (theEvent->key.keysym.sym)
    {
      case SDLK_ESCAPE:
	NSDebugMLLog(@"Pref", @"On se barre");
	[theLoop stopMainLoop];
	break;
    default:
      [super keyDown: theEvent];
      break;
    }
}

@end
