/* 	-*-ObjC-*- */
/*
 *  Resource.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: April 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Resource.h"
#include "Metaobj.h"
#include "GGFont.h"
#include "Preference.h"
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSFileManager.h>

Resource *theResourceManager;


@implementation Resource
+ initResourceManager
{
  if( theResourceManager != nil )
    return theResourceManager;

  //execute only once :
  [Pref loadOption]; //call initialize
  theResourceManager = [self new];
  [theResourceManager addForCleaning: [Pref class]];
  return theResourceManager;
}

+ resourceManager
{
  return [self initResourceManager];
}

+ (void) destroyResourceManager
{
  DESTROY(theResourceManager);
}

- (NSString *) fontPath
{
  return fontPath;
}

- (NSString *) modelPath
{
  return @"data/model";
}

- (NSString *) texturePath
{
  return @"data/textures";
}

- (NSString *)bundlePath
{
  return @"./Library/Bundles/";
}

- (NSArray *) ships
{
  return ships;
}

- (void) initShips
{
  NSArray *temp;
  NSArray *ext = [NSArray arrayWithObject: @"shp"];
  
  temp = [[NSFileManager defaultManager]
	       directoryContentsAtPath: [self modelPath]];
  ships = [temp pathsMatchingExtensions: ext];
  RETAIN(ships);
}

- init
{
  [super init];
  resourceDic = [NSMutableDictionary new];
  toClean = [NSMutableArray new];
  dataPath = @"data";
  fontPath = RETAIN(([NSString pathWithComponents:
				[NSArray arrayWithObjects:  dataPath,
					 @"fonts", nil]]));
  [self initShips];
  addToCheck(self);
  
  return self;
}

- (void) addForCleaning: obj
{
  [toClean addObject: obj];
}  

- (void) addResource: (id) obj
	    withName: (NSString *)name
{
  [resourceDic setObject: obj
	       forKey: name];
}

- (id) resourceWithName: (NSString *)name
{
  return [resourceDic objectForKey: name];
}

- (void) dealloc
{	
  [toClean makeObjectsPerformSelector: @selector(clean)];
  RELEASE(toClean);
  RELEASE(dataPath);
  RELEASE(fontPath);
  RELEASE(defaultFont);
  RELEASE(resourceDic);
  [super dealloc];
}

- (GGFont *) defaultFont
{
  return defaultFont;
}

- (void) setDefaultFont: (GGFont *) aFont
{
  ASSIGN(defaultFont, aFont);
}
@end
