/*
 *  ShipWin.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: April 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>

#include "GGRepere.h"
#include "GGLoop.h"
#include "Metaobj.h"
#include "Tableau.h"
#include "Sched.h"
#include "World.h"
#include "GGWidget.h"
#include "GGShip.h"
#include "utile.h"
#include "Item.h"
#include "Commandant.h"
#include "GGFont.h"
#include "ShipWin.h"
#include "GGConcreteShip.h"
#include "NSWindow.h"
#include "Resource.h"
#include "Texture.h"

static const float scaleFactor = ScaleFactorForShipView;

@implementation ShipWindow
- init
{
  NSRect frame;
  [super init];
  //  [self setOpacity: 1.0];

  frame = [self frame];

  [self enableTextureBackground];

  shipView = [[ShipView alloc]
	       initWithFrame: NSMakeRect(10, 50, frame.size.width-20, 
					 frame.size.height-70)];
  [self addSubview: shipView];
  //  [self addCloseButton];
  //  [self makeFirstResponder: shipView];

  return self;
}

- (void) keyDown: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_n:
      [shipView nextChoice];
      break;
    default:
      [super keyDown: pev];
      break;
    }
}

- (void) dealloc
{
  RELEASE(shipView);
  [super dealloc];
}

@end

@implementation ShipView
- (id) initWithFrame: (NSRect)frameRect
{
  [super initWithFrame: frameRect];

  ennemi = nil;
  angle = 0;
  typeShipView = TypeRotation;

  //  [self nextChoice];
  typeProjection = [GGString new];
  [typeProjection setString: @"Rotation"
		  withFont: [GGFont fontWithFile: DefaultFontName
				    atSize: 10]];
  [typeProjection setFrameOrigin: NSMakePoint(20, 20)];

  [self addSubview: typeProjection];
  

  return self;
}

- (void) dealloc
{
  RELEASE(typeProjection);
  RELEASE(ennemi);
  [super dealloc];
}

- (void) setShip: (GGShip *)shp
{
  ASSIGN(ennemi, shp);
  [ennemi setVisible: YES];
}

- (void) nextChoice
{
  GGShip *bis;  
  NSArray *ships = [theResourceManager ships];
  if( index >= [ships count])
    index = 0;
  bis = [GGShip shipWithFileName: [ships objectAtIndex: index]];
  ASSIGN(ennemi, bis);
  //  [ennemi adjustGlobalCoordinate: nil];
  [ennemi setVisible: YES];
  index++;
}

- (void) myLockFocusInRect
{
  NSRect wrect;

  NSAssert(_window != nil, NSInternalInconsistencyException);

  wrect = [self convertRect: [self bounds] toView: nil];
  NSDebugLLog(@"NSView", @"Displaying rect \n\t%@\n\t window %p", 
	      NSStringFromRect(wrect), _window);
  [_window setGLFrameInRect: wrect];
}

- (void) drawRect: (NSRect) _theRect
{
  NSRect theRect = [self frame];
  float  aspect = theRect.size.width / theRect.size.height;
  GLfloat light_position[] = { -1, 1, 2, 0};

  glColor3f(0.8, 0.5, 0.7);

  glBegin(GL_LINE_LOOP);
  glVertex2f(2,2);
  glVertex2f(theRect.size.width-2, 2);
  glVertex2f(theRect.size.width-2,theRect.size.height-2);
  glVertex2f(2,theRect.size.height-2);
  glEnd();

  if( ennemi == nil )
    return;
  [self myLockFocusInRect];

  glDepthMask(GL_TRUE);
  glClear(GL_DEPTH_BUFFER_BIT);

  //glShadeModel(GL_FLAT);
  glEnable(GL_DEPTH_TEST);
  //  glEnable(GL_CULL_FACE);

  

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  if( typeShipView == TypeRotation )
    gluPerspective(60, aspect, Dimension(ennemi), 10*Dimension(ennemi));
  else
    glOrtho(-aspect * scaleFactor*Dimension(ennemi), 
	    aspect*scaleFactor*Dimension(ennemi),
	    0-scaleFactor*Dimension(ennemi), scaleFactor*Dimension(ennemi), 
	    Dimension(ennemi), 10*Dimension(ennemi));

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
  switch( typeShipView )
    {
    case TypeXY:
      gluLookAt( 0, 0, -3*Dimension(ennemi), 
	     0, 0, 0, 
	     0, 1, 0);
      break;
    case TypeYZ:
      gluLookAt(-3*Dimension(ennemi), 0, 0,
	     0, 0, 0, 
	     0, 0, 1);
      break;
    case TypeRotation:
      gluLookAt( -3*Dimension(ennemi), 0, Dimension(ennemi),
	     0, 0, 0, 
	     0, 1, 0);
      break;
    case TypeZX:
      gluLookAt( 0, 3*Dimension(ennemi), 0,
	     0, 0, 0, 
	     0, 0, 1);
      break;
    }


  if( typeShipView == TypeRotation )
    {
      angle += realDeltaTime*50;
      glRotatef(angle, sin(angle/50), cos(angle/50), 0);
    }

  [ennemi drawWithCameraNoTransform: NULL
	  from: nil];

  glDisable(GL_LIGHTING);

  [_window initWindowGL];

}

- (void) keyDown: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_t:
      [self circulateTypeProjection];
      break;
    default:
      [super keyDown: pev];
      break;
    }
}

- (void) mouseDown: (Event *) pev
{
  double a, b;
  NSPoint pt = NSMakePoint(pev->button.x, pev->button.y);
  NSPoint res = [self convertPoint: pt
		      fromView: nil];
  NSRect frame = [self frame];
  
  if( typeShipView == TypeRotation )
    return;

  a = 2*(res.x - frame.size.width/2.0) *
    Dimension(ennemi)*scaleFactor / frame.size.height;
  b = 2*(res.y - frame.size.height/2.0) / frame.size.height*
    Dimension(ennemi)*scaleFactor;

  NSLog(@"(%g, %g)", a, b);
}

- (void) circulateTypeProjection
{
  switch(typeShipView)
    {
    case TypeRotation:
      typeShipView = TypeXY;
      [typeProjection updateString: @"XY"];
      break;
    case TypeXY:
      typeShipView = TypeYZ;
      [typeProjection updateString: @"YZ"];
      break;
    case TypeYZ:
      typeShipView = TypeZX;
      [typeProjection updateString: @"ZX"];
      break;
    case TypeZX:
      typeShipView = TypeRotation;
      [typeProjection updateString: @"Rotation"];
      break;
    }
}

@end

