/*
 *  Sound.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <math.h>

#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSTask.h>
#include <Foundation/NSFileHandle.h>
#include <Foundation/NSException.h>
#include <Foundation/NSThread.h>
#include <Foundation/NSLock.h>

#include "Sound.h"
#include "Metaobj.h"

#define DEV "/dev/dsp"
#define VIT 0.0001

Sound *theSound=nil;

@implementation Sound
- (void) checkCapability
{
  if( ioctl(fd, SNDCTL_DSP_GETCAPS, &caps) == -1 )
    {
      perror("check cap failed");
      return;
    }
  
  if( caps & DSP_CAP_DUPLEX )
    {
      printf("%s has full duplex !!\n", DEV);
    }
  if( caps & DSP_CAP_REALTIME )
    {
      printf("%s has real time\n", DEV);
    }
  if( caps & DSP_CAP_BATCH )
    {
      printf("%s has batch cap\n", DEV);
    }
  if( caps & DSP_CAP_TRIGGER )
    {
      printf("%s has trigger\n", DEV);
    }
  if( caps & DSP_CAP_MMAP )
    {
      printf("%s can memmap !!\n", DEV);
    }
}

- (void) _setivar
{
  fd = -1;
  buffer = NULL;
  object = nil;
  toWrite = 0;
  shouldEnd = NO;
  running = NO;
  lock = nil;
}

- init
{
  [super init];
  [self _setivar];
  //  fd = open(DEV, O_WRONLY | O_NONBLOCK);
  fd = open(DEV, O_WRONLY);
  if( fd == -1 )
    {
      perror("open dsp failed");
      RELEASE(self);
      return nil;
    }

  {
    int arg = 0x00030006;
    if( ioctl( fd, SNDCTL_DSP_SETFRAGMENT, &arg ) )
      {
	perror("set frag size failed");
      }
  }

  [self checkCapability];
  format = AFMT_S16_LE;
  if( ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1 )
    {
      perror("format dsp failed");
      RELEASE(self);
      return nil;
    }
  if( format != AFMT_S16_LE )
    {
      NSWarnMLog(@"not AFMT_U8 : %d", format);
    }

  channel = 2;
  if( ioctl(fd, SNDCTL_DSP_CHANNELS, &channel) == -1 )
    {
      perror("channel set failed");
      RELEASE(self);
      return nil;
    }
  if( channel != 2 )
    {
      NSWarnMLog(@"bad channel  : %d", format);
    }

  speed = 44000;

  if( ioctl(fd, SNDCTL_DSP_SPEED, &speed) == -1 )
    {
      perror("speed set failed");
      RELEASE(self);
      return nil;
    }
    {
      NSWarnMLog(@"speed  : %d", speed);
    }

	
  if( ioctl( fd, SNDCTL_DSP_GETBLKSIZE, &fragSize) == -1 )
    {
      perror("frag size get failed");
      RELEASE(self);
      return nil;
    }
    {
      signed short *pbuf;
      int i;
      fragSize *= 32;
      sizeBuffer = fragSize;
      NSWarnMLog(@"fragment size is  %d bytes", fragSize);
      buffer = NSZoneMalloc([self zone], fragSize);
      pbuf = (signed short *)buffer;
      for( i = 0; i < fragSize/sizeof(signed short); ++i )
	pbuf[i] = 0;
    }

  theSound = self;
  addToCheck(self);

  lock = [NSLock new];

  [NSThread detachNewThreadSelector: @selector(runSound:) 
	    toTarget: self 
	    withObject: nil];
  return self;
}

- (void) close
{
  if( fd >= 0)
    {
      close(fd);
      fd = -1;
    }
}

- (void) runSound: obj
{
  running = YES;
  while(!shouldEnd)
    {
      NSAutoreleasePool *pool = [NSAutoreleasePool new];
      [self runSample];
      RELEASE(pool);
    }
  running = NO;
}

- (BOOL) running
{
  return running;
}

- (void) runSample
{
  //  audio_buf_info info;
  int i;
  NSData *data;
  int lim;

  static int count = 0;
  if( count ++ >= 0 )
    {
      count = 0;
    }
  else
    return;

//    if( ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1 )
//      {
//        perror("get ospace failed");
//        return;
//      }
//    if( info.fragments >= 2 )
    {
      ssize_t res;
      res = write(fd, buffer+toWrite, sizeBuffer-toWrite);
      if( res == -1 && errno != EAGAIN)
	{
	  perror("write failed");
	  exit(1);
	}
      if( res < sizeBuffer - toWrite )
	{
	  toWrite += res;
	}
      else
	{
	  toWrite = 0;
	  [lock lock];
	  if ( object )
	    {
	      data = [object dataWithSize: fragSize];
	      if( data == nil )
		{
		  NSLog(@"destruction %@", object);
		  DESTROY(object);
		  lim = 0;
		  sizeBuffer = fragSize;
		}
	      else
		{
		  lim = [data length];
		  NSAssert(lim <= fragSize, @"lim > fragSize");
		  [data getBytes: buffer];
		  sizeBuffer = lim;
		}
	      for( i = lim; i < fragSize; ++ i)
		{
		  buffer[i] = 0;
		}
	    }
	  [lock unlock];
	}
    }
}

- (void) playFile: (NSString *)file
{
  NSString *ext;
  [lock lock];
  DESTROY(object);
  ext = [file pathExtension];
  if( [ext isEqualToString: @"mp3"] )
    object = [Mpeg newWithFile: file];
  else
    object = [FileSample newWithFile: file];
  [lock unlock];
}

- (void) dealloc
{
  NSLog(@"Sound Released");
  [self close];
  RELEASE(object);
  RELEASE(lock);
  if( buffer != NULL )
    NSZoneFree([self zone], buffer);
  [super dealloc];
  theSound = nil;
}

- (void) endSound
{
  shouldEnd = YES;
  while(running);
}

@end


@implementation FileSample
- (void) _setivar
{
  file = nil;
}

- initWithFileName: (NSString *) name
{
  [super init];

  [self _setivar];

  file = [NSFileHandle  fileHandleForReadingAtPath: name];
  RETAIN(file);
   if( file == nil )
    {
      NSWarnMLog(@"unable to open %@", name);
      RELEASE(self);
      return nil;
    }
  addToCheck(self);
  return self;
}

+ newWithFile: (NSString *) name
{
  FileSample *f;
  
  f = [self alloc];
  return [f initWithFileName: name];
}

- (void) dealloc
{
  [self closeAndRelease];
  [super dealloc];
}

- (NSData *) dataWithSize: (int) val
{
  NSData *data = nil; 
  if( file != nil )
    {
      data = [file readDataOfLength: val];
      if( [data length] == 0 )
	{
	  [self closeAndRelease];
	  return nil;
	}
    }
  return data;
}


- (void) closeAndRelease
{
  DESTROY(file);
}


@end

@implementation Mpeg
- (void) _Mpegsetivar
{
  theTask = nil;
  readHandle = nil;
}

- initWithFileName: (NSString *) name
{
  NSPipe *pipeOut;
//    NSArray *args = [NSArray arrayWithObjects: @"-s", @"-r", @"16000", @"--mono",
//    			   name, nil];
  NSArray *args = [NSArray arrayWithObjects: @"-s", 
			   name, nil];
  [super init];

  [self _Mpegsetivar];

  theTask = [NSTask new];
  pipeOut = [NSPipe new];

  [theTask setStandardOutput: pipeOut];
  [theTask setLaunchPath: @"mpg123"];
  [theTask setArguments: args];
  

  readHandle = RETAIN([pipeOut fileHandleForReading]);

  RELEASE(pipeOut);

  [theTask launch];

  addToCheck(self);

  return self;
}

+ newWithFile: (NSString *) name
{
  Mpeg *f;
  
  f = [self alloc];
  return [f initWithFileName: name];
}

- (NSData *) dataWithSize: (int) val
{
  NSData *data;
  if( theTask )
    {
//        const signed short *samp;
//        unsigned char *sampChar;
      unsigned length;
//        int i, j;
      //      NSMutableData *result;
      data = [readHandle readDataOfLength: 4*val];
      
      if( (length = [data length]) == 0)
	{
	  [self closeAndRelease];
	  return nil;
	}
      
      return data;
//        samp = (const unsigned short *)[data bytes];
//        result = [NSMutableData dataWithLength: length/4];

//        val = length/4;


//        sampChar = (unsigned char *)[result mutableBytes];

//        for( i = 0, j=0; i < val; ++i, j+=2 )
//  	{
//  	  sampChar[i] = (((long)samp[j]*128)/32768)+128;
//  	}
//        return result;
    }
  else
    return nil;
}

- (void) closeAndRelease
{
  if( theTask != nil )
    {
      [theTask terminate];
      DESTROY(theTask);
    }
  DESTROY(readHandle);
}

- (void) dealloc
{
  [self closeAndRelease];
  [super dealloc];
}
@end
