/*
 *  StarView.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: May 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>

#include "GGRepere.h"
#include "GGLoop.h"
#include "Metaobj.h"
#include "Tableau.h"
#include "Sched.h"
#include "World.h"
#include "GGWidget.h"
#include "GGShip.h"
#include "utile.h"
#include "GGInput.h"
#include "Item.h"
#include "Commandant.h"
#include "GGFont.h"
#include "ShipWin.h"
#include "GGConcreteShip.h"
#include "StarView.h"
#include "Console.h"
#include "System.h"

static const double foc = 60;

static double setColor(float mag)
{
  double val;
  if( mag < 0.1 )
    return 1.1-mag;
  val = (2/mag);
  val = (val < 0 ? 0 : val);
  if( val >= 0.6 )
    val =  sqrt(val);
  //  return val;
  return (val > 1 ? 1 : val);
}

@interface StarView (Private)
- (void) _closestChange: (int) indClosest;
@end

@implementation StarView
- (id) initWithFrame: (NSRect)frameRect
{
  GGSky *sky;
  Star *stars;
  char *names;
  int n, i, index;
  int indexSystemCommandant;
  int indexLocalSystemCommandant=0;
  double xc, yc, zc;

  [super initWithFrame: frameRect];
  [self enableTextureBackground];
  
  sky 		= [theWorld sky];
  stars 	= [sky stars: &n];
  names 	= [sky names];
  indexSystemCommandant = [[theWorld commandant] 
			    indexSystem];

  range = [[theWorld commandant] rangeForHyperJump];

  NSParameterAssert(indexSystemCommandant>=0);
  xc = stars[indexSystemCommandant].xd;
  yc = stars[indexSystemCommandant].yd;
  zc = stars[indexSystemCommandant].zd;

  initVect(&posCommandant, xc, yc, zc);

  fontStar 	= RETAIN([GGFont fontWithFile: DefaultFontName
				 atSize: 18]);

  fontAxe	= RETAIN([GGFont fontWithFile: DefaultFontName
				 atSize: 12]);

  [GGFont setDefaultFont: fontAxe];

  axex = RETAIN([GGGlyphLayout glyphLayoutWithString: @"[Ox)"]);
  axey = RETAIN([GGGlyphLayout glyphLayoutWithString: @"[Oy)"]);
  axez = RETAIN([GGGlyphLayout glyphLayoutWithString: @"[Oz)"]);

  for( i = 0, index = 0; i < n && index < MAXLOCALSTAR; ++i)
    {
      Star s = stars[i];
      if( square(s.xd-xc) + square(s.yd-yc) + square(s.zd-zc) <= square(3000) )
	{
	  initVect(&s.pos, s.xd, s.yd, s.zd);
	  //	  s.pos.y  *= -1;
	  s.intensity = setColor(absoluteMagnitude(&s));
	  s.stringLayout = RETAIN([GGGlyphLayout glyphLayoutWithString: 
						   [NSString stringWithCString:
							       names+s.index]
						 withFont: fontStar]);
	  theStars[index] = s;
	  if( i == indexSystemCommandant )
	    indexLocalSystemCommandant = index;
	  index++;
	}
    }

  nStars = index;
  angle = 0;

  data = gluNewQuadric();
  gluQuadricNormals(data, GLU_SMOOTH);
  gluQuadricOrientation(data, GLU_OUTSIDE);
  gluQuadricTexture(data, GL_FALSE);
  compileListId = glGenLists(1);
  glNewList(compileListId, GL_COMPILE);
  gluSphere(data, 50, 10, 10);
  glEndList();

  [self centerView];

  
  if( i < n )
    NSLog(@"problem, too much local star");

  [self setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];

  [theLoop fastSchedule: self
	   withSelector: @selector(updateView)
	   arg: nil];

  textview = [[Console alloc] initWithFrame: 
				NSMakeRect(0, 0, 0.3*frameRect.size.width,
					   0.3*frameRect.size.height)];
  [textview setTransluentBackground: YES];
  [self addSubview: textview];

  [self _closestChange: indexLocalSystemCommandant];

  [[self titleView]
    updateString: @"Neighbor stars"];
  return self;
}

+ starView
{
  StarView *sv;
  sv = [[self alloc] init];
  return AUTORELEASE(sv);
}

- (void) dealloc
{
  int i;
  [theLoop unSchedule: self];
  for(i = 0; i < nStars; ++i)
    {
      RELEASE(theStars[i].stringLayout);
    }
  RELEASE(axex);
  RELEASE(axey);
  RELEASE(axez);
  RELEASE(textview);
  

  glDeleteLists(compileListId, 1);
  gluDeleteQuadric(data);
  RELEASE(fontStar);
  RELEASE(fontAxe);
  [super dealloc];
}

- (void) open
{
  [super open];
  [_window makeFirstResponder: self];
}

- (void) _closestChange: (int) indClosest
{
  real newDist;
  closest = indClosest;
  [[theWorld commandant] 
    setWishSystemIndex: theStars[indClosest].indexStar];

  newDist = sqrt(distance2Vect(&posCommandant, &theStars[indClosest].pos))/100;

  [textview clear];
  [textview log: @"range: %g light-years", range];
  [textview log: @"System\n@r%s@w", [[theWorld sky] cNameOf: 
	      theStars[closest].indexStar]];
  [textview log: @"economic type %@", [[System systemAt: &theStars[closest]] systemType]];
  [textview log: @"distance is %g", newDist];
  if( newDist > range )
    [textview log: @"Out of range"];
}

- (void) myLockFocusInRect
{
  NSRect wrect;

  NSAssert(_window != nil, NSInternalInconsistencyException);

  wrect = [self convertRect: [self bounds] toView: nil];
  NSDebugLLog(@"NSView", @"Displaying rect \n\t%@\n\t window %p", 
	      NSStringFromRect(wrect), _window);
  [_window setGLFrameInRect: wrect];
}

typedef struct __2dcoord
{
  Vect3D pos;
  BOOL	valid;
}_coord2d;

- (void) drawRect: (NSRect) _theRect
{
  int i;
  NSRect 		theRect 	= [self frame];
  float  		aspect 		=theRect.size.width/theRect.size.height;
  GLfloat 		light_position[] = { -1, 1, 2, 0};
  Camera 		cam;
  ProjectionParam	param;
  _coord2d		coord[MAXLOCALSTAR];
  GGMatrix4 		tempMat;
  int			indClosest=0;
  float			distClosest=0;
  float			xci, yci;
  
  xci		= 	floor(orig.x/400)*400;
  yci 		= 	floor(orig.y/400)*400;


  param.ratio = theRect.size.width/theRect.size.height;
  param.yFoc = tan((double)foc/2/180*M_PI);
  param.xFoc = param.yFoc*param.ratio;
  param.width = theRect.size.width;
  param.height = theRect.size.height;
  param.zmin = 1;
  param.zmax = 10000;
  cam.param = &param;

  glPushAttrib(GL_DEPTH_BUFFER_BIT | GL_VIEWPORT_BIT);
  glDepthFunc(GL_ALWAYS);
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  glDepthRange(1, 1);
  [super drawRect: _theRect];
  glPopAttrib();
  


//    [super drawRect: _theRect];
//    glClear (GL_DEPTH_BUFFER_BIT);


//    glClearColor(0.1, 0.1, 0.5, 1);
//    glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

  glColor3f(0.8, 0.5, 0.7);

  glBegin(GL_LINE_LOOP);
  glVertex2f(2,2);
  glVertex2f(theRect.size.width-2, 2);
  glVertex2f(theRect.size.width-2,theRect.size.height-2);
  glVertex2f(2,theRect.size.height-2);
  glEnd();

  [self myLockFocusInRect];


  //glShadeModel(GL_FLAT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);

  angle += realDeltaTime/4;

  glPushAttrib(GL_ENABLE_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  gluPerspective(foc, aspect, 1, 10000);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  transposeMatriceStrange(&modelView, &tempMat);
  glLoadMatrix((real*)&tempMat);
  glTranslatef(-modelView.position.x, -modelView.position.y,-
	       modelView.position.z);

  transposeMatrice(&modelView, &cam.modelView);


  glColor3f(0, 1, 0);
  glPushMatrix();
  glTranslatef(xci, yci, 0);
  glBegin(GL_LINES);
  for( i = 1; i < 6; ++i)
    {
      glVertex2f((-i)*400, -3000);
      glVertex2f((-i)*400, 3000);
      glVertex2f((+i)*400, -3000);
      glVertex2f((+i)*400, 3000);
      glVertex2f(-3000, (-i)*400);
      glVertex2f(3000, (-i)*400);
      glVertex2f(-3000, (i)*400);
      glVertex2f(3000, (i)*400);
    }

  glColor3f(1, 1, 0);
  glVertex2f(0, -3000);
  glVertex2f(0, 3000);
  glVertex2f(-3000, 0);
  glVertex2f(3000, 0);

  glEnd();
  glPopMatrix();
  glBegin(GL_LINES);

  glColor3f(1, 0, 0);
  for(i = 0; i < nStars; ++i)
    {
      glVertex2fv((real*)&theStars[i].pos);
      glVertex3fv((real*)&theStars[i].pos);
    }
  glEnd();


  glColor3f(0.7, 0.7, 0.7);
  glEnable(GL_LIGHTING);
  glEnable(GL_RESCALE_NORMAL);
  if( hasMoved )
    {
      indClosest = -1;
      distClosest = 1e20;
    }
  for(i = 0; i < nStars; ++i)
    {
      if( getWinCoordinates(&cam, &theStars[i].pos, &coord[i].pos) )
	{
	  glPushMatrix();
	  glTranslatef(theStars[i].pos.x, theStars[i].pos.y, theStars[i].pos.z);
	  glScalef(theStars[i].intensity, theStars[i].intensity,
		   theStars[i].intensity);
	  glCallList(compileListId);
	  glPopMatrix();
	  coord[i].valid = YES;

	  if( hasMoved )
	    {
	      float tmp = fabs(coord[i].pos.x-theRect.size.width/2) + 
		fabs(coord[i].pos.y-theRect.size.height/2);
	      if( tmp < distClosest )
		{
		  indClosest = i;
		  distClosest = tmp;
		}
	    }
	}
      else
	coord[i].valid = NO;
    }

  if( hasMoved )
    {
      hasMoved = NO;
      if( indClosest != closest )
	[self _closestChange: indClosest];
    }

  glDisable(GL_LIGHTING);
  pushStandardRep(theRect.size);
  [fontStar selectFont];
  glColor3f(0, 1, 0);
  glDisable(GL_DEPTH_TEST);
  for(i = 0; i < nStars; ++i)
    if( coord[i].valid && theStars[i].stringLayout != nil && 
	( coord[i].pos.z - theStars[i].intensity  < 0. || i == closest ))
      {
	glLoadIdentity();
	glTranslatef(coord[i].pos.x+10, coord[i].pos.y, 0);
	if( i == closest )
	  {
	    glColor3f(1, 1, 0);
	    [theStars[i].stringLayout fastDraw];
	    glColor3f(0, 1, 0);
	  }
	else
	  [theStars[i].stringLayout fastDraw];
      }

  [fontStar endSelectFont];

  glColor3f(1, 1, 1);

  [fontAxe selectFont];
  draw3DstringFast(&cam, 1000, 0, 0, axex);
  draw3DstringFast(&cam, 0, 1000, 0, axey);
  draw3DstringFast(&cam, 0, 0, 1000, axez);
  [fontAxe endSelectFont];
  popStandardRep();

  glPopAttrib();

  [_window initWindowGL];

}

- (void) centerView
{
  initVect(&orig, posCommandant.x, posCommandant.y, 0);
  distance = 2000;
  initWithIdentityMatrix(&modelView);
  initVect(&modelView.direction, 0, 0, 1);
  initVect(&modelView.haut, 0, 1, 0);
  normaliseMatrice(&modelView);
  addLambdaVect(&orig, -distance, &modelView.direction, &modelView.position);
}

- (void) keyDown: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_c:
      [self centerView];
      break;
    case SDLK_LEFT:
      state |= GGLeft;
      break;	
    case SDLK_RIGHT:    
      state |= GGRight;
      break;
    case SDLK_DOWN:
      state |= GGDown;
      break;
    case SDLK_UP:
      state |= GGUp;
      break;
    case SDLK_PAGEUP:
      state |= GGAccelere;
      break;
    case SDLK_PAGEDOWN:
      state |= GGDecelere;
      break;
    default:
      [super keyDown: pev];
      break;
    }
}

- (void) _obsHasMoved
{
  int i, indClosest;
  float min = 1e20;
  indClosest = -1;
  for( i = 0; i < nStars; ++i)
    {
      float tmp = fabs(theStars[i].pos.x - orig.x) + 
	fabs(theStars[i].pos.y - orig.y);
      if( tmp < min )
	{
	  indClosest = i;
	  min = tmp;
	}
    }
  if( indClosest != closest )
    [self _closestChange: indClosest];
}

- (void) updateView
{
  if( state )
    {
      if( state & GGLeft )
	orig.x += 1000*realDeltaTime;
      if( state & GGRight )
	orig.x -= 1000*realDeltaTime;
      if( state & GGDown )
	orig.y -= 1000*realDeltaTime;
      if( state & GGUp )
	orig.y += 1000*realDeltaTime;
      if( state & GGAccelere && distance > 500 )
	distance -= 1000*realDeltaTime;
      if( state & GGDecelere )
	distance += 1000*realDeltaTime;
      addLambdaVect(&orig, -distance, &modelView.direction, 
		    &modelView.position);
      [self _obsHasMoved];
    }
}

- (void) keyUp: (Event *)pev
{
  switch(pev->key.keysym.sym)
    {
    case SDLK_LEFT:
      state &= ~GGLeft;
      break;	
    case SDLK_RIGHT:    
      state &= ~GGRight;
      break;
    case SDLK_DOWN:
      state &= ~GGDown;
      break;
    case SDLK_UP:
      state &= ~GGUp;
      break;
    case SDLK_PAGEUP:
      state &= ~GGAccelere;
      break;
    case SDLK_PAGEDOWN:
      state &= ~GGDecelere;
      break;
    default:
      [super keyDown: pev];
      break;
    }
}


- (void) mouseDown: (Event *) pev
{
  if( pev->button.button == SDL_BUTTON_RIGHT )
    _drag = YES;
}

- (void) mouseUp: (Event *)pev
{
  if( pev->button.button == SDL_BUTTON_RIGHT )
    _drag = NO;
}


- (void) mouseMoved: (Event *) pev
{
  if( _drag )
    {
      int dx = pev->motion.xrel;
      int dy = pev->motion.yrel;
      //      NSLog(@"dx = %d, dy = %d", dx, dy);
//        addLambdaVect(&modelView.direction, -0.005*dx, &modelView.droite,
//  		    &modelView.direction);
//        addLambdaVect(&modelView.direction, -0.005*dy, &modelView.haut,
//  		    &modelView.direction);
      addLambdaVect(&modelView.haut, -0.005*dx, &modelView.gauche,
		    &modelView.haut);
      addLambdaVect(&modelView.direction, 0.005*dy, &modelView.haut,
		    &modelView.direction);
      normaliseMatrice(&modelView);
      addLambdaVect(&orig, -distance, &modelView.direction, &modelView.position);
      [self _obsHasMoved];
    }
}

@end
