/*
 *  Station.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>


#include "Station.h"
#include "GGRepere.h"
#include "GGLoop.h"
#include "Metaobj.h"
#include "Tableau.h"
#include "Texture.h"
#include "Sched.h"
#include "World.h"
#include "NSWindow.h"
#include "GGWidget.h"
#include "GGShip.h"
#include "GGConcreteShip.h"
#include "utile.h"
#include "parseASC.h"
#include "GGMenu.h"
#include "Item.h"
#include "Commandant.h"
#include "Goodies.h"
#include "GGFont.h"
#include "ShipWin.h"
#include "GGInput.h"
#include "System.h"
#include "Console.h"
#include "Resource.h"

#define SQ(x) ((x)*(x))


@interface StationWindow : TopView
{
  Station *station;
}
+ stationView: (Station *)s;
@end


@interface NewShipWindow: TopView
{
  Station 		*station;
  ShipView		*shipView;
  Console		*details;
  int			current;
}
+ shipViewWith: (Station *) s;
@end

@interface Station (PrivateMethod)
- (void) endSession;
@end

@implementation Station
- (void) calculeTailleSysteme
{
  tailleSysteme = 10000e3;
}

- init
{
  [super init];
  distanceLocal2 = SQ(100e3);
  distanceGlobal2 = SQ(500e3);
  dimension = 1e3;
  masse = 10e6;

  return self;
}

- initWithDictionary: (NSDictionary *)dic
		name: (NSString *)_name
  /* We know that [super initWithDictionary] will call [self init]
   */
{
  [super initWithDictionary: dic
	 name: _name];

  modelName = RETAIN([dic objectForKey:@"Modele"]);
  alternateModelName = RETAIN([dic objectForKey:@"AlternateModele"]);
  ASSIGN(dico, dic);
  return self;
}



- (void) dealloc
{
  RELEASE(visiteur);
  RELEASE(aModel);
  RELEASE(modelName);
  RELEASE(alternateModelName);
  RELEASE(dico);
  [super dealloc];
}

- (id) copyWithZone: (NSZone *)zone
{
  Station *copy = [super copyWithZone: zone];
  copy->visiteur = nil;
  RETAIN(copy->modelName);
  RETAIN(copy->alternateModelName);
  RETAIN(copy->aModel);
  RETAIN(copy->dico);
  return copy;
}

- (BOOL) manageCollision
{
  return YES;
}

- (void)  createModel
{
  if( modelName != nil )
    {
      ModelObject *m = [ModelObject modelWithFileName: modelName
				    withDictionary: dico];
      ASSIGN(aModel, m);
      //      NSDebugMLLog(@"Station", @"path = %@", path);
      NSDebugMLLog(@"dico", @"modele = %@", modelName);
    }

  if( aModel == nil && alternateModelName != nil )
    {
      ModelObject *m = [ModelObject modelWithFileName: alternateModelName
				    withDictionary: dico];
      ASSIGN(aModel, m);
      //      NSDebugMLLog(@"Station", @"path = %@", path);
      NSDebugMLLog(@"dico", @"modele = %@", alternateModelName);
    }
  [self invalidCache];
}


- (void) collisionOccursWith: (GGSpaceObject *) other
{
  if( other == (GGSpaceObject *)[theWorld commandant] )
    {
      GGShip *ship = (GGShip *)other;
      [ship stopEngine];
      [ship centerObject];
      [ship notify: @"wellcome on %@", [self name]];
      ASSIGN(visiteur, ship);
      [theGGInput setInStation: self];
      [theWorld setCheckCollision: NO];
    }
}

- (void) openView
{
  StationWindow *sw;
  sw = [StationWindow stationView: self];
  [sw open];
}

- (void) endSession
{
  mulScalVect(Direction(self), 2*dimension, Point(visiteur));
  mulScalVect(Direction(self), -1, Direction(visiteur));
  *Haut(visiteur) = *Haut(self);
  [visiteur normalise];
  
  mulScalVect(Direction(self), 10, Speed(visiteur));
  DESTROY(visiteur);
  [theGGInput setInStation: nil];
  [theWorld setCheckCollision: YES];
}

- (void) cameraBeginBeingClose
{
  [self createModel];
}

- (void) cameraStopBeingClose
{
  DESTROY(aModel);
  [self invalidCache];
}

- (void) beginBeingSuperLocal
{
  [super beginBeingSuperLocal];
  [[manager repere] translateNodeTo: self];
  [[manager repere] reparent: self];
}

- (void) endBeingSuperLocal
{
  [repereTranslation reparent: self];
  [super endBeingSuperLocal];
  //  [repereTranslation reparent: [theWorld repere]];
}

- (void) rasterise
{
  [aModel rasterise];
}


- (NSArray *) shipMaterialToSel
{
  return [[self surroundingSystem] shipMaterialToSel];
}

- (NSArray *) goodiesToSel
{
  return [[self surroundingSystem] goodiesToSel];
}

@end

@implementation StationWindow
- __initStationWith: (Station *)s
{
  GGMenu *m;

  station = s;
  //  [self setOpacity: 1.0];

  [[self titleView]
    updateString: [NSString stringWithFormat: @"Welcome to %@",
			    s]];

  [self enableTextureBackground];

  [GGFont setDefaultFont: [GGFont fontWithFile: DefaultFontName
				  atSize: 18]];
  m = [GGMenu menu];
  [m setFrameOrigin: NSMakePoint(300, 50)];

  [m addItemWithTitle: @"Stockmarket"
     target: self
     action: @selector(openBourseWindow)];

  [m addItemWithTitle: @"Shipyard\n[You may upgrade you ship, contact the local police]"
     target: self
     action: @selector(openShipYard)];

  [m addItemWithTitle: @"Launch Request"
     target: self
     action: @selector(launch)];

  [m endItem];
  [self  addSubview: m];
  [self setState: GUI_StationView];

  //  [self addCloseButton];

  return self;
}

+ stationView: (Station *) s
{
  StationWindow *sw;
  sw = (StationWindow *)[self new];
  AUTORELEASE(sw);
  return [sw __initStationWith: s];
}

// - (void) open
// {
//   //  NSDebugMLLog(@"Station", @"window below %@", windowBelow);
//   //  [theWorld setDrawWorld: NO];
//   [super open];
// }

// - (void) close
// {
//   [super close];
//   NSDebugMLLog(@"Station", @"closing");
// }

- (void) launch
{
  [station endSession];
  station = nil;
  [self close];
}

- (void) openChantier
{
  MarketPlace *place;
  NSArray *goodies;

  TopView *win = [TopView topview];
  [[win titleView]
    updateString: @"Shipyard"];

  [win enableTextureBackground];

  goodies = [station shipMaterialToSel];

  place = [[MarketPlace alloc] initWithGoodies: goodies];

  [win addSubview: place];
  RELEASE(place);
  //  [theGGInput addTopView: win];
  [win setState: GUI_StationView];
  [win open];
}

- (void) openShipYard
{
  GGMenu *m;
  TopView *win = [TopView topview];
  [[win titleView] updateString: [NSString stringWithFormat: @"%@ Shipyard",
					   station]];
  [win enableTextureBackground];

  //  [self setOpacity: 1.0];

  [[self titleView] updateString: @"Shipyard"];
  [self enableTextureBackground];

  m = [GGMenu menu];
  [m setFrameOrigin: NSMakePoint(300, 50)];

  [m addItemWithTitle: @"Upgrade\n[Buy ship equipment]"
     target: self
     action: @selector(openChantier)];
  [m addItemWithTitle: @"Repair and Servicing"
     target: self
     action: @selector(nothing)];
  [m addItemWithTitle: @"New and Reconditioned ships\n[Ships part exchanged]"
     target: self
     action: @selector(openNewShipWindow)];

  [m endItem];
  [win  addSubview: m];
  
  [win setState: GUI_StationView];
  [win open];
}

- (void) nothing
{
}

- (void) openNewShipWindow
{
  TopView *win = [NewShipWindow shipViewWith: station];
  //  [theGGInput addTopView: win];
  [win open];
}

- (void) openBourseWindow
{
  MarketPlace *place;
  NSArray *goodies;

  TopView *win = [TopView topview];
  [[win titleView]
    updateString: @"Bourse"];
  [win enableTextureBackground];

  goodies = [station goodiesToSel];

  place = [[MarketPlace alloc] initWithGoodies: goodies];
  [win addSubview: place];
  RELEASE(place);

  [win setState: GUI_StationView];
  [win open];
  //  [theGGInput addTopView: win];
}

- (void) dealloc
{
  NSDebugMLLog(@"Station", @"deallocing %@", self);
  [super dealloc];
}



@end

@implementation NewShipWindow
- initNewShipViewWith: (Station *)s
{
  int i, n;
  NSRect shipFrame = {{ 20, 20 }, {200, 200}};
  NSRect detailsFrame = {{ 20, 250 }, {300, 210}};
  NSSize aSize;
  NSArray *ships;

  [super init];
  [[self titleView]
    updateString: @"Choose a ship to buy"];
  [self enableTextureBackground];
  aSize = [self frame].size;
  station = s;
  current = -1;

  ships = [theResourceManager ships];
  n = [ships count];
  for( i = 0; i < n; ++i)
    {
      GGOption *option = [GGOption optionWithStr: [ships objectAtIndex: i]];
      [option setFrameOrigin: NSMakePoint(aSize.width-200, aSize.height - 40-
					  (i*20))];
      [option setTarget: self];
      [option setAction: @selector(chooseShip:)];
      [option setTag: i];
      [self addSubview: option];
    }


  shipView = [[ShipView alloc]
	       initWithFrame: shipFrame];
  [self addSubview: shipView];

  details = [[Console alloc] 
	      initWithFrame: detailsFrame];
  //  [details setTransluentBackground: YES];
  [self addSubview: details];
  [self setState: GUI_StationView];
  
  return self;
}

- (void) chooseShip: (GGBouton *)bo
{
  NSArray *ships = [theResourceManager ships];
  int n = [ships count];
  int i = [bo tag];
  int j;
  GGShip *ship;

  NSParameterAssert(0<= i && i < n);
  if( current == i )
    {
      [bo setState: YES];
      return;
    }
  
  for( j = 0; j < n; ++j )
    {
      if( i == j)
	continue;
      [bo setState: NO];
    }
  ship = [GGShip shipWithFileName: [ships objectAtIndex: i]];
  [shipView setShip: ship];
  [details clear];
  [details log:@"nom du vaisseau %@", ship];
  [details log:@"masse coque : %.5g t", ship->masseCoque/1000.0];
  [details log:@"Capacity : %.5g t", ship->capaciteFret/1000.0];
  [details log:@"Hyper Drive : type %d", ship->hyperDriveType];
  [details log:@"Manufacturer :  %@", 
	   [ship->property objectForKey: @"Manufacturer"]];
  [details log:@"Designer :  %@", 
	   [ship->property objectForKey: @"Designer"]];
  [details log:@"Crew :  %d", 
	   [[ship->property objectForKey: @"Crew"] intValue]];
  [details log:@"Gun Mounting :  %d", ship->pylonLaser];
  [details log:@"Missile Pylons :  %d", ship->pylonRocket];
  [details log:@"MAIN THRUSTER ACCELERATION :  %.2g g", ship->gAcc/10.0];
  [details log:@"Retro THRUSTER ACCELERATION :  %.2g g", ship->gFrein/10.0];
}

+ shipViewWith: (Station *) s
{
  NewShipWindow *sw;
  sw = (NewShipWindow *)[self alloc];
  return AUTORELEASE([sw initNewShipViewWith: s]);
}

- (void) dealloc
{
  RELEASE(shipView);
  RELEASE(details);
  [super dealloc];
}

- (void) close
{
  station = nil;
  [super close];
}


@end


  
