/*
 *  System.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: May 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSObject.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>

#include "Metaobj.h"
#include "System.h"
#include "Planete.h"
#include "Station.h"
#include "utile.h"
#include "Item.h"
#include "Sched.h"
#include "Resource.h"
#include "GGConcreteShip.h"
#include "World.h"

@implementation System
- initWithStar: (Star *)s
{
  float mag;
  mag = absoluteMagnitude(s);

  NSDebugMLLog(@"System", @"mag %g", (double) mag);
  theSystem = self;
  index = s->indexStar;

  if( s->indexStar == 0 )
    {
      systemType = Modern;
      systemSpeciality = 0;
    }
  else
    {
      srand(s->indexStar);
      systemType = (SystemType)(irnd(2)+1);
    }
  addToCheck(self);
  return self;
}

/* the method copy will keep the old theSystem, and this is good so I don't
   redefine it
*/

- copyWithZone: (NSZone *)zone
{
  NSParameterAssert(0);
  return nil;
}

- (void) dealloc
{
  [aTimer invalidate];
  RELEASE(aTimer);
  RELEASE(mainStar);
  [super dealloc];
}

+ systemAt: (Star *)s
{
  return AUTORELEASE([[self alloc] initWithStar: s]);
}

- (GGOrbiteur *) mainStar
{
  if( mainStar == nil )
    {
      if( index == 0 )
	{
	  mainStar = [[Planete alloc] initSystemWithContentOfFile: 
			 @"data/model/solsys.data"];
	}
      else
	{
	  srand(index);
	  mainStar =  [[Planete alloc]initRandomSystemForStar: index];
	}
      mainStar->surroundingSystem = self;
    }
  return mainStar;
}

- (NSArray *) shipMaterialToSel
{
  return [Item itemList];
}
//  - (NSArray *) shipMaterialToSel
//  {
//    return [NSArray arrayWithObjects:
//  		    [Item itemWithName: @"Laser Pulsion 1MW"
//  			  type: LaserPulsion
//  			  arg: 1
//  			  prixAchat: 3000
//  			  prixVente: 2000],
//  		  [Item itemWithName: @"Laser Pulsion 5MW"
//  			type: LaserPulsion
//  			arg: 5
//  			prixAchat: 15000
//  			prixVente: 10000],
//  		  [Item itemWithName: @"Laser Rayon 1MW"
//  			type: LaserRayon
//  			arg: 1
//  			prixAchat: 15000
//  			prixVente: 10000],
//  		  [Item itemWithName: @"Laser Rayon 5MW"
//  			type: LaserRayon
//  			arg: 5
//  			prixAchat: 50000
//  			prixVente: 40000],
//  		  [Item itemWithName: @"Laser Rayon 10MW"
//  			type: LaserRayon
//  			arg: 10
//  			prixAchat: 100000
//  			prixVente: 80000],
//  		  [Item itemWithName: @"Missile"
//  			type: Missile
//  			arg: 1
//  			prixAchat: 700
//  			prixVente: 500],
//  		  [Item itemWithName: @"Missile Naval"
//  			type: Missile
//  			arg: 2
//  			prixAchat: 2000
//  			prixVente: 1000],
//  		  nil];
//  }

- (NSArray *) goodiesToSel
{
  return [NSArray arrayWithObjects:
		    [Item itemWithName: @"Slaves"
			  type: Slaves
			  arg: 0
			  prixAchat: 1000
			  prixVente: 800],
		    [Item itemWithName: @"Gems"
			  type: Gems
			  arg: 0
			  prixAchat: 3000
			  prixVente: 2000],
		    [Item itemWithName: @"Hydrogen"
			  type: Hydrogen
			  arg: 0
			  prixAchat: 10
			  prixVente: 8],
		    [Item itemWithName: @"Computers"
			  type: Computers
			  arg: 0
			  prixAchat: 1000
			  prixVente: 800],
		    [Item itemWithName: @"Foods"
			  type: Foods
			  arg: 0
			  prixAchat: 200
			  prixVente: 100],
		    [Item itemWithName: @"Animals"
			  type: Animals
			  arg: 0
			  prixAchat: 800
			  prixVente: 600],
		    [Item itemWithName: @"narcotics"
			  type: Narcotics
			  arg: 0
			  prixAchat: 4000
			  prixVente: 3000],
		  nil];
}

- (NSString *) systemType
{
  switch(systemType)
    {
    case Modern: return @"Modern";
    case Farmer: return @"Farmer";
    case Intermediate: return @"Intermediate";
    }
  return nil;
}

- (int) index
{
  return index;
}

- (Vect3D) randomVect
{
  Vect3D res;
  real ts;
  ts = [[self mainStar] 
	 tailleSysteme]*0.1;

  res.x = ts*(vrnd()-0.5);
  res.y = ts*(vrnd()-0.5);
  res.z = ts*(vrnd()-0.5);

  return res;
}

- (void) createAShip: (GGTimer *)ggtimer
{
  NSArray *shipList =   [theResourceManager ships];
  Vect3D initPos;
  GGEnnemi *aShip = [GGEnnemi shipWithFileName: 
				[shipList objectAtIndex: irnd([shipList count])]];
  initPos = [self randomVect];
  [aShip setPosition: &initPos];
  [theWorld addShip: aShip];
  [theWorld addSubNode: aShip];
  [aShip scheduleAsGlobal];
  [aShip setDefaultIA];
  NSDebugMLLog(@"IA", @"blabla createAShip : %@", aShip);
}

- (void) makeCurrent
{
  aTimer = RETAIN([GGTimer scheduledTimerWithTimeInterval: 1e4
			   target: self
			   selector: @selector(createAShip:)
			   repeats: NO]);
}

- (void) endCurrent
{
  [aTimer invalidate];
  DESTROY(aTimer);
}

- (NSArray *) allStations
{
  id obj;
  NSEnumerator *enumerator;
  NSMutableArray *res;
  NSArray *allOrb;
  GGOrbiteur *sol;
  sol = [self mainStar];
  res = [NSMutableArray array];

  allOrb = [sol getAllOrbiteur];
  enumerator = [allOrb objectEnumerator];

  
  while( (obj = [enumerator nextObject]) != nil )
    {
      if( [obj isKindOfClass: [Station class]] )
	[res addObject: obj];
    }
  
  return res;
}

@end
