/*
 *  Tableau.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <Foundation/NSObject.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSException.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSScanner.h>
#include <Foundation/NSCalendarDate.h>
#include "Metaobj.h"
#include "GGLoop.h"
#include "GGSpaceObject.h"
#include "GGShip.h"
#include "GGRepere.h"
#include "utile.h"
#include "Tableau.h"
#include "Preference.h"
#include "GGOrbiteur.h"
#include "Commandant.h"
#include "GGFont.h"
#include "Radar.h"
#include "Console.h"
#include "GGInput.h"
#include "World.h"
#include "Sched.h"
#include "GGIcon.h"
#include "GGWidget.h"

GGTableau *theTableau=nil;

#define  xOriginePower (15)
#define  yOriginePower (10+HauteurTableauBord)

void drawPower(float power)
{
  int i, incy, larg;

  static float effCoef = 0.0;
  static float elapsedTime = 0.0;
  float coef = 1.0;

  larg = 10;

  if( elapsedTime >= 0.1 )
    {
      elapsedTime = 0.0;
      if( effCoef + 0.1 <= power )
	effCoef += 0.1;
      else if( effCoef >= power )
	effCoef -= 0.1;
    }

  elapsedTime += deltaTime;

  for( i = 0, incy = 0; i < 9; ++ i, incy += 6)
    {
      float val = (float)(i)/10;
      if( val >= effCoef )
	{
	  coef = 0.4;
	}
      glColor3f(coef * val, coef * val/2, coef * (0.5-val/2));
      glRecti(xOriginePower, yOriginePower+incy, xOriginePower+larg, 
	      yOriginePower+incy+4);
      larg += 3*i;
    }
}


@implementation GGTableau
- init
{
  [super init];
  observeur = nil;
  closestOrbiteur = nil;
  theTableau = self;
  log = nil;
  posCibleCorrect = NO;
  posDestCorrect = NO;
  posVitCorrect = NO;
  posDirCorrect = NO;
  posVitDestCorrect = NO;
  joyXPos = 0.0;
  joyYPos = 0.0;
  string = [StringArray new];
  font = RETAIN([GGFont fontWithFile: DefaultFontName
			atSize: 10]);
  addToCheck(self);
  return self;
}

- (void) dealloc
{
  RELEASE(font);
  RELEASE(closestOrbiteur);
  RELEASE(log);
  RELEASE(observeur);
  RELEASE(string);
  [[NSNotificationCenter defaultCenter] removeObserver: self];
  [super dealloc];
}

- (void) setPositionCible: (int) x 
		      and: (int)y 
		     dist: (float) dist
{
  xCible = x;
  yCible = y;
  distCible = dist;
  posCibleCorrect = YES;
}

- (void) setJoyXPos: (float) val
{
  joyXPos = val;
}
- (void) setJoyYPos: (float) val
{
  joyYPos = val;
}

- (void) setClosestOrbiteur: (GGOrbiteur *)theOrb
{
  ASSIGN(closestOrbiteur, theOrb);
}

- (GGOrbiteur *) closestOrbiteur;
{
  return closestOrbiteur;
}


- (void) setObserveur: (GGSpaceObject *) obj
{
  ASSIGN(observeur, obj);
  [obj enregistre: self
       with: @selector(removeObserveurWithNotification:)];
}

- (void) removeObserveurWithNotification:  (NSNotification *) aNotification
{
  if( [aNotification object] == observeur )
    [self setObserveur: nil];
}


- (void) addLogString: (NSString *)format,...
{
  va_list ap;
  NSString *tmp;
  va_start(ap, format);
  if (format == nil)
    return;
  else
    tmp = AUTORELEASE([[NSString alloc] initWithFormat: format arguments: ap]);
  va_end(ap);
  if( log == nil )
    {
      log = [StringArray new];
    }
  [log addString: tmp];
}


- (void) setFrameRect: (NSRect) aRect
{
  frameRect = aRect;
}

- (void) drawTableau
{
  int width = frameRect.size.width;
  int height = frameRect.size.height;
  int i;
  GGShip *target = nil;
  GGSpaceObject *destination = nil;

  //  abort();
  NSDebugMLog(@"Tableau", @"w %d h %d\n", width, height);

  //  puts("caca");
  if( thePref.GLBlend )
    glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(0.4,0.4,0.4,0.5);
  glRecti(1,height-200,210, height-10);
  
  [string clear];
  if( thePref.frameRate )
    [string addString: @"@r%g fps@w", [theLoop fps]];
#ifdef DEBUG
  if( wizard )
    [string addString: @"nShip = @g%d@w", [[theWorld ships] count]];
#endif
  [string addString: @"%g X", [theLoop timeSpeed]];
  if( observeur != nil )
    {
      int modePilote = [observeur modePilote];
      target = [observeur target];
      destination = [observeur destination];
      switch(modePilote)
	{
	case Manuel:
	  [string addString: @"Manual Mode"];
	  break;
	case  AutoPilote:
	  [string addString: @"Auto Pilot"];
	  break;
	case Hinter:
	  [string addString: @"Hinter Mode"];
	  [string addString: @"wished speed : %g", 
		  [observeur speedWished]];
	  break;
	default:
	  break;
	}
      [string addString: @"fuel %g L", [observeur carburant]];
#ifdef DEVEL
      if( wizard )
	{
	  [string addString: stringOfVect2([observeur getPositionAbsolu])];
	  [string addString: @"distance %5.5e", 
		  sqrt(prodScal( Point(observeur), Point(observeur)))];
	  [string addString: [NSString stringWithFormat: @"camera speed %e",
				       sqrt(prodScal(Speed(observeur), 
						     Speed(observeur)))]];
	}
#endif
      drawPower([(GGShip *)observeur power]);
    }


  if	( target != nil )
    {
      [string addString: 
		[NSString stringWithFormat: @"target :@r%@@w", target]];

      if( [target isKindOfClass: [GGShip class]] )
	{
	  GGShip *val = (GGShip *)target;
	  [string addString: @"shield : %g, hull : %g", 
		  val->bouclier,
		  val->endurance];
	  [string addString: @"masse : %g", val->masse];
	}

      if( wizard )
	{
	  [string addString: 
		    [NSString stringWithFormat: @" pos : %@\nrel to @y%@@w", 
			      stringOfVect(Point(target)),
			      target->nodePere]];
	}

      [string addString: 
		[NSString stringWithFormat: @" distance : %.2fkm",
			  distanceAbsolu(observeur, target)/1000.0]];
    }

  if( destination != nil )
    {
      [string addString: @"destination : @g%@@b", destination];
    }

  glDisable(GL_BLEND);

  [string displayAt: 10 and: height-20];

  [log displayAt: 250 and: height-20];
  ASSIGN(log, nil);

  if( closestOrbiteur != nil )
    {
      glColor3f(0.8, 0.8, 0.8);
      
      [font drawString:
	      [NSString stringWithFormat: @"%@ : %2g km", 
			closestOrbiteur,
			distance/1000]
	    atOrigin: NSMakePoint(width-150, HauteurTableauBord+10)];
      [font drawString:
	      [NSString stringWithFormat: @"speed : %2g km/s", 
			      speed/1000]
	    atOrigin: NSMakePoint(width-150, HauteurTableauBord+20)];
      if( observeur != nil && [observeur nearPlanete] )
	{
	  double hauteur = [observeur hauteurTerrain];
	  double alt = [observeur altitude];
	  [font drawString: [NSString stringWithFormat: 
					@"hauteur terrain : %5g m", 
				      hauteur]
		atOrigin: NSMakePoint(width-150, HauteurTableauBord+30)];
	  [font drawString: [NSString stringWithFormat: @"altitude : %5g m", 
				  alt]
		atOrigin: NSMakePoint(width-150, HauteurTableauBord+40)];
	}
      
    }

  if( posCibleCorrect )
    {
      posCibleCorrect = NO;
      if( destIsTarget )
	glColor3f(1, 0, 1);
      else
	glColor3f(1, 0, 0);
      drawRectBorder(xCible-5, yCible-5, xCible+5, yCible+5);
      [font drawString: 
	      [NSString stringWithFormat: @"%3g km", distCible/1000]
	    atOrigin: NSMakePoint(xCible+10, yCible-20)];
      destIsTarget = NO;
    }
  if( posDestCorrect )
    {
      posDestCorrect = NO;
      glColor3f(0, 1, 0.4);
      [font drawString: 
	      [NSString stringWithFormat: 
				@"%.2f km", distDest/1000]
	    atOrigin: NSMakePoint(xDest, yDest-20)];
      drawRectBorder(xDest-5,yDest-5,xDest+5, yDest+5);
    }
  if( posVitCorrect )
    {
      posVitCorrect = NO;
      glColor3f(1, 1, 1);
      glBegin(GL_LINES);
      glVertex2f(xVit - 2, yVit);
      glVertex2f(xVit - 10, yVit);
      glVertex2f(xVit + 2, yVit);
      glVertex2f(xVit + 10, yVit);
      glVertex2f(xVit, yVit - 2);
      glVertex2f(xVit, yVit - 10);
      glVertex2f(xVit, yVit + 2);
      glVertex2f(xVit, yVit + 10);
      glEnd();
    }

  if( posDirCorrect )
    {
      posDirCorrect = NO;
      glColor3f(1, 1, 1);
      glBegin(GL_LINES);
      glVertex2f(xDir-10, yDir-10);
      glVertex2f(xDir+10, yDir+10);
      glVertex2f(xDir+10, yDir-10);
      glVertex2f(xDir-10, yDir+10);
      glEnd();
    }

  if( posVitDestCorrect )
    {
      posVitDestCorrect = NO;
      glColor3f(1, 0, 0);
      glBegin(GL_LINES);
      glVertex2f(xVitDest - 2, yVitDest);
      glVertex2f(xVitDest - 10, yVitDest);
      glVertex2f(xVitDest + 2, yVitDest);
      glVertex2f(xVitDest + 10, yVitDest);
      glVertex2f(xVitDest, yVitDest - 2);
      glVertex2f(xVitDest, yVitDest - 10);
      glVertex2f(xVitDest, yVitDest + 2);
      glVertex2f(xVitDest, yVitDest + 10);
      glEnd();
    }

  glColor3f(0.1, 0.6, 0.1);
  glBegin(GL_LINES);
  glVertex2f(0.4*width, 0.6*height);
  glVertex2f(0.6*width, 0.6*height);
  glVertex2f(0.6*width, 0.6*height);
  glVertex2f(0.6*width, 0.4*height);
  for( i = 0; i < 8; ++i )
    {
      float x = 0.4+ (float)i*0.2/8;
      glVertex2f(x*width, 0.6*height);
      glVertex2f(x*width, (0.59)*height);
      glVertex2f(0.6*width, x*height);
      glVertex2f(0.59*width, x*height);
    }
  glEnd();
  glBegin(GL_TRIANGLES);
  {
    float x = (0.5+joyXPos/10)*width; 
    float y = (0.5+joyYPos/10)*height; 
    
    glVertex2f(x, 0.6*height);
    glVertex2f(x-5, 0.58*height);
    glVertex2f(x+5, 0.58*height);
    glVertex2f(0.6*width, y);
    glVertex2f(0.58*width, y+5);
    glVertex2f(0.58*width, y-5);
  }
  glEnd();
}

- (GGSpaceObject *) observeur
{
  return observeur;
}

- (void) doIndicator: (ProjectionParam *)ppp
	      camera: (GGSpaceObject *) camera
{
  GGSpaceObject *destination = nil;
  GGShip *target = nil;
  Camera c;
  Vect3D result;
  [observeur rotationMatrixTo: camera
	     in: &c.modelView];
  c.param = ppp;

  if (observeur)
    {
      Vect3D temp;
      destination = [observeur destination];
      target = [observeur target];

      if( closestOrbiteur != nil )
	{
	  PetitGalileen pg;
	  [closestOrbiteur positionSpeedInRepere:  observeur
			   in: &pg];
	  speed = sqrt(prodScal(&pg.speed, &pg.speed));
	  distance = sqrt(prodScal(&pg.point, &pg.point));
	  if( speed > EPS )
	    {
	      mulScalVect(&pg.speed, -10/speed, &pg.speed);
	      if( getWinCoordinates(&c, &pg.speed, &result) )
		{
		  xVit = result.x;
		  yVit = result.y;
		  posVitCorrect = YES;
		}
	    }
	}
      
      initVect(&temp, 0, 0, 10);

      if( getWinCoordinates(&c, &temp, &result) )
	{
	  xDir = result.x;
	  yDir = result.y;
	  posDirCorrect = YES;
	}

      if (target)
	{
	  Vect3D posLoc;
	  double dist;
	  [target positionInRepere: observeur
		  in: &posLoc];
	  dist = sqrt(prodScal(&posLoc, &posLoc));
	  mulScalVect(&posLoc, 10/dist, &posLoc);
	  if( getWinCoordinates(&c, &posLoc, &result) )
	    {
	      [self setPositionCible: result.x 
		    and: result.y 
		    dist: dist];
	    }
	}

      if (destination )
	{
	  PetitGalileen pg;
	  real tmp;

	  if( destination == target )
	    {
	      destIsTarget = YES;
	    }
	  else
	    {
	      destIsTarget = NO;
	      [destination positionSpeedInRepere: observeur
			   in: &pg];
	      distDest = sqrt(prodScal(&pg.point, &pg.point));
	      //      mulScalVect(&pg.point, 10/distDest, &pg.point);
	      if( getWinCoordinates(&c, &pg.point, &result) )
		{
		  posDestCorrect = YES;
		  xDest = result.x;
		  yDest = result.y;
		}
	      if( (tmp = prodScal(&pg.speed, &pg.speed)) > EPS*EPS)
		{
		  tmp = sqrt(tmp);
		  mulScalVect(&pg.speed, -10/tmp, &pg.speed);
		  if( getWinCoordinates(&c, &pg.speed, &result) )
		    {
		      xVitDest = result.x;
		      yVitDest = result.y;
		      posVitDestCorrect = YES;
		    }
		}
	    }
	}
    }
}
@end

@implementation StringArray
+ stringArray
{
  return AUTORELEASE([self new]);
}

- init
{
  [super init];
  theString = [NSMutableString new];
  theFont = RETAIN([GGFont fontWithFile: Fixed6x12]);
  addToCheck(self);
  return self;
}

- (void) dealloc
{
  RELEASE(theString);
  RELEASE(theFont);
  [super dealloc];
}

- (void) _addString: (NSString *)tmp
{
  [theString appendString: @"\n"];
  [theString appendString: tmp];
}


- addString: (NSString *) format,...
{
  va_list ap;
  NSString *tmp;
  va_start(ap, format);
  if (format == nil)
    return self;
  else
    tmp = AUTORELEASE([[NSString alloc] initWithFormat: format arguments: ap]);
  va_end(ap);
  [self _addString: tmp];
  return self;
}


- (void)clear
{
  [theString setString: @""];
}

- (void) displayAt: (int) x and: (int) y
{
  NSPoint origin = NSMakePoint(x, y);
  glColor3f(1.0,1.0,1.0);
  [theFont drawString: theString
	   atOrigin: origin];
}

@end

@interface IconeTouche : NSObject
{}
@end

@implementation IconeTouche
@end

@implementation ViewTableau
- (void) _updateDate: (GGTimer *)timer
{
  NSDate *theDate;
  NSCalendarDate *cal;

  theDate = [NSDate dateWithTimeIntervalSinceReferenceDate: gameTime+
		    (365.25*24*3600*1500)];
  cal = [theDate dateWithCalendarFormat: @"%A %B %d %H:%M:%S %Y UT"
		 timeZone: nil];
		 
  [dateString updateString: [cal description]];
}

- init
{
  float x;
  Console *c;
  GGBouton *bouton;
  Radar *radar;
  [super initWithFrame: NSMakeRect(0, 0, WIDTH, HauteurTableauBord)];
  //  [self setBackgroundColor: GGMakeCol(0.1, 0.1, 0.5, 0.8)];
  radar = [[Radar alloc] initWithFrame: NSMakeRect(200, 0, 200, 100)];
  [self addSubview: radar];
  RELEASE(radar);
  c = [[Console alloc]
	initWithFrame: NSMakeRect(_frame.size.width-300, 0, 300, 
				  _frame.size.height-20)];

  //  [self setNextResponder: theGGInput];

  [self addSubview: c];
  [c setAutoresizingMask: NSViewMinXMargin];
  theConsole = c;
  RELEASE(c);

  dateString = RETAIN([GGGlyphLayout glyphLayoutWithString: @"blabla"
			       withFont: [GGFont fontWithFile: @"6x12.pcf"]]);
  //  [date setAlignement: TopAlignement];
  [self _updateDate: nil];

  timer = [GGTimer scheduledTimerWithTimeInterval: 1
	   target: self
	   selector: @selector(_updateDate:)
	   repeats: YES];
  RETAIN(timer);

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(loadingAGame:)
    name: GGNotificationAfterLoadingGame
    object: nil];

  x = 3;

#define addButton(icon, label) \
  bouton = [GGBouton new];\
  [bouton setIcon: [GGIcon iconWithName: icon\
			   scale: 1]];\
  [bouton setFrameOrigin: NSMakePoint(x, 3)];\
  [bouton setBorder: NO];\
  [bouton setTitle: label \
	  withFont: [GGFont defaultSmallFont]];\
  [self addSubview: bouton];\
  x += [bouton frame].size.width;\
  RELEASE(bouton)

  addButton(@"gmap0.jpg", @"F3");
  addButton(@"ship0.jpg", @"F6");
  addButton(@"cmdr0.jpg", @"F5");
  addButton(@"yard0.jpg", @"F1");



  [self setAutoresizingMask: NSViewWidthSizable];
  return self;
}

- (void) loadingAGame: obj
{
  [theConsole clear];
}


- (void) dealloc
{
  [timer invalidate];
  RELEASE(timer);
  RELEASE(dateString);
  [super dealloc];
}

- (void) drawRect: (NSRect) theRect
{
  NSRect rect = [self frame];
  glColor4f(0.2, 0.2, 0.25,0.4);
  glRectf(0,0 , NSWidth(rect), NSHeight(rect));
  glPolygonMode(GL_FRONT, GL_LINE);
  glColor3f(0.8, 0.5, 0);
  glRectf(2,2 , NSWidth(rect)-2, NSHeight(rect)-2);
  glPolygonMode(GL_FRONT, GL_FILL);
  glColor3f(0.5, 0.7, 0.7);
  //  glRecti(10, HauteurTableauBord-5, 20, HauteurTableauBord+5);
  [dateString drawAtPoint: NSMakePoint(400, 81)];

}

- (Console *) console
{
  return theConsole;
}

@end

