/*
 *  Task.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/NSArray.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSCoder.h>

#include "Task.h"
#include "Metaobj.h"
#include "GGLoop.h"
#include "Sched.h"
#include "Missile.h"
#include "Station.h"
#include "GGShip.h"
#include "World.h"
#include "utile.h"
#include "System.h"

@implementation GGTask
- init
{
  return [self initWithShip: nil];
}

//designated initializer
- initWithShip: (GGShip *)_ship
{
  [super init];
  ship = _ship;
  subTask = [NSMutableArray new];
  endTaskTimer = [GGTimer new];
  addToCheck(self);
  return self;
}

- (void) setShip: (GGShip *) aShip
{
  /*not retained because gnagnanga*/
  ship = aShip;
  [ship enregistre: self 
	with:@selector(removeWithNotification:)];
}

- (void) removeWithNotification: (NSNotification *)aNotification
{
  id obj = [aNotification object];

  if( ship == obj )
    [self invalidate];
}


- (void) dealloc
{
  [self doinvalidate];
  RELEASE(endTaskTimer);
  RELEASE(subTask);
  [super dealloc];
}

- (void) doinvalidate
{
  [fatherTask removeSubTask: self];
  [endTaskTimer invalidate];
  [MessageManager
    removeObserver: self];
  [theLoop unSchedule: self];
  //  ship = nil;
}


- (void) invalidate
{
  RETAIN(self);
  [self doinvalidate];
  AUTORELEASE(self); /*so that, invalidate does not destroy us and some 
		       instances may still be valid for the calling context
		     */
}

- (void) setFatherTask: (GGTask *) task
{
  fatherTask = task;
}

- (GGTask *)currentTask
{
  if( [subTask count] == 0 )
    return self;
  else
    return [[subTask objectAtIndex: 0] currentTask];
}

- (void)addSubTask: (GGTask *) aTask
{
  NSParameterAssert(aTask != nil && aTask->fatherTask == nil);
  [subTask addObject: aTask];
  aTask->fatherTask = self;
}

- (void) removeSubTask: (GGTask *)aTask
{
  NSParameterAssert(aTask != nil && aTask->fatherTask == self);
  aTask->fatherTask = nil;
  [subTask removeObjectIdenticalTo: aTask];
}
  
- (BOOL) shouldSave
{
  return YES;
}

- (BOOL) saveChannel
{
  return YES;
}

- (void) encodeWithCoder: (NSCoder *) encoder
{
  GGEncode(fatherTask);
  GGEncode(subTask);
  GGEncode(ship);
  GGEncode(endTaskTimer);
}

- initWithCoder: (NSCoder *) decoder
{
  fatherTask = [decoder decodeObject];
  GGDecode(subTask);
  ship = [decoder decodeObject];
  GGDecode(endTaskTimer);

  return self;
}

@end


NSString *GGNotificationArriveToStation = @"GGArriveToStation";


@implementation FlightToStation

/*designated initializer
 */
- initWithDestination: (Station *)_station
		 ship: (GGShip *)_ship
{
  id defaultCenter = [MessageManager class];
  [self initWithShip: _ship];
  stationName = RETAIN([_station signature]); 
    
  [defaultCenter
    addObserver: self
    selector: @selector(atDestination:)
    name: GGMissileArrive
    object: ship];

  [self startRunning];

  mode = 0;
  return self;
}

- init
{
  NSLog(@"invalid initializer");
  NSParameterAssert(0);
  abort();
  return nil;
}

+ destination: (Station *)_station
	 ship: (GGShip *)_ship
{
  FlightToStation *fts = [self alloc];
  [fts initWithDestination: _station
       ship: _ship];
  return AUTORELEASE(fts);
}

- (void) encodeWithCoder: (NSCoder *) encoder
{
  [super encodeWithCoder: encoder];
  GGEncode(stationName);
  GGEncode(mode);
}

- initWithCoder: (NSCoder *) decoder
{
  [super initWithCoder: decoder];
  GGDecode(stationName);
  GGDecode(mode);
  return self;
}

- (void) atDestination: (NSNotification *) aNotification
{
  AutoPilotCommon *mis;
  if( mode == 0 )
    {
      mode = 1;
      NSDebugLLog(@"IA", @"%@ reaches the first control point", ship);
      NSParameterAssert(ship != nil && stationName != nil);


      mis = [AutoPilotCommon autoPilotWithTarget: stationName
			     andShip: ship
			     decal: NULL
			     speed: -1
			     distance: 100];
      [self addSubTask: mis];
    }
  else
    {
      [self invalidate];
      [[NSNotificationCenter defaultCenter]
	postNotificationName: GGNotificationArriveToStation
	object: ship];
      NSDebugLLog(@"IA", @"the ship %@ arrived", ship);
    }
}

- (void) startRunning
{
  GGSpaceObject *obj;
  Vect3D v;
  AutoPilotCommon *mis;
  NSParameterAssert(ship != nil && stationName != nil);
  obj = [stationName objectFromSignature];
  NSParameterAssert(obj != nil);
  mulScalVect(Direction(obj), 2e4, &v);
  mis = [AutoPilotCommon autoPilotWithTarget: obj
			 andShip: ship
			 decal: &v
			 speed: -1
			 distance: 100];
  [self addSubTask: mis];
}

/*C'est de la merde, ce code*/

@end

@implementation StupidIA
- initWithShip: (GGShip *)_ship
{
  NSArray *array = [[theWorld currentSystem] allStations];
  int n = [array count];
  Station *s;
  GGTask  *iabla;

  [super initWithShip: _ship];

  if( n > 0 )
    {
      s = (Station *)[array objectAtIndex: irnd(n)];
      NSDebugMLLog(@"IA", @"on cree task destination %@", s);
      iabla = [FlightToStation destination: s
			       ship: _ship];
      [self addSubTask: iabla];
      [MessageManager
	addObserver: self
	selector: @selector(arriveToStation:)
	name: GGNotificationArriveToStation
	object: _ship];
    }
  else
    {
      RELEASE(self);
      self = nil;
    }
  return self;
}

- (void) arriveToStation: (NSNotification *) aNotification
{
  NSDebugMLLog(@"IA", @"%@ at Station, now dying", ship);
  [self invalidate];
  if( ! ship->_flags.activated )
    {
      NSDebugMLLog(@"IA", @"nobody watches! vanishing...");
      [ship destroy];
    }
  else
    {
      NSDebugMLLog(@"IA", @"must wait");
    }
}


+ stupidIA: (GGShip *)_ship
{
  return AUTORELEASE([[self alloc] initWithShip: _ship]);
}
@end
