/*
 *  TopView.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: May 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <Foundation/NSObject.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSNotification.h>

#include "GGRepere.h"
#include "GGLoop.h"
#include "Metaobj.h"
#include "Tableau.h"
#include "Sched.h"
#include "World.h"
#include "GGWidget.h"
#include "utile.h"
#include "GGInput.h"
#include "GGFont.h"
#include "TopView.h"
#include "Texture.h"
#include "Preference.h"
#include "GGIcon.h"

NSString *GGTopViewClosing = @"GGTopViewClosing";

@implementation TopView
+ topview
{
  return AUTORELEASE([[self alloc] init]);
}

- init
{
  NSRect aRect = NSMakeRect(0, HauteurTableauBord, WIDTH, 
			    HEIGHT-HauteurTableauBord);
  [self initWithFrame: aRect];
  return self;
}

- (id) initWithFrame: (NSRect)frameRect
{
  [super initWithFrame: frameRect];
  
  {
    GGBouton *bo;

    bo = AUTORELEASE([GGBouton new]);
    [bo setFrame: NSMakeRect(frameRect.size.width-100, 10, 100, 35)];
    [bo setTitle: @"close"];
    [bo setTarget: self];
    [bo setAction: @selector(close)];
    [bo setAutoresizingMask: NSViewMinXMargin | NSViewMaxYMargin];

    [self addSubview: bo];

    title = [[GGString alloc] initWithFrame: 
				NSMakeRect(300, frameRect.size.height-30,
					   150, 20)];
    [title setString: @"TopView"
	   withFont: [GGFont defaultTitleFont]];
    [self addSubview: title];
    RELEASE(title);
  }
  [self setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];
  [self setState: GUI_OtherView];

  return self;
}

- (void) setTexture: (Texture *)_t
{
  ASSIGN(t, _t);
}

- (void) enableTextureBackground
{
  if( thePref.HighQuality)
    [self setTexture: [Texture textureWithFileName: @"marble.jpg"
			       mipmap: NO]];
}

- (void) drawRect: (NSRect) _theRect
{
  NSRect theRect = [self frame];
  if( t)
    {
      glPushAttrib(GL_ENABLE_BIT);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, [t idTexture]);
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
      glColor3f(0.4, 0.4, 0.4);
      glBegin(GL_QUADS);
      glTexCoord2f(0, 0);
      glVertex2f(0, 0);

      glTexCoord2f(1, 0);
      glVertex2f(NSWidth(theRect), 0);

      glTexCoord2f(1, 1);
      glVertex2f(NSWidth(theRect), NSHeight(theRect));

      glTexCoord2f(0, 1);
      glVertex2f(0, NSHeight(theRect));
      glEnd();

      glPopAttrib();
    }
  else
    {
      //        glClearColor(0, 0, 0, 0);
      //        glClear(GL_COLOR_BUFFER_BIT);
      glColor3f(0,0,0);
      glRecti(0,0,NSWidth(theRect), NSHeight(theRect));
    }
}

- (GGString *) titleView
{
  return title;
}

- (void) dealloc
{
  [[NSNotificationCenter defaultCenter]
    removeObserver: nil 
    name: nil 
    object: self];
  RELEASE(t);
  [super dealloc];
}

- (void) open
{
  [theGGInput addTopView: self];
}

- (void) close
{
  NSDebugMLLog(@"GGInput", @"notification posted");
  
  [theGGInput aTopViewIsClosing: self];
  [theGGInput takeFocus];
  [self removeFromSuperview];
}

- (GUIState) state
{
  return _state;
}

- (void) setState: (GUIState) newState
{
  _state = newState;
}

@end
