/* 	-*-ObjC-*- */
/*
 *  World.h
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __World_h
#define __World_h

#include "GGNode.h"
#include <Foundation/NSGeometry.h>
#include "config.h"

@class GGSky;
@class GGOrbiteur;
@class Sched;
@class PlaneteManager;
@class GGInput;
@class GGShip;
@class Commandant;
@class Laser;
@class Flare;
@class System;
@class NSMutableDictionary;
@class WorldArchiver;

#include "GGNode.h"

struct __state{
  int				indexSystem;
  Commandant   			*commandant;

  //where ship on others system goes :
  NSMutableDictionary		*otherSystem;  
  Vect3D			galacticPosition;
};


@interface World : GGNode <WorldManager>
{
  struct __state	state;
  GGRepere		*leRepereLocal;
  NSArray		*planeteArray;
  GGSpaceObject		*mainCamera, *auxiliaryCamera;
  GGOrbiteur		*theClosest, *theLightSource;
  //  *soleil
  System		*starSytem;
  Vect4D		lightPosition;
  GGSky			*theSky;
  Flare			*flares;
  //  NSMutableArray	*ships;
  NSMutableArray	*lasers;

  
  NSMutableArray	*needToDie;
  int               	isInit;
  ProjectionParam	projectionParam;
  WorldArchiver		*wa;
  BOOL			isPaused;
  BOOL			managingLaser;
  BOOL			checkCollision;
  BOOL			drawWorld;
}

- (void) setCheckCollision: (BOOL) f;
- (void) setDrawWorld: (BOOL) flag;
- (void) setPaused: (BOOL) flag;
- (GGRepere *)repere;

- (void) updateProjection;

- (void) setSystem: (System *)theSystem;
- (System *)currentSystem;
- (void) jumpToSystemIndex: (int) index;

- (void) loadBaseSystem;
- (NSArray *)allOrbiteur;
- (GGOrbiteur *) orbiteurByName: (NSString *)name;
- (Commandant *) commandant;
- (void) setCommandant: (Commandant *) obj;
- (void) setMainCamera:(GGSpaceObject *) obj;
- (void) setAuxiliaryCamera:(GGSpaceObject *)obj;
- (GGSpaceObject *) camera;
- (void) getProjectionParam: (ProjectionParam *)ppp;
//- (PlaneteManager *) systeme;
- (void) setSky: (GGSky *) s;
- (GGSky *) sky;

- (NSArray *) ships;
- (NSMutableArray *) arrayForIndex: (int) index;

- (void) addShip: (GGShip *) aSpaceObject;
- (void) restoreShips: (NSArray *) array;
- (void) setSky: (GGSky *) s;

- (void) drawScene: (NSRect) aFrame;
- (void) checkForCollision;

- (void) manageLaser: (Laser *) l;
- (void) genereExplosionFrom: (GGSpaceObject *) obj
		  withEnergy: (float) val;
- (void) addToKillPool: obj;

- (void) trySaveBordel;
- (void) tryLoadBordel;
- (GGOrbiteur *) closestOrbiteur;
@end

extern World *theWorld;
extern NSString *GGNotificationLeavingSystem;
extern NSString *GGNotificationEnteringSystem;
extern NSString *GGNotificationBeforeLoadingGame;
extern NSString *GGNotificationAfterLoadingGame;

@interface GGArchiver : NSObject <NSCoding>
{
  NSString *name;
}
+ ggArchiver: (NSString *)str;
@end

@interface NSString (GGMethod)
- objectFromSignature;
@end

#endif
