/*
 *  World.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <GL/gl.h>


#include <Foundation/NSObject.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSException.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSScanner.h>
#include <Foundation/NSArchiver.h>
#include <Foundation/NSData.h>

#include "Metaobj.h"
#include "GGLoop.h"
#include "GGSpaceObject.h"
#include "GGInput.h"
#include "GGSky.h"
#include "GGRepere.h"
#include "World.h"
#include "utile.h"
#include "Tableau.h"
#include "GGOrbiteur.h"
#include "Preference.h"
#include "Sched.h"
#include "Explosion.h"
#include "Commandant.h"
#include "Laser.h"
#include "Flare.h"
#include "GGGlyphLayout.h"
#include "Missile.h"
#include "Console.h"
#include "Item.h"
#include "System.h"
#include "GGScene.h"
#include "Resource.h"
#include "GGConcreteShip.h"

World *theWorld=nil;
NSString *GGNotificationLeavingSystem = @"GGNotificationLeavingSystem";
NSString *GGNotificationEnteringSystem = @"GGNotificationEnteringSystem";
NSString *GGNotificationBeforeLoadingGame = @"GGNotificationBeforeLoadingGame";
NSString *GGNotificationAfterLoadingGame = @"GGNotificationAfterLoadingGame";

@interface WorldArchiver : NSObject <NSCoding>
{
@public
  struct __state 	state;
  struct __loopState	loopstate;
  NSString 		*nameLoc;
  GGGalileen		galLoc;
  BOOL			superLocal;
}
+ worldArchiverWithWorld;
- (struct __state *) pstate;
@end

@interface World (PrivateMethod)
- (BOOL) makeFatherCurrentOrbiteur;
- (void) checkClosestOrbiteur;
- (void) unloadSystem;
- getObjectAtSystemIndex: (int) index
		    name: (NSString *)name;
- (GGNode *) nodeForOrbiteur: (GGOrbiteur *)orb;
- objectForName: (NSString *) str;
- (void) setWorldArchiver: (WorldArchiver *)  _wa;
- (void) loadSystemAtIndex: (int) index;
- (void) _anOrbiteurBeingOriginOfLocal;

- (void) makeSonCurrentOrbiteur: (GGOrbiteur *)son;
- (void) initCommandant;
- (struct __state *) pstate;

@end


@implementation WorldArchiver
+ worldArchiverWithWorld
{
  struct __state *pstate;
  WorldArchiver *wa;
  
  wa = [self new];
  pstate = [theWorld pstate];
  wa->state = *pstate;
  wa->loopstate = *[theLoop pstate];
  wa->nameLoc = RETAIN([[[theWorld closestOrbiteur] subSystemFrame] signature]);
  wa->galLoc = *Galileen([theWorld repere]);
  wa->superLocal = ([theWorld repere]->modeRepere == ReperePlanete);
  return AUTORELEASE(wa);
}

- (void) dealloc
{
  RELEASE(nameLoc);
  [super dealloc];
}

- (void) encodeWithCoder: (NSCoder *) encoder
{
  GGEncode(gameTime);
  [encoder encodeObject: nameLoc];
  [encoder encodeValueOfObjCType: @encode(GGGalileen)
	   at: &galLoc];
  [encoder encodeValueOfObjCType: @encode(BOOL)
	   at: &superLocal];
  [encoder encodeValueOfObjCType: @encode(struct __state) 
	   at: &state];
  [encoder encodeValueOfObjCType: @encode(struct __loopState) 
	   at: &loopstate];
  [encoder encodeObject: [MessageManager new]];
}

-  initWithCoder: (NSCoder *) decoder
{
  id fuck;
  state.indexSystem = -2;
  [theWorld setWorldArchiver: self];
  GGDecode(gameTime);
  [decoder decodeValueOfObjCType: @encode(id)
	   at: &nameLoc];
  [decoder decodeValueOfObjCType: @encode(GGGalileen)
	   at: &galLoc];
  [decoder decodeValueOfObjCType: @encode(BOOL)
	   at: &superLocal];
  [decoder decodeValueOfObjCType: @encode(struct __state)
	   at: &state];
  [decoder decodeValueOfObjCType: @encode(struct __loopState)
	   at: &loopstate];
  fuck = [decoder decodeObject];

  //  NSLog(@"Message restoring : %@", [MessageManager class]);

  return self;
}

- (struct __state *) pstate
{
  return &state;
}
@end


@implementation World
- init
{
  if (theWorld)
    {
      NSWarnMLog(@"world allocated more than once...");
      [super dealloc];
      return RETAIN( theWorld );
    }
  [super init];
  [self setManager: self];

  needToDie = [NSMutableArray new];
  //  ships = [NSMutableArray new];
  lasers = [NSMutableArray new];
  flares = [Flare new];
  planeteArray = nil;
  state.otherSystem = [NSMutableDictionary new];

  state.commandant = nil;
  isInit = 1;
  isPaused = NO;
  lightPosition.w = 0;
  checkCollision = YES;
  drawWorld = YES;
  state.galacticPosition = GGZeroVect;

  theWorld = self;

  addToCheck(self);

  leRepereLocal = [GGRepere new];
  [leRepereLocal setDefaultName: @"repere local"];
  [self addSubNode: leRepereLocal];
  [leRepereLocal beginSchedule];

  [theLoop slowSchedule: self
	   withSelector: @selector(checkClosestOrbiteur)
	   arg: nil];

  theSky = [[GGSky alloc] initAtOrigin: GGZeroVect];
  [self updateProjection];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(loadingAGame:)
    name: GGNotificationBeforeLoadingGame
    object: nil];

  return self;
}

- (void) encodeWithCoder: (NSCoder *) encoder
{
}

- (id)replacementObjectForArchiver:(NSArchiver *)anArchiver
{
  NSAssert(_originOrbiter != nil, NSInternalInconsistencyException);

  return [GGArchiver ggArchiver: @"theWorld" ];
}

- (BOOL) shouldSave
{
  return YES;
}

- (BOOL) saveChannel
{
  return YES;
}

- (void) loadingAGame: obj
{
  checkCollision = YES;
  managingLaser = NO;
}


- (NSData *) saveState
{
  WorldArchiver *wa2;
  wa2 = [WorldArchiver worldArchiverWithWorld];
  return [NSArchiver archivedDataWithRootObject: wa2];
}

- (void) unloadState
{
  [state.commandant reallyDestroy];

  [self unloadSystem];
  DESTROY(state.commandant);
  
  //  blabladestroy;

  DESTROY(mainCamera);
  DESTROY(auxiliaryCamera);
  DESTROY(state.otherSystem);

  theClosest = nil;
  DESTROY(theLightSource);
  DESTROY(starSytem);
  [lasers removeAllObjects];
}

- (void) loadState: (NSData *) data
{
  struct __state *pstate;
  WorldArchiver *wa2;
  [self unloadState];
  wa = nil;
  state.indexSystem = -2;
  wa2 = [NSUnarchiver unarchiveObjectWithData: data];
  NSAssert(wa2 != nil, @"Unable to restore saved states");
  NSAssert(wa2 == wa, NSInternalInconsistencyException);
  
  pstate = [wa pstate];

  state = *pstate;

  [self initCommandant];

  //  NSLog(@"%@", wa->loopstate.slowScheduler);
  //  NSLog(@"%@", wa->loopstate.taskSched);
  [theLoop setState: &wa->loopstate];
  [theLoop setTimeSpeed: 10];
  [theLoop setTimeSpeed: 1];
}

- (void) trySaveBordel
{
  NSData *data;
  data = [self saveState];

  [data writeToFile: @"test.sav"
	atomically: NO];
  NSLog(@"sauv'e, taille = %d", [data length]);
}

- (void) tryLoadBordel
{
  NSData *data;
  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGNotificationBeforeLoadingGame
    object: nil];
  data = [NSData dataWithContentsOfFile: @"test.sav"];
  [self loadState: data];

  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGNotificationAfterLoadingGame
    object: nil];


  //  NSLog(@"tree:\n%@", [self displayObject: 0]);
}


- (void) updateProjection
{
  NSRect frameWindow = [[theGGInput contentView] frame];
  projectionParam.ratio = (double) frameWindow.size.width / 
    frameWindow.size.height;
  projectionParam.yFoc = tan((double)Focal/2.0/180*M_PI);
  projectionParam.xFoc = projectionParam.yFoc*projectionParam.ratio;
  projectionParam.zmin = 1;
  projectionParam.zmax = 10000;
}

- (GGSpaceObject *)ancestor
{
  return self;
}

- (void) drawSky: (Camera *)cam
{
  GGMatrix4 fuck;
  Camera skyCam;

  initWithIdentityMatrix(&fuck);
  fuck.gauche.x = -1;
  fuck.direction.z = -1;
  skyCam.modelView = cam->modelView;
  SetZeroVect(skyCam.modelView.position);
  glPushMatrix();
  glLoadMatrix((real *)&fuck);
  glMultMatrix((real *)&skyCam.modelView);

  skyCam.param = cam->param;

  [theSky drawSky: &skyCam];
  glPopMatrix();

  //  debugOrbiteurWindow(@"Terre");
}

- (void) willDrawSons: (Camera *)cam
{
  glDisable(GL_LIGHTING);
  [self drawSky: cam];

  glDepthMask(GL_TRUE);
  glEnable(GL_DEPTH_TEST);
  glDepthRange(0.95, 0.95);
  glDepthFunc(GL_ALWAYS);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, (real*)&lightPosition);


  /*ncessaire pour avoir une ilumination correcte des objets proches.
   */
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,   (int) thePref.LightLocal);
  
  if( thePref.GLSmooth )
    glShadeModel(GL_SMOOTH);
  else
    glShadeModel(GL_FLAT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(Focal,WIDTH/(float)HEIGHT,1,1e20);
  glMatrixMode(GL_MODELVIEW);
}

- (GGRepere *)repere
{
  return leRepereLocal;
}

- (GGOrbiteur *) closestOrbiteur
{
  return theClosest;
}

- (void) dealloc
{
  [theLoop unSchedule: self];
  RELEASE(flares);
  RELEASE(theLightSource);
  RELEASE(needToDie);
  RELEASE(planeteArray);
  RELEASE(state.commandant);
  RELEASE(starSytem);
  RELEASE(lasers);
  RELEASE(mainCamera);
  RELEASE(auxiliaryCamera);
  RELEASE(state.otherSystem);
  RELEASE(leRepereLocal);

  RELEASE(theSky);

  [super dealloc];
  theWorld = nil;
}


- (void) loadBaseSystem
{
  NSAutoreleasePool *pool;    
  pool = [NSAutoreleasePool new];
  {
//      Vect3D posObs = { -150e9+50e6, 0, -1e11};
    Vect3D posObs = { -150e9+50e6, 0, 0};
    Item *i = AUTORELEASE([Item new]);
    Commandant *_commandant;

    i->type = Missile;
    i->arg1 = 7;
    _commandant = [Commandant shipWithFileName: @"gg.shp"];
    [_commandant setName: @"C-007"];
    [_commandant setPosition: &posObs];
    [_commandant setIndexSystem: 0]; //we start in Sol
    initVect(Direction(_commandant), 1, 0, 0);
    initVect(Haut(_commandant), 0, 0, 1);
    SetZeroVect(*Omega(_commandant));
    [_commandant normalise];
    [_commandant addItem: i];
    [_commandant addItem: i];
    [_commandant addItem: i];
    [_commandant addItem: i];

    i = AUTORELEASE([Item new]);
    i->type = LaserRayon;
    i->arg1 = 15;
    [_commandant addItem: i];

    [self setCommandant: _commandant];

#ifdef GGSIM
    {
    GGScene *scene;
    scene = [GGScene new];
    [leRepereLocal addSubNode: scene];
    RELEASE(scene);
    }
#endif

#if 0
    {
      GGScene *scene;
      scene = [GGScene new];
      initVect(Point(scene), 0, 0, -10000);
      //      initVect(Point(scene), 0, 0, 20000);
      [leRepereLocal addSubNode: scene];
      RELEASE(scene);
    }
#endif

    [self loadSystemAtIndex:  [_commandant indexSystem]];
    Point(_commandant)->z = 10;
  }
  RELEASE(pool);
}

- (void) setPaused: (BOOL) flag
{
  isPaused = flag;
  if( isPaused )
    {
      [theLoop setPollMode: WaitForEvent];
      [theLoop setTimeSpeed: 0];
    }
  else
    {
      [theLoop setPollMode: DontWaitForEvent];
      [theLoop setTimeSpeed: 1];
    }
}


- (void) jumpToSystemIndex: (int) index
{
  NSArray *array;
  NSDictionary *aDic;
  [theSky genereSkyForIndex: index];
  [self reparent: leRepereLocal];

  aDic = [NSDictionary dictionaryWithObject: [NSNumber numberWithInt: 
						[starSytem index]]
		       forKey: @"Name"];

  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGNotificationLeavingSystem
    object: self
    userInfo: aDic];

  [self unloadSystem];  //this remove the local frame

  //recenter the local frame around the camera.

  [self loadSystemAtIndex: index];

  *Point(leRepereLocal) = GGZeroVect;
  *Point(state.commandant) = [starSytem randomVect];
  [state.commandant tryGlobalCoordinates: leRepereLocal];

  //add saved ships
  array = [self arrayForIndex: [starSytem index]];
  [self restoreShips: array];
  
  [state.commandant setIndexSystem: [starSytem index]];
  consoleLog(@"jump to system %s", [theSky cNameOf: index]);

  aDic = [NSDictionary dictionaryWithObject: [NSNumber numberWithInt: 
						[starSytem index]]
		       forKey: @"Name"];

  [[NSNotificationCenter defaultCenter]
    postNotificationName: GGNotificationEnteringSystem
    object: self
    userInfo: aDic];
}

- (void) setCheckCollision: (BOOL) f
{
  checkCollision = f;
}

- (BOOL) isPaused
{
  return isPaused;
}

- (NSArray *) ships
{
  return [self arrayForIndex: [starSytem index]];
}

- (Commandant *) commandant
{
  return state.commandant;
}

- (void) setCommandant: (Commandant *) obj
{
  ASSIGN(state.commandant, obj);
  [obj setMasseCoque: 2000];
  [obj setInvincibility: YES];
  [self addSubNode: obj];

  *Point(leRepereLocal) = *Point(obj);
  [obj enterUniverse];
  [obj beginSchedule];
  [self initCommandant];
  [obj tryLocalCoordinates: nil];
}

- (NSMutableArray *) arrayForIndex: (int) index
{
  NSMutableArray *array;
  NSValue *val;

  val = [NSNumber numberWithInt: index];
  array = [state.otherSystem objectForKey: val];
  if( array == nil )
    {
      array = [NSMutableArray array];
      [state.otherSystem setObject: array
		   forKey: val];
    }
  return array;
}

- (void) makeCurrent: (GGShip *) ship
{
  [ship setLocality: 1];
  [ship beginSchedule];
}

- (void) addShip: (GGShip *) aSpaceObject
{
  NSMutableArray *array;
  int index = [starSytem index];
  array = [self arrayForIndex: index];
  NSParameterAssert(array != nil);
  [aSpaceObject setIndexSystem: index];
  [self makeCurrent: aSpaceObject];
  [array addObject: aSpaceObject];
  [aSpaceObject enregistre: self
		with: @selector(removeWithNotification:)];
  [aSpaceObject enterUniverse];
}



- (void) restoreShips: (NSArray *) array
{
  int n = [array count];
  if( n > 0 )
    {
      GGShip *lst[n];
      int i;
      [array getObjects: lst];
      
      for(i = 0; i < n; ++i)
	{
	  [self makeCurrent: lst[i]];
	  [self addSubNode: lst[i]];
	}
    }
}

- (void) removeWithNotification: (NSNotification *) aNotification
{
  GGShip *obj= [aNotification object];
  if( [ obj isKindOfClass: [GGShip class]] )
    {
      int index;
      NSMutableArray *array;

      index = [obj indexSystem];
      array = [self arrayForIndex: index];

      NSDebugMLLog(@"World", @"ship %@ dying", obj);
      
      if( [array indexOfObjectIdenticalTo: obj] == NSNotFound )
	NSWarnMLog(@"object %@ not Found", obj);
      [array removeObjectIdenticalTo: obj];
      if( [array count] == 0 )
	{
	  [state.otherSystem removeObjectForKey: [NSNumber numberWithInt: index]];
	}
    }
}

- (GGOrbiteur *) orbiteurByName: (NSString *)name
{
  int n = [planeteArray count];
  GGOrbiteur *orbs[n];
  int i;

  [planeteArray getObjects: orbs];

  for( i = 0; i < n; ++i )
    {
      if( [[orbs[i] name] isEqualToString: name] )
	return orbs[i];
    }
  return nil;
}

- (NSArray *) allOrbiteur
{
  return planeteArray;
}


- (void) setMainCamera:(GGSpaceObject *) obj
{
  if( mainCamera == obj )
    return;
  [mainCamera desenregistre: self]; //Puique qu'on peut envoyer des messages  nil
  [mainCamera setCameraFlag: 0];
  ASSIGN(mainCamera, obj);
  [mainCamera setCameraFlag: 1];
  [obj enregistre: self
       with:@selector(removeCameraWithNotification:)];
}

- (void) setAuxiliaryCamera:(GGSpaceObject *)obj
  /*on ne regle pas le camera flag de l'objet car il ne concerne que la 
   *camera principale
   */
{
  NSDebugMLLog(@"auxcamera", @"on perd le missile");
  if( auxiliaryCamera == obj )
    return;
  [auxiliaryCamera desenregistre: self]; //Puique qu'on peut envoyer des messages  nil
  ASSIGN(auxiliaryCamera, obj);
  [obj enregistreForRemoving: self
       with:@selector(removeCameraWithNotification:)];
}

- (GGSpaceObject *) camera
{
  return mainCamera;
}

- (void) removeCameraWithNotification: (NSNotification *)aNotification
{
  /* Si l'objet qui est la camra se fait buter, le vaisseau commandant
     devient la nouvelle camra
  */
  id obj = [aNotification object];
  if( obj == mainCamera )
    {
      NSWarnMLog(@"la camera principale meure!!!");
      //      [self setMainCamera: state.commandant];
      [self setMainCamera: nil];
    }
  if( obj == auxiliaryCamera )
    [self setAuxiliaryCamera: nil];
}

- (void) setSky: (GGSky *) s
{
  ASSIGN(theSky, s);
}

- (GGSky *) sky
{
  return theSky;
}

- (void) setDrawWorld: (BOOL) flag
{
  drawWorld = flag;
}

typedef struct __GGRepere_struct
{
  @defs(GGRepere);
}GGRepere_struct;


- (void) drawSceneWithCamera: (GGSpaceObject *)aCamera
			left: (int) x
		       lower: (int) y
		       width: (int) w
		      height: (int) h
		       label: (BOOL) label
{
  //  printf("%d %d %d %d\n", x, y, w, h);
  ProjectionParam p;
  glViewport(x, y, w, h);
  glEnable(GL_SCISSOR_TEST);
  glScissor(x, y, w, h);
  //  clearScene(0);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(Focal,w/(float)h,0.1,50.0);
  glMatrixMode(GL_MODELVIEW);

  p = projectionParam;
  p.x = x;
  p.y = y;
  p.width = w;
  p.height = h;
  p.label = label;

  if(label)
    [theTableau doIndicator: &p	
		camera: aCamera];

//   [self drawGraphFrom: aCamera
// 	projection: &p];
  [aCamera drawGraphProjection: &p];

  if( theLightSource != nil )
    {
      Vect3D pos;
      real x, y;
      [theLightSource positionInRepere: aCamera
		      in: &pos];

      if( pos.z < 0.1 )
	{
	  NSDebugMLLogWindow(@"Flare", @"sun is behind...");

	}
      else
	{
	  x = -pos.x/pos.z/projectionParam.xFoc;
	  y = pos.y/pos.z/projectionParam.yFoc;

	  if( x >= -1 && x <= 1 && y >= -1 && y <= 1 )
	    [flares drawFlaresAt: x
		    and: y];
	}
    }
  glDisable(GL_SCISSOR_TEST);
}

- (void) getProjectionParam: (ProjectionParam *)ppp
{
  *ppp = projectionParam;
}


#ifdef DEBUG
static int sortAux(id o1, id o2, void *bla)
{
  int diff = GSDebugAllocationCount(o2)-GSDebugAllocationCount(o1);

  if(diff > 0 )
    return NSOrderedDescending;
  if(diff < 0 )
    return NSOrderedAscending;
  else
    return NSOrderedSame;
}
#endif

- (void) drawScene: (NSRect) aFrame
{
  if( !isPaused )
    {
      if( checkCollision )
	[self checkForCollision];
      [needToDie makeObjectsPerformSelector: @selector(reallyDestroy)];
      [needToDie removeAllObjects];
    }

  if( !drawWorld )
    return;

  [self drawSceneWithCamera: mainCamera
	left: aFrame.origin.x
	lower: aFrame.origin.y
	width: aFrame.size.width 
	height: aFrame.size.height
	label: YES];



  //  [mainCamera setVisible: YES];
  NSDebugMLLogWindow(@"Tree", [self displayObject: 0]);
  NSDebugMLLogWindow(@"Speed", stringOfVect(Speed(leRepereLocal)));
#ifdef DEBUG
  {
    if( thePref.DebugMemoire )
      {
	NSEnumerator *enumerator;
	Class *lst, *temp;
	Class aClass;
	int count;
	NSMutableArray *array;

	lst = GSDebugAllocationClassList();
	for(count = 0, temp=lst; *temp != nil; count++, temp++);
	
	array = [NSMutableArray arrayWithObjects: lst
				count: count];
	[array sortUsingFunction: &sortAux context: NULL];
	enumerator = [array objectEnumerator];
	while((aClass = [enumerator nextObject]) != nil)
	  {
	    [theTableau addLogString: @"Class %@ count %d total %d peak %d", 
			aClass,
			GSDebugAllocationCount(aClass),
			GSDebugAllocationTotal(aClass),
			GSDebugAllocationPeak(aClass)];
	  }
      }
  }
#endif


  if( auxiliaryCamera != nil )
    {
      [self drawSceneWithCamera: auxiliaryCamera
	    left: WIDTH-250
	    lower: HEIGHT-250
	    width: 250
	    height: 250
	    label: NO];

      //draw  the border :
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0.,1.01,0,1.01,-1.0,1.0);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glColor3f(0.8, 0.8, 1);
      glPolygonMode(GL_FRONT, GL_LINE);
      glRecti(0, 0, 1, 1);
      glPolygonMode(GL_FRONT, GL_FILL);
    }

}

- (void) checkForCollision
{
  NSArray *locals = [leRepereLocal subNodes];
  int n = [locals count];
    

  if( n > 0 )
    {
      GGSpaceObject *objs[n+1];
      int i, j;

      [locals getObjects: objs];
      for( i = 0; i < n; ++i )
	{
	  if( CollisionCounter(objs[i]) )
	    continue;
	  for( j = i+1; j < n; ++j )
	    {
	      Vect3D *v1 = Point(objs[i]);
	      Vect3D *v2 = Point(objs[j]);

	      real d2 = distance2Vect(v1, v2);
	      real dcol = Dimension(objs[i]) + Dimension(objs[j]);

	      if( CollisionCounter(objs[j]) )
		continue;

	      NSDebugMLLogWindow(@"Collision", @"d(%@, %@) = %g",
				 objs[i], objs[j], sqrt(d2));


	      if( d2 < SQ(dcol) )
		{
		  NSDebugMLLog(@"Collision", 
			       @"Collision detected between %@ %@",
			       objs[i], objs[j]);
		  if( [objs[j] manageCollision] )
		    [objs[j] collisionOccursWith: objs[i]];
		  else
		    [objs[i] collisionOccursWith: objs[j]];
		
		}
	    }
	}
    }
}

- (void) manageLaser: (Laser *) l
{
  [l enregistre: self
     with: @selector(removeLaserWithNotification:)];
  [lasers addObject: l];
  if( !managingLaser )
    {
      managingLaser = YES;
      [theLoop fastSchedule: self
	       withSelector: @selector(checkForLaser)
	       arg: nil];
    }
  NSDebugMLLog(@"Laser", @"start managing laser");
}

- (void) removeLaserWithNotification: (NSNotification *)aNotification
{
  id obj = [aNotification object];
  NSDebugMLLog(@"Laser", @"stop managing laser");
  [lasers removeObjectIdenticalTo: obj];
  if( managingLaser && [lasers count] == 0 )
    {
      managingLaser = NO;
      [[theLoop fastScheduler] deleteTarget: self
			      withSelector: @selector(checkForLaser)];
      NSDebugMLLog(@"Laser", @"no more laser");
    }
}

- (void) checkForLaser
{
  NSArray *locals = [leRepereLocal subNodes];
  int n, nl, i , j;
  
  nl = [lasers count];
  NSAssert(nl > 0, NSInternalInconsistencyException);
  n = [locals count];
  if( n > 0 )
    {
      GGSpaceObject *objs[n];
      [locals getObjects: objs];
      for( i = 0; i < nl; ++i)
	{
	  Laser *l = [lasers objectAtIndex: i];
	  Vect3D pos, dir;
	  float power = [l power];
	  //FIXME : hack
	  pos = *Point(l->nodePere);
	  dir = *Direction(l->nodePere);
	  NSDebugMLLogWindow(@"Laser", @"pos %@, dir %@", 
		       stringOfVect(&pos),
		       stringOfVect(&dir));
	  for(j = 0; j < n; ++ j)
	    {
	      real d, xpos;
	      GGSpaceObject *obj = objs[j];
	      if( CollisionCounter(obj) < 0 || obj == l->nodePere )
		continue;

	      d = distanceDroitePoint(&pos, &dir, Point(obj),
				      &xpos);
	      NSDebugMLLogWindow(@"Laser", @"dist with : %@ = %g",
				 objs[i], d);
	      if( xpos <= 10)
		continue;

	      if( d < Dimension(obj) )
		{
		  [obj chocWithEnergy: deltaTime*power];
		}
	    }
	}
    }
}

- (void) genereExplosionFrom: (GGSpaceObject *) obj
		  withEnergy: (float) val
{
  Explosion *e;
  NSArray *locals = [leRepereLocal subNodes];
  int n = [locals count];

  if( [obj locality] != 2 )
    {
      NSWarnMLog(@"%@ is not local", obj);
      return;
    }


  if( n > 0 )
    {
      GGSpaceObject *objs[n];
      int i;
      Vect3D *v1 = Point(obj);

      [locals getObjects: objs];

      for( i = 0; i < n; ++i )
	{
	  if( objs[i] != obj )
	    {
	      Vect3D *v2 = Point(objs[i]);
	      real d2 = distance2Vect(v1, v2);

	      NSDebugMLLog(@"Explosion", @"distance with %@=^2 : %g", objs[i],
			   d2);

	      [objs[i] chocWithEnergy: val/d2];
	    }
	}
    }
  e = [Explosion explosionAt: obj
		 withEnergy: val];
  [e setVisible: YES];
  [e enterUniverse];
  [leRepereLocal addSubNode: e];
  [e beginSchedule];
}

- (void) addToKillPool: obj
{
  [needToDie addObject: obj];
}	

- (void) postLoadingSystem
{
  NSArray	*sons;
  GGOrbiteur *soleil = [starSytem mainStar];
  theClosest = soleil;
  [self addSubNode: soleil];
  [theClosest startBeingOrigineOfLocale: self];
  [self _anOrbiteurBeingOriginOfLocal];
  sons = [soleil sons];
  [self addSubNodeFromArray: sons];
  [soleil beginSchedule];
  [sons makeObjectsPerformSelector: @selector(beginSchedule)];
  [theTableau setClosestOrbiteur: theClosest];
}

- (void) setSystem: (System *)theSystem
{
  GGOrbiteur *unSoleil = [theSystem mainStar];
  NSArray 	 *allPlanetes = [unSoleil getAllOrbiteur];
  int n = [allPlanetes count];
  int i;
  ASSIGN(starSytem, theSystem);
  [self setOriginOrbiteur: unSoleil];
  for( i = 0; i < n; ++i )
    {
      GGOrbiteur *o = (GGOrbiteur *)[allPlanetes objectAtIndex: i];
      [o enterUniverse];
      if( [o lightSource] )
	ASSIGN(theLightSource, o);
    }
  [unSoleil miseAJourSysteme];
  [unSoleil calculeTailleSysteme];
  ASSIGN(planeteArray, allPlanetes);
  [self postLoadingSystem];
}

- (System *)currentSystem
{
  return starSytem;
}
@end

@implementation World (PrivateMethod)

- (void) calcLightSource
{
  Vect3D com, sol;
  if( theLightSource == nil )
    return;
  [theLightSource positionInRepere: self
		  in: &sol];
  [state.commandant positionInRepere: self
	      in: &com];
  mkVectFromPoint(&com, &sol, &lightPosition.v);
  normaliseVect(&lightPosition.v);
}

- (void) unloadSystem
{
  int n;
  
  // Now the problem of the ships
  n = [[self ships] count];
  if( n > 0 )
    {
      GGShip *shipsTab[n];
      int i;
      [[self ships] getObjects: shipsTab];
      
      for( i = 0; i < n ; ++i)
	if (shipsTab[i] != state.commandant )
	  {
	    [self reparent: shipsTab[i]];
	    [shipsTab[i] removeFromSuperNode];
	    [shipsTab[i] endSchedule];
	    [shipsTab[i] looseMainSystem];
	  }
    }
  else
    {
      [state.otherSystem removeObjectForKey: [NSNumber numberWithInt:
						   [starSytem index]]];
    }

  //Now the problem of the orbiteur.
  while( [self makeFatherCurrentOrbiteur] );

  //we destroy the sun
  [theClosest stopBeingOrigineOfLocale];
  NSParameterAssert([theClosest nodePere] == self);
  [[theClosest sons] makeObjectsPerformSelector: @selector(endSchedule)];
  [[theClosest sons] makeObjectsPerformSelector: 
		       @selector(removeFromSuperNode)];
  [theClosest removeFromSuperNode];


  [[starSytem mainStar] reallyDestroy];
  DESTROY(starSytem);
  DESTROY(theLightSource);
  theClosest = nil;

  [self removeAllSons];
  [self addSubNode: leRepereLocal]; //we put it back

}

- (void) _anOrbiteurBeingOriginOfLocal
{
  [MessageManager
    removeObserver: self
    name: GGOrbiteurBeingSuperLocal
    object: nil];
  [MessageManager
    addObserver: self
    selector: @selector(anOrbiteurBeingSuperLocal:)
    name: nil
    object: theClosest];
}

- (void) anOrbiteurBeingSuperLocal: (NSNotification *) aNotification
{
  // HACK:
  if( [[aNotification name] isEqualToString: GGOrbiteurBeingSuperLocal] )
    {
      GGOrbiteur *theOrb = [aNotification object];
      [[theOrb nodePere] reparent: leRepereLocal];
      leRepereLocal->modeRepere = ReperePlanete;
    }
  else if([[aNotification name] 
	    isEqualToString: GGOrbiteurEndBeingSuperLocal])
    {
      GGOrbiteur *theOrb = [aNotification object];
      [[theOrb subSystemFrame] reparent: leRepereLocal];
      leRepereLocal->modeRepere = RepereNormal;
    }
}

- (BOOL) makeFatherCurrentOrbiteur
{
  GGNode 	*newNode;
  GGNode 	*oldNode;
  GGOrbiteur	*orbiteurPere = [theClosest orbiteurPere];

  if( orbiteurPere == nil )
    return NO;

  [theClosest stopBeingOrigineOfLocale];
  oldNode = [theClosest nodePere];
  NSParameterAssert(oldNode != nil);
  newNode = [oldNode nodePere];
  NSParameterAssert(newNode != nil);
  
  [newNode reparent: leRepereLocal];
  [newNode reparent: theClosest];

  //we remove all the suborbiteur from the current universe :
  [[theClosest sons] makeObjectsPerformSelector: @selector(endSchedule)];
  [[theClosest sons] makeObjectsPerformSelector: 
		       @selector(removeFromSuperNode)];

  //but don't remove other object (like ships)
  //transfer them to the father node.

  [oldNode removeFromSuperNodeAndReparentSons];

  theClosest =  orbiteurPere;
  NSParameterAssert(theClosest != nil);

  [theClosest startBeingMainLocale];
  [self _anOrbiteurBeingOriginOfLocal];
  [theTableau setClosestOrbiteur: theClosest];
  return YES;
}

- (void) makeSonCurrentOrbiteur: (GGOrbiteur *)son
{
  GGNode 	*newNode;
  GGNode 	*currentNode;
  NSArray 	*sons;

  [theClosest stopBeingMainLocale];
  currentNode = [theClosest nodePere];

  newNode = [GGNode new];
  [newNode setOriginOrbiteur: son];
  [newNode setName: [NSString stringWithFormat: 
				@"repere translation from %@",
			      son]];

  *Point(newNode) = *Point(son);
  *Speed(newNode) = *Speed(son);
    //  *Galileen(newNode) = *Galileen(son);

  
  [currentNode addSubNode: newNode];
  [newNode reparent: son];
  [newNode reparent: leRepereLocal];
  theClosest = son;

  [theClosest startBeingOrigineOfLocale: newNode];
  [self _anOrbiteurBeingOriginOfLocal];
  sons = [son sons];
  [newNode addSubNodeFromArray: sons];
  [son beginSchedule];
  [sons makeObjectsPerformSelector: @selector(beginSchedule)];
  [theTableau setClosestOrbiteur: theClosest];


  RELEASE(newNode);
}

- (void) setClosestOrbiteurRec
{
  GGOrbiteur	*oldClosest = theClosest;

  Vect3D 		pos;
  GGOrbiteur	*son;
  [state.commandant positionInRepere: [theClosest subSystemFrame]
	in: &pos];
  NSDebugMLLog(@"WorldHard", @"position orb %@", stringOfVect(&pos));
  if( prodScal(&pos, &pos) > square([theClosest tailleSysteme]) )
    [self makeFatherCurrentOrbiteur];
  else if ((son = [theClosest closestFromRelative: &pos] ) != theClosest )
    [self makeSonCurrentOrbiteur: son];
  /* else do nothing */

  if( oldClosest != theClosest )
    {
      [self setClosestOrbiteurRec];
    }
}


- (void) checkClosestOrbiteur
{
  NSAssert(theClosest != nil, NSInternalInconsistencyException);

  [self setClosestOrbiteurRec];
  [self calcLightSource];
}

- getObjectAtSystemIndex: (int) index
		    name: (NSString *)name
{
  NSAssert([starSytem index] == state.indexSystem,
	   NSInternalInconsistencyException);
  if( index == [starSytem index] )
    return [self orbiteurByName: name];
  else
    {
      NSWarnMLog(@"index = %d current = %d", index,
		 [starSytem index]);
      return nil;
    }
}

- (void) loadSystemAtIndex: (int) index
{
  state.indexSystem = index;
  
  [starSytem endCurrent];
  [self setSystem: [System systemAt: [theSky stars: NULL]+index]];
  [starSytem makeCurrent];
}

- (void) setWorldArchiver: (WorldArchiver *) _wa
{
  wa = _wa;
}

- (GGNode *) nodeForOrbiteur: (GGOrbiteur *)orb
{
  if( [orb isOriginOfNode] )
    return [orb nodePere];
  else
    {
      GGNode *node = nil;
      GGOrbiteur *orbPere;
      GGNode *pereNode;
      orbPere = [orb orbiteurPere];
      NSAssert(orbPere != nil, NSInternalInconsistencyException);
      if( orbPere != nil )
	{
	  pereNode = [self nodeForOrbiteur: orbPere];
	  NSAssert(pereNode != nil, NSInternalInconsistencyException);
	  [self makeSonCurrentOrbiteur: orb];
	  node = [orb nodePere];
	  NSAssert(node != nil, NSInternalInconsistencyException);
	}
      return node;
    }
}

- objectForName: (NSString *) str
{
  NSCharacterSet *semicolonSet;
  NSScanner *theScanner;
  NSString 	*name;

  //  NSLog(@"%@ is requested %@", self, str);

  if( state.indexSystem == -2 )
    {
      GGNode *aNode;
      NSAssert(wa != nil, NSInternalInconsistencyException);
      state.indexSystem = wa->state.indexSystem;

      NSAssert(state.indexSystem >= -1, NSInternalInconsistencyException);
      [self loadSystemAtIndex: state.indexSystem];
      

      aNode = [self objectForName: wa->nameLoc];
      //after this call, normally, the localFrame has been moved to its save 
      //position, we check it now

      NSAssert(aNode == [leRepereLocal nodePere], 
	       NSInternalInconsistencyException);

      if( wa->superLocal )
	[theClosest beginBeingSuperLocal];

      *Galileen(leRepereLocal) = wa->galLoc;
    }

  semicolonSet = [NSCharacterSet
		   characterSetWithCharactersInString:@";"];
  theScanner = [NSScanner scannerWithString: str];

  if( [str isEqualToString: @"theWorld"] )
    return self;
  else if ([theScanner scanString: @"rep:" intoString:NULL] )
    {
      if ( [theScanner scanUpToCharactersFromSet: semicolonSet 
		       intoString: &name])
	{
	  GGOrbiteur *orb;
	  orb = [self orbiteurByName: name];
	  NSAssert1(orb != nil, @"can't recover orbiteur from name \"%@\"",
		    name);
	  return [self nodeForOrbiteur: orb];
	}
      NSAssert(0, @"can't parse node name");
    }
  else if( [theScanner scanString: @"orb:" intoString:NULL] )
    {
      if ( [theScanner scanUpToCharactersFromSet: semicolonSet 
		       intoString: &name])
	{
	  GGOrbiteur *orb;
	  orb = [self orbiteurByName: name];
	  NSAssert1(orb != nil, @"can't recover orbiteur from name \"%@\"",
		    name);
	  return orb;
	}
      NSAssert(0, @"can't parse node name");
    }
  else if( [theScanner scanString: @"reploc:" intoString:NULL] )
    {
      return leRepereLocal;
    }

  NSLog(@"%@", str);
  NSAssert(0, NSInternalInconsistencyException);
  
  return nil;
}

- (struct __state *) pstate
{
  return &state;
}


- (void) initCommandant
{
  Commandant *obj = state.commandant;

  [[theLoop input] setShip: obj];
  if( mainCamera == nil )
    [self setMainCamera: state.commandant];
  [theTableau setObserveur: mainCamera];
}

@end

@implementation NSString (GGMethod)
- objectFromSignature
{
  NSCharacterSet *semicolonSet;
  NSScanner *theScanner;
  int 		index;
  NSString 	*name;

  semicolonSet = [NSCharacterSet
		   characterSetWithCharactersInString:@";"];
  theScanner = [NSScanner scannerWithString: self];

  if ([theScanner scanString: @"orb:" intoString:NULL] &&
      [theScanner scanInt: &index] &&
      [theScanner scanString:@":" intoString:NULL] &&
      [theScanner scanUpToCharactersFromSet: semicolonSet
		  intoString: &name])
    {
      return [theWorld getObjectAtSystemIndex: index
		       name: name];
    }
  else
    NSWarnMLog(@"could not identify %@", self);
  
  return self;
}

@end

@implementation GGArchiver
+ ggArchiver: (NSString *)str
{
  GGArchiver *arch;
  arch = [[self alloc] init];
  ASSIGN(arch->name, str);
  return AUTORELEASE(arch);
}

- (void) dealloc
{
  RELEASE(name);
  [super dealloc];
}

- (void) encodeWithCoder: (NSCoder *)aCoder
{
  [aCoder encodeObject: name];
}

- initWithCoder: (NSCoder *)decoder
{
  NSString *str = [decoder decodeObject];
  RELEASE(self);
  return RETAIN([theWorld objectForName: str]);
}

@end
