/*
 *  main.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <SDL/SDL.h>
#include <stdio.h>
#include <stdlib.h>

#include <Foundation/Foundation.h>
#include <Foundation/NSObjCRuntime.h>
#include <Foundation/NSException.h>
#include "GGInput.h"
#include "GGSky.h"
#include "GGScene.h"
#include "GGLoop.h"
#include "GGShip.h"
#include "World.h"
#include "Station.h"
#include "Texture.h"
#include "Metaobj.h"
#include "Preference.h"
#include "utile.h"
#include "Commandant.h"
#include "GGConcreteShip.h"
#include "Item.h"
#include "Resource.h"


static void quit_ggelite( int code )
{
  /*
   * Quit SDL so we can release the fullscreen
   * mode and restore the previous video settings,
   * etc.
   */
  SDL_Quit( );

  /* Exit program. */
  exit( code );
}

static void createGlContext()
{
  /* Information about the current video settings. */
  const SDL_VideoInfo* info = NULL;
  /* Dimensions of our window. */
  int width = 0;
  int height = 0;
  /* Color depth in bits of our window. */
  int bpp = 0;
  /* Flags we will pass into SDL_SetVideoMode. */
  int flags = 0;

  /* First, initialize SDL's video subsystem. */
  if( SDL_Init( SDL_INIT_VIDEO| SDL_INIT_JOYSTICK | SDL_INIT_NOPARACHUTE ) < 0 ) {
    /* Failed, exit. */
    fprintf( stderr, "Video initialization failed: %s\n",
             SDL_GetError( ) );
    quit_ggelite( 1 );
  }

  /* Let's get some video information. */
  info = SDL_GetVideoInfo( );

  if( !info ) {
    /* This should probably never happen. */
    fprintf( stderr, "Video query failed: %s\n",
             SDL_GetError( ) );
    quit_ggelite( 1 );
  }

  /*
     * Set our width/height to 640/480 (you would
     * of course let the user decide this in a normal
     * app). We get the bpp we will request from
     * the display. On X11, VidMode can't change
     * resolution, so this is probably being overly
     * safe. Under Win32, ChangeDisplaySettings
     * can change the bpp.
     */
  width = WIDTH;
  height = HEIGHT;
  bpp = info->vfmt->BitsPerPixel;

  /*
     * Now, we want to setup our requested
     * window attributes for our OpenGL window.
     * We want *at least* 5 bits of red, green
     * and blue. We also want at least a 16-bit
     * depth buffer.
     *
     * The last thing we do is request a double
     * buffered window. '1' turns on double
     * buffering, '0' turns it off.
     *
     * Note that we do not use SDL_DOUBLEBUF in
     * the flags to SDL_SetVideoMode. That does
     * not affect the GL attribute state, only
     * the standard 2D blitting setup.
     */
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

  /*
     * We want to request that SDL provide us
     * with an OpenGL window, in a fullscreen
     * video mode.
     *
     * EXERCISE:
     * Make starting windowed an option, and
     * handle the resize events properly with
     * glViewport.
     */
  flags = SDL_OPENGL | (thePref.FullScreen ? SDL_FULLSCREEN : 0);
  //flags = SDL_OPENGL | SDL_FULLSCREEN;

  /*
     * Set the video mode
     */
  if( SDL_SetVideoMode( width, height, bpp, flags ) == 0 ) {
    /* 
     * This could happen for a variety of reasons,
     * including DISPLAY not being set, the specified
     * resolution not being available, etc.
     */
    fprintf( stderr, "Video mode set failed: %s\n",
             SDL_GetError( ) );
    quit_ggelite( 1 );
  }
  SDL_ShowCursor(SDL_DISABLE);

}


int main( int argc, char *argv[] )
{
  GGLoop *ggLoop;
  NSAutoreleasePool *pool, *pool2;    
  START;
#ifdef DEBUG
  GSDebugAllocationActive(YES);
#endif
  pool = [NSAutoreleasePool new];
  [Resource initResourceManager];

  createGlContext();

  //  setMode(CheckMarked | LogFree);

  ggLoop = [GGLoop new: argc
		   arg: argv];

  pool2 = [NSAutoreleasePool new];

  [theWorld loadBaseSystem];

  [ggLoop runLoop];

  [Resource destroyResourceManager];

  RELEASE(pool2);
  RELEASE(ggLoop);
  RELEASE(pool);


  SDL_Quit();
  END;
  return 0;
}
