#include <lib3ds/file.h>
#include <lib3ds/chunk.h>
#include <lib3ds/node.h>
#include <lib3ds/mesh.h>
#include <lib3ds/vector.h>
#include <lib3ds/quat.h>
#include <lib3ds/matrix.h>
#include <lib3ds/material.h>
#include <lib3ds/camera.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <GL/gl.h>

#include "parse3ds.h"

static void
rasteriseNodeAndCompile(Lib3dsFile *file, Lib3dsNode *node)
{
  Lib3dsNode *subNode;
  for (subNode=node->childs; subNode!=0; subNode=subNode->next) {
    rasteriseNodeAndCompile(file, subNode);
  }

    if (!node->user.d) {
      float s[1];
      unsigned p;
      Lib3dsMatrix invMeshMatrix;
      Lib3dsMesh *mesh=lib3ds_file_mesh_by_name(file, node->name);
      ASSERT(mesh);
      if (!mesh) {
        return;
      }

      node->user.d=glGenLists(1);
      glNewList(node->user.d, GL_COMPILE);

      lib3ds_matrix_copy(invMeshMatrix, mesh->matrix);
      lib3ds_matrix_inv(invMeshMatrix);

      for (p=0; p<mesh->faces; ++p) {
        Lib3dsFace f=mesh->faceL[p];
        Lib3dsMaterial *mat=0;
        if (f.material[0]) {
          mat=lib3ds_file_material_by_name(file, f.material);
        }
        if (mat) {
          s[0]=1.0;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat->ambient);
          glMaterialfv(GL_FRONT, GL_DIFFUSE, mat->diffuse);
          glMaterialfv(GL_FRONT, GL_SPECULAR, mat->specular);
          glMaterialf(GL_FRONT, GL_SHININESS, 11.0-0.1*mat->shininess);
        }
        else {
          Lib3dsRgba a={0.2, 0.2, 0.2, 1.0};
          Lib3dsRgba d={0.8, 0.8, 0.8, 1.0};
          Lib3dsRgba s={0.0, 0.0, 0.0, 1.0};
          glMaterialfv(GL_FRONT, GL_AMBIENT, a);
          glMaterialfv(GL_FRONT, GL_DIFFUSE, d);
          glMaterialfv(GL_FRONT, GL_SPECULAR, s);
          glMaterialf(GL_FRONT, GL_SHININESS, 0.0);
        }
        {
          int i;
          Lib3dsVector v[3];
          for (i=0; i<3; ++i) {
            lib3ds_vector_transform(v[i], invMeshMatrix, mesh->pointL[f.points[i]].pos);
          }
          glBegin(GL_TRIANGLES);
            glNormal3fv(f.normal);
            glVertex3fv(v[0]);
            glVertex3fv(v[1]);
            glVertex3fv(v[2]);
          glEnd();
        }
      }

      glEndList();
    }

    if (node->user.d) {
      Lib3dsObjectData *d=&node->data.object;
      glPushMatrix();
      glMultMatrixf(&node->matrix[0][0]);
      glTranslatef(-d->pivot[0], -d->pivot[1], -d->pivot[2]);
      glCallList(node->user.d);
      glPopMatrix();
    }
}

static void
rasteriseNode(Lib3dsFile *file, Lib3dsNode *node)
{
  Lib3dsNode *subNode;
  for (subNode=node->childs; subNode!=0; subNode=subNode->next) {
    rasteriseNode(file, subNode);
  }

{      float s[1];
      unsigned p;
      Lib3dsMatrix invMeshMatrix;
      Lib3dsMesh *mesh=lib3ds_file_mesh_by_name(file, node->name);
      ASSERT(mesh);
      if (!mesh) {
        return;
      }


      lib3ds_matrix_copy(invMeshMatrix, mesh->matrix);
      lib3ds_matrix_inv(invMeshMatrix);

      for (p=0; p<mesh->faces; ++p) {
        Lib3dsFace f=mesh->faceL[p];
        Lib3dsMaterial *mat=0;
        if (f.material[0]) {
          mat=lib3ds_file_material_by_name(file, f.material);
        }
        if (mat) {
          s[0]=1.0;
          glMaterialfv(GL_FRONT, GL_AMBIENT, mat->ambient);
          glMaterialfv(GL_FRONT, GL_DIFFUSE, mat->diffuse);
          glMaterialfv(GL_FRONT, GL_SPECULAR, mat->specular);
          glMaterialf(GL_FRONT, GL_SHININESS, 11.0-0.1*mat->shininess);
        }
        else {
          Lib3dsRgba a={0.2, 0.2, 0.2, 1.0};
          Lib3dsRgba d={0.8, 0.8, 0.8, 1.0};
          Lib3dsRgba s={0.0, 0.0, 0.0, 1.0};
          glMaterialfv(GL_FRONT, GL_AMBIENT, a);
          glMaterialfv(GL_FRONT, GL_DIFFUSE, d);
          glMaterialfv(GL_FRONT, GL_SPECULAR, s);
          glMaterialf(GL_FRONT, GL_SHININESS, 0.0);
        }
        {
          int i;
          Lib3dsVector v[3];
          for (i=0; i<3; ++i) {
            lib3ds_vector_transform(v[i], invMeshMatrix, mesh->pointL[f.points[i]].pos);
          }
          glBegin(GL_TRIANGLES);
            glNormal3fv(f.normal);
            glVertex3fv(v[0]);
            glVertex3fv(v[1]);
            glVertex3fv(v[2]);
          glEnd();
        }
      }

    }
}



void destroy(void *_f)
{
  Lib3dsFile *file;
  file = (Lib3dsFile *)_f;
  lib3ds_close(file);
}
void rasterise(void *_f)
{
  Lib3dsFile *file;
  Lib3dsNode *p;
  file = (Lib3dsFile *)_f;

  glPushAttrib(GL_ENABLE_BIT | GL_LIGHTING_BIT);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_CULL_FACE);
  glEnable(GL_NORMALIZE);
  for (p=file->nodes; p!=0; p=p->next) {
    rasteriseNode(file, p);
  }
  glPopAttrib();
}
  

void * parse3ds(const char *filename)
{
  Lib3dsFile *file;
  file=lib3ds_open(filename);
  if (!file) {
    return NULL;
  }

  return(void*)file;
}
