/*
 *  parseASC.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Foundation/Foundation.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSZone.h>
#include <Foundation/NSException.h>
#include <GL/gl.h>
#include <stdlib.h>
#include <string.h>
#include "Metaobj.h"
#include "GG3D.h"
#include "parseASC.h"
#include "GGSpaceObject.h"
#include "utile.h"
#include "Resource.h"
#include "Model3DS.h"

NSString *EndOfString=@"myEndOfString";
NSString *WordTooBig=@"myWordTooBig";

#define RAISEENDFOFSTRING     [NSException raise: EndOfString  \
					   format: @"Unexpected End Of File"]

static int numLigne;

static Class model3DSClass;

void skipSpace(const char **buffer)
{
  while( **buffer != '\0' && isspace( **buffer ) )
    {
      if( **buffer == '\n' )
	numLigne++;
      (*buffer)++;
    }
  if( **buffer == '\0' )
    RAISEENDFOFSTRING;
}

void getWord(char word[50], const char **buffer)
{
  int i = 0;
  skipSpace(buffer);
  while( **buffer != '\0' && isgraph(**buffer) && i < 49 )
    word[i++] = *((*buffer)++);
  if( i == 49 )
    {
      [NSException raise: WordTooBig format: @""];
    }
  word[i] = '\0';
  //  fprintf(stderr, "str : '%s'\n", word);
}

//  void getIdent(char word[50], const char **buffer)
//  {
//    skipSpace(buffer);
  
//  }

int skipToEndOfLine(const char **buffer)
{
  while( **buffer != '\0' && **buffer != '\n'  )
    (*buffer)++;
  if( **buffer == '\n' )
    {
      (*buffer)++;
      numLigne++;
      return 0;
    }
  else /* end of file */
    return -1;
}

void skipUntilLineBegin(const char **buffer, const char *begin)
{
  int len = strlen(begin);
  while( strncasecmp(*buffer, begin, len) != 0 )
    {
      if( skipToEndOfLine(buffer) < 0 )
	RAISEENDFOFSTRING;
    }
}

void skipUntilWord(const char **buffer, const char *word)
{
  char buf[50];
  while(1)
    {
      getWord(buf, buffer);
      if( strcasecmp(buf, word) == 0 )
	return;
    }
}

static NSMutableDictionary	*texture;


@implementation ModelObject
+ (void) initialize
{
  if( self == [ModelObject class])
    {
      texture = [NSMutableDictionary new];
    }
}

- init
{
  [super init];
  nameObject = nil;
  vertices = NULL;
  faces = NULL;
  addToCheck(self);
  return self;
}

+(VectCol) getCouleurFromString: (NSString *)str
{
  VectCol cl, *pcl;
  NSData *theData;
  if( (theData = [texture objectForKey: str]) != nil )
    {
      pcl = (VectCol *)[theData bytes];
      return *pcl;
    }
  else
    {
      const char *chaine = [str cString];
      unsigned int val;
      int r, g, b;
      cl.a = 1;
      if( sscanf(chaine, "r%dg%db%d", &r, &g, &b) == 3 )
	{
	  cl.r = (float)r/255;
	  cl.g = (float)g/255;
	  cl.b = (float)b/255;
	}
      else if (sscanf(chaine, "#%x", &val) == 1 )
	{
	  cl.r = (float) ((val>>16) & 0xff) / 255;
	  cl.g = (float) ((val >> 8) & 0xff) / 255;
	  cl.b = (float) ((val) & 0xff) / 255;
	}
      else
	{
	  NSWarnMLog(@"can't parse %@", str);
	  cl.r = vrnd();
	  cl.g = vrnd();
	  cl.b = vrnd();
	}
      theData = [NSData dataWithBytes:&cl length: sizeof(cl)];
      [texture setObject: theData forKey: str];
      return cl;
    }
    
}

- (void) dealloc
{
  RELEASE(nameObject);
  if( vertices != NULL )
    NSZoneFree([self zone], vertices);
  if( faces != NULL )
    NSZoneFree([self zone], faces);
  [super dealloc];
}

- (void) initNormal
{
  int i;
  for( i = 0; i < nface; ++i)
    {
      Vect3D tp1, tp2, tp3;
      Face theFace = faces[i];
      mkVectFromPoint(&(vertices[theFace.A]), &(vertices[theFace.B]), &tp1);
      mkVectFromPoint(&(vertices[theFace.A]), &(vertices[theFace.C]), &tp2);

      prodVect(&tp1, &tp2, &tp3);
      normaliseVect(&tp3);
      faces[i].normal = tp3;
    }
}

- (void) calcBoundingBox
{
  int i;
  Vect3D _max = { -1e20, -1e10, -1e20 };
  Vect3D _min = { 1e20, 1e10, 1e20 };
		 
  min = _min;
  max = _max;

  for( i = 0; i < nvert; ++i )
    {
      Vect3D pos = vertices[i];

      min.x = (min.x < pos.x ? min.x : pos.x);
      min.y = (min.y < pos.y ? min.y : pos.y);
      min.z = (min.z < pos.z ? min.z : pos.z);

      max.x = (max.x > pos.x ? max.x : pos.x);
      max.y = (max.y > pos.y ? max.y : pos.y);
      max.z = (max.z > pos.z ? max.z : pos.z);
    }

  NSDebugMLLog(@"parseur", @"min = %@, max = %@", stringOfVect2(min),
	       stringOfVect2(max));
}

- (Vect3D *) boundingBox: (Vect3D *)pax
{
  *pax = max;
  return &min;
}

- (void) rasterise
{
  int i;
  //  glDisable(GL_CULL_FACE);
  glBegin(GL_TRIANGLES);
#if 1
  for (i = 0; i < nface; ++i)
    {
      Face theFace = faces[i];
      glColor3f(faces[i].couleur.r,
		faces[i].couleur.g,
		faces[i].couleur.b);
      glNormal3f(faces[i].normal.x,
		 faces[i].normal.y,faces[i].normal.z);
      glVertex3f(vertices[theFace.A].x,
		 vertices[theFace.A].y,
		 vertices[theFace.A].z);
      glVertex3f(vertices[theFace.B].x,
		 vertices[theFace.B].y,
		 vertices[theFace.B].z);
      glVertex3f(vertices[theFace.C].x,
		 vertices[theFace.C].y,
		 vertices[theFace.C].z);
    }
#endif
#if 0
  glColor3d(0.8, 0.8, 0.8);
  glVertex3d(1, 0, 0);
  glVertex3d(0, 1, 0);
  glVertex3d(0, 0, 1);
#endif
  glEnd();
  //  glEnable(GL_CULL_FACE);
}


- (void) parseASC: (NSString *)name
{
  NSString *data = [NSString stringWithContentsOfFile: name];

  if( data == nil )
    {
      [NSException raise: @"FileNotFound" format: @""];
    }
  else
    {
      const char *buffer = [data cString];
      VectCol whiteCol = {1, 1, 1, 1};
      char word[50];
      char *ptr;
      int res, i;
      
      skipUntilLineBegin(&buffer, "Named object");

      getWord(word, &buffer);
      getWord(word, &buffer);
      getWord(word, &buffer);
      
      ptr = strchr(word+1, '"');
      NSAssert(ptr != NULL, @"no Quote");


      *ptr='\0';

      nameObject = RETAIN([NSString stringWithCString: word+1]);
      NSDebugMLLog(@"parseur", @"nom : %@", nameObject);
      skipToEndOfLine(&buffer);
      res = sscanf(buffer, "Tri-mesh, Vertices: %d Faces: %d", &nvert, &nface);
      NSAssert(res == 2, @"can't parse nvert, nface");
      NSDebugMLLog(@"parseur", @"nvert = %d nface = %d", nvert, nface);

      vertices = NSZoneMalloc([self zone], sizeof(Vect3D)*nvert);
      faces = NSZoneMalloc([self zone], sizeof(Face)*nface);

      skipToEndOfLine(&buffer);

      skipUntilLineBegin(&buffer, "Vertex list:");
      skipToEndOfLine(&buffer);

      for( i = 0; i < nvert; ++i )
	{
	  double t1, t2, t3;
	  skipUntilLineBegin(&buffer, "Vertex");
	  if( sscanf(buffer, "Vertex %*d: X:%lf Y:%lf Z:%lf", &t1, &t2, &t3) < 3 )
	    {
	      NSLog(@"merdouille %d", i);
	      [NSException raise:@"SyntaxError" format: @"vertices expected"];
	    }
	  vertices[i].x = t1;
	  vertices[i].y = t2;
	  vertices[i].z = t3;
	  skipToEndOfLine(&buffer);
	}

      skipUntilLineBegin(&buffer, "Face list:");
      skipToEndOfLine(&buffer);


      for( i = 0; i < nface; ++i )
	{
	  int ab, bc, ca;
	  skipUntilLineBegin(&buffer, "Face");
	  if( sscanf(buffer, 
		     "Face %*d:    A:%hd B:%hd C:%hd AB:%d BC:%d CA:%d",
		     &(faces[i].A),&(faces[i].B),&(faces[i].C),
		     &ab, &bc, &ca) < 6 )
	    {
	      NSLog(@"chie level %d, %s", i, buffer);
	      [NSException raise:@"SyntaxError" format: @"face expected"];
	    }
	  else
	    {
	      faces[i].AB = !!ab;
	      faces[i].BC = !!bc;
	      faces[i].CA = !!ca;
	      NSAssert(0<= faces[i].A && faces[i].A < nvert, 
		       @"vertex out of range");
	      NSAssert(0<= faces[i].B && faces[i].B < nvert, 
		       @"vertex out of range");
	      NSAssert(0<= faces[i].C && faces[i].C < nvert, 
		       @"vertex out of range");
	    }

	  skipToEndOfLine(&buffer);

	  faces[i].couleur = whiteCol;
	  for(;;) {
	    if( strncasecmp(buffer, "Face", 4) == 0 )
	      goto endOfLoop;
	    else if( strncasecmp(buffer, "Material", 8) == 0 )
	      break;
	    if( skipToEndOfLine(&buffer) < 0 )
	      goto endOfLoop;
	  }
	    
	  skipUntilLineBegin(&buffer, "Material");
	  {	      
	    char *ptr2;
	    NSString *tex;
	    ptr = strchr(buffer, '"');
	    NSAssert(ptr != NULL, @"no Quote");
	    ptr2 = strchr(ptr+1, '"');
	    NSAssert(ptr2 != NULL, @"no Quote");
	    *ptr2 = '\0';
	    buffer = ptr2+1;
	    tex = [NSString stringWithCString: ptr+1];
	    faces[i].couleur = [ModelObject getCouleurFromString: tex];
	    //	      NSLog(@"texture : %@", tex);
	  }
	  skipToEndOfLine(&buffer);
	endOfLoop:
	}
		     
    }
}

- (void) _setOptionWithDictionary: (NSDictionary *)dico
{
  NSString *val;
  if( dico == nil )
    return;

  if( (val = [dico objectForKey:@"Scale"]) != nil )
    {
      real fact;
      fact = [val doubleValue];
      if( fact <= 0 )
	NSWarnMLog(@"negative scale factor !!");
      else
	{
	  int i;
	  NSDebugMLLog(@"parseur", @"scaling model by %g", (double) fact);
	  for( i = 0; i < nvert; ++i)
	    {
	      vertices[i].x *= fact;
	      vertices[i].y *= fact;
	      vertices[i].z *= fact;
	    }
	}
    }
}

+ (void) load3DSBundle
{
#ifdef HAVE_LIB3DS
  model3DSClass = [Model3DS class];
#else
  NSBundle *theBundle;

  theBundle = [NSBundle bundleWithPath: [[theResourceManager bundlePath]
					  stringByAppendingPathComponent: 
					    @"Model3DS.bundle"]];
  model3DSClass = [theBundle classNamed: @"Model3DS"];
#endif
}


+ modelWithFileName: (NSString *) name
     withDictionary: (NSDictionary *)dico
{
  NSString *ext;
  ModelObject *m = nil;
  NS_DURING
    {
      numLigne = 0;
      ext = [name pathExtension];
      name = [[theResourceManager modelPath] stringByAppendingPathComponent:
					       name];
      if( [ext isEqualToString: @"asc"] )
	{
	  m = [self new];
	  [m parseASC: name];
	  [m _setOptionWithDictionary: dico];
	  [m calcBoundingBox];
	  [m initNormal];
	}
      else if ( [ext  isEqualToString: @"3ds"] )
	{
	  id <NSObject> obj;
	  if( model3DSClass == nil )
	    {
	      [self load3DSBundle];
	      if( model3DSClass == nil )
		{
		  NSWarnMLog(@"Unable to load 3DS Bundle");
		  NS_VALUERETURN(nil, id);
		}
	      NSDebugMLLog(@"Bundle", @"Model3ds succesfully loaded");
	    }
	  obj = [model3DSClass modelWith3DStudioFile: name
			       withDictionary: dico];
	  m = (ModelObject *)RETAIN(obj);
	}
      else
	{
	  NSWarnMLog(@"unknow file format %@", name);
	  NS_VALUERETURN(nil, id);
	}
      
    }
    NS_HANDLER
    {
      NSWarnMLog(@"ca chie grave avec %@ ligne %d :%@ %@", name, numLigne,
		 [localException name],
		 [localException reason]);
      if( m != nil )
	RELEASE(m);
      return nil;
    }
    NS_ENDHANDLER
  return AUTORELEASE(m);
}

+ modelWithFileName: (NSString *) name
{
  return [self modelWithFileName: name
	       withDictionary: nil];
}

- (NSString *)description
{
  int i;
  NSMutableString *str = [NSMutableString string];
  
  [str appendString: @"Vertices\n"];
  for( i = 0; i < nvert; ++i )
    [str appendFormat: @"A: %g B: %g C: %g\n", vertices[i].x,
	 vertices[i].y, vertices[i].z];

  [str appendString: @"Faces\n"];
  for( i = 0; i < nface; ++i)
    [str appendFormat: @"A:%hd B:%hd C:%hd\n", faces[i].A,
	 faces[i].B,faces[i].C];
  return str;
}
@end

#if 0
int main(int argc, char **argv)
{
  NSAutoreleasePool *pool;
  ModelObject *m;
  START;
  pool = [NSAutoreleasePool new];
  m = [ModelObject modelWithFileName: [NSString stringWithCString: 
						    argv[argc-1]]];
  if( m != nil )
    {
      puts("ok dukki");
      NSLog(@"string : %@", m);
    }
  else
    {
      NSLog(@"ca chie\n");
      puts("merde");
    }
  RELEASE(pool);
  END;
  return 0;
}


#endif
