/*
 *  utile.m
 *
 *  Copyright (c) 2001 by Frdric De Jaeger
 *  
 *  Date: Mars 2001
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <jpeglib.h>

#include <Foundation/NSException.h>
#include <Foundation/NSDebug.h>
#include <Foundation/NSString.h>
#include <Foundation/NSArray.h>

#include "config.h"
#include "utile.h"
#include "GGSpaceObject.h"
#include "Preference.h"
#include "GGRepere.h"
#include "Tableau.h"
#include "GGSky.h"
#include "GGGlyphLayout.h"

const real EPS=0.00000001;

double gettimesec(void)
{
  double res;
  struct timeval ret;
  if( gettimeofday(&ret, NULL) == 0 )
    {
      res = ret.tv_sec;
      res += (double)(ret.tv_usec)/1e6;
      return res;
    }
  else
    {	
      perror("bug!!");
      return 0.0;
    }
}


double vrnd(void)
{
  return(((double)rand())/RAND_MAX);
}

unsigned irnd(unsigned max)
{
  unsigned res = (unsigned)floor(vrnd()*(max));
  if( res >= max-1 )
    res = max-1;
  return res;
}


real distanceAbsolu2(GGSpaceObject *o1, GGSpaceObject *o2)
{
  Vect3D v;
  real temp;
  [o1 positionInRepere: o2
      in: &v];
  temp = prodScal(&v, &v);

  NSDebugLLogWindow(@"distance", @"%@ %@ %g", o1, o2, temp);

  return temp;
}

NSString *stringOfVect(Vect3D *v)
{
  return [NSString stringWithFormat: @"(%g, %g, %g)", 
		   v->x, v->y, v->z];
}

NSString *stringOfVect2(Vect3D v)
{
  return stringOfVect(&v);
}

NSString *stringOfMatrice(GGMatrix4 *aMat)
{
  NSMutableString *res = [NSMutableString string];
  pmat mat = (pmat)aMat;
  int i;

  for( i = 0; i < 4; ++i)
    {
      [res appendFormat: @"| %+5.5e %+5.5e %+5.5e %+5.5e |\n",
		      mat[0][i], mat[1][i], mat[2][i], mat[3][i]];
    }
  return res;
}

void logArray(NSArray *array)
{
  int n = [array count];
  NSString *str[n];
  int i;
  [array getObjects: str];

  for(i = 0; i < n; ++i)
    NSLog(str[i]);
}

void drawRectBorder(short int x1, short int y1, short int x2, short int y2)
{
  glBegin(GL_LINE_LOOP);
  glVertex2f(x1, y1);
  glVertex2f(x1, y2);
  glVertex2f(x2, y2);
  glVertex2f(x2, y1);
  //  glVertex2f(x1, y1);
  glEnd();
}

double ggmod(double x, double y)
{
  double res = fmod(x, y);

  if( res < 0 )
    res += y;
  return res;
}

void pushStandardRep(NSSize size)
{
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0,size.width,0,size.height,-1.0,1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
}

void popStandardRep()
{
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
}

static int count;

static void outPutPixel(unsigned char r, unsigned char g, 
			unsigned char b, FILE *f)
{
  fprintf(f, "%d %d %d", (int)r, (int)g, (int)b);

  if( ++count > 4 )
    {
      count = 0;
      fprintf(f, "\n");
    }
  else
    fprintf(f, " ");
}


void writeRGBTofile(int width, int height, const unsigned char *data)
{
  int i, j;
  static int indFile = 1;
  NSString *fileName = [NSString stringWithFormat: @"img%d.raw", indFile];
  FILE *outfile;

  indFile++;
  count = 0;

  if ((outfile = fopen([fileName cString], "w")) == NULL) {
    NSLog(@"can't open %@", fileName);
    exit(1);
  }

  fprintf(outfile, "P3\n");
  fprintf(outfile, "# CREATOR: XV Version 3.10a  Rev: 12/29/94 (PNG patch 1.2)\n");
  fprintf(outfile, "# CREATOR: XV Version 3.10a  Rev: 12/29/94  Quality = 75, Smoothing = 0\n");
  fprintf(outfile, "%d %d\n255\n", width, height);


  for( i = 0; i < height; ++i)
    for( j = 0; j < width; ++j)
      {
	int index = 3*(width*(height -1 - i)+j);
	outPutPixel(data[index], data[index+1], data[index+2], outfile);
      }

  fclose(outfile);
}

void writeRGBTofileJPG(int width, int height, const unsigned char *data)
{
  static int ind = 1;
  NSString *fileName;
  int quality = 90;

  
  /* This struct contains the JPEG compression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   * It is possible to have several such structures, representing multiple
   * compression/decompression processes, in existence at once.  We refer
   * to any one struct (and its associated working data) as a "JPEG object".
   */
  struct jpeg_compress_struct cinfo;
  /* This struct represents a JPEG error handler.  It is declared separately
   * because applications often want to supply a specialized error handler
   * (see the second half of this file for an example).  But here we just
   * take the easy way out and use the standard error handler, which will
   * print a message on stderr and call exit() if compression fails.
   * Note that this struct must live as long as the main JPEG parameter
   * struct, to avoid dangling-pointer problems.
   */
  struct jpeg_error_mgr jerr;
  /* More stuff */
  FILE * outfile;               /* target file */
  JSAMPROW row_pointer[1];      /* pointer to JSAMPLE row[s] */
  int row_stride;               /* physical row width in image buffer */

  /* Step 1: allocate and initialize JPEG compression object */

  /* We have to set up the error handler first, in case the initialization
   * step fails.  (Unlikely, but it could happen if you are out of memory.)
   * This routine fills in the contents of struct jerr, and returns jerr's
   * address which we place into the link field in cinfo.
   */

  fileName = [NSString stringWithFormat: @"pic%d.jpg", ind];
  ind ++;
  cinfo.err = jpeg_std_error(&jerr);
  /* Now we can initialize the JPEG compression object. */
  jpeg_create_compress(&cinfo);

  /* Step 2: specify data destination (eg, a file) */
  /* Note: steps 2 and 3 can be done in either order. */

  /* Here we use the library-supplied code to send compressed data to a
   * stdio stream.  You can also write your own code to do something else.
   * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
   * requires it in order to write binary files.
   */
  if ((outfile = fopen([fileName cString], "wb")) == NULL) {
    NSLog(@"can't open %@\n", fileName);
    exit(1);
  }
  jpeg_stdio_dest(&cinfo, outfile);

  /* Step 3: set parameters for compression */

  /* First we supply a description of the input image.
   * Four fields of the cinfo struct must be filled in:
   */
  cinfo.image_width = width;      /* image width and height, in pixels */
  cinfo.image_height = height;
  cinfo.input_components = 3;           /* # of color components per pixel */
  cinfo.in_color_space = JCS_RGB;       /* colorspace of input image */
  /* Now use the library's routine to set default compression parameters.
   * (You must set at least cinfo.in_color_space before calling this,
   * since the defaults depend on the source color space.)
   */
  jpeg_set_defaults(&cinfo);
  /* Now you can set any non-default parameters you wish to.
   * Here we just illustrate the use of quality (quantization table) scaling:
   */
  jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);

  /* Step 4: Start compressor */

  /* TRUE ensures that we will write a complete interchange-JPEG file.
   * Pass TRUE unless you are very sure of what you're doing.
   */
  jpeg_start_compress(&cinfo, TRUE);

  /* Step 5: while (scan lines remain to be written) */
  /*           jpeg_write_scanlines(...); */

  /* Here we use the library's state variable cinfo.next_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   * To keep things simple, we pass one scanline per call; you can pass
   * more if you wish, though.
   */
  row_stride = width * 3; /* JSAMPLEs per row in image_buffer */

  while (cinfo.next_scanline < cinfo.image_height) {
    /* jpeg_write_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could pass
     * more than one scanline at a time if that's more convenient.
     */
    row_pointer[0] = (unsigned char *)& data[(height - 1 - cinfo.next_scanline )* row_stride];
    (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }

  /* Step 6: Finish compression */

  jpeg_finish_compress(&cinfo);
  /* After finish_compress, we can close the output file. */
  fclose(outfile);

  /* Step 7: release JPEG compression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_compress(&cinfo);

  /* And we're done! */
}


BOOL getWinCoordinates(const Camera *c, const Vect3D *vec, Vect3D *res)
{
  Vect3D pos;
  real	x, y;
  transformePoint4D(&c->modelView, vec, &pos);
  if( pos.z > EPS )
    {
      x = -pos.x / pos.z / c->param->xFoc;
      y = pos.y / pos.z / c->param->yFoc;

      if( x >= -1 && x <= 1 && y >= -1 && y <= 1 )
	{
	  res->x = (x+1) * c->param->width/2;
	  res->y = (y+1) * c->param->height/2;
	  res->z = (pos.z - c->param->zmin) / (c->param->zmax-c->param->zmin);
	  return YES;
	}
    }
  return NO;
}

float absoluteMagnitude(struct __star *s)
     /*distance is in 0.01 * LightYear*/
{
  double dist2 = square(s->xd)+square(s->yd) +  square(s->zd);
  if (dist2 <= 1 )
    return s->mag0;
  else
    return s->mag0+5-2.5*log10(dist2/(326*326));;
}

void draw3DstringFast(Camera *c, real x, real y, real z, GGGlyphLayout *str)
{
  Vect3D pos = {x, y, z};
  Vect3D res;

  if( getWinCoordinates(c, &pos, &res) )
    {
      glLoadIdentity();
      glTranslatef(res.x, res.y, 0);
      [str fastDraw];
    }
}

real omegaOrbite(real radius, real orbiteRadius)
{
  return 2e-7*pow((150e9/1e9)*radius/orbiteRadius, 1.5);
}

void normaliseMatrice(GGMatrix4 *mat)
{
  real tmp;
  normaliseVect(&mat->direction);
  tmp = prodScal(&mat->direction, &mat->haut);
  addLambdaVect(&mat->haut, -tmp, &mat->direction, &mat->haut);
  normaliseVect(&mat->haut);

  prodVect(&mat->haut, &mat->direction, &mat->gauche);
}

void printGLError(id str)
{
  GLenum err;
  if( str )
    NSWarnLog(@"id : %@", str);
  if( (err=glGetError() ) != GL_NO_ERROR )
    {
      NSWarnLog(@"gl Error : %s", gluErrorString(err));
      abort();
    }
}


#define MAGIC1 0xdeadface
#define MAGIC2 0xfacedead



typedef struct __block
{
  int magic1;
  int count;
  int magic2;
  void (*destructor)(void *);
#ifdef DebugMem
  char *s;
#endif
}blockdeb;

static inline void __checkHeader(blockdeb *b)
{
  NSCParameterAssert(b->magic1 == MAGIC1 && b->magic2 == MAGIC2);
}

#ifdef DebugMem
#define MaxCache 3000

void setName(void *ptr, char *name)
{
  blockdeb *b = ((blockdeb *)ptr) - 1;
  __checkHeader(b);
  b->s = strdup(name);
}

char *stringOfPtr(void *ptr)
{
  blockdeb *b = ((blockdeb *)ptr) - 1;
  __checkHeader(b);
  return b->s;
}

static void * checker[MaxCache];
void addPtr(void *ptr)
{
  int i;
  for( i = 0; i < MaxCache && checker[i] != NULL; ++i );

  if( i < MaxCache )
    checker[i] = ptr;
  else
    puts("too many blocks allocated");
}

void removePtr(void *ptr)
{
  int i;
  for( i = 0; i < MaxCache && checker[i] != ptr; ++i);

  if( i < MaxCache )
    checker[i] = NULL;
  else
    puts("pointer not found");
}

void checkMem(void)
{
  int i;
  puts("dbut du contrle mmoire");
  for( i = 0; i < MaxCache; ++i )
    if( checker[i] )
      {
	fprintf(stderr, "object %p not deallocated\nstring value %s\n",
		checker[i], stringOfPtr(checker[i]));
      }
  puts("fin du contrle mmoire");
}

#undef ggalloc
void *ggalloc(int size);
void *__ggalloc(int size, char *chaine)
{
  void *ptr;
  ptr = ggalloc(size);
  setName(ptr, chaine);
  return ptr;
}

#endif



void *ggalloc(int size)
{
  blockdeb *b;
  int newSize = size + sizeof(blockdeb);
  b = (blockdeb *)malloc(newSize);
  b->magic1 = MAGIC1;
  b->magic2 = MAGIC2;
  b->count = 1;
  b->destructor = NULL;
#ifdef DebugMem
  b->s = NULL;
  addPtr(&b[1]);
#endif
  return &b[1];
}

void ggretain(void *ptr)
{
  blockdeb *b = ((blockdeb *)ptr) - 1;
  __checkHeader(b);

  (b->count)++;
}

void ggrelease(void *ptr)
{
  blockdeb *b = ((blockdeb *)ptr) - 1;

  if( ptr == NULL )
    return;

  __checkHeader(b);

//  fprintf(stderr, "ggrelease called\n");

  if( b->count <= 0 )
    {
      fprintf(stderr, "bock is already dead\n");
      abort();
    }

  (b->count)--;
  if( b->count == 0)
    {
      if( b->destructor != NULL )
	b->destructor(ptr);
      b->magic1 = 0;
      b->magic2 = 0;
#ifdef DebugMem
      if( b->s != NULL )
	free(b->s);
      removePtr(&b[1]);
#endif
      free(b);
    }
}
  
void ggSetDestructor(void *ptr, void (*f)(void *))
{
  blockdeb *b = ((blockdeb *)ptr) - 1;
  __checkHeader(b);
  b->destructor = f;
}


static inline VectCol moyenne(float val, VectCol c2, VectCol c1)
{
  VectCol ret;
  ret.r = val*c1.r + (1-val)*c2.r;
  ret.g = val*c1.g + (1-val)*c2.g;
  ret.b = val*c1.b + (1-val)*c2.b;
  ret.a = val*c1.a + (1-val)*c2.a;
  return ret;
}

VectCol genColFromMapHeight(MapCol map[], int size, float param)
{
  int i;
  for( i = 0; i < size; ++i )
    {
      if( param <= map[i].val )
	{
	  if( i > 0 )
	    {
	      float alpha;
	      alpha = ((float)param - map[i-1].val) / 
		((float)(map[i].val - map[i-1].val));
	      return moyenne(alpha, map[i-1].couleur, map[i].couleur);
	    }
	  else
	    return map[0].couleur;
	}
    }
  return map[size-1].couleur;
}

unsigned char *rawImageFromJPGFileMalloc(const char *fileName, int *pwidth,
					 int *pheight, int *pncomp)
{
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  int width, height;
  int	row_stride;
  FILE * infile;		/* source file */
  JSAMPROW row_pointer[1];
  unsigned char *dataChar;


  if ((infile = fopen(fileName, "rb")) == NULL) {
    NSWarnLog(@"can't open %@", fileName);
    return NULL;
  }

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);

  jpeg_stdio_src(&cinfo, infile);

  jpeg_read_header(&cinfo, TRUE);

  jpeg_start_decompress(&cinfo);

  *pwidth = width = cinfo.output_width;
  *pheight = height = cinfo.output_height;
  *pncomp = cinfo.output_components;
  row_stride = cinfo.output_components*cinfo.output_width;

  dataChar = malloc(width*height*cinfo.output_components);
  while (cinfo.output_scanline < cinfo.output_height) 
    {
      row_pointer[0] = dataChar+(cinfo.output_height - 1 - 
				 cinfo.output_scanline)*row_stride;
      jpeg_read_scanlines(&cinfo, row_pointer, 1);
    }
  (void) jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(infile);

  return dataChar;
}
